#include "libgamelibgame.h"

#include "files.h"
#include "init.h"
#include "tools.h"
#include "tape.h"

#include "p_gfxinit.h"

#define CHUNK_ID_LEN		4	/* IFF style chunk id length  */

#define FILE_VERS_CHUNK_SIZE	8	/* size of file version chunk */

#define LEVEL_HEADER_SIZE	80	/* size of level file header  */
#define LEVEL_HEADER_UNUSED	1	/* unused level header bytes  */
#define LEVEL_CHUNK_CNT2_SIZE	160	/* size of level CNT2 chunk   */ 
#define LEVEL_CHUNK_CNT2_UNUSED	11	/* unused CNT2 chunk bytes    */
#define LEVEL_CHUNK_CNT3_HEADER	16	/* size of level CNT3 header  */
#define LEVEL_CHUNK_CNT3_UNUSED	10	/* unused CNT3 chunk bytes    */

#define TAPE_HEADER_SIZE	20	/* size of tape file header   */
#define TAPE_HEADER_UNUSED	3	/* unused tape header bytes   */

#define LEVEL_CHUNK_CNT3_SIZE(x) (LEVEL_CHUNK_CNT3_HEADER + (x))

/* file identifier strings */
#define LEVEL_COOKIE_TMPL	"ROCKSNDIAMONDS_LEVEL_FILE_VERSION_x.x"
#define TAPE_COOKIE_TMPL	"ROCKSNDIAMONDS_TAPE_FILE_VERSION_x.x"

#define IS_SINGLE_LEVEL_FILE(x)		(x < 10)
#define IS_PACKED_LEVEL_FILE(x)		(x > 10)

//------start
/*
LocalID			any_dbid;
UInt16			any_card;
*/
struct
{
	MemPtr	ptr;
	UInt16	cur_pos, size;
}	LikeFile;

inline void InitLikeFile(MemPtr ptr)
{
	LikeFile.ptr=ptr;
	LikeFile.size = MemPtrSize(ptr);
	LikeFile.cur_pos=0;
}

inline void CloseLikeFile(void)
{
	MemPtrUnlock(LikeFile.ptr);
};

INIT_SECTION UInt16 ReadLikeFile(char* where, UInt16 size)
{
	UInt16 i;
	i=0;
	while(LikeFile.cur_pos<LikeFile.size&&i<size)
	{
		where[i]=((char*)LikeFile.ptr)[LikeFile.cur_pos];
		i++;
		LikeFile.cur_pos++;	
		if(where[i-1]==0xA)
		{
			where[i]=0;
			return i;
		}
	}
	where[i]=0;
	return i;
};

INIT_SECTION UInt16 ReadUnusedBytesFromLikeFile(UInt16 size)
{
	UInt16 to_read;
	to_read=LikeFile.size-LikeFile.cur_pos;
	if(to_read>size)
		to_read=size;
	LikeFile.cur_pos+=to_read;
	return to_read;
};

INIT_SECTION unsigned char getLikeFile8Bit(void)
{
	if(LikeFile.size<=LikeFile.cur_pos)
		return 0;
	LikeFile.cur_pos++;
	return ((char*)LikeFile.ptr)[LikeFile.cur_pos-1];
};

INIT_SECTION unsigned short int getLikeFile16BitBE(void)
{
	unsigned char* src;
	if(LikeFile.size<=LikeFile.cur_pos+1)
		return 0;
	src=(unsigned char*)LikeFile.ptr;
	src+=LikeFile.cur_pos;
	LikeFile.cur_pos+=2;
	return ((src[0]<<8)|src[1]);
};

INIT_SECTION unsigned long getLikeFile32BitBE(void)
{
	unsigned char* src;
	long result;
	if(LikeFile.size<=LikeFile.cur_pos+3)
		return 0;
	src=(unsigned char*)LikeFile.ptr;
	src+=LikeFile.cur_pos;
	LikeFile.cur_pos+=4;
	((char*)&result)[0]=src[0];
	((char*)&result)[1]=src[1];
	((char*)&result)[2]=src[2];
	((char*)&result)[3]=src[3];
	return result;
};

INIT_SECTION long getLikeFileVersion(void)
{
	if(LikeFile.size<=LikeFile.cur_pos+3)
		return 0;
	{
		int version_major = ((char*)LikeFile.ptr)[LikeFile.cur_pos];
		int version_minor = ((char*)LikeFile.ptr)[LikeFile.cur_pos+1];
		int version_patch = ((char*)LikeFile.ptr)[LikeFile.cur_pos+2];
		int version_build = ((char*)LikeFile.ptr)[LikeFile.cur_pos+3];
		
		LikeFile.cur_pos+=4;
		
		return VERSION_IDENT(version_major, version_minor, version_patch,
				version_build);
	} 
};

INIT_SECTION boolean getLikeFileChunkBE(char* chunk_name, long* chunk_size)
{
	unsigned char TmpBuf[4];
	unsigned char* src;
	if(ReadLikeFile(chunk_name,4)!=4)
		return FALSE;
	if(LikeFile.size<=LikeFile.cur_pos+3)
		return FALSE;
	src=(unsigned char*)LikeFile.ptr;
	src+=LikeFile.cur_pos;
	TmpBuf[0]=*src++;
	TmpBuf[1]=*src++;
	TmpBuf[2]=*src++;
	TmpBuf[3]=*src;
	LikeFile.cur_pos+=4;
	*chunk_size=*(long*)TmpBuf;
	return TRUE;
};

INIT_SECTION boolean SeekLikeFileCurPos(Int16 where)
{
	if(LikeFile.cur_pos+where>LikeFile.size||LikeFile.cur_pos+where<0)
		return FALSE;
	LikeFile.cur_pos+=where;
	return TRUE;
};
//------end

// =========================================================================
// tape file functions                                                      
// =========================================================================

INIT_SECTION static void setTapeInfoToDefaults()
{
  int i;
  
  // always start with reliable default values (empty tape)
  TapeErase();

  // default values (also for pre-1.2 tapes) with only the first player
  tape.player_participates[0] = TRUE;
  tape.player_participates[1] = 
  tape.player_participates[2] = 
  tape.player_participates[3] = FALSE;
	
  // at least one (default: the first) player participates in every tape
  tape.num_participating_players = 1;

  tape.level_nr = level_nr;
  tape.counter = 0;
  tape.changed = FALSE;

  tape.recording = FALSE;
  tape.playing = FALSE;
  tape.pausing = FALSE;

  tape.no_valid_file = FALSE;
}

static int LoadTape_VERS(int chunk_size)
//was static int LoadTape_VERS(FILE *file, int chunk_size, struct TapeInfo *tape)
{
	tape.file_version = getLikeFileVersion();
	tape.game_version = getLikeFileVersion();

	return chunk_size;
}

INIT_SECTION static int LoadTape_HEAD(int chunk_size)
//was static int LoadTape_HEAD(FILE *file, int chunk_size, struct TapeInfo *tape)
{
	int i;
	
	tape.random_seed = getLikeFile32BitBE();
	tape.date        = getLikeFile32BitBE();
	tape.length      = getLikeFile32BitBE();

	// read header fields that are new since version 1.2
	if (tape.file_version >= FILE_VERSION_1_2)
	{
		byte store_participating_players = getLikeFile8Bit();
		long engine_version;

		// since version 1.2, tapes store which players participate in the tape
		tape.num_participating_players = 0;
		for (i = 0; i < 4; i++)
		{
			tape.player_participates[i] = FALSE;

			if (store_participating_players & (1 << i))
			{
				tape.player_participates[i] = TRUE;
				tape.num_participating_players++;
			}
		}
		 
		ReadUnusedBytesFromLikeFile(TAPE_HEADER_UNUSED);

		engine_version = getLikeFileVersion();

		if (engine_version > 0)
			tape.engine_version = engine_version;
		else
			tape.engine_version = tape.game_version;
	}

	return chunk_size;
}

INIT_SECTION static int LoadTape_INFO(int chunk_size)
//was static int LoadTape_INFO(FILE *file, int chunk_size, struct TapeInfo *tape)
{
	int level_identifier_size;
	int i;

	level_identifier_size = getLikeFile16BitBE();

/*	tape->level_identifier =
			checked_realloc(tape->level_identifier, level_identifier_size);
*/
	for (i = 0; i < level_identifier_size; i++)
		/*tape->level_identifier[i] = */getLikeFile8Bit();

	tape.level_nr = getLikeFile16BitBE();

	chunk_size = 2 + level_identifier_size + 2;

	return chunk_size;
}

INIT_SECTION static int LoadTape_BODY(int chunk_size)
//was static int LoadTape_BODY(FILE *file, int chunk_size, struct TapeInfo *tape)
{
	int i, j;
	int chunk_size_expected =
			(tape.num_participating_players + 1) * tape.length;

	if (chunk_size_expected != chunk_size)
	{
		ReadUnusedBytesFromLikeFile(chunk_size);
		return chunk_size_expected;
	}

	for (i = 0; i < tape.length; i++)
	{
		if (i >= MAX_TAPELEN)
			break;

		for (j = 0; j < 4; j++)
		{
			byte action = MV_NO_MOVING;

			if (tape.player_participates[j])
				action = getLikeFile8Bit();

			if (j==0)
				tape.pos_action_[i] = action;
		}

		tape.pos_delay[i] = getLikeFile8Bit();

		if (tape.file_version == FILE_VERSION_1_0)
		{
			// eliminate possible diagonal moves in old tapes
			// this is only for backward compatibility

			byte mov_dir[4] = { MV_LEFT, MV_RIGHT, MV_UP, MV_DOWN };
			byte action = tape.pos_action_[i];
			int k, num_moves = 0;

			for (k = 0; k<4; k++)
			{
				if (action & mov_dir[k])
				{
					tape.pos_action_[i + num_moves] = mov_dir[k];
					if (num_moves > 0)
						tape.pos_delay[i + num_moves] = 0;
					num_moves++;
				}
			}

			if (num_moves > 1)
			{
				num_moves--;
				i += num_moves;
				tape.length += num_moves;
			}
		}
		else if (tape.file_version < FILE_VERSION_2_0)
		{
			// convert pre-2.0 tapes to new tape format

			if (tape.pos_delay[i] > 1)
			{
				// action part
				tape.pos_action_[i + 1] = tape.pos_action_[i];
				tape.pos_delay[i + 1] = 1;

				// delay part
				tape.pos_action_[i] = MV_NO_MOVING;
				tape.pos_delay[i]--;

				i++;
				tape.length++;
			}
		}

		if (LikeFile.cur_pos>=LikeFile.size)
			break;
	}

	if (i != tape.length)
		chunk_size = (tape.num_participating_players + 1) * i;

	return chunk_size;
}

INIT_SECTION void LoadTapeFromLevNumber(int lev_nr)
{
	char cookie[MAX_LINE_LEN];
	char chunk_name[CHUNK_ID_LEN + 1];

	DmOpenRef dbRef;
	LocalID TapeID;
	MemHandle mhRecord;
	char tape_name[32];

	long chunk_size;

	// always start with reliable default values
	setTapeInfoToDefaults();

	StrCopy(tape_name, LevelSets->name[(int)LevelSets->cur_levset]);
	StrCat(tape_name, "_t"); 
	
	TapeID = DmFindDatabase(LevelSets->card[(int)LevelSets->cur_levset],
								tape_name);
	
	if(	TapeID == 0)
		dbRef = (DmOpenRef)0;
	else
		dbRef = DmOpenDatabase(LevelSets->card[(int)LevelSets->cur_levset],
					TapeID, dmModeReadOnly); 

	if(dbRef == 0)
	{
		tape.no_valid_file = TRUE;
		return;
	}

	if(DmNumRecords (dbRef)!=1)
	{
		DmCloseDatabase(dbRef);
		tape.no_valid_file = TRUE;
		return;
	}
	
	mhRecord = DmGetRecord (dbRef, 0 );

	InitLikeFile ( MemHandleLock (mhRecord ) );

	if(ReadLikeFile(chunk_name,4)!=4)
		goto bye_bye2;

	if (strcmp(chunk_name, "RND1") == 0)
	{
		if(!SeekLikeFileCurPos(4))
			goto bye_bye2;
		if(ReadLikeFile(chunk_name,4)!=4)
			goto bye_bye2;
		
		if (strcmp(chunk_name, "TAPE") != 0)
		{
			tape.no_valid_file = TRUE;
			goto bye_bye2;
		}
	}
	else	// check for pre-2.0 file format with cookie string
	{
		int sle;
		
		if(!SeekLikeFileCurPos(-4))
			goto bye_bye2;

		ReadLikeFile(cookie,MAX_LINE_LEN);
		
		sle=StrLen(cookie);

		if (sle > 0 && cookie[sle - 1] == '\n')
			cookie[sle - 1] = '\0';

		if (!checkCookieString(cookie, TAPE_COOKIE_TMPL))
		{
			tape.no_valid_file = TRUE;
			goto bye_bye2;
		}

		if ((tape.file_version = getFileVersionFromCookieString(cookie)) == -1)
		{
			tape.no_valid_file = TRUE;
			goto bye_bye2;
		}

		// pre-2.0 tape files have no game version, so use file version here
		tape.game_version = tape.file_version;
	}

	if (tape.file_version < FILE_VERSION_1_2)
	{
		// tape files from versions before 1.2.0 without chunk structure
		LoadTape_HEAD(TAPE_HEADER_SIZE);
		LoadTape_BODY(2 * tape.length);
	}
	else
	{
		const struct
		{
			char *name;
			int size;
			int (*loader)(int);
		}
		chunk_info[] =
		{
			{ "VERS", FILE_VERS_CHUNK_SIZE,	LoadTape_VERS },
			{ "HEAD", TAPE_HEADER_SIZE,	LoadTape_HEAD },
			{ "INFO", -1,			LoadTape_INFO },
			{ "BODY", -1,			LoadTape_BODY },
			{  NULL,  0,			NULL }
		};

		while (getLikeFileChunkBE(chunk_name, &chunk_size))
		{
			int i = 0;

			while (chunk_info[i].name != NULL &&
								strcmp(chunk_name, chunk_info[i].name) != 0)
				i++;

			if (chunk_info[i].name == NULL)
			{
//				Error(ERR_WARN, "unknown chunk '%s' in tape file '%s'",
//									chunk_name, filename);
				ReadUnusedBytesFromLikeFile(chunk_size);
			}
			else if (chunk_info[i].size != -1 &&
									chunk_info[i].size != chunk_size)
			{
//				Error(ERR_WARN, "wrong size (%d) of chunk '%s' in tape file '%s'",
//									chunk_size, chunk_name, filename);
				ReadUnusedBytesFromLikeFile(chunk_size);
			}
			else
			{
				// call function to load this tape chunk
				int chunk_size_expected =
								(chunk_info[i].loader)(chunk_size);

				// the size of some chunks cannot be checked before reading other
				// chunks first (like "HEAD" and "BODY") that contain some header
				// information, so check them here
				if (chunk_size_expected != chunk_size)
				{
//					Error(ERR_WARN, "wrong size (%d) of chunk '%s' in tape file '%s'",
//											chunk_size, chunk_name, filename);
				}
			}
		}
	}
//	tape.length_seconds = GetTapeLength();
bye_bye2:
	CloseLikeFile();
	DmReleaseRecord(dbRef, 0, false);
	DmCloseDatabase(dbRef);
}

void LoadTape(int nr)
{
	LoadTapeFromLevNumber(nr);
}
/*
static void SaveTape_VERS(FILE *file, struct TapeInfo *tape)
{
  putFileVersion(file, tape->file_version);
  putFileVersion(file, tape->game_version);
}

static void SaveTape_HEAD(FILE *file, struct TapeInfo *tape)
{
  int i;
  byte store_participating_players = 0;

  // set bits for participating players for compact storage
  for (i = 0; i < MAX_PLAYERS; i++)
    if (tape->player_participates[i])
      store_participating_players |= (1 << i);

  putFile32BitBE(file, tape->random_seed);
  putFile32BitBE(file, tape->date);
  putFile32BitBE(file, tape->length);

  putFile8Bit(file, store_participating_players);

  // unused bytes not at the end here for 4-byte alignment of engine_version
  WriteUnusedBytesToFile(file, TAPE_HEADER_UNUSED);

  putFileVersion(file, tape->engine_version);
}

static void SaveTape_INFO(FILE *file, struct TapeInfo *tape)
{
  int level_identifier_size = strlen(tape->level_identifier) + 1;
  int i;

  putFile16BitBE(file, level_identifier_size);

  for (i = 0; i < level_identifier_size; i++)
    putFile8Bit(file, tape->level_identifier[i]);

  putFile16BitBE(file, tape->level_nr);
}

static void SaveTape_BODY(FILE *file, struct TapeInfo *tape)
{
  int i, j;

  for (i = 0; i < tape->length; i++)
  {
    for (j = 0; j < MAX_PLAYERS; j++)
      if (tape->player_participates[j])
	putFile8Bit(file, tape->pos[i].action[j]);

    putFile8Bit(file, tape->pos[i].delay);
  }
}
*/
void SaveTape(int nr)
{/*
  char *filename = getTapeFilename(nr);
  FILE *file;*/
	boolean new_tape = TRUE;
	int num_participating_players = 0;
	int info_chunk_size;
	int body_chunk_size;
	int i;

//	InitTapeDirectory(leveldir_current->subdir);
/*
	// if a tape still exists, ask to overwrite it
	if (access(filename, F_OK) == 0)
	{
		new_tape = FALSE;
		if (!Request("Replace old tape ?", REQ_ASK))
			return;
	}

	if (!(file = fopen(filename, MODE_WRITE)))
	{
		Error(ERR_WARN, "cannot save level recording file '%s'", filename);
		return;
	}

	tape.file_version = FILE_VERSION_ACTUAL;
	tape.game_version = GAME_VERSION_ACTUAL;

	// count number of participating players
	if (tape.player_participates_)
		num_participating_players++;

	info_chunk_size = 2 + (strlen(tape.level_identifier) + 1) + 2;
	body_chunk_size = (num_participating_players + 1) * tape.length;

	putFileChunkBE(file, "RND1", CHUNK_SIZE_UNDEFINED);
	putFileChunkBE(file, "TAPE", CHUNK_SIZE_NONE);

	putFileChunkBE(file, "VERS", FILE_VERS_CHUNK_SIZE);
	SaveTape_VERS(file, &tape);

	putFileChunkBE(file, "HEAD", TAPE_HEADER_SIZE);
	SaveTape_HEAD(file, &tape);

	putFileChunkBE(file, "INFO", info_chunk_size);
	SaveTape_INFO(file, &tape);

	putFileChunkBE(file, "BODY", body_chunk_size);
	SaveTape_BODY(file, &tape);

	fclose(file);

	SetFilePermissions(filename, PERMS_PRIVATE);

	tape.changed = FALSE;

	if (new_tape)
		Request("tape saved !", REQ_CONFIRM);*/
}

INIT_SECTION static void setLevelInfoToDefaults(void)
//was setLevelInfoToDefaults(struct LevelInfo *level)
{
//	static boolean clipboard_elements_initialized = FALSE;

	int i, /*j,*/ x, y;

	level.file_version = FILE_VERSION_ACTUAL;
	level.game_version = GAME_VERSION_ACTUAL;

	level.encoding_16bit_field  = FALSE;	// default: only 8-bit elements
	level.encoding_16bit_yamyam = FALSE;	// default: only 8-bit elements
	level.encoding_16bit_amoeba = FALSE;	// default: only 8-bit elements

	level.fieldx = STD_LEV_FIELDX;
	level.fieldy = STD_LEV_FIELDY;

	for (x = 0; x < MAX_LEV_FIELDX; x++)
	{
		short int xy=A2TO1(x,0);
		for (y = 0; y < MAX_LEV_FIELDY; y++,xy+=A2TO1(0,1))
			Feld[xy] = EL_SAND; // was level->field
	}

	level.time = 100;
	level.gems_needed = 0;

	level.amoeba_speed = 10;

	level.time_magic_wall = 10;
	level.time_wheel = 10;
	level.time_light = 10;
	level.time_timegate = 10;

	level.amoeba_content = EL_DIAMOND;

	level.double_speed = FALSE;
	level.initial_gravity = FALSE;
	level.em_slippery_gems = FALSE;

	level.block_last_field = FALSE;
	level.sp_block_last_field = TRUE;
	level.instant_relocation = FALSE;
	level.can_pass_to_walkable = FALSE;
	level.grow_into_diggable = TRUE;

	level.can_move_into_acid_bits = ~0;	// everything can move into acid
	level.dont_collide_with_bits = ~0;	// always deadly when colliding

	level.use_spring_bug = FALSE;
	level.use_step_counter = FALSE;

	level.use_custom_template = FALSE;

	for (i = 0; i < MAX_LEVEL_NAME_LEN; i++)
		level.name[i] = '\0';
/*	for (i = 0; i < MAX_LEVEL_AUTHOR_LEN; i++)
		level.author[i] = '\0';
*/
	StrCopy(level.name, NAMELESS_LEVEL_NAME); // for palm only, was strcpy
/*	StrCopy(level.author, ANONYMOUS_NAME); // for palm only, was strcpy

	for (i = 0; i < 4; i++)
	{
		level.envelope_text[i][0] = '\0';
		level.envelope_xsize[i] = MAX_ENVELOPE_XSIZE;
		level.envelope_ysize[i] = MAX_ENVELOPE_YSIZE;
	}
*/
	for (i = 0; i < LEVEL_SCORE_ELEMENTS; i++)
		level.score[i] = 10;

	level.num_yamyam_contents = STD_ELEMENT_CONTENTS;
	for (i = 0; i < MAX_ELEMENT_CONTENTS; i++)
		for (x = 0; x < 3; x++)
			for (y = 0; y < 3; y++)
				level.yamyam_content[i][x][y] =
								(i < STD_ELEMENT_CONTENTS ? EL_ROCK : EL_EMPTY);

	Feld[A2TO1(0,0)] = EL_PLAYER_1; //was level->field
	Feld[A2TO1(STD_LEV_FIELDX - 1,STD_LEV_FIELDY - 1)] = EL_EXIT_CLOSED; //was level->field
/*
  for (i = 0; i < MAX_NUM_ELEMENTS; i++)
  {
    int element = i;

    // never initialize clipboard elements after the very first time
    if (IS_CLIPBOARD_ELEMENT(element) && clipboard_elements_initialized)
      continue;

    setElementChangePages(&element_info[element], 1);
    setElementChangeInfoToDefaults(element_info[element].change);

    if (IS_CUSTOM_ELEMENT(element) ||
	IS_GROUP_ELEMENT(element) ||
	IS_INTERNAL_ELEMENT(element))
    {
      for (j = 0; j < MAX_ELEMENT_NAME_LEN + 1; j++)
	element_info[element].description[j] = '\0';

      if (element_info[element].custom_description != NULL)
	strncpy(element_info[element].description,
		element_info[element].custom_description,MAX_ELEMENT_NAME_LEN);
      else
	strcpy(element_info[element].description,
	       element_info[element].editor_description);

      element_info[element].use_gfx_element = FALSE;
      element_info[element].gfx_element = EL_EMPTY_SPACE;

      element_info[element].modified_settings = FALSE;
    }

    if (IS_CUSTOM_ELEMENT(element) ||
	IS_INTERNAL_ELEMENT(element))
    {
      element_info[element].access_direction = MV_ALL_DIRECTIONS;

      element_info[element].collect_score = 10;		// special default
      element_info[element].collect_count = 1;		// special default

      element_info[element].push_delay_fixed = -1;	// initialize later
      element_info[element].push_delay_random = -1;	// initialize later
      element_info[element].drop_delay_fixed = 0;
      element_info[element].drop_delay_random = 0;
      element_info[element].move_delay_fixed = 0;
      element_info[element].move_delay_random = 0;

      element_info[element].move_pattern = MV_ALL_DIRECTIONS;
      element_info[element].move_direction_initial = MV_START_AUTOMATIC;
      element_info[element].move_stepsize = TILEX / 8;

      element_info[element].move_enter_element = EL_EMPTY_SPACE;
      element_info[element].move_leave_element = EL_EMPTY_SPACE;
      element_info[element].move_leave_type = LEAVE_TYPE_UNLIMITED;

      element_info[element].slippery_type = SLIPPERY_ANY_RANDOM;

      element_info[element].explosion_type = EXPLODES_3X3;
      element_info[element].explosion_delay = 16;
      element_info[element].ignition_delay = 8;

      for (x = 0; x < 3; x++)
	for (y = 0; y < 3; y++)
	  element_info[element].content[x][y] = EL_EMPTY_SPACE;

      element_info[element].access_type = 0;
      element_info[element].access_layer = 0;
      element_info[element].access_protected = 0;
      element_info[element].walk_to_action = 0;
      element_info[element].smash_targets = 0;
      element_info[element].deadliness = 0;


      element_info[element].can_explode_by_fire = FALSE;
      element_info[element].can_explode_smashed = FALSE;
      element_info[element].can_explode_impact = FALSE;

      element_info[element].current_change_page = 0;

      // start with no properties at all
      for (j = 0; j < NUM_EP_BITFIELDS; j++)
	Properties[element][j] = EP_BITMASK_DEFAULT;

      // now set default properties
      SET_PROPERTY(element, EP_CAN_MOVE_INTO_ACID, TRUE);
    }

    if (IS_GROUP_ELEMENT(element) ||
	IS_INTERNAL_ELEMENT(element))
    {
      // initialize memory for list of elements in group
      if (element_info[element].group == NULL)
	element_info[element].group =
	  checked_malloc(sizeof(struct ElementGroupInfo));

      for (j = 0; j < MAX_ELEMENTS_IN_GROUP; j++)
	element_info[element].group->element[j] = EL_EMPTY_SPACE;

      // default: only one element in group
      element_info[element].group->num_elements = 1;

      element_info[element].group->choice_mode = ANIM_RANDOM;
    }
  }

  clipboard_elements_initialized = TRUE;

*/
	BorderElement = EL_STEELWALL;

	level.no_valid_file = FALSE;
/*
	if (leveldir_current == NULL)		// only when dumping level
		return;

	// try to determine better author name than 'anonymous'
	if (strcmp(leveldir_current.author, ANONYMOUS_NAME) != 0)
	{
		strncpy(level.author, leveldir_current.author, MAX_LEVEL_AUTHOR_LEN);
		level.author[MAX_LEVEL_AUTHOR_LEN] = '\0';
	}
	else
	{
		switch (LEVELCLASS(leveldir_current))
		{
			case LEVELCLASS_TUTORIAL:
				strcpy(level.author, PROGRAM_AUTHOR_STRING);
				break;

			case LEVELCLASS_CONTRIB:
				strncpy(level.author, leveldir_current->name, MAX_LEVEL_AUTHOR_LEN);
				level.author[MAX_LEVEL_AUTHOR_LEN] = '\0';
				break;

			case LEVELCLASS_PRIVATE:
				strncpy(level.author, getRealName(), MAX_LEVEL_AUTHOR_LEN);
				level.author[MAX_LEVEL_AUTHOR_LEN] = '\0';
				break;

			default:
				// keep default value
				break;
		}
	}*/
};

INIT_SECTION static short int getMappedElement(short int element)//was int-int
{
	// map some (historic, now obsolete) elements

	switch (element)
	{
		case EL_PLAYER_OBSOLETE:
			element = EL_PLAYER_1;
			break;

		case EL_KEY_OBSOLETE:
			element = EL_KEY_1;

		case EL_EM_KEY_1_FILE_OBSOLETE:
			element = EL_EM_KEY_1;
			break;

		case EL_EM_KEY_2_FILE_OBSOLETE:
			element = EL_EM_KEY_2;
			break;

		case EL_EM_KEY_3_FILE_OBSOLETE:
			element = EL_EM_KEY_3;
			break;

		case EL_EM_KEY_4_FILE_OBSOLETE:
			element = EL_EM_KEY_4;
			break;

		case EL_ENVELOPE_OBSOLETE:
			element = EL_ENVELOPE_1;
			break;

		case EL_SP_EMPTY:
			element = EL_EMPTY;
			break;

		default:
			if (element >= NUM_FILE_ELEMENTS)
			{
//				Error(ERR_WARN, "invalid level element %d", element);

				element = EL_UNKNOWN;
			}
			break;
	}
	
	return element;
};

INIT_SECTION int getMappedElementByVersion(int element, long game_version)
{
	// remap some elements due to certain game version

	if (game_version <= VERSION_IDENT(2,2,0,0))
	{
		// map game font elements
		element = (element == EL_CHAR('[')  ? EL_CHAR_AUMLAUT :
					element == EL_CHAR('\\') ? EL_CHAR_OUMLAUT :
					element == EL_CHAR(']')  ? EL_CHAR_UUMLAUT :
					element == EL_CHAR('^')  ? EL_CHAR_COPYRIGHT : element);
	}

	if (game_version < VERSION_IDENT(3,0,0,0))
	{
		// map Supaplex gravity tube elements
		element = (element == EL_SP_GRAVITY_PORT_LEFT  ? EL_SP_PORT_LEFT  :
					element == EL_SP_GRAVITY_PORT_RIGHT ? EL_SP_PORT_RIGHT :
					element == EL_SP_GRAVITY_PORT_UP    ? EL_SP_PORT_UP    :
					element == EL_SP_GRAVITY_PORT_DOWN  ? EL_SP_PORT_DOWN  :
					element);
	}

	return element;
};

INIT_SECTION static int LoadLevel_VERS(int chunk_size)
//was static int LoadLevel_VERS(FILE *file, int chunk_size, struct LevelInfo *level)
{
	level.file_version = getLikeFileVersion();
	level.game_version = getLikeFileVersion();

	return chunk_size;
};

INIT_SECTION static int LoadLevel_HEAD(int chunk_size)
//was static int LoadLevel_HEAD(FILE *file, int chunk_size, struct LevelInfo *level)
{
	int i, x, y;

	level.fieldx = getLikeFile8Bit();
	level.fieldy = getLikeFile8Bit();

//----------------------------------------start
	if(level.fieldx>MAX_LEV_FIELDX)
		{bugd("ERROR! level.fieldx is",level.fieldx);level.fieldx=MAX_LEV_FIELDX;}
	if(level.fieldy>MAX_LEV_FIELDY)
		{bugd("ERROR! level.fieldy is",level.fieldy);level.fieldy=MAX_LEV_FIELDY;}
//----------------------------------------end
	
	level.time		= getLikeFile16BitBE();
	level.gems_needed	= getLikeFile16BitBE();

	for (i = 0; i < MAX_LEVEL_NAME_LEN; i++)
		level.name[i] = getLikeFile8Bit();
	level.name[MAX_LEVEL_NAME_LEN] = 0;

	ShowLevelName();
	
	for (i = 0; i < LEVEL_SCORE_ELEMENTS; i++)
		level.score[i] = getLikeFile8Bit();

	level.num_yamyam_contents = STD_ELEMENT_CONTENTS;
	for (i = 0; i < STD_ELEMENT_CONTENTS; i++)
		for (y = 0; y < 3; y++)
			for (x = 0; x < 3; x++)
				level.yamyam_content[i][x][y] = getMappedElement(getLikeFile8Bit());

	level.amoeba_speed		= getLikeFile8Bit();
	level.time_magic_wall	= getLikeFile8Bit();
	level.time_wheel		= getLikeFile8Bit();
	level.amoeba_content	= getMappedElement(getLikeFile8Bit());
	level.double_speed		= (getLikeFile8Bit() == 1 ? TRUE : FALSE);
	level.initial_gravity	= (getLikeFile8Bit() == 1 ? TRUE : FALSE);
	level.encoding_16bit_field	= (getLikeFile8Bit() == 1 ? TRUE : FALSE);
	level.em_slippery_gems	= (getLikeFile8Bit() == 1 ? TRUE : FALSE);

	level.use_custom_template	= (getLikeFile8Bit() == 1 ? TRUE : FALSE);

	level.block_last_field	= (getLikeFile8Bit() == 1 ? TRUE : FALSE);
	level.sp_block_last_field	= (getLikeFile8Bit() == 1 ? TRUE : FALSE);
	level.can_move_into_acid_bits = getLikeFile32BitBE();
	level.dont_collide_with_bits = getLikeFile8Bit();

	level.use_spring_bug		= (getLikeFile8Bit() == 1 ? TRUE : FALSE);
	level.use_step_counter	= (getLikeFile8Bit() == 1 ? TRUE : FALSE);

	level.instant_relocation	= (getLikeFile8Bit() == 1 ? TRUE : FALSE);
	level.can_pass_to_walkable	= (getLikeFile8Bit() == 1 ? TRUE : FALSE);
	level.grow_into_diggable	= (getLikeFile8Bit() == 1 ? TRUE : FALSE);

	ReadUnusedBytesFromLikeFile(LEVEL_HEADER_UNUSED);

	return chunk_size;
};

INIT_SECTION static int LoadLevel_BODY(int chunk_size)
//was static int LoadLevel_BODY(FILE *file, int chunk_size, struct LevelInfo *level)
{
	int x, y;
	int chunk_size_expected = level.fieldx * level.fieldy;

	// Note: "chunk_size" was wrong before version 2.0 when elements are
	// stored with 16-bit encoding (and should be twice as big then).
	// Even worse, playfield data was stored 16-bit when only yamyam content
	// contained 16-bit elements and vice versa. 

	if (level.encoding_16bit_field && level.file_version >= FILE_VERSION_2_0)
		chunk_size_expected *= 2;

	if (chunk_size_expected != chunk_size)
	{
		ReadUnusedBytesFromLikeFile(chunk_size);
		return chunk_size_expected;
	}

	for (y = 0; y < level.fieldy; y++)
	{
		register short int xy=A2TO1(0,y);
		for (x = 0; x < level.fieldx; x++,xy+=A2TO1(1,0))
			Feld[xy] = getMappedElement(level.encoding_16bit_field ? 
									  getLikeFile16BitBE() : getLikeFile8Bit());
	}

	return chunk_size;
};

INIT_SECTION static int LoadLevel_CONT(int chunk_size)
//was static int LoadLevel_CONT(FILE *file, int chunk_size, struct LevelInfo *level)
{
	int i, x, y;
	int header_size = 4;
	int content_size = MAX_ELEMENT_CONTENTS * 3 * 3;
	int chunk_size_expected = header_size + content_size;

	// Note: "chunk_size" was wrong before version 2.0 when elements are
	// stored with 16-bit encoding (and should be twice as big then).
	// Even worse, playfield data was stored 16-bit when only yamyam content
	// contained 16-bit elements and vice versa.

	if (level.encoding_16bit_field && level.file_version >= FILE_VERSION_2_0)
		chunk_size_expected += content_size;

	if (chunk_size_expected != chunk_size)
	{
		ReadUnusedBytesFromLikeFile(chunk_size);
		return chunk_size_expected;
	}

	SeekLikeFileCurPos(1);
	level.num_yamyam_contents = getLikeFile8Bit();
	SeekLikeFileCurPos(2);

	// correct invalid number of content fields -- should never happen
	if (level.num_yamyam_contents < 1 ||
					level.num_yamyam_contents > MAX_ELEMENT_CONTENTS)
		level.num_yamyam_contents = STD_ELEMENT_CONTENTS;

	for (i = 0; i < MAX_ELEMENT_CONTENTS; i++)
		for (y = 0; y < 3; y++)
			for (x = 0; x < 3; x++)
				level.yamyam_content[i][x][y] =
							getMappedElement(level.encoding_16bit_field ?
								getLikeFile16BitBE() : getLikeFile8Bit());
	return chunk_size;
};

INIT_SECTION static int LoadLevel_CNT2(int chunk_size)
//was static int LoadLevel_CNT2(FILE *file, int chunk_size, struct LevelInfo *level)
{
	int i, x, y;
	int element;
	int num_contents, content_xsize, content_ysize;
	int content_array[MAX_ELEMENT_CONTENTS][3][3];

	element = getMappedElement(getLikeFile16BitBE());
	num_contents = getLikeFile8Bit();
	content_xsize = getLikeFile8Bit();
	content_ysize = getLikeFile8Bit();

	ReadUnusedBytesFromLikeFile(LEVEL_CHUNK_CNT2_UNUSED);

	for (i = 0; i < MAX_ELEMENT_CONTENTS; i++)
		for (y = 0; y < 3; y++)
			for (x = 0; x < 3; x++)
				content_array[i][x][y] = getMappedElement(getLikeFile16BitBE());

	// correct invalid number of content fields -- should never happen
	if (num_contents < 1 || num_contents > MAX_ELEMENT_CONTENTS)
		num_contents = STD_ELEMENT_CONTENTS;

	if (element == EL_YAMYAM)
	{
		level.num_yamyam_contents = num_contents;

		for (i = 0; i < num_contents; i++)
			for (y = 0; y < 3; y++)
				for (x = 0; x < 3; x++)
					level.yamyam_content[i][x][y] = content_array[i][x][y];
	}
	else if (element == EL_BD_AMOEBA)
	{
		level.amoeba_content = content_array[0][0][0];
	}
	else
	{
//		Error(ERR_WARN, "cannot load content for element '%d'", element);
	}

	return chunk_size;
};

INIT_SECTION static int LoadLevel_CNT3(int chunk_size)
//was static int LoadLevel_CNT3(FILE *file, int chunk_size, struct LevelInfo *level)
{
//	int i;
	int element;
	int envelope_nr;
	int envelope_len;
	int chunk_size_expected;

	element = getMappedElement(getLikeFile16BitBE());
	if (!IS_ENVELOPE(element))
		element = EL_ENVELOPE_1;

	envelope_nr = element - EL_ENVELOPE_1;

	envelope_len = getLikeFile16BitBE();

	level.envelope_xsize[envelope_nr] = getLikeFile8Bit();
	level.envelope_ysize[envelope_nr] = getLikeFile8Bit();

	ReadUnusedBytesFromLikeFile(LEVEL_CHUNK_CNT3_UNUSED);

	chunk_size_expected = LEVEL_CHUNK_CNT3_SIZE(envelope_len);
	if (chunk_size_expected != chunk_size)
	{
		ReadUnusedBytesFromLikeFile(chunk_size - LEVEL_CHUNK_CNT3_HEADER);
		return chunk_size_expected;
	}

//	for (i = 0; i < envelope_len; i++)
//		level->envelope_text[envelope_nr][i] = getFile8Bit(file);
	ReadUnusedBytesFromLikeFile(envelope_len);
	
	return chunk_size;
};

INIT_SECTION static int LoadLevel_CUS1(int chunk_size)
//was static int LoadLevel_CUS1(FILE *file, int chunk_size, struct LevelInfo *level)
{
	bugd("Loading level section CUS1 not supported!",0);
	ReadUnusedBytesFromLikeFile(chunk_size);
	return chunk_size;
};

INIT_SECTION static int LoadLevel_CUS2(int chunk_size)
//was static int LoadLevel_CUS2(FILE *file, int chunk_size, struct LevelInfo *level)
{
	bugd("Loading level section CUS2 not supported",0);
	ReadUnusedBytesFromLikeFile(chunk_size);
	return chunk_size;
};

INIT_SECTION static int LoadLevel_CUS3(int chunk_size)
//was static int LoadLevel_CUS3(FILE *file, int chunk_size, struct LevelInfo *level)
{
	bugd("Loading level section CUS3 not supported!",0);
	ReadUnusedBytesFromLikeFile(chunk_size);
	return chunk_size;
};

INIT_SECTION static int LoadLevel_CUS4(int chunk_size)
//was static int LoadLevel_CUS4(FILE *file, int chunk_size, struct LevelInfo *level)
{
	bugd("Loading level section CUS4 not supported!",0);
	ReadUnusedBytesFromLikeFile(chunk_size);
	return chunk_size;
};

INIT_SECTION void LoadLevelFromFileInfo_RND(int level_number)
{
	char cookie[MAX_LINE_LEN];
	char chunk_name[CHUNK_ID_LEN+1];
	long chunk_size;

	DmOpenRef dbRef;
	UInt16	  NumRecs;
	MemHandle mhRecord;

	dbRef = (DmOpenRef)0;
 
	if(LevelSets->cur_levset<0||LevelSets->cur_levset>MAXLEVSETS-1)
	{
		level.no_valid_file = TRUE;
		bugd("RND ERROR! cur_levset",0);
		return;
	}

 	{
		int i=LevelSets->cur_levset;
		LocalID	c_dbid = LevelSets->dbid[i];
		UInt16	c_card = LevelSets->card[i];
//		bugd(LevelSets->name[i],c_dbid);
//		bugd("And card",c_card);
		
		if(c_dbid==0)
		{
			level.no_valid_file = TRUE;
			bugd("RND ERROR! dbid",0);
			return;
		}
		dbRef = DmOpenDatabase(c_card, c_dbid, dmModeReadOnly);
	} 

	if(dbRef == 0)
	{
		level.no_valid_file = TRUE;
		bugd("RND ERROR! openDB",0);
		return;
	}

	NumRecs = DmNumRecords (dbRef);
	if(NumRecs<=level_number+1)
	{
		DmCloseDatabase(dbRef);
		level.no_valid_file = TRUE;
		return;
	}
	
	mhRecord = DmGetRecord (dbRef, 1+level_number );
	//skip 1st record as future levelseries descr
	InitLikeFile ( MemHandleLock (mhRecord ) );

	if(ReadLikeFile(chunk_name,4)!=4)
		goto bye_bye;

	if (strcmp(chunk_name, "RND1") == 0)
	{
		if(!SeekLikeFileCurPos(4))
			goto bye_bye;
		if(ReadLikeFile(chunk_name,4)!=4)
			goto bye_bye;
		
		if (strcmp(chunk_name, "CAVE") != 0)
		{
			level.no_valid_file = TRUE;
			goto bye_bye;
		}
	}
	else	// check for pre-2.0 file format with cookie string
	{
		int sle;
		
		if(!SeekLikeFileCurPos(-4))
			goto bye_bye;

		ReadLikeFile(cookie,MAX_LINE_LEN);
		
		sle=StrLen(cookie);

		if (sle > 0 && cookie[sle - 1] == '\n')
			cookie[sle - 1] = '\0';

		if (!checkCookieString(cookie, LEVEL_COOKIE_TMPL))
		{
			level.no_valid_file = TRUE;
			goto bye_bye;
		}

		if ((level.file_version = getFileVersionFromCookieString(cookie)) == -1)
		{
			level.no_valid_file = TRUE;
			goto bye_bye;
		}

		// pre-2.0 level files have no game version, so use file version here
		level.game_version = level.file_version;
	}

	if (level.file_version < FILE_VERSION_1_2)
	{
		// level files from versions before 1.2.0 without chunk structure
		LoadLevel_HEAD( LEVEL_HEADER_SIZE);
		LoadLevel_BODY( level.fieldx * level.fieldy);
	}
	else
	{
		static struct
		{
			char *name;
			int size;
			int (*loader)( int );
		}
		chunk_info[] =
		{
			{ "VERS", FILE_VERS_CHUNK_SIZE,	LoadLevel_VERS },
			{ "HEAD", LEVEL_HEADER_SIZE,	LoadLevel_HEAD },
//			{ "AUTH", MAX_LEVEL_AUTHOR_LEN,	LoadLevel_AUTH },
			{ "BODY", -1,			LoadLevel_BODY },
			{ "CONT", -1,			LoadLevel_CONT },
			{ "CNT2", LEVEL_CHUNK_CNT2_SIZE,	LoadLevel_CNT2 },
			{ "CNT3", -1,			LoadLevel_CNT3 },
			{ "CUS1", -1,			LoadLevel_CUS1 },
			{ "CUS2", -1,			LoadLevel_CUS2 },
			{ "CUS3", -1,			LoadLevel_CUS3 },
			{ "CUS4", -1,			LoadLevel_CUS4 },
			{  NULL,  0,			NULL }
		};

		while (getLikeFileChunkBE(chunk_name, &chunk_size))
		{
			int i = 0;

			while (chunk_info[i].name != NULL &&
							strcmp(chunk_name, chunk_info[i].name) != 0)
				i++;

			if (chunk_info[i].name == NULL)
			{
//				Error(ERR_WARN, "unknown chunk '%s' in level file '%s'",
//												  chunk_name, filename);
				ReadUnusedBytesFromLikeFile(chunk_size);
			}
			else if (chunk_info[i].size != -1 &&
									chunk_info[i].size != chunk_size)
			{
//				Error(ERR_WARN, "wrong size (%d) of chunk '%s' in level file '%s'",
//												 chunk_size, chunk_name, filename);
				ReadUnusedBytesFromLikeFile(chunk_size);
			}
			else
			{
				// call function to load this level chunk
				int chunk_size_expected =
								(chunk_info[i].loader)( chunk_size );

				// the size of some chunks cannot be checked before reading other
				// chunks first (like "HEAD" and "BODY") that contain some header
				// information, so check them here
				if (chunk_size_expected != chunk_size)
				{
//					Error(ERR_WARN, "wrong size (%d) of chunk '%s' in level file '%s'",
//													 chunk_size, chunk_name, filename);
				}
			}
		}
	}

bye_bye:
	CloseLikeFile();
	DmReleaseRecord(dbRef, 1+level_number, false);
	DmCloseDatabase(dbRef);

};

// -------------------------------------------------------------------------
// functions for loading EM level                                           
// -------------------------------------------------------------------------

INIT_SECTION static short int map_em_element_yam(byte element)
{
  switch (element)
  {
    case 0x00:	return EL_EMPTY;
    case 0x01:	return EL_EMERALD;
    case 0x02:	return EL_DIAMOND;
    case 0x03:	return EL_ROCK;
    case 0x04:	return EL_ROBOT;
    case 0x05:	return EL_SPACESHIP_UP;
    case 0x06:	return EL_BOMB;
    case 0x07:	return EL_BUG_UP;
    case 0x08:	return EL_AMOEBA_DROP;
    case 0x09:	return EL_NUT;
    case 0x0a:	return EL_YAMYAM;
    case 0x0b:	return EL_QUICKSAND_FULL;
    case 0x0c:	return EL_SAND;
    case 0x0d:	return EL_WALL_SLIPPERY;
    case 0x0e:	return EL_STEELWALL;
    case 0x0f:	return EL_WALL;
    case 0x10:	return EL_EM_KEY_1;
    case 0x11:	return EL_EM_KEY_2;
    case 0x12:	return EL_EM_KEY_4;
    case 0x13:	return EL_EM_KEY_3;
    case 0x14:	return EL_MAGIC_WALL;
    case 0x15:	return EL_ROBOT_WHEEL;
    case 0x16:	return EL_DYNAMITE;

    case 0x17:	return EL_EM_KEY_1;			/* EMC */
    case 0x18:	return EL_BUG_UP;			/* EMC */
    case 0x1a:	return EL_DIAMOND;			/* EMC */
    case 0x1b:	return EL_EMERALD;			/* EMC */
    case 0x25:	return EL_NUT;				/* EMC */
    case 0x80:	return EL_EMPTY;			/* EMC */
    case 0x85:	return EL_EM_KEY_1;			/* EMC */
    case 0x86:	return EL_EM_KEY_2;			/* EMC */
    case 0x87:	return EL_EM_KEY_4;			/* EMC */
    case 0x88:	return EL_EM_KEY_3;			/* EMC */
    case 0x94:	return EL_QUICKSAND_EMPTY;		/* EMC */
    case 0x9a:	return EL_AMOEBA_WET;			/* EMC */
    case 0xaf:	return EL_DYNAMITE;			/* EMC */
    case 0xbd:	return EL_SAND;				/* EMC */

    default:
	  bugd("invalid em_element_yam", element);
      //Error(ERR_WARN, "invalid level element %d", element);
      return EL_UNKNOWN;
  }
}

INIT_SECTION static short int map_em_element_field(byte element)
{
  if (element >= 0xc8 && element <= 0xe1)
    return EL_CHAR_A + (element - 0xc8);
  else if (element >= 0xe2 && element <= 0xeb)
    return EL_CHAR_0 + (element - 0xe2);

  switch (element)
  {
    case 0x00:	return EL_ROCK;
    case 0x01:	return EL_ROCK;				/* EMC */
    case 0x02:	return EL_DIAMOND;
    case 0x03:	return EL_DIAMOND;
    case 0x04:	return EL_ROBOT;
    case 0x05:	return EL_ROBOT;			/* EMC */
    case 0x06:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x07:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x08:	return EL_SPACESHIP_UP;
    case 0x09:	return EL_SPACESHIP_RIGHT;
    case 0x0a:	return EL_SPACESHIP_DOWN;
    case 0x0b:	return EL_SPACESHIP_LEFT;
    case 0x0c:	return EL_SPACESHIP_UP;
    case 0x0d:	return EL_SPACESHIP_RIGHT;
    case 0x0e:	return EL_SPACESHIP_DOWN;
    case 0x0f:	return EL_SPACESHIP_LEFT;

    case 0x10:	return EL_BOMB;
    case 0x11:	return EL_BOMB;				/* EMC */
    case 0x12:	return EL_EMERALD;
    case 0x13:	return EL_EMERALD;
    case 0x14:	return EL_BUG_UP;
    case 0x15:	return EL_BUG_RIGHT;
    case 0x16:	return EL_BUG_DOWN;
    case 0x17:	return EL_BUG_LEFT;
    case 0x18:	return EL_BUG_UP;
    case 0x19:	return EL_BUG_RIGHT;
    case 0x1a:	return EL_BUG_DOWN;
    case 0x1b:	return EL_BUG_LEFT;
    case 0x1c:	return EL_AMOEBA_DROP;
    case 0x1d:	return EL_AMOEBA_DROP;			/* EMC */
    case 0x1e:	return EL_AMOEBA_DROP;			/* EMC */
    case 0x1f:	return EL_AMOEBA_DROP;			/* EMC */

    case 0x20:	return EL_ROCK;
    case 0x21:	return EL_BOMB;				/* EMC */
    case 0x22:	return EL_DIAMOND;			/* EMC */
    case 0x23:	return EL_EMERALD;			/* EMC */
    case 0x24:	return EL_MAGIC_WALL;
    case 0x25:	return EL_NUT;
    case 0x26:	return EL_NUT;				/* EMC */
    case 0x27:	return EL_NUT;				/* EMC */

      // looks like magic wheel, but is _always_ activated
    case 0x28:	return EL_ROBOT_WHEEL;			/* EMC */

    case 0x29:	return EL_YAMYAM;	/* up */
    case 0x2a:	return EL_YAMYAM;	/* down */
    case 0x2b:	return EL_YAMYAM;	/* left */	/* EMC */
    case 0x2c:	return EL_YAMYAM;	/* right */	/* EMC */
    case 0x2d:	return EL_QUICKSAND_FULL;
    case 0x2e:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x2f:	return EL_EMPTY_SPACE;			/* EMC */

    case 0x30:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x31:	return EL_SAND;				/* EMC */
    case 0x32:	return EL_SAND;				/* EMC */
    case 0x33:	return EL_SAND;				/* EMC */
    case 0x34:	return EL_QUICKSAND_FULL;		/* EMC */
    case 0x35:	return EL_QUICKSAND_FULL;		/* EMC */
    case 0x36:	return EL_QUICKSAND_FULL;		/* EMC */
    case 0x37:	return EL_SAND;				/* EMC */
    case 0x38:	return EL_ROCK;				/* EMC */
    case 0x39:	return EL_EXPANDABLE_WALL_HORIZONTAL;	/* EMC */
    case 0x3a:	return EL_EXPANDABLE_WALL_VERTICAL;	/* EMC */
    case 0x3b:	return EL_DYNAMITE_ACTIVE;	/* 1 */
    case 0x3c:	return EL_DYNAMITE_ACTIVE;	/* 2 */
    case 0x3d:	return EL_DYNAMITE_ACTIVE;	/* 3 */
    case 0x3e:	return EL_DYNAMITE_ACTIVE;	/* 4 */
    case 0x3f:	return EL_ACID_POOL_BOTTOM;

    case 0x40:	return EL_EXIT_OPEN;	/* 1 */
    case 0x41:	return EL_EXIT_OPEN;	/* 2 */
    case 0x42:	return EL_EXIT_OPEN;	/* 3 */
    case 0x43:	return EL_BALLOON;			/* EMC */
    case 0x44:	return EL_UNKNOWN;			/* EMC ("plant") */
    case 0x45:	return EL_SPRING;			/* EMC */
    case 0x46:	return EL_SPRING;	/* falling */	/* EMC */
    case 0x47:	return EL_SPRING;	/* left */	/* EMC */
    case 0x48:	return EL_SPRING;	/* right */	/* EMC */
    case 0x49:	return EL_UNKNOWN;			/* EMC ("ball 1") */
    case 0x4a:	return EL_UNKNOWN;			/* EMC ("ball 2") */
    case 0x4b:	return EL_UNKNOWN;			/* EMC ("android") */
    case 0x4c:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x4d:	return EL_UNKNOWN;			/* EMC ("android") */
    case 0x4e:	return EL_INVISIBLE_WALL;		/* EMC (? "android") */
    case 0x4f:	return EL_UNKNOWN;			/* EMC ("android") */

    case 0x50:	return EL_UNKNOWN;			/* EMC ("android") */
    case 0x51:	return EL_UNKNOWN;			/* EMC ("android") */
    case 0x52:	return EL_UNKNOWN;			/* EMC ("android") */
    case 0x53:	return EL_UNKNOWN;			/* EMC ("android") */
    case 0x54:	return EL_UNKNOWN;			/* EMC ("android") */
    case 0x55:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x56:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x57:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x58:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x59:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x5a:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x5b:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x5c:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x5d:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x5e:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x5f:	return EL_EMPTY_SPACE;			/* EMC */

    case 0x60:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x61:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x62:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x63:	return EL_SPRING;	/* left */	/* EMC */
    case 0x64:	return EL_SPRING;	/* right */	/* EMC */
    case 0x65:	return EL_ACID;		/* 1 */		/* EMC */
    case 0x66:	return EL_ACID;		/* 2 */		/* EMC */
    case 0x67:	return EL_ACID;		/* 3 */		/* EMC */
    case 0x68:	return EL_ACID;		/* 4 */		/* EMC */
    case 0x69:	return EL_ACID;		/* 5 */		/* EMC */
    case 0x6a:	return EL_ACID;		/* 6 */		/* EMC */
    case 0x6b:	return EL_ACID;		/* 7 */		/* EMC */
    case 0x6c:	return EL_ACID;		/* 8 */		/* EMC */
    case 0x6d:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x6e:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x6f:	return EL_EMPTY_SPACE;			/* EMC */

    case 0x70:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x71:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x72:	return EL_NUT;		/* left */	/* EMC */
    case 0x73:	return EL_SAND;				/* EMC (? "nut") */
    case 0x74:	return EL_STEELWALL;
    case 0x75:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x76:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x77:	return EL_BOMB;		/* left */	/* EMC */
    case 0x78:	return EL_BOMB;		/* right */	/* EMC */
    case 0x79:	return EL_ROCK;		/* left */	/* EMC */
    case 0x7a:	return EL_ROCK;		/* right */	/* EMC */
    case 0x7b:	return EL_ACID;				/* (? EMC "blank") */
    case 0x7c:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x7d:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x7e:	return EL_EMPTY_SPACE;			/* EMC */
    case 0x7f:	return EL_EMPTY_SPACE;			/* EMC */

    case 0x80:	return EL_EMPTY;
    case 0x81:	return EL_WALL_SLIPPERY;
    case 0x82:	return EL_SAND;
    case 0x83:	return EL_STEELWALL;
    case 0x84:	return EL_WALL;
    case 0x85:	return EL_EM_KEY_1;
    case 0x86:	return EL_EM_KEY_2;
    case 0x87:	return EL_EM_KEY_4;
    case 0x88:	return EL_EM_KEY_3;
    case 0x89:	return EL_EM_GATE_1;
    case 0x8a:	return EL_EM_GATE_2;
    case 0x8b:	return EL_EM_GATE_4;
    case 0x8c:	return EL_EM_GATE_3;
    case 0x8d:	return EL_INVISIBLE_WALL;		/* EMC (? "dripper") */
    case 0x8e:	return EL_EM_GATE_1_GRAY;
    case 0x8f:	return EL_EM_GATE_2_GRAY;

    case 0x90:	return EL_EM_GATE_4_GRAY;
    case 0x91:	return EL_EM_GATE_3_GRAY;
    case 0x92:	return EL_MAGIC_WALL;
    case 0x93:	return EL_ROBOT_WHEEL;
    case 0x94:	return EL_QUICKSAND_EMPTY;		/* (? EMC "sand") */
    case 0x95:	return EL_ACID_POOL_TOPLEFT;
    case 0x96:	return EL_ACID_POOL_TOPRIGHT;
    case 0x97:	return EL_ACID_POOL_BOTTOMLEFT;
    case 0x98:	return EL_ACID_POOL_BOTTOMRIGHT;
    case 0x99:	return EL_ACID;			/* (? EMC "fake blank") */
    case 0x9a:	return EL_AMOEBA_DEAD;		/* 1 */
    case 0x9b:	return EL_AMOEBA_DEAD;		/* 2 */
    case 0x9c:	return EL_AMOEBA_DEAD;		/* 3 */
    case 0x9d:	return EL_AMOEBA_DEAD;		/* 4 */
    case 0x9e:	return EL_EXIT_CLOSED;
    case 0x9f:	return EL_CHAR_LESS;		/* arrow left */

      // looks like normal sand, but behaves like wall
    case 0xa0:	return EL_UNKNOWN;		/* EMC ("fake grass") */
    case 0xa1:	return EL_UNKNOWN;		/* EMC ("lenses") */
    case 0xa2:	return EL_UNKNOWN;		/* EMC ("magnify") */
    case 0xa3:	return EL_UNKNOWN;		/* EMC ("fake blank") */
    case 0xa4:	return EL_UNKNOWN;		/* EMC ("fake grass") */
    case 0xa5:	return EL_UNKNOWN;		/* EMC ("switch") */
    case 0xa6:	return EL_UNKNOWN;		/* EMC ("switch") */
    case 0xa7:	return EL_EMPTY_SPACE;			/* EMC */
    case 0xa8:	return EL_EMC_WALL_1;			/* EMC ("decor 8") */
    case 0xa9:	return EL_EMC_WALL_2;			/* EMC ("decor 9") */
    case 0xaa:	return EL_EMC_WALL_3;			/* EMC ("decor 10") */
    case 0xab:	return EL_EMC_WALL_7;			/* EMC ("decor 5") */
    case 0xac:	return EL_CHAR_COMMA;			/* EMC */
    case 0xad:	return EL_CHAR_QUOTEDBL;		/* EMC */
    case 0xae:	return EL_CHAR_MINUS;			/* EMC */
    case 0xaf:	return EL_DYNAMITE;

    case 0xb0:	return EL_EMC_STEELWALL_1;		/* EMC ("steel 3") */
    case 0xb1:	return EL_EMC_WALL_8;			/* EMC ("decor 6") */
    case 0xb2:	return EL_UNKNOWN;			/* EMC ("decor 7") */
    case 0xb3:	return EL_STEELWALL;		/* 2 */	/* EMC */
    case 0xb4:	return EL_WALL_SLIPPERY;	/* 2 */	/* EMC */
    case 0xb5:	return EL_EMC_WALL_6;			/* EMC ("decor 2") */
    case 0xb6:	return EL_EMC_WALL_5;			/* EMC ("decor 4") */
    case 0xb7:	return EL_EMC_WALL_4;			/* EMC ("decor 3") */
    case 0xb8:	return EL_BALLOON_SWITCH_ANY;		/* EMC */
    case 0xb9:	return EL_BALLOON_SWITCH_RIGHT;		/* EMC */
    case 0xba:	return EL_BALLOON_SWITCH_DOWN;		/* EMC */
    case 0xbb:	return EL_BALLOON_SWITCH_LEFT;		/* EMC */
    case 0xbc:	return EL_BALLOON_SWITCH_UP;		/* EMC */
    case 0xbd:	return EL_SAND;				/* EMC ("dirt") */
    case 0xbe:	return EL_UNKNOWN;			/* EMC ("plant") */
    case 0xbf:	return EL_UNKNOWN;			/* EMC ("key 5") */

    case 0xc0:	return EL_UNKNOWN;			/* EMC ("key 6") */
    case 0xc1:	return EL_UNKNOWN;			/* EMC ("key 7") */
    case 0xc2:	return EL_UNKNOWN;			/* EMC ("key 8") */
    case 0xc3:	return EL_UNKNOWN;			/* EMC ("door 5") */
    case 0xc4:	return EL_UNKNOWN;			/* EMC ("door 6") */
    case 0xc5:	return EL_UNKNOWN;			/* EMC ("door 7") */
    case 0xc6:	return EL_UNKNOWN;			/* EMC ("door 8") */
    case 0xc7:	return EL_UNKNOWN;			/* EMC ("bumper") */

      // characters: see above

    case 0xec:	return EL_CHAR_PERIOD;
    case 0xed:	return EL_CHAR_EXCLAM;
    case 0xee:	return EL_CHAR_COLON;
    case 0xef:	return EL_CHAR_QUESTION;

    case 0xf0:	return EL_CHAR_GREATER;			/* arrow right */
    case 0xf1:	return EL_CHAR_COPYRIGHT;		/* EMC: "decor 1" */
    case 0xf2:	return EL_UNKNOWN;		/* EMC ("fake door 5") */
    case 0xf3:	return EL_UNKNOWN;		/* EMC ("fake door 6") */
    case 0xf4:	return EL_UNKNOWN;		/* EMC ("fake door 7") */
    case 0xf5:	return EL_UNKNOWN;		/* EMC ("fake door 8") */
    case 0xf6:	return EL_EMPTY_SPACE;			/* EMC */
    case 0xf7:	return EL_EMPTY_SPACE;			/* EMC */

    case 0xf8:	return EL_EMPTY_SPACE;			/* EMC */
    case 0xf9:	return EL_EMPTY_SPACE;			/* EMC */
    case 0xfa:	return EL_EMPTY_SPACE;			/* EMC */
    case 0xfb:	return EL_EMPTY_SPACE;			/* EMC */
    case 0xfc:	return EL_EMPTY_SPACE;			/* EMC */
    case 0xfd:	return EL_EMPTY_SPACE;			/* EMC */

    case 0xfe:	return EL_PLAYER_1;			/* EMC: "blank" */
    case 0xff:	return EL_PLAYER_2;			/* EMC: "blank" */

    default:
      // should never happen (all 8-bit value cases should be handled)
	  bugd("invalid em_element_field",element);
      //Error(ERR_WARN, "invalid level element %d", element);
      return EL_UNKNOWN;
  }
};

#define EM_LEVEL_SIZE			2106
#define EM_LEVEL_XSIZE			64
#define EM_LEVEL_YSIZE			32

INIT_SECTION static void LoadLevelFromFileInfo_EM(int nr)
{
//  char *filename = level_file_info->filename;
//  FILE *file;
	unsigned char leveldata[EM_LEVEL_SIZE];
	unsigned char *header = &leveldata[EM_LEVEL_XSIZE * EM_LEVEL_YSIZE];
	unsigned char code0 = 0x65;
	unsigned char code1 = 0x11;
	boolean level_is_crypted = FALSE;
//	int nr = level_file_info->nr;
	int i, x, y;

	DmOpenRef dbRef;
	UInt16	  NumRecs;
	MemHandle mhRecord;

	dbRef = (DmOpenRef)0;

	if(LevelSets->cur_levset<0||LevelSets->cur_levset>MAXLEVSETS-1)
	{
		level.no_valid_file = TRUE;
		bugd("EM ERROR! cur_levset",0);
		return;
	}

 	{
		int i=LevelSets->cur_levset;
		LocalID	c_dbid = LevelSets->dbid[i];
		UInt16	c_card = LevelSets->card[i];
//		bugd(LevelSets->name[i],(long)c_dbid+c_card*256*256);
		
		if(c_dbid==0)
		{
			level.no_valid_file = TRUE;
			bugd("EM ERROR! dbid",0);
			return;
		}
		dbRef = DmOpenDatabase(c_card, c_dbid, dmModeReadOnly);
	} 

	if(dbRef == 0)
	{
		level.no_valid_file = TRUE;
		bugd("EM ERROR! openDB",0);
		return;
	}
	
	NumRecs = DmNumRecords (dbRef);
	if(NumRecs<=nr)
	{
		DmCloseDatabase(dbRef);
		level.no_valid_file = TRUE;
		return;
	}
	
	mhRecord = DmGetRecord (dbRef, nr );
	InitLikeFile ( MemHandleLock (mhRecord ) );
	
	for(i = 0; i < EM_LEVEL_SIZE; i++)
		leveldata[i] = getLikeFile8Bit();

	CloseLikeFile();
	DmReleaseRecord(dbRef, nr, false);
	DmCloseDatabase(dbRef);

  // check if level data is crypted by testing against known starting bytes
  //   of the few existing crypted level files (from Emerald Mine 1 + 2)

	if ((leveldata[0] == 0xf1 ||
			leveldata[0] == 0xf5) && leveldata[2] == 0xe7 && leveldata[3] == 0xee)
	{
		level_is_crypted = TRUE;

		if (leveldata[0] == 0xf5)	// error in crypted Emerald Mine 2 levels
			leveldata[0] = 0xf1;
	}

	if (level_is_crypted)		// decode crypted level data
	{
		for(i = 0; i < EM_LEVEL_SIZE; i++)
		{
			leveldata[i] ^= code0;
			leveldata[i] -= code1;

			code0  = (code0 + 7) & 0xff;
		}
	}

	level.fieldx	= EM_LEVEL_XSIZE;
	level.fieldy	= EM_LEVEL_YSIZE;

	level.time		= header[46] * 10;
	level.gems_needed	= header[47];

	// The original Emerald Mine levels have their level number stored
	// at the second byte of the level file...
	// Do not trust this information at other level files, e.g. EMC,
	// but correct it anyway (normally the first row is completely
	// steel wall, so the correction does not hurt anyway).

	if (leveldata[1] == nr + GetEM_LevelOffset())
		leveldata[1] = leveldata[2];	// correct level number field

//	sprintf(level->name, "Level %d", nr);		// set level name

	level.score[SC_EMERALD]	= header[36];
	level.score[SC_DIAMOND]	= header[37];
	level.score[SC_ROBOT]	= header[38];
	level.score[SC_SPACESHIP]	= header[39];
	level.score[SC_BUG]		= header[40];
	level.score[SC_YAMYAM]	= header[41];
	level.score[SC_NUT]		= header[42];
	level.score[SC_DYNAMITE]	= header[43];
	level.score[SC_TIME_BONUS]	= header[44];

	level.num_yamyam_contents = 4;

	for(i = 0; i < level.num_yamyam_contents; i++)
		for(y = 0; y < 3; y++)
			for(x = 0; x < 3; x++)
				level.yamyam_content[i][x][y] =
								map_em_element_yam(header[i * 9 + y * 3 + x]);

	level.amoeba_speed		= (header[52] * 256 + header[53]) % 256;
	level.time_magic_wall	= (header[54] * 256 + header[55]) * 16 / 100;
	level.time_wheel		= (header[56] * 256 + header[57]) * 16 / 100;
	level.amoeba_content		= EL_DIAMOND;

	for (y = 0; y < level.fieldy; y++) for (x = 0; x < level.fieldx; x++)
	{
		int new_element = map_em_element_field(leveldata[y * EM_LEVEL_XSIZE + x]);
//bugd("element at x,y",(long)new_element*256*256+x*256+y);
		if (new_element == EL_AMOEBA_DEAD && level.amoeba_speed)
			new_element = EL_AMOEBA_WET;
//if(y==30)bugd("original at x, y=30",(long)leveldata[y * EM_LEVEL_XSIZE + x]*256+x);
		Feld[A2TO1(x,y)] = new_element;
//PALM!!!
		if(new_element==EL_ACID_POOL_BOTTOM&&y>=1)
		{
			register short int upper_place=A2TO1(x,y-1);
			if(Feld[upper_place]!=EL_ACID)
			{
				if(Feld[upper_place]==EL_SAND||Feld[upper_place]==EL_EMPTY)
					Feld[upper_place]=EL_ACID;
//				else bugd("EM file load: ACIDPOOL without ACID like SAND!",0);
			}			
		}
	}

	x = (header[48] * 256 + header[49]) % EM_LEVEL_XSIZE;
	y = (header[48] * 256 + header[49]) / EM_LEVEL_XSIZE;
	Feld[A2TO1(x,y)] = EL_PLAYER_1;

	x = (header[50] * 256 + header[51]) % EM_LEVEL_XSIZE;
	y = (header[50] * 256 + header[51]) / EM_LEVEL_XSIZE;
	Feld[A2TO1(x,y)] = EL_PLAYER_2;
}

// -------------------------------------------------------------------------
// functions for loading generic level                                      
// -------------------------------------------------------------------------

static void LoadLevelFromFileInfo(int nr)
//was (struct LevelInfo *level, struct LevelFileInfo *level_file_info)
{
	// always start with reliable default values
	setLevelInfoToDefaults();

	switch (LevelSets->levset_type[(int)LevelSets->cur_levset])
	{
		case LEVEL_FILE_TYPE_RND:
			LoadLevelFromFileInfo_RND(nr);
			break;

		case LEVEL_FILE_TYPE_EM:
			LoadLevelFromFileInfo_EM(nr);
			break;
/*
		case LEVEL_FILE_TYPE_SP:
			LoadLevelFromFileInfo_SP(level, level_file_info);
			break;
*/
		default:
			LoadLevelFromFileInfo_RND(nr);
			break;
	}
}

static void LoadLevel_InitVersion(int level_number)
// was static void LoadLevel_InitVersion(struct LevelInfo *level, char *filename) 
{
//bugd("level.game_version",level.game_version);
//bugd("level.file_version",level.file_version);
/*	if (leveldir_current == NULL)		// only when dumping level
		return;
*/
	// determine correct game engine version of current level
//	if (!leveldir_current->latest_engine)
//PALM
	if( !UseLatestEngine() )
	{
    //   For all levels which are not forced to use the latest game engine
    //   version (normally user contributed, private and undefined levels),
    //   use the version of the game engine the levels were created for.

    //   Since 2.0.1, the game engine version is now directly stored
    //   in the level file (chunk "VERS"), so there is no need anymore
    //   to set the game version from the file version (except for old,
    //   pre-2.0 levels, where the game version is still taken from the
    //   file format version used to store the level -- see above).

		// player was faster than enemies in 1.0.0 and before
		if (level.file_version == FILE_VERSION_1_0)
			level.double_speed = TRUE;

		// Default behaviour for EM style gems was "slippery" only in 2.0.1
		if (level.game_version == VERSION_IDENT(2,0,1,0))
			level.em_slippery_gems = TRUE;

		if (level.game_version < VERSION_IDENT(2,2,0,0))
			level.use_spring_bug = TRUE;

		if (level.game_version < VERSION_IDENT(3,1,0,0))
		{
//			int i, j;

			level.can_move_into_acid_bits = 0; // nothing can move into acid
			level.dont_collide_with_bits = 0; // nothing is deadly when colliding

			setMoveIntoAcidProperty(&level, EL_ROBOT,     TRUE);
			setMoveIntoAcidProperty(&level, EL_SATELLITE, TRUE);
			setMoveIntoAcidProperty(&level, EL_PENGUIN,   TRUE);
			setMoveIntoAcidProperty(&level, EL_BALLOON,   TRUE);
//bugd("can_move_into_acid_bits",level.can_move_into_acid_bits);
/*
			for (i = 0; i < NUM_CUSTOM_ELEMENTS; i++)
				SET_PROPERTY(EL_CUSTOM_START + i, EP_CAN_MOVE_INTO_ACID, TRUE);

			for (i = 0; i < NUM_CUSTOM_ELEMENTS; i++)
			{
				int element = EL_CUSTOM_START + i;
				struct ElementInfo *ei = &element_info[element];

				for (j = 0; j < ei->num_change_pages; j++)
				{
					struct ElementChangeInfo *change = &ei->change_page[j];

					change->trigger_player = CH_PLAYER_ANY;
					change->trigger_page = CH_PAGE_ANY;
				}
			}
*/		}
	}
	else
	{

    //   For all levels which are forced to use the latest game engine version
    //   (normally all but user contributed, private and undefined levels), set
    //   the game engine version to the actual version; this allows for actual
    //   corrections in the game engine to take effect for existing, converted
    //   levels (from "classic" or other existing games) to make the emulation
    //   of the corresponding game more accurate, while (hopefully) not breaking
    //   existing levels created from other players.

		level.game_version = GAME_VERSION_ACTUAL;

    //   Set special EM style gems behaviour: EM style gems slip down from
    //   normal, steel and growing wall. As this is a more fundamental change,
    //   it seems better to set the default behaviour to "off" (as it is more
    //   natural) and make it configurable in the level editor (as a property
    //   of gem style elements). Already existing converted levels (neither
    //   private nor contributed levels) are changed to the new behaviour.

		if (level.file_version < FILE_VERSION_2_0)
			level.em_slippery_gems = TRUE;
	}

};

static void LoadLevel_InitElements(int level_number)
//was static void LoadLevel_InitElements(struct LevelInfo *level, char *filename)
{
	int i, /*j,*/ x, y;
/*
	// map custom element change events that have changed in newer versions
	// (these following values were accidentally changed in version 3.0.1)
	if (level.game_version <= VERSION_IDENT(3,0,0,0))
	{
		for (i = 0; i < NUM_CUSTOM_ELEMENTS; i++)
		{
			int element = EL_CUSTOM_START + i;

			// order of checking and copying events to be mapped is important
			for (j = CE_BY_OTHER_ACTION; j >= CE_BY_PLAYER_OBSOLETE; j--)
			{
				if (HAS_CHANGE_EVENT(element, j - 2))
				{
					SET_CHANGE_EVENT(element, j - 2, FALSE);
					SET_CHANGE_EVENT(element, j, TRUE);
				}
			}

			// order of checking and copying events to be mapped is important
			for (j = CE_OTHER_GETS_COLLECTED; j >= CE_HITTING_SOMETHING; j--)
			{
				if (HAS_CHANGE_EVENT(element, j - 1))
				{
					SET_CHANGE_EVENT(element, j - 1, FALSE);
					SET_CHANGE_EVENT(element, j, TRUE);
				}
			}
		}
	}

	// some custom element change events get mapped since version 3.0.3
	for (i = 0; i < NUM_CUSTOM_ELEMENTS; i++)
	{
		int element = EL_CUSTOM_START + i;

		if (HAS_CHANGE_EVENT(element, CE_BY_PLAYER_OBSOLETE) ||
						HAS_CHANGE_EVENT(element, CE_BY_COLLISION_OBSOLETE))
		{
			SET_CHANGE_EVENT(element, CE_BY_PLAYER_OBSOLETE, FALSE);
			SET_CHANGE_EVENT(element, CE_BY_COLLISION_OBSOLETE, FALSE);

			SET_CHANGE_EVENT(element, CE_BY_DIRECT_ACTION, TRUE);
		}
	}

	// initialize "can_change" field for old levels with only one change page
	if (level.game_version <= VERSION_IDENT(3,0,2,0))
	{
		for (i = 0; i < NUM_CUSTOM_ELEMENTS; i++)
		{
			int element = EL_CUSTOM_START + i;

			if (CAN_CHANGE(element))
				element_info[element].change->can_change = TRUE;
		}
	}
	// correct custom element values (for old levels without these options)
	for (i = 0; i < NUM_CUSTOM_ELEMENTS; i++)
	{
		int element = EL_CUSTOM_START + i;
		struct ElementInfo *ei = &element_info[element];

		if (ei->access_direction == MV_NO_MOVING)
			ei->access_direction = MV_ALL_DIRECTIONS;

		for (j = 0; j < ei->num_change_pages; j++)
		{
			struct ElementChangeInfo *change = &ei->change_page[j];

			if (change->trigger_side == CH_SIDE_NONE)
				change->trigger_side = CH_SIDE_ANY;
		}
	}

	// initialize "can_explode" field for old levels which did not store this
	if (level->game_version <= VERSION_IDENT(3,1,0,0))
	{
		for (i = 0; i < NUM_CUSTOM_ELEMENTS; i++)
		{
			int element = EL_CUSTOM_START + i;

			if (EXPLODES_1X1_OLD(element))
				element_info[element].explosion_type = EXPLODES_1X1;

			SET_PROPERTY(element, EP_CAN_EXPLODE, (EXPLODES_BY_FIRE(element) ||
													EXPLODES_SMASHED(element) ||
													EXPLODES_IMPACT(element)));
		}
	}
*/
	level.amoeba_content = getMappedElementByVersion(level.amoeba_content,
													level.game_version);
	for (i = 0; i < MAX_ELEMENT_CONTENTS; i++)
		for (x = 0; x < 3; x++)
			for (y = 0; y < 3; y++)
				level.yamyam_content[i][x][y] =
						getMappedElementByVersion(level.yamyam_content[i][x][y],
															level.game_version);

	// initialize element properties for level editor etc.
	InitElementPropertiesEngine(level.game_version);

};

static void LoadLevel_InitPlayfield(int level_number)
//was static void LoadLevel_InitPlayfield(struct LevelInfo *level, char *filename)
{
	int x, y;

	// map elements that have changed in newer versions
	for (y = 0; y < level.fieldy; y++)
	{
		short int xy=A2TO1(0,y);
		for (x = 0; x < level.fieldx; x++,xy+=A2TO1(1,0))
		{
			int element = Feld[xy]; // was level.field

			element = getMappedElementByVersion(element, level.game_version);

			Feld[xy] = element;// was level.field
		}
	}

	// initialize level size variables for faster access
	lev_fieldx = level.fieldx;
	lev_fieldy = level.fieldy;

	// determine border element for this level
	SetBorderElement();
};

void LoadLevel(int nr)
{
/*	struct LevelFileInfo *level_file_info = getLevelFileInfo(nr);
	char *filename = level_file_info->filename;
*/
	LoadLevelFromFileInfo(nr);
//	was   LoadLevelFromFileInfo(&level, level_file_info);

/*
	if (level.use_custom_template)
		LoadLevelTemplate(-1);
*/
//	for palm only, was LoadLevel_InitVersion(&level, filename);
	LoadLevel_InitVersion(nr);
//	for palm only, was LoadLevel_InitElements(&level, filename);
	LoadLevel_InitElements(nr);
//	for palm only, was LoadLevel_InitPlayfield(&level, filename);
	LoadLevel_InitPlayfield(nr);
};

INIT_SECTION static void setSetupInfoToDefaults(struct SetupInfo *si)
{/*
  int i;

  si->player_name = get_corrected_login_name(getLoginName());

  si->sound = TRUE;
  si->sound_loops = TRUE;
  si->sound_music = TRUE;
  si->sound_simple = TRUE;
  si->toons = TRUE;
  si->double_buffering = TRUE;
  si->direct_draw = !si->double_buffering;*/
	si->scroll_delay = TRUE;
	si->soft_scrolling = FALSE; //was TRUE
/*  si->fading = FALSE;
  si->autorecord = TRUE;
  si->quick_doors = FALSE;
  si->team_mode = FALSE;
  si->handicap = TRUE;*/
	si->time_limit = TRUE;
/*  si->fullscreen = FALSE;
  si->ask_on_escape = TRUE;

  si->graphics_set = getStringCopy(GFX_CLASSIC_SUBDIR);
  si->sounds_set = getStringCopy(SND_CLASSIC_SUBDIR);
  si->music_set = getStringCopy(MUS_CLASSIC_SUBDIR);
  si->override_level_graphics = FALSE;
  si->override_level_sounds = FALSE;
  si->override_level_music = FALSE;

  si->editor.el_boulderdash = TRUE;
  si->editor.el_emerald_mine = TRUE;
  si->editor.el_more = TRUE;
  si->editor.el_sokoban = TRUE;
  si->editor.el_supaplex = TRUE;
  si->editor.el_diamond_caves = TRUE;
  si->editor.el_dx_boulderdash = TRUE;
  si->editor.el_chars = TRUE;
  si->editor.el_custom = TRUE;
  si->editor.el_custom_more = FALSE;

  si->editor.el_headlines = TRUE;
  si->editor.el_user_defined = FALSE;

  si->shortcut.save_game = DEFAULT_KEY_SAVE_GAME;
  si->shortcut.load_game = DEFAULT_KEY_LOAD_GAME;
  si->shortcut.toggle_pause = DEFAULT_KEY_TOGGLE_PAUSE;

  for (i = 0; i < MAX_PLAYERS; i++)
  {
    si->input[i].use_joystick = FALSE;
    si->input[i].joy.device_name=getStringCopy(getDeviceNameFromJoystickNr(i));
    si->input[i].joy.xleft   = JOYSTICK_XLEFT;
    si->input[i].joy.xmiddle = JOYSTICK_XMIDDLE;
    si->input[i].joy.xright  = JOYSTICK_XRIGHT;
    si->input[i].joy.yupper  = JOYSTICK_YUPPER;
    si->input[i].joy.ymiddle = JOYSTICK_YMIDDLE;
    si->input[i].joy.ylower  = JOYSTICK_YLOWER;
    si->input[i].joy.snap  = (i == 0 ? JOY_BUTTON_1 : 0);
    si->input[i].joy.drop  = (i == 0 ? JOY_BUTTON_2 : 0);
    si->input[i].key.left  = (i == 0 ? DEFAULT_KEY_LEFT  : KSYM_UNDEFINED);
    si->input[i].key.right = (i == 0 ? DEFAULT_KEY_RIGHT : KSYM_UNDEFINED);
    si->input[i].key.up    = (i == 0 ? DEFAULT_KEY_UP    : KSYM_UNDEFINED);
    si->input[i].key.down  = (i == 0 ? DEFAULT_KEY_DOWN  : KSYM_UNDEFINED);
    si->input[i].key.snap  = (i == 0 ? DEFAULT_KEY_SNAP  : KSYM_UNDEFINED);
    si->input[i].key.drop  = (i == 0 ? DEFAULT_KEY_DROP  : KSYM_UNDEFINED);
  }

  si->system.sdl_audiodriver = getStringCopy(ARG_DEFAULT);
  si->system.audio_fragment_size = DEFAULT_AUDIO_FRAGMENT_SIZE;

  si->options.verbose = FALSE;*/
};

void LoadSetup()
{
/*  char *filename = getSetupFilename();
  SetupFileHash *setup_file_hash = NULL;
*/
	// always start with reliable default values
	setSetupInfoToDefaults(&setup);
/*
  setup_file_hash = loadSetupFileHash(filename);

  if (setup_file_hash)
  {
    char *player_name_new;

    checkSetupFileHashIdentifier(setup_file_hash, getCookie("SETUP"));
    decodeSetupFileHash(setup_file_hash);

    setup.direct_draw = !setup.double_buffering;

    freeSetupFileHash(setup_file_hash);

    // needed to work around problems with fixed length strings
    player_name_new = get_corrected_login_name(setup.player_name);
    free(setup.player_name);
    setup.player_name = player_name_new;
  }
  else
    Error(ERR_WARN, "using default setup values");*/
};