#ifndef LIBGAMESYSTEM_H_INCLUDED
#define LIBGAMESYSTEM_H_INCLUDED

#include "libgametypes.h"
#include "code_sections.h"

/* values for move directions */
#define MV_BIT_LEFT		0
#define MV_BIT_RIGHT		1
#define MV_BIT_UP		2
#define MV_BIT_DOWN	       	3

#define NUM_DIRECTIONS		4

/* values for special "button" bitmasks */
#define BUTTON_1		4
#define BUTTON_2		5

/* values for move direction and special "button" key bitmasks */
#define MV_NO_MOVING		0
#define MV_LEFT			(1 << MV_BIT_LEFT)
#define MV_RIGHT		(1 << MV_BIT_RIGHT)
#define MV_UP			(1 << MV_BIT_UP)
#define MV_DOWN	       		(1 << MV_BIT_DOWN)

#define KEY_BUTTON_1		(1 << BUTTON_1)
#define KEY_BUTTON_2		(1 << BUTTON_2)
#define KEY_MOTION		(MV_LEFT | MV_RIGHT | MV_UP | MV_DOWN)
#define KEY_BUTTON		(KEY_BUTTON_1 | KEY_BUTTON_2)
#define KEY_ACTION		(KEY_MOTION | KEY_BUTTON) 

#define MV_DIR_BIT(x)		((x) == MV_LEFT  ? MV_BIT_LEFT  :	\
							 (x) == MV_RIGHT ? MV_BIT_RIGHT :	\
							 (x) == MV_UP    ? MV_BIT_UP    : MV_BIT_DOWN)

#define MV_DIR_OPPOSITE(x)	((x) == MV_LEFT  ? MV_RIGHT :		\
							 (x) == MV_RIGHT ? MV_LEFT  :		\
							 (x) == MV_UP    ? MV_DOWN  :		\
							 (x) == MV_DOWN  ? MV_UP    : MV_NO_MOVING)

/* values for animation mode (frame order and direction) */
#define ANIM_NONE		0
#define ANIM_LOOP		(1 << 0)
#define ANIM_LINEAR		(1 << 1)
#define ANIM_PINGPONG		(1 << 2)
#define ANIM_PINGPONG2		(1 << 3)
#define ANIM_RANDOM		(1 << 4)
#define ANIM_REVERSE		(1 << 5)

/* values for special (non game element) animation modes */
#define ANIM_HORIZONTAL		(1 << 6)
#define ANIM_VERTICAL		(1 << 7)

#define ANIM_DEFAULT		ANIM_LOOP

/* values for redraw_mask */
#define REDRAW_NONE		(0)
#define REDRAW_ALL		(1 << 0)
#define REDRAW_FIELD		(1 << 1)
#define REDRAW_TILES		(1 << 2)
#define REDRAW_DOOR_1		(1 << 3) 

#define REDRAW_FROM_BACKBUFFER	(1 << 9)

#define REDRAW_MAIN		(REDRAW_FIELD | \
				 REDRAW_TILES )/*| \
for palm only no				 REDRAW_MICROLEVEL)*/ 

#define REDRAWTILES_THRESHOLD	(SCR_FIELDX * SCR_FIELDY / 2)

/* maximum number of parallel players supported by libgame functions */
#define MAX_PLAYERS		1	//for palm set 1 not 4

/* default name for new levels */
#define NAMELESS_LEVEL_NAME	"nameless level" 

/* default name for unknown player names */
#define ANONYMOUS_NAME		"anonymous"

#define VERSION_IDENT(a,b,c,d)	(long)((a) * 1000000 + (b) * 10000 + (c) * 100 + (d))

struct VideoSystemInfo
{
//	int default_depth;
	int width, height, depth;
//	boolean fullscreen_available;
//	boolean fullscreen_enabled;
};

struct SetupInfo
{/*
  char *player_name;

  boolean sound;
  boolean sound_loops;
  boolean sound_music;
  boolean sound_simple;
  boolean toons;
  boolean double_buffering;
  boolean direct_draw;		// !double_buffering (redundant!)
*/	boolean scroll_delay;
	boolean soft_scrolling;
/*  boolean fading;
  boolean autorecord;
  boolean quick_doors;
  boolean team_mode;
  boolean handicap;*/
	boolean time_limit;
/*  boolean fullscreen;
  boolean ask_on_escape;

  char *graphics_set;
  char *sounds_set;
  char *music_set;
  boolean override_level_graphics;
  boolean override_level_sounds;
  boolean override_level_music;

  struct SetupEditorInfo editor;
  struct SetupShortcutInfo shortcut;
  struct SetupInputInfo input[MAX_PLAYERS];
  struct SetupSystemInfo system;
  struct OptionInfo options;*/
};

extern struct VideoSystemInfo	video;

extern struct SetupInfo		setup;

extern int			level_nr;

extern int			redraw_mask;
extern int			redraw_tiles;

extern int			FrameCounter;

void SetDrawBackgroundMask(int);

void InitVideoDisplay(void) INIT_SECTION;
void CloseVideoDisplay(void) INIT_SECTION;

void FlushDisplay(void);

#endif
