#ifndef MAIN__H_INCLUDED
#define MAIN__H_INCLUDED

#include "libgamelibgame.h"

#include "conf_gfx.h"	/* include auto-generated data structure definitions */

#define DELAY_MILLION	250

#define IMG_EMPTY		IMG_EMPTY_SPACE

#define SCR_FIELDX		10	//was 17
#define SCR_FIELDY		10	//was 17
#define MAX_BUF_XSIZE		(SCR_FIELDX + 2)
#define MAX_BUF_YSIZE		(SCR_FIELDY + 2)

#define STD_LEV_FIELDX		64
#define STD_LEV_FIELDY		32
#define MAX_LEV_FIELDX		64	//was 128 and for palm 64
#define MAX_LEV_FIELDY		32	//was 128 and for palm 32

#define SCREENX(a)		((a) - scroll_x)
#define SCREENY(a)		((a) - scroll_y)
#define LEVELX(a)		((a) + scroll_x)
#define LEVELY(a)		((a) + scroll_y)
#define IN_VIS_FIELD(x,y) ((x)>=0 && (x)<SCR_FIELDX && (y)>=0 &&(y)<SCR_FIELDY)
#define IN_SCR_FIELD(x,y) ((x)>=BX1 && (x)<=BX2 && (y)>=BY1 &&(y)<=BY2)
#define IN_LEV_FIELD(x,y) ((x)>=0 && (x)<lev_fieldx && (y)>=0 &&(y)<lev_fieldy) 

/* values for configurable properties (custom elem's only, else pre-defined) */
#define EP_DIGGABLE		0
#define EP_COLLECTIBLE_ONLY	1
#define EP_DONT_RUN_INTO	2
#define EP_DONT_COLLIDE_WITH	3
#define EP_DONT_TOUCH		4
#define EP_INDESTRUCTIBLE	5
#define EP_SLIPPERY		6
#define EP_CAN_CHANGE		7
#define EP_CAN_MOVE		8
#define EP_CAN_FALL		9
#define EP_CAN_SMASH_PLAYER	10
#define EP_CAN_SMASH_ENEMIES	11
#define EP_CAN_SMASH_EVERYTHING	12
#define EP_EXPLODES_BY_FIRE	13
#define EP_EXPLODES_SMASHED	14
#define EP_EXPLODES_IMPACT	15
#define EP_WALKABLE_OVER	16
#define EP_WALKABLE_INSIDE	17
#define EP_WALKABLE_UNDER	18
#define EP_PASSABLE_OVER	19
#define EP_PASSABLE_INSIDE	20
#define EP_PASSABLE_UNDER	21
#define EP_DROPPABLE		22
#define EP_EXPLODES_1X1_OLD	23
#define EP_PUSHABLE		24
#define EP_EXPLODES_CROSS_OLD	25
#define EP_PROTECTED		26
#define EP_CAN_MOVE_INTO_ACID	27
#define EP_THROWABLE		28
#define EP_CAN_EXPLODE		29
#define EP_GRAVITY_REACHABLE	30

/* values for pre-defined properties */
#define EP_PLAYER		32
#define EP_CAN_PASS_MAGIC_WALL	33
#define EP_SWITCHABLE		34
#define EP_BD_ELEMENT		35
#define EP_SP_ELEMENT		36
#define EP_SB_ELEMENT		37
#define EP_GEM			38
#define EP_FOOD_DARK_YAMYAM	39
#define EP_FOOD_PENGUIN		40
#define EP_FOOD_PIG		41
#define EP_HISTORIC_WALL	42
#define EP_HISTORIC_SOLID	43
#define EP_CLASSIC_ENEMY	44
#define EP_BELT			45
#define EP_BELT_ACTIVE		46
#define EP_BELT_SWITCH		47
#define EP_TUBE			48
#define EP_KEYGATE		49
#define EP_AMOEBOID		50
#define EP_AMOEBALIVE		51
#define EP_HAS_CONTENT		52
#define EP_CAN_TURN_EACH_MOVE	53
#define EP_CAN_GROW		54
#define EP_ACTIVE_BOMB		55
#define EP_INACTIVE		56

/* values for special configurable properties (depending on level settings) */
#define EP_EM_SLIPPERY_WALL	57

/* values for special graphics properties (no effect on game engine) */
#define EP_GFX_CRUMBLED		58

/* values for derived properties (determined from properties above) */
#define EP_ACCESSIBLE_OVER	59
#define EP_ACCESSIBLE_INSIDE	60
#define EP_ACCESSIBLE_UNDER	61
#define EP_WALKABLE		62
#define EP_PASSABLE		63
#define EP_ACCESSIBLE		64
#define EP_COLLECTIBLE		65
#define EP_SNAPPABLE		66
#define EP_WALL			67
#define EP_SOLID_FOR_PUSHING	68
#define EP_DRAGONFIRE_PROOF	69
#define EP_EXPLOSION_PROOF	70
#define EP_CAN_SMASH		71
#define EP_EXPLODES_3X3_OLD	72
#define EP_CAN_EXPLODE_BY_FIRE	73
#define EP_CAN_EXPLODE_SMASHED	74
#define EP_CAN_EXPLODE_IMPACT	75
#define EP_SP_PORT		76
#define EP_CAN_EXPLODE_BY_DRAGONFIRE	77
#define EP_CAN_EXPLODE_BY_EXPLOSION	78
#define EP_COULD_MOVE_INTO_ACID		79
#define EP_MAYBE_DONT_COLLIDE_WITH	80

/* values for internal purpose only (level editor) */
#define EP_WALK_TO_OBJECT	81
#define EP_DEADLY		82

//PALM
#define EP_CAN_HOLD_FALLING_DOWN		83
#define EP_CAN_HOLD_BLOCKED_FALLING		84
#define EP_INDESTR_INACTIVE_NOANIM		85

#define NUM_ELEMENT_PROPERTIES	86//was 83

#define NUM_EP_BITFIELDS	((NUM_ELEMENT_PROPERTIES + 7) / 8) // changed for palm
															//8 was 32 

#define PROPERTY_BIT(p)		(1 << ((p) % 8))		// changed for palm 8 was 32
#define PROPERTY_VAR(e,p)	(PropertiesF[(p) / 8][e])	// changed for palm
														// 8 was 32
#define HAS_PROPERTY(e,p)	((PROPERTY_VAR(e, p) & PROPERTY_BIT(p)) != 0) 
#define SET_PROPERTY(e,p,v)	((v) ?					   \
							(PROPERTY_VAR(e,p) |=  PROPERTY_BIT(p)) : \
							(PROPERTY_VAR(e,p) &= ~PROPERTY_BIT(p))) 

/* values for change events for custom elements (stored in level file) */
#define CE_DELAY		0
#define CE_TOUCHED_BY_PLAYER	1
#define CE_PRESSED_BY_PLAYER	2
#define CE_PUSHED_BY_PLAYER	3
#define CE_DROPPED_BY_PLAYER	4
#define CE_HITTING_SOMETHING	5
#define CE_IMPACT		6
#define CE_SMASHED		7
#define CE_OTHER_IS_TOUCHING	8
#define CE_OTHER_IS_CHANGING	9
#define CE_OTHER_IS_EXPLODING	10
#define CE_OTHER_GETS_TOUCHED	11
#define CE_OTHER_GETS_PRESSED	12
#define CE_OTHER_GETS_PUSHED	13
#define CE_OTHER_GETS_COLLECTED	14
#define CE_OTHER_GETS_DROPPED	15
#define CE_BY_PLAYER_OBSOLETE	16	/* obsolete; now CE_BY_DIRECT_ACTION */
#define CE_BY_COLLISION_OBSOLETE 17	/* obsolete; now CE_BY_DIRECT_ACTION */
#define CE_BY_OTHER_ACTION	18	/* activates other element events */
#define CE_BY_DIRECT_ACTION	19	/* activates direct element events */
#define CE_OTHER_GETS_DIGGED	20
#define CE_ENTERED_BY_PLAYER	21
#define CE_LEFT_BY_PLAYER	22
#define CE_OTHER_GETS_ENTERED	23
#define CE_OTHER_GETS_LEFT	24
#define CE_SWITCHED		25
#define CE_OTHER_IS_SWITCHING	26
#define CE_HIT_BY_SOMETHING	27
#define CE_OTHER_IS_HITTING	28
#define CE_OTHER_GETS_HIT	29
#define CE_BLOCKED		30

#define NUM_CHANGE_EVENTS	31

#define HAS_ANY_CHANGE_EVENT(e,c) (IS_CUSTOM_ELEMENT(e) &&		  \
				 (CH_ANY_EVENT_VAR(e) & CH_EVENT_BIT(c)) != 0) 

/* values for change side for custom elements */
#define CH_SIDE_NONE		MV_NO_MOVING
#define CH_SIDE_LEFT		MV_LEFT
#define CH_SIDE_RIGHT		MV_RIGHT
#define CH_SIDE_TOP		MV_UP
#define CH_SIDE_BOTTOM		MV_DOWN
#define CH_SIDE_LEFT_RIGHT	MV_HORIZONTAL
#define CH_SIDE_TOP_BOTTOM	MV_VERTICAL
#define CH_SIDE_ANY		MV_ANY_DIRECTION 

/* values for custom move patterns (bits 0 - 3: basic move directions) */
#define MV_BIT_TOWARDS_PLAYER	4
#define MV_BIT_AWAY_FROM_PLAYER	5
#define MV_BIT_ALONG_LEFT_SIDE	6
#define MV_BIT_ALONG_RIGHT_SIDE	7
#define MV_BIT_TURNING_LEFT	8
#define MV_BIT_TURNING_RIGHT	9
#define MV_BIT_WHEN_PUSHED	10
#define MV_BIT_MAZE_RUNNER	11
#define MV_BIT_MAZE_HUNTER	12
#define MV_BIT_WHEN_DROPPED	13
#define MV_BIT_TURNING_LEFT_RIGHT 14
#define MV_BIT_TURNING_RIGHT_LEFT 15
#define MV_BIT_TURNING_RANDOM	16

/* values for custom move patterns */
#define MV_HORIZONTAL		(MV_LEFT | MV_RIGHT)
#define MV_VERTICAL		(MV_UP   | MV_DOWN)
#define MV_ALL_DIRECTIONS	(MV_LEFT | MV_RIGHT | MV_UP | MV_DOWN)
#define MV_ANY_DIRECTION	(MV_ALL_DIRECTIONS)
#define MV_TOWARDS_PLAYER	(1 << MV_BIT_TOWARDS_PLAYER)
#define MV_AWAY_FROM_PLAYER	(1 << MV_BIT_AWAY_FROM_PLAYER)
#define MV_ALONG_LEFT_SIDE	(1 << MV_BIT_ALONG_LEFT_SIDE)
#define MV_ALONG_RIGHT_SIDE	(1 << MV_BIT_ALONG_RIGHT_SIDE)
#define MV_TURNING_LEFT		(1 << MV_BIT_TURNING_LEFT)
#define MV_TURNING_RIGHT	(1 << MV_BIT_TURNING_RIGHT)
#define MV_WHEN_PUSHED		(1 << MV_BIT_WHEN_PUSHED)
#define MV_MAZE_RUNNER		(1 << MV_BIT_MAZE_RUNNER)
#define MV_MAZE_HUNTER		(1 << MV_BIT_MAZE_HUNTER)
#define MV_MAZE_RUNNER_STYLE	(MV_MAZE_RUNNER | MV_MAZE_HUNTER)
#define MV_WHEN_DROPPED		(1 << MV_BIT_WHEN_DROPPED)
#define MV_TURNING_LEFT_RIGHT	(1 << MV_BIT_TURNING_LEFT_RIGHT)
#define MV_TURNING_RIGHT_LEFT	(1 << MV_BIT_TURNING_RIGHT_LEFT)
#define MV_TURNING_RANDOM	(1 << MV_BIT_TURNING_RANDOM)

/* values for initial move direction (bits 0 - 3: basic move directions) */
#define MV_START_BIT_PREVIOUS	4

/* values for initial move direction */
#define MV_START_AUTOMATIC	(MV_NO_MOVING)
#define MV_START_LEFT		(MV_LEFT)
#define MV_START_RIGHT		(MV_RIGHT)
#define MV_START_UP		(MV_UP)
#define MV_START_DOWN		(MV_DOWN)
#define MV_START_RANDOM		(MV_ALL_DIRECTIONS)
#define MV_START_PREVIOUS	(1 << MV_START_BIT_PREVIOUS)

/* values for explosion type for custom elements */
#define EXPLODES_3X3		0
#define EXPLODES_1X1		1
#define EXPLODES_CROSS		2

/* macros for configurable properties */
#define IS_DIGGABLE(e)		HAS_PROPERTY(e, EP_DIGGABLE)
#define IS_COLLECTIBLE_ONLY(e)	HAS_PROPERTY(e, EP_COLLECTIBLE_ONLY)
#define DONT_RUN_INTO(e)	HAS_PROPERTY(e, EP_DONT_RUN_INTO)
#define DONT_COLLIDE_WITH(e)	HAS_PROPERTY(e, EP_DONT_COLLIDE_WITH)
#define DONT_TOUCH(e)		HAS_PROPERTY(e, EP_DONT_TOUCH)
#define IS_INDESTRUCTIBLE(e)	HAS_PROPERTY(e, EP_INDESTRUCTIBLE)
#define IS_SLIPPERY(e)		HAS_PROPERTY(e, EP_SLIPPERY)
#define CAN_CHANGE(e)		HAS_PROPERTY(e, EP_CAN_CHANGE)
#define CAN_MOVE(e)		HAS_PROPERTY(e, EP_CAN_MOVE)
#define CAN_FALL(e)		HAS_PROPERTY(e, EP_CAN_FALL)
#define CAN_SMASH_PLAYER(e)	HAS_PROPERTY(e, EP_CAN_SMASH_PLAYER)
#define CAN_SMASH_ENEMIES(e)	HAS_PROPERTY(e, EP_CAN_SMASH_ENEMIES)
#define CAN_SMASH_EVERYTHING(e)	HAS_PROPERTY(e, EP_CAN_SMASH_EVERYTHING)
#define EXPLODES_BY_FIRE(e)	HAS_PROPERTY(e, EP_EXPLODES_BY_FIRE)
#define EXPLODES_SMASHED(e)	HAS_PROPERTY(e, EP_EXPLODES_SMASHED)
#define EXPLODES_IMPACT(e)	HAS_PROPERTY(e, EP_EXPLODES_IMPACT)
#define IS_WALKABLE_OVER(e)	HAS_PROPERTY(e, EP_WALKABLE_OVER)
#define IS_WALKABLE_INSIDE(e)	HAS_PROPERTY(e, EP_WALKABLE_INSIDE)
#define IS_WALKABLE_UNDER(e)	HAS_PROPERTY(e, EP_WALKABLE_UNDER)
#define IS_PASSABLE_OVER(e)	HAS_PROPERTY(e, EP_PASSABLE_OVER)
#define IS_PASSABLE_INSIDE(e)	HAS_PROPERTY(e, EP_PASSABLE_INSIDE)
#define IS_PASSABLE_UNDER(e)	HAS_PROPERTY(e, EP_PASSABLE_UNDER)
#define IS_DROPPABLE(e)		HAS_PROPERTY(e, EP_DROPPABLE)
#define EXPLODES_1X1_OLD(e)	HAS_PROPERTY(e, EP_EXPLODES_1X1_OLD)
#define IS_PUSHABLE(e)		HAS_PROPERTY(e, EP_PUSHABLE)
#define EXPLODES_CROSS_OLD(e)	HAS_PROPERTY(e, EP_EXPLODES_CROSS_OLD)
#define IS_PROTECTED(e)		HAS_PROPERTY(e, EP_PROTECTED)
#define CAN_MOVE_INTO_ACID(e)	HAS_PROPERTY(e, EP_CAN_MOVE_INTO_ACID)
#define IS_THROWABLE(e)		HAS_PROPERTY(e, EP_THROWABLE)
#define CAN_EXPLODE(e)		HAS_PROPERTY(e, EP_CAN_EXPLODE)
#define IS_GRAVITY_REACHABLE(e)	HAS_PROPERTY(e, EP_GRAVITY_REACHABLE)

/* macros for special configurable properties */
#define IS_EM_SLIPPERY_WALL(e)	HAS_PROPERTY(e, EP_EM_SLIPPERY_WALL)

/* macros for special graphics properties */
#define GFX_CRUMBLED(e)		HAS_PROPERTY(GFX_ELEMENT(e), EP_GFX_CRUMBLED)

/* macros for pre-defined properties */
#define ELEM_IS_PLAYER(e)	HAS_PROPERTY(e, EP_PLAYER)
#define CAN_PASS_MAGIC_WALL(e)	HAS_PROPERTY(e, EP_CAN_PASS_MAGIC_WALL)
#define IS_SWITCHABLE(e)	HAS_PROPERTY(e, EP_SWITCHABLE)
#define IS_BD_ELEMENT(e)	HAS_PROPERTY(e, EP_BD_ELEMENT)
#define IS_SP_ELEMENT(e)	HAS_PROPERTY(e, EP_SP_ELEMENT)
#define IS_SB_ELEMENT(e)	HAS_PROPERTY(e, EP_SB_ELEMENT)
#define IS_GEM(e)		HAS_PROPERTY(e, EP_GEM)
#define IS_FOOD_DARK_YAMYAM(e)	HAS_PROPERTY(e, EP_FOOD_DARK_YAMYAM)
#define IS_FOOD_PENGUIN(e)	HAS_PROPERTY(e, EP_FOOD_PENGUIN)
#define IS_FOOD_PIG(e)		HAS_PROPERTY(e, EP_FOOD_PIG)
#define IS_HISTORIC_WALL(e)	HAS_PROPERTY(e, EP_HISTORIC_WALL)
#define IS_HISTORIC_SOLID(e)	HAS_PROPERTY(e, EP_HISTORIC_SOLID)
#define IS_CLASSIC_ENEMY(e)	HAS_PROPERTY(e, EP_CLASSIC_ENEMY)
#define IS_BELT(e)		HAS_PROPERTY(e, EP_BELT)
#define IS_BELT_ACTIVE(e)	HAS_PROPERTY(e, EP_BELT_ACTIVE)
#define IS_BELT_SWITCH(e)	HAS_PROPERTY(e, EP_BELT_SWITCH)
#define IS_TUBE(e)		HAS_PROPERTY(e, EP_TUBE)
#define IS_KEYGATE(e)		HAS_PROPERTY(e, EP_KEYGATE)
#define IS_AMOEBOID(e)		HAS_PROPERTY(e, EP_AMOEBOID)
#define IS_AMOEBALIVE(e)	HAS_PROPERTY(e, EP_AMOEBALIVE)
#define HAS_CONTENT(e)		HAS_PROPERTY(e, EP_HAS_CONTENT)
#define CAN_TURN_EACH_MOVE(e)	HAS_PROPERTY(e, EP_CAN_TURN_EACH_MOVE)
#define CAN_GROW(e)		HAS_PROPERTY(e, EP_CAN_GROW)
#define IS_ACTIVE_BOMB(e)	HAS_PROPERTY(e, EP_ACTIVE_BOMB)
#define IS_INACTIVE(e)		HAS_PROPERTY(e, EP_INACTIVE)

/* macros for derived properties */
#define IS_ACCESSIBLE_OVER(e)	HAS_PROPERTY(e, EP_ACCESSIBLE_OVER)
#define IS_ACCESSIBLE_INSIDE(e)	HAS_PROPERTY(e, EP_ACCESSIBLE_INSIDE)
#define IS_ACCESSIBLE_UNDER(e)	HAS_PROPERTY(e, EP_ACCESSIBLE_UNDER)
#define IS_WALKABLE(e)		HAS_PROPERTY(e, EP_WALKABLE)
#define IS_PASSABLE(e)		HAS_PROPERTY(e, EP_PASSABLE)
#define IS_ACCESSIBLE(e)	HAS_PROPERTY(e, EP_ACCESSIBLE)
#define IS_COLLECTIBLE(e)	HAS_PROPERTY(e, EP_COLLECTIBLE)
#define IS_SNAPPABLE(e)		HAS_PROPERTY(e, EP_SNAPPABLE)
#define IS_WALL(e)		HAS_PROPERTY(e, EP_WALL)
#define IS_SOLID_FOR_PUSHING(e)	HAS_PROPERTY(e, EP_SOLID_FOR_PUSHING)
#define IS_DRAGONFIRE_PROOF(e)	HAS_PROPERTY(e, EP_DRAGONFIRE_PROOF)
#define IS_EXPLOSION_PROOF(e)	HAS_PROPERTY(e, EP_EXPLOSION_PROOF)
#define CAN_SMASH(e)		HAS_PROPERTY(e, EP_CAN_SMASH)
#define EXPLODES_3X3_OLD(e)	HAS_PROPERTY(e, EP_EXPLODES_3X3_OLD)
#define CAN_EXPLODE_BY_FIRE(e)	HAS_PROPERTY(e, EP_CAN_EXPLODE_BY_FIRE)
#define CAN_EXPLODE_SMASHED(e)	HAS_PROPERTY(e, EP_CAN_EXPLODE_SMASHED)
#define CAN_EXPLODE_IMPACT(e)	HAS_PROPERTY(e, EP_CAN_EXPLODE_IMPACT)
#define IS_SP_PORT(e)		HAS_PROPERTY(e, EP_SP_PORT)
#define CAN_EXPLODE_BY_DRAGONFIRE(e)	\
				HAS_PROPERTY(e, EP_CAN_EXPLODE_BY_DRAGONFIRE)
#define CAN_EXPLODE_BY_EXPLOSION(e)	\
				HAS_PROPERTY(e, EP_CAN_EXPLODE_BY_EXPLOSION)
#define COULD_MOVE_INTO_ACID(e)	HAS_PROPERTY(e, EP_COULD_MOVE_INTO_ACID)
#define MAYBE_DONT_COLLIDE_WITH(e) HAS_PROPERTY(e, EP_MAYBE_DONT_COLLIDE_WITH)

//PALM
#define CAN_HOLD_FALLING_DOWN(e)	HAS_PROPERTY(e, EP_CAN_HOLD_FALLING_DOWN)
#define CAN_HOLD_BLOCKED_FALLING(e)	HAS_PROPERTY(e, EP_CAN_HOLD_BLOCKED_FALLING)
#define IS_INDESTR_INACTIVE_NOANIM(e)	HAS_PROPERTY(e, EP_INDESTR_INACTIVE_NOANIM)

/* special macros used in game engine */
#define IS_CUSTOM_ELEMENT(e)	((e) >= EL_CUSTOM_START &&		\
								 (e) <= EL_CUSTOM_END) 

#define IS_GROUP_ELEMENT(e)	((e) >= EL_GROUP_START &&		\
	 			 (e) <= EL_GROUP_END)

#define IS_CLIPBOARD_ELEMENT(e)	((e) >= EL_INTERNAL_CLIPBOARD_START &&	\
	 			 (e) <= EL_INTERNAL_CLIPBOARD_END)

#define IS_INTERNAL_ELEMENT(e)	((e) >= EL_INTERNAL_START &&		\
	 			 (e) <= EL_INTERNAL_END)

#define IS_ENVELOPE(e)		((e) >= EL_ENVELOPE_1 &&		\
							 (e) <= EL_ENVELOPE_4)

#define IS_GATE(e)		((e) >= EL_GATE_1 &&			\
	 			 (e) <= EL_GATE_4)

#define IS_GATE_GRAY(e)		((e) >= EL_GATE_1_GRAY &&		\
	 			 (e) <= EL_GATE_4_GRAY)

#define IS_EM_GATE(e)		((e) >= EL_EM_GATE_1 &&			\
	 			 (e) <= EL_EM_GATE_4)

#define IS_EM_GATE_GRAY(e)	((e) >= EL_EM_GATE_1_GRAY &&		\
	 			 (e) <= EL_EM_GATE_4_GRAY)

#define GFX_ELEMENT(e)			(e)
/*palm, was		(element_info[e].use_gfx_element ?	\
				 element_info[e].gfx_element : e)*/
 
#define IS_PLAYER(xy)		(StorePlayerXY==xy)
//was		(ELEM_IS_PLAYER(SAGetValue(&StorePlayer,x,y)))

#define IS_FREE(xy)		(Feld[xy] == EL_EMPTY && !IS_PLAYER(xy)) 
#define IS_FREE_OR_PLAYER(xy)	(Feld[xy] == EL_EMPTY)

#define IS_MOVING(xy)		(MovPos[xy] != 0)
#define IS_FALLING(xy)		(MovPos[xy] != 0 && MovDir[xy] == MV_DOWN)
#define IS_BLOCKED(xy)		(Feld[xy] == EL_BLOCKED) 

#define EL_CHANGED(e)		((e) == EL_ROCK           ? EL_EMERALD :    \
				 (e) == EL_BD_ROCK        ? EL_BD_DIAMOND : \
				 (e) == EL_EMERALD        ? EL_DIAMOND :    \
				 (e) == EL_EMERALD_YELLOW ? EL_DIAMOND :    \
				 (e) == EL_EMERALD_RED    ? EL_DIAMOND :    \
				 (e) == EL_EMERALD_PURPLE ? EL_DIAMOND :    \
				 EL_ROCK)
#define EL_CHANGED2(e)		((e) == EL_ROCK           ? EL_BD_DIAMOND : \
				 (e) == EL_BD_ROCK        ? EL_BD_DIAMOND : \
				 EL_BD_ROCK) 
#define IS_DRAWABLE(e)		((e) < EL_BLOCKED)
#define IS_NOT_DRAWABLE(e)	((e) >= EL_BLOCKED)
#define TAPE_IS_EMPTY(x)	((x).length == 0)
#define TAPE_IS_STOPPED(x)	(!(x).recording && !(x).playing)

#define PLAYERINFO(x,y)		(&stored_player_)
//was (&stored_player[SAGetValue(&StorePlayer,x,y)-EL_PLAYER_1])

#define SHIELD_ON(p)		((p)->shield_normal_time_left > 0)

#define ENEMY_PROTECTED_FIELD(xy)	(IS_PROTECTED(Feld[xy]) ||       \
					 IS_PROTECTED(Back[xy]))
#define EXPLOSION_PROTECTED_FIELD(xy)  (IS_EXPLOSION_PROOF(Feld[xy]))
#define PLAYER_ENEMY_PROTECTED(x,y,xy)     (SHIELD_ON(PLAYERINFO(x, y)) ||	   \
					 ENEMY_PROTECTED_FIELD(xy))
#define PLAYER_EXPLOSION_PROTECTED(x,y,xy) (SHIELD_ON(PLAYERINFO(x, y)) ||	   \
					 EXPLOSION_PROTECTED_FIELD(xy))
				 
#define PLAYER_SWITCHING(p,x,y)	((p)->is_switching &&			\
				 (p)->switch_x == (x) && (p)->switch_y == (y))  

//#define ANIM_FRAMES(g)		(graphic_infoF->anim_frames[g])
#define ANIM_DELAY(ga)		(anim_infoF->anim_delay_[ga]<DELAY_MILLION?	\
	anim_infoF->anim_delay_[ga]:1000000)
#define ANIM_INDEX(g)		(graphic_infoF->anim_index[g])

#define IS_ANIMATED(g)		(ANIM_INDEX(g)!=1)//was (ANIM_FRAMES(g) > 1)
#define IS_NEW_DELAY(f, ga)	((f) % ANIM_DELAY(ga) == 0)
#define IS_NEW_FRAME(f, g)	(IS_ANIMATED(g) && IS_NEW_DELAY(f, ANIM_INDEX(g)))

/* values for game_status (must match special image configuration suffixes) */
#define GAME_MODE_DEFAULT			0
#define GAME_MODE_MAIN				1
#define GAME_MODE_LEVELS			2
#define GAME_MODE_SCORES			3
#define GAME_MODE_EDITOR			4
#define GAME_MODE_INFO				5
#define GAME_MODE_SETUP				6
#define GAME_MODE_PLAYING			7
#define GAME_MODE_PSEUDO_DOOR		8
#define GAME_MODE_PSEUDO_PREVIEW	9
#define GAME_MODE_PSEUDO_CRUMBLED	10

/* there are no special config file suffixes for these modes */
#define GAME_MODE_PSEUDO_TYPENAME	11
#define GAME_MODE_QUIT				12

#define TAPE_IS_STOPPED(x)	(!(x).recording && !(x).playing)

#define SHIELD_ON(p)		((p)->shield_normal_time_left > 0)

/* fundamental game speed values */
#define ONE_SECOND_DELAY	1000	/* delay value for one second */
#define GAME_FRAME_DELAY	20	/* frame delay in milliseconds */
#define FFWD_FRAME_DELAY	10	/* 200% speed for fast forward */
#define FRAMES_PER_SECOND	(ONE_SECOND_DELAY / GAME_FRAME_DELAY)
//#define MICROLEVEL_SCROLL_DELAY	50	/* delay for scrolling micro level */
//#define MICROLEVEL_LABEL_DELAY	250	/* delay for micro level label */ 
 
/* boundaries of arrays etc. */
#define MAX_LEVEL_NAME_LEN	32
#define MAX_TAPELEN		(30 * FRAMES_PER_SECOND) /* max.time x fps */
// was 1000 and palm now is 30

#define MAX_NUM_AMOEBA		100
#define MAX_INVENTORY_SIZE	100	//for palm set 100 not 1000
#define MAX_KEYS		4
#define NUM_BELTS		4
#define NUM_BELT_PARTS		3

/* values for elements with content */
#define MIN_ELEMENT_CONTENTS	1
#define STD_ELEMENT_CONTENTS	4
#define MAX_ELEMENT_CONTENTS	4	//was 8 and for palm 4

#define LEVEL_SCORE_ELEMENTS	16	/* level elements with score */ 

/* often used screen positions */
#define SX			8
#define SY			8

#define TILEX			16//was 32
#define TILEY			16//was 32

#define MIDPOSX			(SCR_FIELDX / 2)
#define MIDPOSY			(SCR_FIELDY / 2)

/* score for elements */
#define SC_EMERALD		0
#define SC_DIAMOND		1
#define SC_BUG			2
#define SC_SPACESHIP		3
#define SC_YAMYAM		4
#define SC_ROBOT		5
#define SC_PACMAN		6
#define SC_NUT			7
#define SC_DYNAMITE		8
#define SC_KEY			9
#define SC_TIME_BONUS		10
#define SC_CRYSTAL		11
#define SC_PEARL		12
#define SC_SHIELD		13


/* "real" level file elements */
#define EL_UNDEFINED			-1

#define EL_EMPTY_SPACE			0
#define EL_EMPTY			EL_EMPTY_SPACE
#define EL_SAND				1
#define EL_WALL				2
#define EL_WALL_SLIPPERY		3
#define EL_ROCK				4
#define EL_KEY_OBSOLETE			5	/* obsolete; now EL_KEY_1 */
#define EL_EMERALD			6
#define EL_EXIT_CLOSED			7
#define EL_PLAYER_OBSOLETE		8	/* obsolete; now EL_PLAYER_1 */
#define EL_BUG				9
#define EL_SPACESHIP			10
#define EL_YAMYAM			11
#define EL_ROBOT			12
#define EL_STEELWALL			13
#define EL_DIAMOND			14
#define EL_AMOEBA_DEAD			15
#define EL_QUICKSAND_EMPTY		16
#define EL_QUICKSAND_FULL		17
#define EL_AMOEBA_DROP			18
#define EL_BOMB				19
#define EL_MAGIC_WALL			20
#define EL_SPEED_PILL			21
#define EL_ACID				22
#define EL_AMOEBA_WET			23
#define EL_AMOEBA_DRY			24
#define EL_NUT				25
#define EL_GAME_OF_LIFE			26
#define EL_BIOMAZE			27
#define EL_DYNAMITE_ACTIVE		28
#define EL_STONEBLOCK			29
#define EL_ROBOT_WHEEL			30
#define EL_ROBOT_WHEEL_ACTIVE		31
#define EL_KEY_1			32
#define EL_KEY_2			33
#define EL_KEY_3			34
#define EL_KEY_4			35
#define EL_GATE_1			36
#define EL_GATE_2			37
#define EL_GATE_3			38
#define EL_GATE_4			39
#define EL_GATE_1_GRAY			40
#define EL_GATE_2_GRAY			41
#define EL_GATE_3_GRAY			42
#define EL_GATE_4_GRAY			43
#define EL_DYNAMITE			44
#define EL_PACMAN			45
#define EL_INVISIBLE_WALL		46
#define EL_LAMP				47
#define EL_LAMP_ACTIVE			48
#define EL_WALL_EMERALD			49
#define EL_WALL_DIAMOND			50
#define EL_AMOEBA_FULL			51
#define EL_BD_AMOEBA			52
#define EL_TIME_ORB_FULL		53
#define EL_TIME_ORB_EMPTY		54
#define EL_EXPANDABLE_WALL		55
#define EL_BD_DIAMOND			56
#define EL_EMERALD_YELLOW		57
#define EL_WALL_BD_DIAMOND		58
#define EL_WALL_EMERALD_YELLOW		59
#define EL_DARK_YAMYAM			60
#define EL_BD_MAGIC_WALL		61
#define EL_INVISIBLE_STEELWALL		62
#define EL_SOKOBAN_FIELD_PLAYER		63
#define EL_DYNABOMB_INCREASE_NUMBER	64
#define EL_DYNABOMB_INCREASE_SIZE	65
#define EL_DYNABOMB_INCREASE_POWER	66
#define EL_SOKOBAN_OBJECT		67
#define EL_SOKOBAN_FIELD_EMPTY		68
#define EL_SOKOBAN_FIELD_FULL		69
#define EL_BD_BUTTERFLY_RIGHT		70
#define EL_BD_BUTTERFLY_UP		71
#define EL_BD_BUTTERFLY_LEFT		72
#define EL_BD_BUTTERFLY_DOWN		73
#define EL_BD_FIREFLY_RIGHT		74
#define EL_BD_FIREFLY_UP		75
#define EL_BD_FIREFLY_LEFT		76
#define EL_BD_FIREFLY_DOWN		77
#define EL_BD_BUTTERFLY_1		EL_BD_BUTTERFLY_DOWN
#define EL_BD_BUTTERFLY_2		EL_BD_BUTTERFLY_LEFT
#define EL_BD_BUTTERFLY_3		EL_BD_BUTTERFLY_UP
#define EL_BD_BUTTERFLY_4		EL_BD_BUTTERFLY_RIGHT
#define EL_BD_FIREFLY_1			EL_BD_FIREFLY_LEFT
#define EL_BD_FIREFLY_2			EL_BD_FIREFLY_DOWN
#define EL_BD_FIREFLY_3			EL_BD_FIREFLY_RIGHT
#define EL_BD_FIREFLY_4			EL_BD_FIREFLY_UP
#define EL_BD_BUTTERFLY			78
#define EL_BD_FIREFLY			79
#define EL_PLAYER_1			80
#define EL_PLAYER_2			81
#define EL_PLAYER_3			82
#define EL_PLAYER_4			83
#define EL_BUG_RIGHT			84
#define EL_BUG_UP			85
#define EL_BUG_LEFT			86
#define EL_BUG_DOWN			87
#define EL_SPACESHIP_RIGHT		88
#define EL_SPACESHIP_UP			89
#define EL_SPACESHIP_LEFT		90
#define EL_SPACESHIP_DOWN		91
#define EL_PACMAN_RIGHT			92
#define EL_PACMAN_UP			93
#define EL_PACMAN_LEFT			94
#define EL_PACMAN_DOWN			95
#define EL_EMERALD_RED			96
#define EL_EMERALD_PURPLE		97
#define EL_WALL_EMERALD_RED		98
#define EL_WALL_EMERALD_PURPLE		99
#define EL_ACID_POOL_TOPLEFT		100
#define EL_ACID_POOL_TOPRIGHT		101
#define EL_ACID_POOL_BOTTOMLEFT		102
#define EL_ACID_POOL_BOTTOM		103
#define EL_ACID_POOL_BOTTOMRIGHT	104
#define EL_BD_WALL			105
#define EL_BD_ROCK			106
#define EL_EXIT_OPEN			107
#define EL_BLACK_ORB			108
#define EL_AMOEBA_TO_DIAMOND		109
#define EL_MOLE				110
#define EL_PENGUIN			111
#define EL_SATELLITE			112
#define EL_ARROW_LEFT			113
#define EL_ARROW_RIGHT			114
#define EL_ARROW_UP			115
#define EL_ARROW_DOWN			116
#define EL_PIG				117
#define EL_DRAGON			118

#define EL_EM_KEY_1_FILE_OBSOLETE	119	/* obsolete; now EL_EM_KEY_1 */

#define EL_CHAR_START			120
#define EL_CHAR_ASCII0			(EL_CHAR_START  - 32)
#define EL_CHAR_ASCII0_START		(EL_CHAR_ASCII0 + 32)

#include "conf_chr.h"	/* include auto-generated data structure definitions */

#define EL_CHAR_ASCII0_END		(EL_CHAR_ASCII0 + 111)
#define EL_CHAR_END			(EL_CHAR_START  + 79)

#define EL_CHAR(c)			(EL_CHAR_ASCII0 + MAP_FONT_ASCII(c))

#define EL_EXPANDABLE_WALL_HORIZONTAL	200
#define EL_EXPANDABLE_WALL_VERTICAL	201
#define EL_EXPANDABLE_WALL_ANY		202

#define EL_EM_GATE_1			203
#define EL_EM_GATE_2			204
#define EL_EM_GATE_3			205
#define EL_EM_GATE_4			206

#define EL_EM_KEY_2_FILE_OBSOLETE	207	/* obsolete; now EL_EM_KEY_2 */
#define EL_EM_KEY_3_FILE_OBSOLETE	208	/* obsolete; now EL_EM_KEY_3 */
#define EL_EM_KEY_4_FILE_OBSOLETE	209	/* obsolete; now EL_EM_KEY_4 */

#define EL_SP_START			210
#define EL_SP_EMPTY_SPACE		(EL_SP_START + 0)
#define EL_SP_EMPTY			EL_SP_EMPTY_SPACE
#define EL_SP_ZONK			(EL_SP_START + 1)
#define EL_SP_BASE			(EL_SP_START + 2)
#define EL_SP_MURPHY			(EL_SP_START + 3)
#define EL_SP_INFOTRON			(EL_SP_START + 4)
#define EL_SP_CHIP_SINGLE		(EL_SP_START + 5)
#define EL_SP_HARDWARE_GRAY		(EL_SP_START + 6)
#define EL_SP_EXIT_CLOSED		(EL_SP_START + 7)
#define EL_SP_DISK_ORANGE		(EL_SP_START + 8)
#define EL_SP_PORT_RIGHT		(EL_SP_START + 9)
#define EL_SP_PORT_DOWN			(EL_SP_START + 10)
#define EL_SP_PORT_LEFT			(EL_SP_START + 11)
#define EL_SP_PORT_UP			(EL_SP_START + 12)
#define EL_SP_GRAVITY_PORT_RIGHT	(EL_SP_START + 13)
#define EL_SP_GRAVITY_PORT_DOWN		(EL_SP_START + 14)
#define EL_SP_GRAVITY_PORT_LEFT		(EL_SP_START + 15)
#define EL_SP_GRAVITY_PORT_UP		(EL_SP_START + 16)
#define EL_SP_SNIKSNAK			(EL_SP_START + 17)
#define EL_SP_DISK_YELLOW		(EL_SP_START + 18)
#define EL_SP_TERMINAL			(EL_SP_START + 19)
#define EL_SP_DISK_RED			(EL_SP_START + 20)
#define EL_SP_PORT_VERTICAL		(EL_SP_START + 21)
#define EL_SP_PORT_HORIZONTAL		(EL_SP_START + 22)
#define EL_SP_PORT_ANY			(EL_SP_START + 23)
#define EL_SP_ELECTRON			(EL_SP_START + 24)
#define EL_SP_BUGGY_BASE		(EL_SP_START + 25)
#define EL_SP_CHIP_LEFT			(EL_SP_START + 26)
#define EL_SP_CHIP_RIGHT		(EL_SP_START + 27)
#define EL_SP_HARDWARE_BASE_1		(EL_SP_START + 28)
#define EL_SP_HARDWARE_GREEN		(EL_SP_START + 29)
#define EL_SP_HARDWARE_BLUE		(EL_SP_START + 30)
#define EL_SP_HARDWARE_RED		(EL_SP_START + 31)
#define EL_SP_HARDWARE_YELLOW		(EL_SP_START + 32)
#define EL_SP_HARDWARE_BASE_2		(EL_SP_START + 33)
#define EL_SP_HARDWARE_BASE_3		(EL_SP_START + 34)
#define EL_SP_HARDWARE_BASE_4		(EL_SP_START + 35)
#define EL_SP_HARDWARE_BASE_5		(EL_SP_START + 36)
#define EL_SP_HARDWARE_BASE_6		(EL_SP_START + 37)
#define EL_SP_CHIP_TOP			(EL_SP_START + 38)
#define EL_SP_CHIP_BOTTOM		(EL_SP_START + 39)
#define EL_SP_END			(EL_SP_START + 39)

#define EL_EM_GATE_1_GRAY		250
#define EL_EM_GATE_2_GRAY		251
#define EL_EM_GATE_3_GRAY		252
#define EL_EM_GATE_4_GRAY		253

#define EL_UNUSED_254			254
#define EL_UNUSED_255			255

#define EL_PEARL			256
#define EL_CRYSTAL			257
#define EL_WALL_PEARL			258
#define EL_WALL_CRYSTAL			259
#define EL_DOOR_WHITE			260
#define EL_DOOR_WHITE_GRAY		261
#define EL_KEY_WHITE			262
#define EL_SHIELD_NORMAL		263
#define EL_EXTRA_TIME			264
#define EL_SWITCHGATE_OPEN		265
#define EL_SWITCHGATE_CLOSED		266
#define EL_SWITCHGATE_SWITCH_UP		267
#define EL_SWITCHGATE_SWITCH_DOWN	268

#define EL_UNUSED_269			269
#define EL_UNUSED_270			270

#define EL_CONVEYOR_BELT_1_LEFT		 271
#define EL_CONVEYOR_BELT_1_MIDDLE	 272
#define EL_CONVEYOR_BELT_1_RIGHT	 273
#define EL_CONVEYOR_BELT_1_SWITCH_LEFT	 274
#define EL_CONVEYOR_BELT_1_SWITCH_MIDDLE 275
#define EL_CONVEYOR_BELT_1_SWITCH_RIGHT	 276
#define EL_CONVEYOR_BELT_2_LEFT		 277
#define EL_CONVEYOR_BELT_2_MIDDLE	 278
#define EL_CONVEYOR_BELT_2_RIGHT	 279
#define EL_CONVEYOR_BELT_2_SWITCH_LEFT	 280
#define EL_CONVEYOR_BELT_2_SWITCH_MIDDLE 281
#define EL_CONVEYOR_BELT_2_SWITCH_RIGHT	 282
#define EL_CONVEYOR_BELT_3_LEFT		 283
#define EL_CONVEYOR_BELT_3_MIDDLE	 284
#define EL_CONVEYOR_BELT_3_RIGHT	 285
#define EL_CONVEYOR_BELT_3_SWITCH_LEFT	 286
#define EL_CONVEYOR_BELT_3_SWITCH_MIDDLE 287
#define EL_CONVEYOR_BELT_3_SWITCH_RIGHT	 288
#define EL_CONVEYOR_BELT_4_LEFT		 289
#define EL_CONVEYOR_BELT_4_MIDDLE	 290
#define EL_CONVEYOR_BELT_4_RIGHT	 291
#define EL_CONVEYOR_BELT_4_SWITCH_LEFT	 292
#define EL_CONVEYOR_BELT_4_SWITCH_MIDDLE 293
#define EL_CONVEYOR_BELT_4_SWITCH_RIGHT	 294
#define EL_LANDMINE			295
#define EL_ENVELOPE_OBSOLETE		296   /* obsolete; now EL_ENVELOPE_1 */
#define EL_LIGHT_SWITCH			297
#define EL_LIGHT_SWITCH_ACTIVE		298
#define EL_SIGN_EXCLAMATION		299
#define EL_SIGN_RADIOACTIVITY		300
#define EL_SIGN_STOP			301
#define EL_SIGN_WHEELCHAIR		302
#define EL_SIGN_PARKING			303
#define EL_SIGN_ONEWAY			304
#define EL_SIGN_HEART			305
#define EL_SIGN_TRIANGLE		306
#define EL_SIGN_ROUND			307
#define EL_SIGN_EXIT			308
#define EL_SIGN_YINYANG			309
#define EL_SIGN_OTHER			310
#define EL_MOLE_LEFT			311
#define EL_MOLE_RIGHT			312
#define EL_MOLE_UP			313
#define EL_MOLE_DOWN			314
#define EL_STEELWALL_SLIPPERY		315
#define EL_INVISIBLE_SAND		316
#define EL_DX_UNKNOWN_15		317
#define EL_DX_UNKNOWN_42		318

#define EL_UNUSED_319			319
#define EL_UNUSED_320			320

#define EL_SHIELD_DEADLY		321
#define EL_TIMEGATE_OPEN		322
#define EL_TIMEGATE_CLOSED		323
#define EL_TIMEGATE_SWITCH_ACTIVE	324
#define EL_TIMEGATE_SWITCH		325

#define EL_BALLOON			326
#define EL_BALLOON_SWITCH_LEFT		327
#define EL_BALLOON_SWITCH_RIGHT		328
#define EL_BALLOON_SWITCH_UP		329
#define EL_BALLOON_SWITCH_DOWN		330
#define EL_BALLOON_SWITCH_ANY		331

#define EL_EMC_STEELWALL_1		332
#define EL_EMC_STEELWALL_2		333
#define EL_EMC_STEELWALL_3 		334
#define EL_EMC_STEELWALL_4		335
#define EL_EMC_WALL_1			336
#define EL_EMC_WALL_2			337
#define EL_EMC_WALL_3			338
#define EL_EMC_WALL_4			339
#define EL_EMC_WALL_5			340
#define EL_EMC_WALL_6			341
#define EL_EMC_WALL_7			342
#define EL_EMC_WALL_8			343

#define EL_TUBE_ANY			344
#define EL_TUBE_VERTICAL		345
#define EL_TUBE_HORIZONTAL		346
#define EL_TUBE_VERTICAL_LEFT		347
#define EL_TUBE_VERTICAL_RIGHT		348
#define EL_TUBE_HORIZONTAL_UP		349
#define EL_TUBE_HORIZONTAL_DOWN		350
#define EL_TUBE_LEFT_UP			351
#define EL_TUBE_LEFT_DOWN		352
#define EL_TUBE_RIGHT_UP		353
#define EL_TUBE_RIGHT_DOWN		354
#define EL_SPRING			355
#define EL_TRAP				356
#define EL_DX_SUPABOMB			357

#define EL_UNUSED_358			358
#define EL_UNUSED_359			359

/* ---------- begin of custom elements section ----------------------------- */
#define EL_CUSTOM_START			360

//#include "conf_cus.h"	/* include auto-generated data structure definitions */

#define NUM_CUSTOM_ELEMENTS		256
#define EL_CUSTOM_END			615
/* ---------- end of custom elements section ------------------------------- */

#define EL_EM_KEY_1			616
#define EL_EM_KEY_2			617
#define EL_EM_KEY_3			618
#define EL_EM_KEY_4			619
#define EL_ENVELOPE_1			620
#define EL_ENVELOPE_2			621
#define EL_ENVELOPE_3			622
#define EL_ENVELOPE_4			623

/* ---------- begin of group elements section ------------------------------ */
#define EL_GROUP_START			624

//#include "conf_grp.h"	/* include auto-generated data structure definitions */

#define NUM_GROUP_ELEMENTS		32
#define EL_GROUP_END			655
/* ---------- end of custom elements section ------------------------------- */

#define EL_UNKNOWN			656
#define EL_TRIGGER_ELEMENT		657
#define EL_TRIGGER_PLAYER		658

#define EL_SP_GRAVITY_ON_PORT_RIGHT	659
#define EL_SP_GRAVITY_ON_PORT_DOWN	660
#define EL_SP_GRAVITY_ON_PORT_LEFT	661
#define EL_SP_GRAVITY_ON_PORT_UP	662
#define EL_SP_GRAVITY_OFF_PORT_RIGHT	663
#define EL_SP_GRAVITY_OFF_PORT_DOWN	664
#define EL_SP_GRAVITY_OFF_PORT_LEFT	665
#define EL_SP_GRAVITY_OFF_PORT_UP	666

#define NUM_FILE_ELEMENTS		667

/* "real" (and therefore drawable) runtime elements */
#define EL_FIRST_RUNTIME_REAL		NUM_FILE_ELEMENTS

#define EL_DYNABOMB_PLAYER_1_ACTIVE	(EL_FIRST_RUNTIME_REAL + 0)
#define EL_DYNABOMB_PLAYER_2_ACTIVE	(EL_FIRST_RUNTIME_REAL + 1)
#define EL_DYNABOMB_PLAYER_3_ACTIVE	(EL_FIRST_RUNTIME_REAL + 2)
#define EL_DYNABOMB_PLAYER_4_ACTIVE	(EL_FIRST_RUNTIME_REAL + 3)
#define EL_SP_DISK_RED_ACTIVE		(EL_FIRST_RUNTIME_REAL + 4)
#define EL_SWITCHGATE_OPENING		(EL_FIRST_RUNTIME_REAL + 5)
#define EL_SWITCHGATE_CLOSING		(EL_FIRST_RUNTIME_REAL + 6)
#define EL_TIMEGATE_OPENING		(EL_FIRST_RUNTIME_REAL + 7)
#define EL_TIMEGATE_CLOSING		(EL_FIRST_RUNTIME_REAL + 8)
#define EL_PEARL_BREAKING		(EL_FIRST_RUNTIME_REAL + 9)
#define EL_TRAP_ACTIVE			(EL_FIRST_RUNTIME_REAL + 10)
#define EL_INVISIBLE_STEELWALL_ACTIVE	(EL_FIRST_RUNTIME_REAL + 11)
#define EL_INVISIBLE_WALL_ACTIVE	(EL_FIRST_RUNTIME_REAL + 12)
#define EL_INVISIBLE_SAND_ACTIVE	(EL_FIRST_RUNTIME_REAL + 13)
#define EL_CONVEYOR_BELT_1_LEFT_ACTIVE	 (EL_FIRST_RUNTIME_REAL + 14)
#define EL_CONVEYOR_BELT_1_MIDDLE_ACTIVE (EL_FIRST_RUNTIME_REAL + 15)
#define EL_CONVEYOR_BELT_1_RIGHT_ACTIVE	 (EL_FIRST_RUNTIME_REAL + 16)
#define EL_CONVEYOR_BELT_2_LEFT_ACTIVE	 (EL_FIRST_RUNTIME_REAL + 17)
#define EL_CONVEYOR_BELT_2_MIDDLE_ACTIVE (EL_FIRST_RUNTIME_REAL + 18)
#define EL_CONVEYOR_BELT_2_RIGHT_ACTIVE	 (EL_FIRST_RUNTIME_REAL + 19)
#define EL_CONVEYOR_BELT_3_LEFT_ACTIVE	 (EL_FIRST_RUNTIME_REAL + 20)
#define EL_CONVEYOR_BELT_3_MIDDLE_ACTIVE (EL_FIRST_RUNTIME_REAL + 21)
#define EL_CONVEYOR_BELT_3_RIGHT_ACTIVE	 (EL_FIRST_RUNTIME_REAL + 22)
#define EL_CONVEYOR_BELT_4_LEFT_ACTIVE	 (EL_FIRST_RUNTIME_REAL + 23)
#define EL_CONVEYOR_BELT_4_MIDDLE_ACTIVE (EL_FIRST_RUNTIME_REAL + 24)
#define EL_CONVEYOR_BELT_4_RIGHT_ACTIVE	 (EL_FIRST_RUNTIME_REAL + 25)
#define EL_EXIT_OPENING			(EL_FIRST_RUNTIME_REAL + 26)
#define EL_EXIT_CLOSING			(EL_FIRST_RUNTIME_REAL + 27)
#define EL_SP_EXIT_OPENING		(EL_FIRST_RUNTIME_REAL + 28)
#define EL_SP_EXIT_CLOSING		(EL_FIRST_RUNTIME_REAL + 29)
#define EL_SP_EXIT_OPEN			(EL_FIRST_RUNTIME_REAL + 30)
#define EL_SP_TERMINAL_ACTIVE		(EL_FIRST_RUNTIME_REAL + 31)
#define EL_SP_BUGGY_BASE_ACTIVATING	(EL_FIRST_RUNTIME_REAL + 32)
#define EL_SP_BUGGY_BASE_ACTIVE		(EL_FIRST_RUNTIME_REAL + 33)
#define EL_SP_MURPHY_CLONE		(EL_FIRST_RUNTIME_REAL + 34)
#define EL_AMOEBA_DROPPING		(EL_FIRST_RUNTIME_REAL + 35)
#define EL_QUICKSAND_EMPTYING		(EL_FIRST_RUNTIME_REAL + 36)
#define EL_MAGIC_WALL_ACTIVE		(EL_FIRST_RUNTIME_REAL + 37)
#define EL_BD_MAGIC_WALL_ACTIVE		(EL_FIRST_RUNTIME_REAL + 38)
#define EL_MAGIC_WALL_FULL		(EL_FIRST_RUNTIME_REAL + 39)
#define EL_BD_MAGIC_WALL_FULL		(EL_FIRST_RUNTIME_REAL + 40)
#define EL_MAGIC_WALL_EMPTYING		(EL_FIRST_RUNTIME_REAL + 41)
#define EL_BD_MAGIC_WALL_EMPTYING	(EL_FIRST_RUNTIME_REAL + 42)
#define EL_MAGIC_WALL_DEAD		(EL_FIRST_RUNTIME_REAL + 43)
#define EL_BD_MAGIC_WALL_DEAD		(EL_FIRST_RUNTIME_REAL + 44)

/* "unreal" (and therefore not drawable) runtime elements */
#define EL_FIRST_RUNTIME_UNREAL		(EL_FIRST_RUNTIME_REAL + 45)

#define EL_BLOCKED			(EL_FIRST_RUNTIME_UNREAL + 0)
#define EL_EXPLOSION			(EL_FIRST_RUNTIME_UNREAL + 1)
#define EL_NUT_BREAKING			(EL_FIRST_RUNTIME_UNREAL + 2)
#define EL_DIAMOND_BREAKING		(EL_FIRST_RUNTIME_UNREAL + 3)
#define EL_ACID_SPLASH_LEFT		(EL_FIRST_RUNTIME_UNREAL + 4)
#define EL_ACID_SPLASH_RIGHT		(EL_FIRST_RUNTIME_UNREAL + 5)
#define EL_AMOEBA_GROWING		(EL_FIRST_RUNTIME_UNREAL + 6)
#define EL_AMOEBA_SHRINKING		(EL_FIRST_RUNTIME_UNREAL + 7)
#define EL_EXPANDABLE_WALL_GROWING	(EL_FIRST_RUNTIME_UNREAL + 8)
#define EL_FLAMES			(EL_FIRST_RUNTIME_UNREAL + 9)
#define EL_PLAYER_IS_LEAVING		(EL_FIRST_RUNTIME_UNREAL + 10)
#define EL_PLAYER_IS_EXPLODING_1	(EL_FIRST_RUNTIME_UNREAL + 11)
#define EL_PLAYER_IS_EXPLODING_2	(EL_FIRST_RUNTIME_UNREAL + 12)
#define EL_PLAYER_IS_EXPLODING_3	(EL_FIRST_RUNTIME_UNREAL + 13)
#define EL_PLAYER_IS_EXPLODING_4	(EL_FIRST_RUNTIME_UNREAL + 14)
#define EL_QUICKSAND_FILLING		(EL_FIRST_RUNTIME_UNREAL + 15)
#define EL_MAGIC_WALL_FILLING		(EL_FIRST_RUNTIME_UNREAL + 16)
#define EL_BD_MAGIC_WALL_FILLING	(EL_FIRST_RUNTIME_UNREAL + 17)

/* dummy elements (never used as game elements, only used as graphics) */
#define EL_FIRST_DUMMY			(EL_FIRST_RUNTIME_UNREAL + 18)

#define EL_STEELWALL_TOPLEFT			(EL_FIRST_DUMMY + 0)
#define EL_STEELWALL_TOPRIGHT			(EL_FIRST_DUMMY + 1)
#define EL_STEELWALL_BOTTOMLEFT			(EL_FIRST_DUMMY + 2)
#define EL_STEELWALL_BOTTOMRIGHT		(EL_FIRST_DUMMY + 3)
#define EL_STEELWALL_HORIZONTAL			(EL_FIRST_DUMMY + 4)
#define EL_STEELWALL_VERTICAL			(EL_FIRST_DUMMY + 5)
#define EL_INVISIBLE_STEELWALL_TOPLEFT		(EL_FIRST_DUMMY + 6)
#define EL_INVISIBLE_STEELWALL_TOPRIGHT		(EL_FIRST_DUMMY + 7)
#define EL_INVISIBLE_STEELWALL_BOTTOMLEFT	(EL_FIRST_DUMMY + 8)
#define EL_INVISIBLE_STEELWALL_BOTTOMRIGHT	(EL_FIRST_DUMMY + 9)
#define EL_INVISIBLE_STEELWALL_HORIZONTAL	(EL_FIRST_DUMMY + 10)
#define EL_INVISIBLE_STEELWALL_VERTICAL		(EL_FIRST_DUMMY + 11)
#define EL_DYNABOMB				(EL_FIRST_DUMMY + 12)
#define EL_DYNABOMB_ACTIVE			(EL_FIRST_DUMMY + 13)
#define EL_DYNABOMB_PLAYER_1			(EL_FIRST_DUMMY + 14)
#define EL_DYNABOMB_PLAYER_2			(EL_FIRST_DUMMY + 15)
#define EL_DYNABOMB_PLAYER_3			(EL_FIRST_DUMMY + 16)
#define EL_DYNABOMB_PLAYER_4			(EL_FIRST_DUMMY + 17)
#define EL_SHIELD_NORMAL_ACTIVE			(EL_FIRST_DUMMY + 18)
#define EL_SHIELD_DEADLY_ACTIVE			(EL_FIRST_DUMMY + 19)
#define EL_AMOEBA				(EL_FIRST_DUMMY + 20)
#define EL_DEFAULT				(EL_FIRST_DUMMY + 21)
#define EL_BD_DEFAULT				(EL_FIRST_DUMMY + 22)
#define EL_SP_DEFAULT				(EL_FIRST_DUMMY + 23)
#define EL_SB_DEFAULT				(EL_FIRST_DUMMY + 24)

/* internal elements (only used for internal purposes like copying) */
#define EL_FIRST_INTERNAL			(EL_FIRST_DUMMY + 25)

#define EL_INTERNAL_CLIPBOARD_CUSTOM		(EL_FIRST_INTERNAL + 0)
#define EL_INTERNAL_CLIPBOARD_CHANGE		(EL_FIRST_INTERNAL + 1)
#define EL_INTERNAL_CLIPBOARD_GROUP		(EL_FIRST_INTERNAL + 2)
#define EL_INTERNAL_DUMMY			(EL_FIRST_INTERNAL + 3)

#define EL_INTERNAL_CLIPBOARD_START		(EL_FIRST_INTERNAL + 0)
#define EL_INTERNAL_CLIPBOARD_END		(EL_FIRST_INTERNAL + 2)
#define EL_INTERNAL_START			(EL_FIRST_INTERNAL + 0)
#define EL_INTERNAL_END				(EL_FIRST_INTERNAL + 3)

#define MAX_NUM_ELEMENTS			(EL_FIRST_INTERNAL + 4)

/* values for graphics/sounds action types */
#define ACTION_DEFAULT				0
#define ACTION_WAITING				1
#define ACTION_FALLING				2
#define ACTION_MOVING				3
#define ACTION_DIGGING				4
#define ACTION_SNAPPING				5
#define ACTION_COLLECTING			6
#define ACTION_DROPPING				7
#define ACTION_PUSHING				8
#define ACTION_WALKING				9
#define ACTION_PASSING				10
#define ACTION_IMPACT				11
#define ACTION_BREAKING				12
#define ACTION_ACTIVATING			13
#define ACTION_DEACTIVATING			14
#define ACTION_OPENING				15
#define ACTION_CLOSING				16
#define ACTION_ATTACKING			17
#define ACTION_GROWING				18
#define ACTION_SHRINKING			19
#define ACTION_ACTIVE				20
#define ACTION_FILLING				21
#define ACTION_EMPTYING				22
#define ACTION_CHANGING				23
#define ACTION_EXPLODING			24
#define ACTION_BORING				25
#define ACTION_BORING_1				26
#define ACTION_BORING_2				27
#define ACTION_BORING_3				28
#define ACTION_BORING_4				29
#define ACTION_BORING_5				30
#define ACTION_BORING_6				31
#define ACTION_BORING_7				32
#define ACTION_BORING_8				33
#define ACTION_BORING_9				34
#define ACTION_BORING_10			35
#define ACTION_SLEEPING				36
#define ACTION_SLEEPING_1			37
#define ACTION_SLEEPING_2			38
#define ACTION_SLEEPING_3			39
#define ACTION_AWAKENING			40
#define ACTION_DYING				41
#define ACTION_TURNING				42
#define ACTION_TURNING_FROM_LEFT		43
#define ACTION_TURNING_FROM_RIGHT		44
#define ACTION_TURNING_FROM_UP			45
#define ACTION_TURNING_FROM_DOWN		46
#define ACTION_OTHER				47

#define NUM_ACTIONS				48

#define ACTION_BORING_LAST			ACTION_BORING_10
#define ACTION_SLEEPING_LAST			ACTION_SLEEPING_3 

/* program information and versioning definitions */

#define PROGRAM_VERSION_MAJOR	3
#define PROGRAM_VERSION_MINOR	1
#define PROGRAM_VERSION_PATCH	0
#define PROGRAM_VERSION_BUILD	5

/* file version numbers for resource files (levels, tapes, score, setup, etc.)
** currently supported/known file version numbers:
**	1.0 (old)
**	1.2 (still in use)
**	1.4 (still in use)
**	2.0 (actual)
*/
#define FILE_VERSION_1_0	VERSION_IDENT(1,0,0,0)
#define FILE_VERSION_1_2	VERSION_IDENT(1,2,0,0)
#define FILE_VERSION_1_4	VERSION_IDENT(1,4,0,0)
#define FILE_VERSION_2_0	VERSION_IDENT(2,0,0,0)

/* file version does not change for every program version, but is changed
   when new features are introduced that are incompatible with older file
   versions, so that they can be treated accordingly */
#define FILE_VERSION_ACTUAL	FILE_VERSION_2_0

#define GAME_VERSION_ACTUAL	VERSION_IDENT(PROGRAM_VERSION_MAJOR, \
										  PROGRAM_VERSION_MINOR, \
										  PROGRAM_VERSION_PATCH, \
										  PROGRAM_VERSION_BUILD)

/* values for game_emulation */
#define EMU_NONE		0
#define EMU_BOULDERDASH		1
#define EMU_SOKOBAN		2
#define EMU_SUPAPLEX		3

struct DoorInfo
{
  int step_offset;
  int step_delay;
  int anim_mode;
};

struct PlayerInfo
{
  boolean present;		/* player present in level playfield */
  boolean connected;		/* player connected (locally or via network) */
  boolean active;		/* player present and connected */

  int index_nr;			/* player number (0 to 3) */
  int index_bit;		/* player number bit (1 << 0 to 1 << 3) */
  int element_nr;		/* element (EL_PLAYER_1 to EL_PLAYER_4) */
//  int client_nr;		/* network client identifier */
//PALM no

  byte action;			/* action from local input device */
  byte effective_action;	/* action acknowledged from network server
				   or summarized over all configured input
				   devices when in single player mode */
  byte programmed_action;	/* action forced by game itself (like moving
				   through doors); overrides other actions */

  int jx, jy, last_jx, last_jy;
  int MovDir, MovPos, GfxDir, GfxPos;
  int Frame, StepFrame;

  int GfxAction;

  boolean use_murphy_graphic;

  boolean block_last_field;
  boolean can_fall_into_acid;

  boolean LevelSolved, GameOver;

  int last_move_dir;

  boolean is_waiting;
  boolean is_moving;
  boolean is_auto_moving;
  boolean is_digging;
  boolean is_snapping;
  boolean is_collecting;
  boolean is_pushing;
  boolean is_switching;
  boolean is_dropping;

  boolean is_bored;
  boolean is_sleeping;

  int frame_counter_bored;
  int frame_counter_sleeping;

  int anim_delay_counter;
  int post_delay_counter;

  int action_waiting, last_action_waiting;
  int special_action_bored;
  int special_action_sleeping;

  int num_special_action_bored;
  int num_special_action_sleeping;

  int switch_x, switch_y;

  int show_envelope;

  unsigned long move_delay;
  int move_delay_value;

  int move_delay_reset_counter;

  unsigned long push_delay;
  unsigned long push_delay_value;

  unsigned long actual_frame_counter;

  int drop_delay;

  int step_counter;

  int score;
  int gems_still_needed;
  int sokobanfields_still_needed;
  int lights_still_needed;
  int friends_still_needed;
  int key[4];
  int dynabomb_count, dynabomb_size, dynabombs_left, dynabomb_xl;
  int shield_normal_time_left;
  int shield_deadly_time_left;

  int inventory_element[MAX_INVENTORY_SIZE];
  int inventory_infinite_element;
  int inventory_size;
};

struct LevelInfo
{
  long file_version;	/* file format version the level is stored with    */
// for palm, was int
  long game_version;	/* game release version the level was created with */
// for palm, was int
  
  boolean encoding_16bit_field;		/* level contains 16-bit elements  */
  boolean encoding_16bit_yamyam;	/* yamyam contains 16-bit elements */
  boolean encoding_16bit_amoeba;	/* amoeba contains 16-bit elements */

  byte fieldx, fieldy;
// for palm, was int

  int time;
  int gems_needed;

  char name[MAX_LEVEL_NAME_LEN + 1];
/*  char author[MAX_LEVEL_AUTHOR_LEN + 1];

  char envelope_text[4][MAX_ENVELOPE_TEXT_LEN + 1];*/
  int envelope_xsize[4], envelope_ysize[4];

  int score[LEVEL_SCORE_ELEMENTS];

  int yamyam_content[MAX_ELEMENT_CONTENTS][3][3];
  int num_yamyam_contents;

  int amoeba_speed;
  int amoeba_content;

  int time_magic_wall;
  int time_wheel;
  int time_light;
  int time_timegate;

  long can_move_into_acid_bits;	/* bitfield to store property for elements */
  long dont_collide_with_bits;	/* bitfield to store property for elements */

  boolean double_speed;
  boolean initial_gravity;
  boolean em_slippery_gems;	// EM style "gems slip from wall" behaviour
  boolean block_last_field;	/* player blocks previous field while moving */
  boolean sp_block_last_field;	/* player blocks previous field while moving */
  boolean use_spring_bug;	/* for compatibility with old levels */
  boolean instant_relocation;	/* no visual delay when relocating player */
  boolean can_pass_to_walkable;	/* player can pass to empty or walkable tile */
  boolean grow_into_diggable;	/* amoeba can grow into anything diggable */

  /* ('int' instead of 'boolean' because used as selectbox value in editor) */
  int use_step_counter;		/* count steps instead of seconds for level */

//  short field[MAX_LEV_FIELDX][MAX_LEV_FIELDY];
//for palm only use Feld instead of this

  boolean use_custom_template;	// use custom properties from template file

  boolean no_valid_file;	/* set when level file missing or invalid */
};

struct TapeInfo
{
  long file_version;	/* file format version the tape is stored with    */
  //changed for palm for long was int
  long game_version;	/* game release version the tape was created with */
  //changed for palm for long was int
  long engine_version;	/* game engine version the tape was recorded with */
  //changed for palm for long was int

  char *level_identifier;
  int level_nr;
  unsigned long random_seed;
  unsigned long date;
  unsigned long counter;
  unsigned long length;
  unsigned long length_seconds;
  unsigned int delay_played;
  boolean pause_before_death;
  boolean recording, playing, pausing;
  boolean fast_forward;
  boolean warp_forward;
  boolean deactivate_display;
//  boolean auto_play;				// no autoplaying in palm
//  boolean auto_play_level_solved;	// no autoplaying in palm
  boolean quick_resume;
  boolean single_step;
  boolean changed;
  boolean player_participates[4];
  int num_participating_players;

  byte pos_action_[MAX_TAPELEN];
  byte pos_delay[MAX_TAPELEN];

  boolean no_valid_file;	// set when tape file missing or invalid
}; 

struct GameInfo
{
  /* values for engine initialization */
  int default_push_delay_fixed;
  int default_push_delay_random;

  /* constant within running game */
  long engine_version; //long for palm, was int
  int emulation;
  int initial_move_delay;
  int initial_move_delay_value;
  int initial_push_delay_value;

  /* variable within running game */
  int yamyam_content_nr;
  boolean magic_wall_active;
  int magic_wall_time_left;
  int light_time_left;
  int timegate_time_left;
  int belt_dir[4];
  int belt_dir_nr[4];
  int switchgate_pos;
  int balloon_dir;
  boolean gravity;
  boolean explosions_delayed;
  boolean envelope_active;

  /* values for player idle animation (no effect on engine) */
  int player_boring_delay_fixed;
  int player_boring_delay_random;
  int player_sleeping_delay_fixed;
  int player_sleeping_delay_random;
};

struct GlobalInfo
{
	char *autoplay_leveldir;
	int autoplay_level_nr;

//	char *convert_leveldir;
//	int convert_level_nr;

	int num_toons;

//  float frames_per_second;
	boolean fps_slowdown;
	int fps_slowdown_factor;
};

struct ElementInfo
{/*
  // ---------- token and description strings ----------

  char *token_name;		// element token used in config files
  char *class_name;		// element class used in config files
  char *editor_description;	// pre-defined description for level editor
  char *custom_description;	// alternative description from config file
  char description[MAX_ELEMENT_NAME_LEN + 1];	// for custom/group elements

  // ---------- graphic and sound definitions ----------
*/
	short int	def_graphic[MAX_NUM_ELEMENTS + 1];
//was  int graphic[NUM_ACTIONS];	// default graphics for several actions

	byte other_graphic_index[MAX_NUM_ELEMENTS + 1];
	byte other_graphic_size[MAX_NUM_ELEMENTS + 1];
/*  int direction_graphic[NUM_ACTIONS][NUM_DIRECTIONS];
				// special graphics for left/right/up/down 

  int crumbled[NUM_ACTIONS];	// crumbled graphics for several actions
  int direction_crumbled[NUM_ACTIONS][NUM_DIRECTIONS];
				// crumbled graphics for left/right/up/down

  int special_graphic[NUM_SPECIAL_GFX_ARGS];
  				// special graphics for certain screens

  int sound[NUM_ACTIONS];	// default sounds for several actions

  // ---------- special element property values ----------

  boolean use_gfx_element;	// use custom graphic element
  short gfx_element;		// optional custom graphic element

  int access_direction;		// accessible from which direction

  int collect_score;		// score value for collecting
  int collect_count;		// count value for collecting

  int push_delay_fixed;		// constant delay before pushing
  int push_delay_random;	// additional random delay before pushing
  int drop_delay_fixed;		// constant delay after dropping
  int drop_delay_random;	// additional random delay after dropping
  int move_delay_fixed;		// constant delay after moving
  int move_delay_random;	// additional random delay after moving

  int move_pattern;		// direction movable element moves to
  int move_direction_initial;	// initial direction element moves to
  int move_stepsize;		// step size element moves with

  int move_enter_element;	// element that can be entered (and removed)
  int move_leave_element;	// element that can be left behind
  int move_leave_type;		// change (limited) or leave (unlimited)

  int slippery_type;		// how/where other elements slip away

  int content[3][3];		// new elements after explosion

  int explosion_type;		// type of explosion, like 3x3, 3+3 or 1x1
  int explosion_delay;		// duration of explosion of this element
  int ignition_delay;		// delay for explosion by other explosion

  struct ElementChangeInfo *change_page; // actual list of change pages
  struct ElementChangeInfo *change;	 // pointer to current change page

  int num_change_pages;		// actual number of change pages
  int current_change_page;	// currently edited change page

  struct ElementGroupInfo *group;	// pointer to element group info

  // ---------- internal values used at runtime when playing ----------

  unsigned long change_events;	// bitfield for combined change events

  int event_page_nr[NUM_CHANGE_EVENTS]; // page number for each event
  struct ElementChangeInfo *event_page[NUM_CHANGE_EVENTS]; // page for event

  boolean in_group[NUM_GROUP_ELEMENTS];

  // ---------- internal values used in level editor ----------

  int access_type;		// walkable or passable
  int access_layer;		// accessible over/inside/under
  int access_protected;		// protection against deadly elements
  int walk_to_action;		// diggable/collectible/pushable
  int smash_targets;		// can smash player/enemies/everything
  int deadliness;		// deadly when running/colliding/touching

  boolean can_explode_by_fire;	// element explodes by fire
  boolean can_explode_smashed;	// element explodes when smashed
  boolean can_explode_impact;	// element explodes on impact

  boolean modified_settings;	// set for all modified custom elements*/
};

#define NUM_EG	150

struct ElementGraphicInfo
{
	char action[NUM_EG+2];
	char direction[NUM_EG+2];
	short int graphic[NUM_EG+2];
};

struct GraphicInfo
{
//  Bitmap *bitmap;
	byte	bmp_no[NUM_IMAGE_FILES];
	byte	x_y[NUM_IMAGE_FILES];
/*  int src_x, src_y;		// start position of animation frames
  int width, height;		// width/height of each animation frame
  int offset_x, offset_y;	// x/y offset to next animation frame*/
	byte anim_index[NUM_IMAGE_FILES];//was int anim_frames
/*  int anim_frames_per_line;
  int anim_start_frame;
  int anim_delay;		// important: delay of 1 means "no delay"!
  int anim_mode;
  boolean anim_global_sync;
  int crumbled_like;		// element for cloning crumble graphics
  int diggable_like;		// element for cloning digging graphics
  int border_size;		// border size for "crumbled" graphics

  int anim_delay_fixed;		// optional delay values for bored and
  int anim_delay_random;	// sleeping player animations (animation
  int post_delay_fixed;		// intervall and following pause before
  int post_delay_random;	// next intervall (bored animation only)

  int step_offset;		// optional step offset of toon animations
  int step_delay;		// optional step delay of toon animations

  int draw_x, draw_y;		// optional offset for drawing fonts chars

  int draw_masked;		// optional setting for drawing envelope gfx
*/
};

#define NUM_ANIMATIONS	100

struct AnimationInfo
{
	byte anim_frames[NUM_ANIMATIONS];
	byte anim_start_frame[NUM_ANIMATIONS];
	byte anim_delay_[NUM_ANIMATIONS];//if DELAY_MILLION then 1000000
	byte anim_mode[NUM_ANIMATIONS];
	boolean anim_global_sync[NUM_ANIMATIONS];
};

extern int			game_status;

extern char				redraw_x1, redraw_y1;//was int 

extern short			Feld[MAX_LEV_FIELDX*MAX_LEV_FIELDY];

extern unsigned char	FeldYStart[MAX_LEV_FIELDY];
extern unsigned char	FeldYEnd[MAX_LEV_FIELDY];

extern char				MovPos[MAX_LEV_FIELDX*MAX_LEV_FIELDY]; //was short 
extern unsigned char	MovDir[MAX_LEV_FIELDX*MAX_LEV_FIELDY];//was short
extern unsigned short 	* MovDelay;

extern short			* Store;
extern short			* Back;
extern short			* Store2;
extern unsigned char 	FFlags[MAX_LEV_FIELDX*MAX_LEV_FIELDY];

#define FF_STOP		1<<0	//Stop[x][y]
#define FF_PUSHED	1<<1	//Pushed[x][y]
#define FF_REDRAW	1<<2	//redraw[x][y]
#define FF_WASJUST_STEP		1<<3
#define FF_WASJUST_MASK		3<<3	//WasJustMoving[x][x]
#define FF_WASJUST_FALL		1<<5	//WasJustFalling or WasJustMoving

extern unsigned char	StopYBegin[MAX_LEV_FIELDY];
extern unsigned char	StopYEnd[MAX_LEV_FIELDY];

#define setStopTRUE(x,y)	{FFlags[A2TO1(x,y)]|= FF_STOP;		\
	if(StopYBegin[y]>x)StopYBegin[y]=x; if(StopYEnd[y]<x)StopYEnd[y]=x;}

#define setStopFALSE(xy)	{FFlags[xy]&= ~FF_STOP;}
#define getStop(xy)			(FFlags[xy]&FF_STOP)

#define setPushedTRUE(xy)	{FFlags[xy]|= FF_PUSHED;}
#define setPushedFALSE(xy)	{FFlags[xy]&= ~FF_PUSHED;}
#define getPushed(xy)		(FFlags[xy]&FF_PUSHED)

#define setRedrawTRUE(xy)	{FFlags[xy]|= FF_REDRAW;}
#define setRedrawFALSE(xy)	{FFlags[xy]&= ~FF_REDRAW;}
#define getRedraw(xy)		(FFlags[xy]&FF_REDRAW)

#define IsWasJustMoving(xy)		(FFlags[xy]&FF_WASJUST_MASK)
#define IsWasJustFalling(xy)  ((FFlags[xy]&FF_WASJUST_MASK)&&(FFlags[xy]&FF_WASJUST_FALL))

extern unsigned char	WasJustYBegin[MAX_LEV_FIELDY];
extern unsigned char	WasJustYEnd[MAX_LEV_FIELDY];

extern unsigned char RedrawYCnt[MAX_BUF_YSIZE+2];
//extern unsigned char RedrawXCnt[MAX_BUF_XSIZE+2];

extern unsigned char	*AmoebaNr;
extern short			AmoebaCnt[MAX_NUM_AMOEBA];
extern short			AmoebaCnt2[MAX_NUM_AMOEBA];

extern unsigned char 	* ExplodeFlags;

extern unsigned char	FeldYExplosion[MAX_LEV_FIELDY];
extern unsigned char	ExplodeFieldY[MAX_LEV_FIELDY];

#define EF_EXP_PHASE	31
#define	EF_EXP_FIELD_SHIFT	5
#define EF_EXP_FIELD	3<<EF_EXP_FIELD_SHIFT

#define getExplodePhase(xy) (MovDelay[xy]) 
//seems to be OK, was (ExplodeFlags[xy]&EF_EXP_PHASE)
#define setExplodePhase(xy,val) {MovDelay[xy]=val;}
//seems to be OK, was {ExplodeFlags[xy]=(ExplodeFlags[xy]&EF_EXP_FIELD)|val;}
#define setExplodeField(xy,y,val)	\
		{ExplodeFlags[xy]=(ExplodeFlags[xy]&EF_EXP_PHASE)|(val<<EF_EXP_FIELD_SHIFT);	\
		ExplodeFieldY[y]=1;}
#define getExplodeField(xy)	((ExplodeFlags[xy]&EF_EXP_FIELD)>>EF_EXP_FIELD_SHIFT)

#define A2TO1(x,y)		(x+((y)*MAX_LEV_FIELDX))//x=64
#define A1TOX(xy)		(xy&63)
#define A1TOY(xy)		(xy>>6)

extern unsigned char	PropertiesF[NUM_EP_BITFIELDS][MAX_NUM_ELEMENTS];
//changed for palm was long and now char

extern short int	*GfxElement;
extern char			GfxAction[MAX_LEV_FIELDX*MAX_LEV_FIELDY];
//changed for palm was int and now char
//extern short int	*Gfx;
extern unsigned short int		*GfxSFrame;
extern unsigned char			*GfxRandom;

extern short int 		StorePlayerXY;

extern int			lev_fieldx, lev_fieldy;
extern int			scroll_x, scroll_y;

extern int			FX, FY;
extern int			ScrollStepSize;
extern int			ScreenMovDir, ScreenMovPos, ScreenGfxPos; 
extern int			BorderElement;
extern int			GameFrameDelay;
extern int			FfwdFrameDelay;
extern int			BX1, BY1;
extern int			BX2, BY2;
extern int			SBX_Left, SBX_Right;
extern int			SBY_Upper, SBY_Lower;
extern int			ZX, ZY;
extern int			ExitX, ExitY; 
extern int			AllPlayersGone;
extern int			TimeFrames, TimePlayed, TimeLeft, TapeTime;

extern struct LevelInfo		level/*, level_template*/;
extern struct PlayerInfo	stored_player_, *local_player;//was stored_player[]
extern struct TapeInfo		tape;
extern struct GameInfo		game;
extern struct GlobalInfo	global;

extern struct DoorInfo		door_1, door_2;
extern struct ElementInfo	*element_infoF;//was []
extern struct ElementGraphicInfo	*element_graphic_infoF;

extern struct GraphicInfo      *graphic_infoF;
extern struct AnimationInfo		*anim_infoF;

extern long gfx_anim_random_frame;

#define		MDL_OTHER				0
#define		MDL_INACTIVE_NOANIM		1
#define		MDL_CLASSIC_FALL		2
#define		MDL_FALL_OR_MOVE		3
#define		MDL_ANIM_ONLY			4
#define		MDL_CONT_MOVE_ONLY		5
#define		MDL_GAME_OF_LIFE		6
#define		MDL_EXP_WALL_GROW		7
#define		MDL_EXP_WALL			8
#define		MDL_EXIT_CLOSED			9
#define		MDL_SP_EXIT_CLOSED		10
#define		MDL_FLAMES				11
#define		MDL_EXPLOSION			12
#define		MDL_AMOEBA_GROW			13
#define		MDL_AMOEBA_SHRINK		14
#define		MDL_AMOEBALIVE			15
#define		MDL_CHANGING			16			

extern unsigned char MainDoLoop[MAX_NUM_ELEMENTS + 1];

void bugd(char*,long);

//my need

#define MAXSA	20
/*
struct SmallArray
{
	short int size;
	short int x[MAXSA];
	short int y[MAXSA];
	short int value[MAXSA];
};

#define SAQuickClear(Arr)	{Arr.size=0;}

short int SAGetValue(struct SmallArray *,short int, short int);
void SASetValue(struct SmallArray *,short int, short int, short int);
*/
extern unsigned char frames50;
extern unsigned long oldf50,newf50,time50;
extern unsigned long countStartMoving,last50StartMoving;
extern unsigned long countBigStartMoving,last50BigStartMoving;
extern unsigned long total50tiles,last50tiles;

extern boolean DS_Cheating,DS_MDL_Other;
extern int DS_TestRender,DS_TestMainLoop;
extern int waiter;

extern long for_bugd;

void TellGameWon(void) INIT_SECTION;

#define SCROLL_SCREEN_X			SCR_FIELDX*2/5
#define SCROLL_SCREEN_Y			SCR_FIELDY*2/5

#endif
