#include "tape.h"
#include "game.h"
#include "tools.h"
#include "files.h"

/* values for tape handling */
#define TAPE_PAUSE_SECONDS_BEFORE_DEATH		5 

static void TapeStopWarpForward(void);

static void TapeDeactivateDisplayOn(void)
{
//	SetDrawDeactivationMask(REDRAW_FIELD);
}

static void TapeDeactivateDisplayOff(boolean redraw_display)
{
//	SetDrawDeactivationMask(REDRAW_NONE);

	if (redraw_display)
	{
//		RedrawPlayfield(TRUE, 0,0,0,0);
		DrawLevel();
		DrawAllPlayers();

		BackToFront();
	}
}

// ========================================================================= 
// tape control functions                                                    
// ========================================================================= 

void TapeErase(void)
{
	tape.length = 0;
	tape.counter = 0;
/*
	if (leveldir_current)
		setString(&tape.level_identifier, leveldir_current->identifier);
*/
	tape.level_nr = level_nr;
	tape.pos_delay[tape.counter] = 0;
	tape.changed = TRUE;

//	tape.date = 10000*(time->tm_year % 100) + 100*time->tm_mon + time->tm_mday;
	tape.random_seed = InitRND(NEW_RANDOMIZE);

	tape.file_version = FILE_VERSION_ACTUAL;
	tape.game_version = GAME_VERSION_ACTUAL;
	tape.engine_version = level.game_version;

//	TapeSetDate();
	
	tape.player_participates[0] = 
	tape.player_participates[1] =
	tape.player_participates[2] =
	tape.player_participates[3] = FALSE;
}

static void TapeRewind(void)
{
	tape.counter = 0;
	tape.delay_played = 0;
	tape.pause_before_death = FALSE;
	tape.recording = FALSE;
	tape.playing = FALSE;
	tape.fast_forward = FALSE;
	tape.warp_forward = FALSE;
//	tape.deactivate_display = FALSE;
//	tape.auto_play = (global.autoplay_leveldir != NULL);
//	tape.auto_play_level_solved = FALSE;
	tape.quick_resume = FALSE;
	tape.single_step = FALSE;

	InitRND(tape.random_seed);
}

void TapeStartRecording(void)
{
	if (!TAPE_IS_STOPPED(tape))
		TapeStop();

	TapeErase();
	TapeRewind();

	tape.recording = TRUE;

/*	DrawVideoDisplay(VIDEO_STATE_REC_ON, 0);
	DrawVideoDisplay(VIDEO_STATE_DATE_ON, tape.date);
	DrawVideoDisplay(VIDEO_STATE_TIME_ON, 0);
	MapTapeWarpButton();

	SetDrawDeactivationMask(REDRAW_NONE);*/
}

void TapeAppendRecording(void)
{
	if (!tape.playing || !tape.pausing)
		return;

	tape.pos_delay[tape.counter] = tape.delay_played;
	tape.playing = FALSE;
	tape.recording = TRUE;
	tape.changed = TRUE;
/*
	TapeSetDate();

	DrawVideoDisplay(VIDEO_STATE_DATE_ON, tape.date);
	DrawVideoDisplay(VIDEO_STATE_PLAY_OFF | VIDEO_STATE_REC_ON, 0);*/
}

void TapeHaltRecording(void)
{
	if (!tape.recording)
		return;

	tape.counter++;
	tape.pos_delay[tape.counter] = 0;

	tape.length = tape.counter;
//	tape.length_seconds = GetTapeLength();
}

void TapeStopRecording(void)
{
	if (!tape.recording)
		return;

	TapeHaltRecording();

	tape.recording = FALSE;
	tape.pausing = FALSE;

/*	DrawVideoDisplay(VIDEO_STATE_REC_OFF, 0);
	MapTapeEjectButton();*/
}

void TapeRecordAction(byte action_)
{
	if (!tape.recording)		// record action even when tape is paused!
		return;

	if (tape.counter >= MAX_TAPELEN - 1)
	{
		TapeStopRecording();
		return;
	}

	if (tape.pos_delay[tape.counter] > 0)		// already stored action
	{
		boolean changed_events = FALSE;

		if (tape.pos_action_[tape.counter] != action_)
			changed_events = TRUE;

		if (changed_events || tape.pos_delay[tape.counter] >= 255)
		{
			tape.counter++;
			tape.pos_delay[tape.counter] = 0;
		}
		else
			tape.pos_delay[tape.counter]++;
	}

	if (tape.pos_delay[tape.counter] == 0)	// store new action
	{
		tape.pos_action_[tape.counter] = action_;

		tape.pos_delay[tape.counter]++;
	}
}

void TapeTogglePause(boolean toggle_manual)
{
/*	int state;

	if (tape.pause_before_death)
		state |= VIDEO_STATE_PBEND_OFF;
	else if (tape.fast_forward)
		state |= VIDEO_STATE_FFWD_OFF;
*/
	tape.pausing = !tape.pausing;
	tape.fast_forward = FALSE;
	tape.pause_before_death = FALSE;

	if (tape.single_step && toggle_manual)
		tape.single_step = FALSE;
/*
	state |= (tape.pausing ? VIDEO_STATE_PAUSE_ON : VIDEO_STATE_PAUSE_OFF);
	if (tape.playing)
		state |= VIDEO_STATE_PLAY_ON;
	else
		state |= (tape.single_step ? VIDEO_STATE_1STEP_ON : VIDEO_STATE_1STEP_OFF);

	DrawVideoDisplay(state, 0);
*/
	if (tape.warp_forward)
	{
		TapeStopWarpForward();

		if (tape.quick_resume)
		{
			tape.quick_resume = FALSE;

			TapeAppendRecording();
		}
	}
}

static void TapeStartPlaying(void)
{
	if (TAPE_IS_EMPTY(tape))
		return;

	if (!TAPE_IS_STOPPED(tape))
		TapeStop();

	TapeRewind();

	tape.playing = TRUE;
/*
	DrawVideoDisplay(VIDEO_STATE_PLAY_ON, 0);
	DrawVideoDisplay(VIDEO_STATE_DATE_ON, tape.date);
	DrawVideoDisplay(VIDEO_STATE_TIME_ON, 0);
	MapTapeWarpButton();

	SetDrawDeactivationMask(REDRAW_NONE);*/
}

static void TapeStartGamePlaying(void)
{
	TapeStartPlaying();

	game_status = GAME_MODE_PLAYING;
//	StopAnimation();
	InitGame();
}

static void TapeStopPlaying(void)
{
	if (!tape.playing)
		return;

	tape.playing = FALSE;
	tape.pausing = FALSE;

	if (tape.warp_forward)
		TapeStopWarpForward();

/*	DrawVideoDisplay(VIDEO_STATE_PLAY_OFF, 0);
	MapTapeEjectButton();*/
}

byte TapePlayAction(void)
{
	static byte action_;

	if (!tape.playing || tape.pausing)
		return 0;

	if (tape.pause_before_death)	// STOP 10s BEFORE PLAYER GETS KILLED...
	{
/*		if (!(FrameCounter % 20))
		{
			if ((FrameCounter / 20) % 2)
				DrawVideoDisplay(VIDEO_STATE_PBEND_ON, VIDEO_DISPLAY_LABEL_ONLY);
			else
				DrawVideoDisplay(VIDEO_STATE_PBEND_OFF, VIDEO_DISPLAY_LABEL_ONLY);
		}

		if (tape.warp_forward)
		{
			if (tape.deactivate_display)
				DrawVideoDisplay(VIDEO_STATE_WARP_ON, VIDEO_DISPLAY_SYMBOL_ONLY);
			else
				DrawVideoDisplay(VIDEO_STATE_WARP2_ON, VIDEO_DISPLAY_SYMBOL_ONLY);
		}
*/
		if (TapeTime > tape.length_seconds - TAPE_PAUSE_SECONDS_BEFORE_DEATH)
		{
			TapeTogglePause(TAPE_TOGGLE_MANUAL);
			return 0;
		}
	}
	else if (tape.fast_forward)
	{
/*    if ((FrameCounter / 20) % 2)
      DrawVideoDisplay(VIDEO_STATE_FFWD_ON, VIDEO_DISPLAY_LABEL_ONLY);
    else
      DrawVideoDisplay(VIDEO_STATE_FFWD_OFF, VIDEO_DISPLAY_LABEL_ONLY);

    if (tape.warp_forward)
    {
      if (tape.deactivate_display)
	DrawVideoDisplay(VIDEO_STATE_WARP_ON, VIDEO_DISPLAY_SYMBOL_ONLY);
      else
	DrawVideoDisplay(VIDEO_STATE_WARP2_ON, VIDEO_DISPLAY_SYMBOL_ONLY);
    }
*/	}

	if (tape.counter >= tape.length)	// end of tape reached
	{
		if (tape.warp_forward /*&& !tape.auto_play*/)
			TapeTogglePause(TAPE_TOGGLE_MANUAL);
		else
			TapeStop();

		return 0;
	}

	action_ = tape.pos_action_[tape.counter];

	tape.delay_played++;
	if (tape.delay_played >= tape.pos_delay[tape.counter])
	{
		tape.counter++;
		tape.delay_played = 0;
	}

	return action_;
}

void TapeStop(void)
{
	TapeStopRecording();
	TapeStopPlaying();

/*	DrawVideoDisplay(VIDEO_STATE_OFF, 0);
	if (tape.date && tape.length)
	{
		DrawVideoDisplay(VIDEO_STATE_DATE_ON, tape.date);
		DrawVideoDisplay(VIDEO_STATE_TIME_ON, tape.length_seconds);
	}*/
}

static void TapeStartWarpForward(void)
{
	tape.warp_forward = TRUE;

	if (!tape.fast_forward && !tape.pause_before_death)
	{
		tape.pausing = FALSE;
		tape.deactivate_display = TRUE;

		TapeDeactivateDisplayOn();
	}
/*
	if (tape.fast_forward || tape.pause_before_death)
		DrawVideoDisplay(VIDEO_STATE_WARP_ON, VIDEO_DISPLAY_SYMBOL_ONLY);
	else
		DrawVideoDisplay(VIDEO_STATE_WARP_ON, 0);*/
}

static void TapeStopWarpForward(void)
{
//	int state = (tape.pausing ? VIDEO_STATE_PAUSE_ON : VIDEO_STATE_PAUSE_OFF);

	tape.warp_forward = FALSE;
	tape.deactivate_display = FALSE;

	TapeDeactivateDisplayOff(game_status == GAME_MODE_PLAYING);
/*
	if (tape.pause_before_death)
		state |= VIDEO_STATE_WARP_OFF | VIDEO_STATE_PBEND_ON;
	else if (tape.fast_forward)
		state |= VIDEO_STATE_WARP_OFF | VIDEO_STATE_FFWD_ON;
	else
		state |= VIDEO_STATE_WARP_OFF | VIDEO_STATE_PLAY_ON;

	DrawVideoDisplay(state, 0);*/
}

void TapeQuickSave(void)
{
	if (game_status == GAME_MODE_PLAYING)
	{
		if (tape.recording)
			TapeHaltRecording();	// prepare tape for saving on-the-fly

		if (TAPE_IS_EMPTY(tape))
			bugd("No tape!",0);//Request("No tape that can be saved !", REQ_CONFIRM);
		else
			SaveTape(tape.level_nr);
	}
	else if (game_status == GAME_MODE_MAIN)
		bugd("Lazy",0);//Request("No game that can be saved !", REQ_CONFIRM);
}

void TapeQuickLoad(void)
{
/*	if (tape.recording && !Request("Stop recording and load tape ?",
				REQ_ASK | REQ_STAY_CLOSED))
	{
//		BlitBitmap(bitmap_db_door, bitmap_db_door,
//						DOOR_GFX_PAGEX2, DOOR_GFX_PAGEY1, DXSIZE, DYSIZE,
//						DOOR_GFX_PAGEX1, DOOR_GFX_PAGEY1);
		OpenDoor(DOOR_OPEN_1);

		return;
	}
*/
	if (game_status == GAME_MODE_PLAYING || game_status == GAME_MODE_MAIN)
	{
		TapeStop();
		TapeErase();
		LoadTape(level_nr);

		if (!TAPE_IS_EMPTY(tape))
		{
			TapeStartGamePlaying();
			TapeStartWarpForward();

			tape.quick_resume = TRUE;
		}
		else
			bugd("No tape for this level",level_nr);
			//Request("No tape for this level !", REQ_CONFIRM);
	}
}