#include "libgamelibgame.h"

#include "game.h"
#include "init.h"
#include "tools.h"
#include "tape.h"

#include "video.h"//PALM drawing routines

/* this switch controls how rocks move horizontally */
#define OLD_GAME_BEHAVIOUR	FALSE

/* EXPERIMENTAL STUFF */
#define USE_NEW_AMOEBA_CODE	FALSE

/* for DigField() */
#define DF_NO_PUSH		0
#define DF_DIG			1
#define DF_SNAP			2

/* for MovePlayer() */
#define MF_NO_ACTION	0
#define MF_MOVING		1
#define MF_ACTION		2

/* for ScrollPlayer() */
#define SCROLL_INIT		0
#define SCROLL_GO_ON	1

/* for Explode() */
#define EX_PHASE_START		0
#define EX_NO_EXPLOSION		0
#define EX_NORMAL		1
#define EX_CENTER		2
#define EX_BORDER		3
/* new
#define EX_TYPE_NONE		0
#define EX_TYPE_NORMAL		(1 << 0)
#define EX_TYPE_CENTER		(1 << 1)
#define EX_TYPE_BORDER		(1 << 2)
#define EX_TYPE_CROSS		(1 << 3)
#define EX_TYPE_SINGLE_TILE	(EX_TYPE_CENTER | EX_TYPE_BORDER)
*/

/* values for initial player move delay (initial delay counter value) */
#define INITIAL_MOVE_DELAY_OFF	-1
#define INITIAL_MOVE_DELAY_ON	0

/* values for player movement speed (which is in fact a delay value) */
#define MOVE_DELAY_NORMAL_SPEED	8
#define MOVE_DELAY_HIGH_SPEED	4

#define DOUBLE_MOVE_DELAY(x)	(x = (x <= MOVE_DELAY_HIGH_SPEED ? x * 2 : x))
#define HALVE_MOVE_DELAY(x)	(x = (x >= MOVE_DELAY_HIGH_SPEED ? x / 2 : x))
#define DOUBLE_PLAYER_SPEED(p)	(HALVE_MOVE_DELAY((p)->move_delay_value))
#define HALVE_PLAYER_SPEED(p)	(DOUBLE_MOVE_DELAY((p)->move_delay_value)) 

/* values for other actions */
#define MOVE_STEPSIZE_NORMAL	(TILEX / MOVE_DELAY_NORMAL_SPEED)

#define GET_DX_FROM_DIR(d)	((d) == MV_LEFT ? -1 : (d) == MV_RIGHT ? 1 : 0)
#define GET_DY_FROM_DIR(d)	((d) == MV_UP   ? -1 : (d) == MV_DOWN  ? 1 : 0)

#define CAN_GROW_INTO(e)						\
	((e) == EL_SAND || (IS_DIGGABLE(e) && level.grow_into_diggable))
 
#define ELEMENT_CAN_ENTER_FIELD_BASE_X(x, y, condition, xy)		\
		(IN_LEV_FIELD(x, y) && (IS_FREE(xy) ||		\
					(condition)))

#define ELEMENT_CAN_ENTER_FIELD_BASE_2(e, x, y, condition, xy)	\
		(IN_LEV_FIELD(x, y) && (IS_FREE(xy) ||		\
					(CAN_MOVE_INTO_ACID(e) &&	\
					 Feld[xy] == EL_ACID) ||	\
					(condition)))

#define ELEMENT_CAN_ENTER_FIELD_BASE_3(e, x, y, condition, xy)	\
		(IN_LEV_FIELD(x, y) && (IS_FREE_OR_PLAYER(xy) ||	\
					(CAN_MOVE_INTO_ACID(e) &&	\
					 Feld[xy] == EL_ACID) ||	\
					(condition)))

#define ELEMENT_CAN_ENTER_FIELD_BASE_4(e, x, y, condition, xy)	\
		(IN_LEV_FIELD(x, y) && (IS_FREE(xy) ||		\
					(condition) ||			\
					(CAN_MOVE_INTO_ACID(e) &&	\
					 Feld[xy] == EL_ACID) ||	\
					(DONT_COLLIDE_WITH(e) &&	\
					 IS_PLAYER(xy) &&		\
					 !PLAYER_ENEMY_PROTECTED(x, y, xy))))

#define ENEMY_CAN_ENTER_FIELD(e, x, y, xy)					\
						ELEMENT_CAN_ENTER_FIELD_BASE_2(e, x, y, 0, xy)

#define YAMYAM_CAN_ENTER_FIELD(e, x, y, xy)					\
	ELEMENT_CAN_ENTER_FIELD_BASE_3(e, x, y, Feld[xy] == EL_DIAMOND, xy)

#define DARK_YAMYAM_CAN_ENTER_FIELD(e, x, y, xy)			\
	ELEMENT_CAN_ENTER_FIELD_BASE_3(e, x, y, IS_FOOD_DARK_YAMYAM(Feld[xy]), xy)

#define PACMAN_CAN_ENTER_FIELD(e, x, y, xy)					\
	ELEMENT_CAN_ENTER_FIELD_BASE_3(e, x, y, IS_AMOEBOID(Feld[xy]), xy)

#define PIG_CAN_ENTER_FIELD(e, x, y, xy)					\
	ELEMENT_CAN_ENTER_FIELD_BASE_2(e, x, y, IS_FOOD_PIG(Feld[xy]), xy)

#define PENGUIN_CAN_ENTER_FIELD(e, x, y, xy)				\
	ELEMENT_CAN_ENTER_FIELD_BASE_2(e, x, y, (Feld[xy] == EL_EXIT_OPEN ||\
						 IS_FOOD_PENGUIN(Feld[xy])), xy)

#define DRAGON_CAN_ENTER_FIELD(e, x, y, xy)					\
	ELEMENT_CAN_ENTER_FIELD_BASE_2(e, x, y, 0, xy)

#define MOLE_CAN_ENTER_FIELD(e, x, y, condition, xy)			\
	ELEMENT_CAN_ENTER_FIELD_BASE_2(e, x, y, (condition), xy)

#define SATELLITE_CAN_ENTER_FIELD(x, y, xy)					\
	ELEMENT_CAN_ENTER_FIELD_BASE_2(EL_SATELLITE, x, y, 0, xy)

#define SPRING_CAN_ENTER_FIELD(e, x, y, xy)					\
	ELEMENT_CAN_ENTER_FIELD_BASE_2(e, x, y, 0, xy)

#define IN_LEV_FIELD_AND_IS_FREE(x, y)  (IN_LEV_FIELD(x, y) &&  IS_FREE(A2TO1(x,y)))
#define IN_LEV_FIELD_AND_NOT_FREE(x, y) (IN_LEV_FIELD(x, y) && !IS_FREE(A2TO1(x,y))) 

/* forward declaration for internal use */
static void CheckGravityMovement(struct PlayerInfo *) LOGIC_SECTION;
#define CheckGravityMovementWhenNotMoving(p)	CheckGravityMovement(p)

static boolean MovePlayerOneStep(struct PlayerInfo *, int, int, int, int) LOGIC_SECTION;
static boolean MovePlayer(struct PlayerInfo *, int, int) LOGIC_SECTION;
static void ScrollPlayer(struct PlayerInfo *, int);
static void ScrollScreen(struct PlayerInfo *, int);

int DigField(struct PlayerInfo *, int, int, int, int, int, int, int) LOGIC_SECTION;

static void InitBeltMovement(void) INIT_SECTION;
static void CloseAllOpenTimegates(void);

#define KillHeroUnlessEnemyProtected(x, y,xy)	\
		{if (!PLAYER_ENEMY_PROTECTED(x, y,xy)) KillHero(PLAYERINFO(x, y));}

#define KillHeroUnlessExplosionProtected(x, y,xy)	\
		{if (!PLAYER_EXPLOSION_PROTECTED(x, y,xy)) KillHero(PLAYERINFO(x, y));}

static void ChangeElement(int, int, int);

void TestIfGoodThingHitsBadThing(int, int, int);
void TestIfBadThingHitsGoodThing(int, int, int);
#define TestIfHeroTouchesBadThing(x,y)		TestIfGoodThingHitsBadThing(x, y, MV_NO_MOVING)
#define TestIfHeroRunsIntoBadThing(x,y,m)	TestIfGoodThingHitsBadThing(x, y, m)
#define TestIfBadThingTouchesHero(x,y)		TestIfBadThingHitsGoodThing(x, y, MV_NO_MOVING)
#define TestIfBadThingRunsIntoHero(x,y,m)	TestIfBadThingHitsGoodThing(x, y, m)
#define TestIfFriendTouchesBadThing(x,y)	TestIfGoodThingHitsBadThing(x, y, MV_NO_MOVING)
#define TestIfBadThingTouchesFriend(x,y)	TestIfBadThingHitsGoodThing(x, y, MV_NO_MOVING)
void TestIfBadThingTouchesOtherBadThing(int, int);

#define	INIT_GFX_RANDOM()	(SimpleRND(100))
//was 1000000 

#define ResetRandomAnimationValue(xy)	{GfxRandom[xy] = INIT_GFX_RANDOM();}

#define ResetGfxAnimation(xy)	\
{	GfxSFrame[xy] = FrameCounter;	GfxAction[xy] = ACTION_DEFAULT;}
//	GfxDir[x][y] = MovDir[x][y];

/*	1000000 random
bd_amoeba
amoeba_wet
amoeba_dropping
amoeba_dry
amoeba_full
amoeba_dead
*/

#define getBeltNrFromBeltElement(element)	\
			(element < EL_CONVEYOR_BELT_2_LEFT ? 0 :	\
			element < EL_CONVEYOR_BELT_3_LEFT ? 1 :		\
			element < EL_CONVEYOR_BELT_4_LEFT ? 2 : 3)

#define getBeltNrFromBeltActiveElement(element)		\
			(element < EL_CONVEYOR_BELT_2_LEFT_ACTIVE ? 0 :		\
			element < EL_CONVEYOR_BELT_3_LEFT_ACTIVE ? 1 :		\
			element < EL_CONVEYOR_BELT_4_LEFT_ACTIVE ? 2 : 3)

#define getBeltNrFromBeltSwitchElement(element)			\
			(element < EL_CONVEYOR_BELT_2_SWITCH_LEFT ? 0 :		\
			element < EL_CONVEYOR_BELT_3_SWITCH_LEFT ? 1 :		\
			element < EL_CONVEYOR_BELT_4_SWITCH_LEFT ? 2 : 3)

LOGIC_SECTION static int getBeltDirNrFromBeltSwitchElement(int element)
{
	const int belt_base_element[4] =
	{
		EL_CONVEYOR_BELT_1_SWITCH_LEFT,
		EL_CONVEYOR_BELT_2_SWITCH_LEFT,
		EL_CONVEYOR_BELT_3_SWITCH_LEFT,
		EL_CONVEYOR_BELT_4_SWITCH_LEFT
	};

	int belt_nr = getBeltNrFromBeltSwitchElement(element);
	int belt_dir_nr = element - belt_base_element[belt_nr];

	return (belt_dir_nr % 3);
};

LOGIC_SECTION static int getBeltDirFromBeltSwitchElement(int element)
{
	const int belt_move_dir[3] =
	{
		MV_LEFT,
		MV_NO_MOVING,
		MV_RIGHT
	};

	int belt_dir_nr = getBeltDirNrFromBeltSwitchElement(element);

	return belt_move_dir[belt_dir_nr];
};

INIT_SECTION static void InitPlayerField(int x, int y, int element, boolean init_game)
{
	short int xy=A2TO1(x,y);
	
	if (element == EL_SP_MURPHY)
	{
		if (init_game)
		{
			if (stored_player_.present)
			{
				Feld[xy] = EL_SP_MURPHY_CLONE;

				return;
			}
			else
			{
				stored_player_.use_murphy_graphic = TRUE;
			}

			Feld[xy] = EL_PLAYER_1;
		}
	}

	if (init_game)
	{	
		if(Feld[xy]==EL_PLAYER_1)
		{
			struct PlayerInfo *player = &stored_player_;
//			int jx = player->jx, jy = player->jy;

			player->present = TRUE;
			player->block_last_field = (element == EL_SP_MURPHY ?
										level.sp_block_last_field :
										level.block_last_field);

			if (player->connected)
			{
				player->active = TRUE;

				// remove potentially duplicate players
/*				if (StorePlayer[jx][jy] == Feld[x][y])
					StorePlayer[jx][jy] = 0;

				StorePlayer[x][y] = Feld[x][y];
*/
			}
			player->jx = player->last_jx = x;
			player->jy = player->last_jy = y;
		}
		Feld[xy] = EL_EMPTY;
	}
}

static void InitField(int x, int y, boolean init_game)
{
	short int xy=A2TO1(x,y);
	int element = Feld[xy];

	switch (element)
	{
		case EL_SP_MURPHY:
		case EL_PLAYER_1:
		case EL_PLAYER_2:
		case EL_PLAYER_3:
		case EL_PLAYER_4:
			InitPlayerField(x, y, element, init_game);
			break;

		case EL_SOKOBAN_FIELD_PLAYER:
			element = Feld[xy] = EL_PLAYER_1;
			InitField(x, y, init_game);

			element = Feld[xy] = EL_SOKOBAN_FIELD_EMPTY;
			InitField(x, y, init_game);
			break;

		case EL_SOKOBAN_FIELD_EMPTY:
			local_player->sokobanfields_still_needed++;
			break;

		case EL_STONEBLOCK:
			if (x < lev_fieldx-1 && Feld[xy+A2TO1(1,0)] == EL_ACID)
				Feld[xy] = EL_ACID_POOL_TOPLEFT;
			else if (x > 0 && Feld[xy-A2TO1(1,0)] == EL_ACID)
				Feld[xy] = EL_ACID_POOL_TOPRIGHT;
			else if (y > 0 && Feld[xy-A2TO1(0,1)] == EL_ACID_POOL_TOPLEFT)
				Feld[xy] = EL_ACID_POOL_BOTTOMLEFT;
			else if (y > 0 && Feld[xy-A2TO1(0,1)] == EL_ACID)
				Feld[xy] = EL_ACID_POOL_BOTTOM;
			else if (y > 0 && Feld[xy-A2TO1(0,1)] == EL_ACID_POOL_TOPRIGHT)
				Feld[xy] = EL_ACID_POOL_BOTTOMRIGHT;
			break;

		case EL_BUG_RIGHT:
		case EL_BUG_UP:
		case EL_BUG_LEFT:
		case EL_BUG_DOWN:
		case EL_BUG:
		case EL_SPACESHIP_RIGHT:
		case EL_SPACESHIP_UP:
		case EL_SPACESHIP_LEFT:
		case EL_SPACESHIP_DOWN:
		case EL_SPACESHIP:
		case EL_BD_BUTTERFLY_RIGHT:
		case EL_BD_BUTTERFLY_UP:
		case EL_BD_BUTTERFLY_LEFT:
		case EL_BD_BUTTERFLY_DOWN:
		case EL_BD_BUTTERFLY:
		case EL_BD_FIREFLY_RIGHT:
		case EL_BD_FIREFLY_UP:
		case EL_BD_FIREFLY_LEFT:
		case EL_BD_FIREFLY_DOWN:
		case EL_BD_FIREFLY:
		case EL_PACMAN_RIGHT:
		case EL_PACMAN_UP:
		case EL_PACMAN_LEFT:
		case EL_PACMAN_DOWN:
		case EL_YAMYAM:
		case EL_DARK_YAMYAM:
		case EL_ROBOT:
		case EL_PACMAN:
		case EL_SP_SNIKSNAK:
		case EL_SP_ELECTRON:
		case EL_MOLE_LEFT:
		case EL_MOLE_RIGHT:
		case EL_MOLE_UP:
		case EL_MOLE_DOWN:
		case EL_MOLE:
			InitMovDir(x, y);
			break;

		case EL_AMOEBA_FULL:
		case EL_BD_AMOEBA:
			InitAmoebaNr(x, y);
			break;

		case EL_AMOEBA_DROP:
			if (y == lev_fieldy - 1)
			{
				Feld[xy] = EL_AMOEBA_GROWING;
				Store[xy]=EL_AMOEBA_WET;
			}
			break;

		case EL_DYNAMITE_ACTIVE:
		case EL_SP_DISK_RED_ACTIVE:
		case EL_DYNABOMB_PLAYER_1_ACTIVE:
		case EL_DYNABOMB_PLAYER_2_ACTIVE:
		case EL_DYNABOMB_PLAYER_3_ACTIVE:
		case EL_DYNABOMB_PLAYER_4_ACTIVE:
			MovDelay[xy] = 96;
			break;

		case EL_LAMP:
			local_player->lights_still_needed++;
			break;

		case EL_PENGUIN:
			local_player->friends_still_needed++;
			break;

		case EL_PIG:
		case EL_DRAGON:
/*			GfxDir[x][y] = */MovDir[xy] = 1 << RND(4);
			break;

		case EL_CONVEYOR_BELT_1_SWITCH_LEFT:
		case EL_CONVEYOR_BELT_1_SWITCH_MIDDLE:
		case EL_CONVEYOR_BELT_1_SWITCH_RIGHT:
		case EL_CONVEYOR_BELT_2_SWITCH_LEFT:
		case EL_CONVEYOR_BELT_2_SWITCH_MIDDLE:
		case EL_CONVEYOR_BELT_2_SWITCH_RIGHT:
		case EL_CONVEYOR_BELT_3_SWITCH_LEFT:
		case EL_CONVEYOR_BELT_3_SWITCH_MIDDLE:
		case EL_CONVEYOR_BELT_3_SWITCH_RIGHT:
		case EL_CONVEYOR_BELT_4_SWITCH_LEFT:
		case EL_CONVEYOR_BELT_4_SWITCH_MIDDLE:
		case EL_CONVEYOR_BELT_4_SWITCH_RIGHT:
			if (init_game)
			{
				int belt_nr = getBeltNrFromBeltSwitchElement(Feld[xy]);
				int belt_dir = getBeltDirFromBeltSwitchElement(Feld[xy]);
				int belt_dir_nr = getBeltDirNrFromBeltSwitchElement(Feld[xy]);

				if (game.belt_dir_nr[belt_nr] == 3)	// initial value
				{
					game.belt_dir[belt_nr] = belt_dir;
					game.belt_dir_nr[belt_nr] = belt_dir_nr;
				}
				else	// more than one switch -- set it like the first switch 
				{
					Feld[xy] = Feld[xy] - belt_dir_nr + game.belt_dir_nr[belt_nr];
				}
			}
			break;

		case EL_SWITCHGATE_SWITCH_DOWN:	// always start with same switch pos 
			if (init_game)
				Feld[xy] = EL_SWITCHGATE_SWITCH_UP;
			break;

		case EL_LIGHT_SWITCH_ACTIVE:
			if (init_game)
				game.light_time_left = level.time_light * FRAMES_PER_SECOND;
			break;
/*
//PALM!!!
		case EL_SP_BUGGY_BASE:
			AddDelayChanging(xy,
					2*FRAMES_PER_SECOND + RND(5 * FRAMES_PER_SECOND) -
						FRAMES_PER_SECOND/4 ,EL_SP_BUGGY_BASE_ACTIVATING);
			break;
		case EL_SP_BUGGY_BASE_ACTIVATING:
			AddDelayChanging(xy,FRAMES_PER_SECOND/4 ,EL_SP_BUGGY_BASE_ACTIVE);
			break;
		case EL_SP_BUGGY_BASE_ACTIVE:
			AddDelayChanging(xy,FRAMES_PER_SECOND + RND(FRAMES_PER_SECOND),
					EL_SP_BUGGY_BASE);
			break;
		case EL_TRAP:
			AddDelayChanging(xy,2*FRAMES_PER_SECOND + RND(5*FRAMES_PER_SECOND),
					EL_TRAP_ACTIVE);
			break;
		case EL_TRAP_ACTIVE:
			AddDelayChanging(xy,31,EL_TRAP);
			break;
*/		default:
/*      if (IS_CUSTOM_ELEMENT(element) && CAN_MOVE(element))
	InitMovDir(x, y);
      else if (IS_GROUP_ELEMENT(element))
      {
	struct ElementGroupInfo *group = element_info[element].group;
	int last_anim_random_frame = gfx.anim_random_frame;
	int element_pos;

	if (group->choice_mode == ANIM_RANDOM)
	  gfx.anim_random_frame = RND(group->num_elements_resolved);

	element_pos = getAnimationFrame(group->num_elements_resolved, 1,
					group->choice_mode, 0,
					group->choice_pos);

	if (group->choice_mode == ANIM_RANDOM)
	  gfx.anim_random_frame = last_anim_random_frame;

	group->choice_pos++;

	Feld[x][y] = group->element_resolved[element_pos];

	InitField(x, y, init_game);
      }
*/			break;
	}
	
	InitGfxValue(xy);
}

#define InitField_WithBug1(x,y,init_game,xy)				\
	{  InitField(x, y, init_game);						\
		if (game.engine_version < VERSION_IDENT(3,1,0,0) && CAN_MOVE(Feld[xy]))\
			InitMovDir(x, y);}
// not needed to call InitMovDir() -- already done by InitField()!

#define DrawGameValue_Emeralds(value)	PalmDrawGemsLeft(value)
//was	DrawText(DX_EMERALDS, DY_EMERALDS, int2str(value, 3), FONT_TEXT_2);

#define DrawGameValue_Dynamite(value)	PalmDrawDynamiteLeft(value)
//was	DrawText(DX_DYNAMITE, DY_DYNAMITE, int2str(value, 3), FONT_TEXT_2);

void DrawGameValue_Keys(struct PlayerInfo *player)
{
/*	int i;

	for (i = 0; i < MAX_KEYS; i++)
		if (player->key[i])
			DrawMiniGraphicExt(drawto, DX_KEYS + i * MINI_TILEX, DY_KEYS,
													el2edimg(EL_KEY_1 + i));*/
}

#define DrawGameValue_Score(value)		PalmDrawScore(value)
//was	DrawText(DX_SCORE, DY_SCORE, int2str(value, 5), FONT_TEXT_2);

#define DrawGameValue_Time(value)		PalmDrawTime(value)
/*was	if (value < 1000)
		DrawText(DX_TIME1, DY_TIME, int2str(value, 3), FONT_TEXT_2);
	else
		DrawText(DX_TIME2, DY_TIME, int2str(value, 4), FONT_LEVEL_NUMBER);*/

void DrawGameValue_Level(int value)
{/*
  if (level_nr < 100)
    DrawText(DX_LEVEL, DY_LEVEL, int2str(value, 2), FONT_TEXT_2);
  else
  {
    // misuse area for displaying emeralds to draw bigger level number
    DrawTextExt(drawto, DX_EMERALDS, DY_EMERALDS,
		int2str(value, 3), FONT_LEVEL_NUMBER, BLIT_OPAQUE);

    // now copy it to the area for displaying level number
    BlitBitmap(drawto, drawto,
	       DX_EMERALDS, DY_EMERALDS + 1,
	       getFontWidth(FONT_LEVEL_NUMBER) * 3,
	       getFontHeight(FONT_LEVEL_NUMBER) - 1,
	       DX_LEVEL - 1, DY_LEVEL + 1);

    // restore the area for displaying emeralds
    DrawGameValue_Emeralds(local_player->gems_still_needed);

    // yes, this is all really ugly :-)
  }*/
}

void DrawGameDoorValues()
{
	DrawGameValue_Level(level_nr);

	DrawGameValue_Keys(&stored_player_);

	DrawGameValue_Emeralds(local_player->gems_still_needed);
	DrawGameValue_Dynamite(local_player->inventory_size);
	DrawGameValue_Score(local_player->score);
	DrawGameValue_Time(TimeLeft);
};

/*
  =============================================================================
  InitGameEngine()
  -----------------------------------------------------------------------------
  initialize game engine due to level / tape version number
  =============================================================================
*/
INIT_SECTION static void InitGameEngine()
{
//	int i, j, k;

	// set game engine from tape file when re-playing, else from level file
	game.engine_version = (tape.playing ? tape.engine_version :
												level.game_version);
	// dynamically adjust element properties according to game engine version
	InitElementPropertiesEngine(game.engine_version);
/*
	// ---------- recursively resolve group elements -------------------------

	for (i = 0; i < MAX_NUM_ELEMENTS; i++)
		for (j = 0; j < NUM_GROUP_ELEMENTS; j++)
			element_info[i].in_group[j] = FALSE;

	for (i = 0; i < NUM_GROUP_ELEMENTS; i++)
		resolve_group_element(EL_GROUP_START + i, 0);
*/
	// ---------- initialize player's initial move delay ---------------------

	// dynamically adjust player properties according to game engine version
	game.initial_move_delay =
			(game.engine_version <= VERSION_IDENT(2,0,1,0) ? INITIAL_MOVE_DELAY_ON :
															INITIAL_MOVE_DELAY_OFF);

	// dynamically adjust player properties according to level information
	game.initial_move_delay_value =
			(level.double_speed ? MOVE_DELAY_HIGH_SPEED : MOVE_DELAY_NORMAL_SPEED);

	// ---------- initialize player's initial push delay ---------------------

	// dynamically adjust player properties according to game engine version 
	game.initial_push_delay_value =
			(game.engine_version < VERSION_IDENT(3,0,7,1) ? 5 : -1);

	// ---------- initialize changing elements -------------------------------
/*
	// initialize changing elements information
	for (i = 0; i < MAX_NUM_ELEMENTS; i++)
	{
		struct ElementInfo *ei = &element_info[i];

		// this pointer might have been changed in the level editor
		ei->change = &ei->change_page[0];

		if (!IS_CUSTOM_ELEMENT(i))
		{
			ei->change->target_element = EL_EMPTY_SPACE;
			ei->change->delay_fixed = 0;
			ei->change->delay_random = 0;
			ei->change->delay_frames = 1;
		}

		ei->change_events = CE_BITMASK_DEFAULT;
		for (j = 0; j < NUM_CHANGE_EVENTS; j++)
		{
			ei->event_page_nr[j] = 0;
			ei->event_page[j] = &ei->change_page[0];
		}
	}

  // add changing elements from pre-defined list
  for (i = 0; change_delay_list[i].element != EL_UNDEFINED; i++)
  {
    struct ChangingElementInfo *ch_delay = &change_delay_list[i];
    struct ElementInfo *ei = &element_info[ch_delay->element];

    ei->change->target_element       = ch_delay->target_element;
    ei->change->delay_fixed          = ch_delay->change_delay;

    ei->change->pre_change_function  = ch_delay->pre_change_function;
    ei->change->change_function      = ch_delay->change_function;
    ei->change->post_change_function = ch_delay->post_change_function;

    ei->change_events |= CH_EVENT_BIT(CE_DELAY);

    SET_PROPERTY(ch_delay->element, EP_CAN_CHANGE, TRUE);
  }

  // add change events from custom element configuration
  for (i = 0; i < NUM_CUSTOM_ELEMENTS; i++)
  {
    struct ElementInfo *ei = &element_info[EL_CUSTOM_START + i];

    for (j = 0; j < ei->num_change_pages; j++)
    {
      if (!ei->change_page[j].can_change)
	continue;

      for (k = 0; k < NUM_CHANGE_EVENTS; k++)
      {
	// only add event page for the first page found with this event
	if (ei->change_page[j].events & CH_EVENT_BIT(k) &&
	    !(ei->change_events & CH_EVENT_BIT(k)))
	{
	  ei->change_events |= CH_EVENT_BIT(k);
	  ei->event_page_nr[k] = j;
	  ei->event_page[k] = &ei->change_page[j];
	}
      }
    }
  }
  // ---------- initialize run-time trigger player and element -------------

  for (i = 0; i < NUM_CUSTOM_ELEMENTS; i++)
  {
    struct ElementInfo *ei = &element_info[EL_CUSTOM_START + i];

    for (j = 0; j < ei->num_change_pages; j++)
    {
      ei->change_page[j].actual_trigger_element = EL_EMPTY;
      ei->change_page[j].actual_trigger_player = EL_PLAYER_1;
    }
  }


	// ---------- initialize trigger events ----------------------------------

	// initialize trigger events information
	for (i = 0; i < MAX_NUM_ELEMENTS; i++)
		trigger_events[i] = EP_BITMASK_DEFAULT;

  // add trigger events from element change event properties
  for (i = 0; i < MAX_NUM_ELEMENTS; i++)
  {
    struct ElementInfo *ei = &element_info[i];

    for (j = 0; j < ei->num_change_pages; j++)
    {
      if (!ei->change_page[j].can_change)
	continue;

      if (ei->change_page[j].events & CH_EVENT_BIT(CE_BY_OTHER_ACTION))
      {
	int trigger_element = ei->change_page[j].trigger_element;

	if (IS_GROUP_ELEMENT(trigger_element))
	{
	  struct ElementGroupInfo *group = element_info[trigger_element].group;

	  for (k = 0; k < group->num_elements_resolved; k++)
	    trigger_events[group->element_resolved[k]]
	      |= ei->change_page[j].events;
	}
	else
	trigger_events[trigger_element] |= ei->change_page[j].events;
      }
    }
  }

	// ---------- initialize push delay --------------------------------------

  // initialize push delay values to default
  for (i = 0; i < MAX_NUM_ELEMENTS; i++)
  {
    if (!IS_CUSTOM_ELEMENT(i))
    {
      element_info[i].push_delay_fixed  = game.default_push_delay_fixed;
      element_info[i].push_delay_random = game.default_push_delay_random;
    }
  }

  // set push delay value for certain elements from pre-defined list 
  for (i = 0; push_delay_list[i].element != EL_UNDEFINED; i++)
  {
    int e = push_delay_list[i].element;

    element_info[e].push_delay_fixed  = push_delay_list[i].push_delay_fixed;
    element_info[e].push_delay_random = push_delay_list[i].push_delay_random;
  }

  // set push delay value for Supaplex elements for newer engine versions
  if (game.engine_version >= VERSION_IDENT(3,1,0,0))
  {
    for (i = 0; i < MAX_NUM_ELEMENTS; i++)
    {
      if (IS_SP_ELEMENT(i))
      {
	element_info[i].push_delay_fixed  = 6;	// just enough to escape ...
	element_info[i].push_delay_random = 0;	// ... from falling zonk
      }
    }
  }

  // ---------- initialize move stepsize -----------------------------------

  // initialize move stepsize values to default
  for (i = 0; i < MAX_NUM_ELEMENTS; i++)
    if (!IS_CUSTOM_ELEMENT(i))
      element_info[i].move_stepsize = MOVE_STEPSIZE_NORMAL;

  // set move stepsize value for certain elements from pre-defined list
  for (i = 0; move_stepsize_list[i].element != EL_UNDEFINED; i++)
  {
    int e = move_stepsize_list[i].element;

    element_info[e].move_stepsize = move_stepsize_list[i].move_stepsize;
  }

  // ---------- initialize access direction --------------------------------

  // initialize access direction values to default (access from every side)
  for (i = 0; i < MAX_NUM_ELEMENTS; i++)
    if (!IS_CUSTOM_ELEMENT(i))
      element_info[i].access_direction = MV_ALL_DIRECTIONS;

  // set access direction value for certain elements from pre-defined list
  for (i = 0; access_direction_list[i].element != EL_UNDEFINED; i++)
    element_info[access_direction_list[i].element].access_direction =
      access_direction_list[i].direction;
*/
};

//PALM only!!!!
LOGIC_SECTION static byte FindElementCollectCount(short int element)
{
	if( element==EL_EMERALD || element==EL_BD_DIAMOND ||
				element==EL_EMERALD_YELLOW || element==EL_EMERALD_RED ||
				element==EL_EMERALD_PURPLE || element==EL_SP_INFOTRON)
		return 1;
	if( element==EL_DIAMOND)
		return 3;
	if( element==EL_PEARL)
		return 5;
	if( element==EL_CRYSTAL)
		return 8;
	return 0;
}

//PALM!! if no PLAYER_1 but PLAYER_2 is present, let it be PLAYER_1
INIT_SECTION void Player2or1(void)
{
	int x,y;
	short int pl_2_XY=-1;
	for(y = 0; y < lev_fieldy; y++)
	{
		short int register xy=A2TO1(0,y);
		for (x = 0; x < lev_fieldx; x++,xy+=A2TO1(1,0))
		{
			if(Feld[xy]==EL_PLAYER_1||Feld[xy]==EL_SP_MURPHY)
				return;
			if(Feld[xy]==EL_PLAYER_2)
				pl_2_XY=xy;
		}
	}
	if(pl_2_XY!=-1)
		Feld[pl_2_XY]=EL_PLAYER_1;
	else
		bugd("HE-HE, no PLAYER 1 or 2:)",0);
}

/*
  =============================================================================
  InitGame()
  -----------------------------------------------------------------------------
  initialize and start new game
  =============================================================================
*/

void InitGame()
{
	boolean emulate_bd = TRUE;	// unless non-BOULDERDASH elements found
	boolean emulate_sb = TRUE;	// unless non-SOKOBAN     elements found
	boolean emulate_sp = TRUE;	// unless non-SUPAPLEX    elements found
	int i, j, k, x, y;
	short int xy;

	InitGameEngine();

	{
		struct PlayerInfo *player = &stored_player_;

		player->index_nr = 0;
		player->index_bit = 0;
		player->element_nr = EL_PLAYER_1;

		player->present = FALSE;
		player->active = FALSE;

		player->action = 0;
		player->effective_action = 0;
		player->programmed_action = 0;

		player->score = 0;
		player->gems_still_needed = level.gems_needed;
		player->sokobanfields_still_needed = 0;
		player->lights_still_needed = 0;
		player->friends_still_needed = 0;

		for (j = 0; j < MAX_KEYS; j++)
			player->key[j] = FALSE;

		player->dynabomb_count = 0;
		player->dynabomb_size = 1;
		player->dynabombs_left = 0;
		player->dynabomb_xl = FALSE;

		player->MovDir = MV_NO_MOVING;
		player->MovPos = 0;
		player->GfxPos = 0;
		player->GfxDir = MV_NO_MOVING;
		player->GfxAction = ACTION_DEFAULT;
		player->Frame = 0;
		player->StepFrame = 0;

		player->use_murphy_graphic = FALSE;

		player->block_last_field = FALSE;
		player->can_fall_into_acid = CAN_MOVE_INTO_ACID(player->element_nr);

		player->actual_frame_counter = 0;

		player->step_counter = 0;

		player->last_move_dir = MV_NO_MOVING;

		player->is_waiting = FALSE;
		player->is_moving = FALSE;
		player->is_auto_moving = FALSE;
		player->is_digging = FALSE;
		player->is_snapping = FALSE;
		player->is_collecting = FALSE;
		player->is_pushing = FALSE;
		player->is_switching = FALSE;
		player->is_dropping = FALSE;

		player->is_bored = FALSE;
		player->is_sleeping = FALSE;

		player->frame_counter_bored = -1;
		player->frame_counter_sleeping = -1;

		player->anim_delay_counter = 0;
		player->post_delay_counter = 0;

		player->action_waiting = ACTION_DEFAULT;
		player->last_action_waiting = ACTION_DEFAULT;
		player->special_action_bored = ACTION_DEFAULT;
		player->special_action_sleeping = ACTION_DEFAULT;

		player->num_special_action_bored = 0;
		player->num_special_action_sleeping = 0;

		// determine number of special actions for bored and sleeping animation
		for (j = ACTION_BORING_1; j <= ACTION_BORING_LAST; j++)
		{
			boolean found = FALSE;

			for (k = 0; k < NUM_DIRECTIONS; k++)
				if (el_act_dir2img(player->element_nr, j, k) !=
							el_act_dir2img(player->element_nr, ACTION_DEFAULT, k))
					found = TRUE;

			if (found)
				player->num_special_action_bored++;
			else
				break;
		}
		for (j = ACTION_SLEEPING_1; j <= ACTION_SLEEPING_LAST; j++)
		{
			boolean found = FALSE;

			for (k = 0; k < NUM_DIRECTIONS; k++)
				if (el_act_dir2img(player->element_nr, j, k) !=
							el_act_dir2img(player->element_nr, ACTION_DEFAULT, k))
					found = TRUE;

			if (found)
				player->num_special_action_sleeping++;
			else
				break;
		}

		player->switch_x = -1;
		player->switch_y = -1;

		player->show_envelope = 0;

		player->move_delay       = game.initial_move_delay;
		player->move_delay_value = game.initial_move_delay_value;

		player->move_delay_reset_counter = 0;

		player->push_delay = 0;
		player->push_delay_value = game.initial_push_delay_value;

		player->last_jx = player->last_jy = 0;
		player->jx = player->jy = 0;

		player->drop_delay = 0;

		player->shield_normal_time_left = 0;
		player->shield_deadly_time_left = 0;

		player->inventory_infinite_element = EL_UNDEFINED;
		player->inventory_size = 0;

		DigField(player, 0, 0, 0, 0, 0, 0, DF_NO_PUSH);
		SnapField(player, 0, 0);

		player->LevelSolved = FALSE;
		player->GameOver = FALSE;
	}

	ZX = ZY = -1;

	FrameCounter = 0;
	TimeFrames = 0;
	TimePlayed = 0;
	TimeLeft = level.time;
	TapeTime = 0;

	ScreenMovDir = MV_NO_MOVING;
	ScreenMovPos = 0;
	ScreenGfxPos = 0;

	ScrollStepSize = 0;	// will be correctly initialized by ScrollScreen()

	AllPlayersGone = FALSE;

	game.yamyam_content_nr = 0;
	game.magic_wall_active = FALSE;
	game.magic_wall_time_left = 0;
	game.light_time_left = 0;
	game.timegate_time_left = 0;
	game.switchgate_pos = 0;
	game.balloon_dir = MV_NO_MOVING;
	game.gravity = level.initial_gravity;
	game.explosions_delayed = TRUE;

	game.envelope_active = FALSE;

	for (i = 0; i < NUM_BELTS; i++)
	{
		game.belt_dir[i] = MV_NO_MOVING;
		game.belt_dir_nr[i] = 3;		// not moving, next moving left
	}

	for (i = 0; i < MAX_NUM_AMOEBA; i++)
		AmoebaCnt[i] = AmoebaCnt2[i] = 0;


	for (x = 0; x < lev_fieldx; x++)
	{
		xy=A2TO1(x,0);
		for (y = 0; y < lev_fieldy; y++,xy+=A2TO1(0,1))
		{
// for palm only - Feld is already initialized instead of level.field
			MovPos[xy] = MovDir[xy] = MovDelay[xy] = 0;
/*			ChangeDelay[x][y] = 0;
			ChangePage[x][y] = -1;*/
			Store[xy] = Store2[xy] = Back[xy] = 0;
			AmoebaNr[xy] = 0;
//			CheckCollision[x][y] = 0;
			FFlags[xy] = 0;//Stop[x][y]= Pushed[x][y] = FALSE;
			//WasJustMoving[x][y] = WasJustFalling[x][y] = 0;
/*
			Changed[x][y] = CE_BITMASK_DEFAULT;
			ChangeEvent[x][y] = CE_BITMASK_DEFAULT;
*/
			ExplodeFlags[xy]=0;
/*			ExplodeDelay[x][y] = 0;
			ExplodeField[x][y] = EX_TYPE_NONE;
			ExplodePhase[x][y] = 0;

			RunnerVisit[x][y] = 0;
			PlayerVisit[x][y] = 0;
*/
			GfxSFrame[xy] = 0;//GfxFrame[x][y] = 0;
			GfxRandom[xy] = INIT_GFX_RANDOM();
			GfxElement[xy] = EL_UNDEFINED;
			GfxAction[xy] = ACTION_DEFAULT;
//			Gfx[xy] = IMG_UNKNOWN; //PALM, new 
//			GfxDir[x][y] = MV_NO_MOVING;
		}
	}

	for (y = 0; y < lev_fieldy; y++)
	{
		unsigned char brr;
		StopYBegin[y] = MAX_LEV_FIELDX;
		StopYEnd[y] = 0; 
		WasJustYBegin[y] = MAX_LEV_FIELDX;
		WasJustYEnd[y] = 0; 
		FeldYExplosion[y] = ExplodeFieldY[y] = 0;

		brr=0;
		xy=A2TO1(0,y);
		while(brr<lev_fieldx-1 && IS_INDESTR_INACTIVE_NOANIM(Feld[xy]))
			{xy+=A2TO1(1,0); brr++;}
		FeldYStart[y]=brr;
//bugd("FeldYStart and y are",y+brr*256);
		brr=lev_fieldx-1;
		xy=A2TO1(brr,y);
		while(brr>0 && IS_INDESTR_INACTIVE_NOANIM(Feld[xy]))
			{xy-=A2TO1(1,0); brr--;}
		FeldYEnd[y]=brr;
//bugd("FeldYEnd and y are",y+brr*256);
	}
	
//	ChangeList.size=0;
	
	InitMainDoLoop();
	
	StorePlayerXY = -1;
	
	Player2or1();

	for (y = 0; y < lev_fieldy; y++)
	{
		short int register xy=A2TO1(0,y);
		for (x = 0; x < lev_fieldx; x++,xy+=A2TO1(1,0))
		{
			short int el=Feld[xy];
			if (emulate_bd && !IS_BD_ELEMENT(el))
				emulate_bd = FALSE;
			if (emulate_sb && !IS_SB_ELEMENT(el))
				emulate_sb = FALSE;
			if (emulate_sp && !IS_SP_ELEMENT(el))
				emulate_sp = FALSE;

			InitField(x, y, TRUE);
		}
	}

	InitBeltMovement();

	game.emulation = (emulate_bd ? EMU_BOULDERDASH :
						emulate_sb ? EMU_SOKOBAN :
						emulate_sp ? EMU_SUPAPLEX : EMU_NONE);

	// initialize explosion and ignition delay
/*
  for (i = 0; i < MAX_NUM_ELEMENTS; i++)
  {
    if (!IS_CUSTOM_ELEMENT(i))
    {
      int num_phase = 8;
      int delay = (((IS_SP_ELEMENT(i) && i != EL_EMPTY_SPACE) &&
		    game.engine_version >= VERSION_IDENT(3,1,0,0)) ||
		   game.emulation == EMU_SUPAPLEX ? 3 : 2);
      int last_phase = (num_phase + 1) * delay;
      int half_phase = (num_phase / 2) * delay;

      element_info[i].explosion_delay = last_phase - 1;
      element_info[i].ignition_delay = half_phase;

      if (i == EL_BLACK_ORB)
	element_info[i].ignition_delay = 1;
    }
  }
*/
	// correct non-moving belts to start moving left
	for (i = 0; i < NUM_BELTS; i++)
		if (game.belt_dir[i] == MV_NO_MOVING)
			game.belt_dir_nr[i] = 3;		// not moving, next moving left

	// check if any connected player was not found in playfield

	{
		struct PlayerInfo *player = &stored_player_;
//palm only-I don't know, why not active!!!!
		if(player->present)player->active = TRUE;
		StorePlayerXY=A2TO1(player->jx,player->jy);
		if (player->connected && !player->present)
		{

			{
				struct PlayerInfo *some_player = &stored_player_;
				int jx = some_player->jx, jy = some_player->jy;

				// assign first free player found that is present in the playfield
				if (some_player->present && !some_player->connected)
				{
					player->present = TRUE;
					player->active = TRUE;
					some_player->present = FALSE;
					some_player->active = FALSE;

//					SASetValue(&StorePlayer,jx,jy,player->element_nr);
					player->jx = player->last_jx = jx;
					player->jy = player->last_jy = jy;

				}
			}
		}
	}

	if (tape.playing)
	{
		// when playing a tape, eliminate all players which do not participate

		{
			if (stored_player_.active && !tape.player_participates[0])
			{
				struct PlayerInfo *player = &stored_player_;
				int jx = player->jx, jy = player->jy;

				player->active = FALSE;
				player->present = FALSE;
//				SASetValue(&StorePlayer, jx, jy, 0);
				Feld[A2TO1(jx,jy)] = EL_EMPTY;
			}
		}
	}

	// when recording the game, store which players take part in the game
	if (tape.recording)
	{
			if (stored_player_.active)
				tape.player_participates[0] = TRUE;
	}

	if (BorderElement == EL_EMPTY)
	{
		SBX_Left = 0;
		SBX_Right = lev_fieldx - SCR_FIELDX;
		SBY_Upper = 0;
		SBY_Lower = lev_fieldy - SCR_FIELDY;
	}
	else
	{
		SBX_Left = -1;
		SBX_Right = lev_fieldx - SCR_FIELDX + 1;
		SBY_Upper = -1;
		SBY_Lower = lev_fieldy - SCR_FIELDY + 1;
	}

	if (lev_fieldx + (SBX_Left == -1 ? 2 : 0) <= SCR_FIELDX)
		SBX_Left = SBX_Right = -1 * (SCR_FIELDX - lev_fieldx) / 2;

	if (lev_fieldy + (SBY_Upper == -1 ? 2 : 0) <= SCR_FIELDY)
		SBY_Upper = SBY_Lower = -1 * (SCR_FIELDY - lev_fieldy) / 2;

	// if local player not found, look for custom element that might create
	//   the player (make some assumptions about the right custom element)
	if (!local_player->present)
	{
		int start_x = 0, start_y = 0;
/*		int found_rating = 0;
		int found_element = EL_UNDEFINED;
*/
		for (y = 0; y < lev_fieldy; y++) for (x = 0; x < lev_fieldx; x++)
		{
			int element = Feld[A2TO1(x,y)];
//			int content;
			int xx, yy;
//			boolean is_player;

			if (!IS_CUSTOM_ELEMENT(element))
				continue;

			if (CAN_CHANGE(element))
			{
/*				for (i = 0; i < element_info[element].num_change_pages; i++)
				{
					content = element_info[element].change_page[i].target_element;
					is_player = ELEM_IS_PLAYER(content);

					if (is_player && (found_rating < 3 || element < found_element))
					{
						start_x = x;
						start_y = y;

						found_rating = 3;
						found_element = element;
					}
				}
*/			}

			for (yy = 0; yy < 3; yy++) for (xx = 0; xx < 3; xx++)
			{
/*				content = element_info[element].content[xx][yy];
				is_player = ELEM_IS_PLAYER(content);

				if (is_player && (found_rating < 2 || element < found_element))
				{
					start_x = x + xx - 1;
					start_y = y + yy - 1;

					found_rating = 2;
					found_element = element;
				}

				if (!CAN_CHANGE(element))
					continue;

				for (i = 0; i < element_info[element].num_change_pages; i++)
				{
					content = element_info[element].change_page[i].target_content[xx][yy];
					is_player = ELEM_IS_PLAYER(content);

					if (is_player && (found_rating < 1 || element < found_element))
					{
						start_x = x + xx - 1;
						start_y = y + yy - 1;

						found_rating = 1;
						found_element = element;
					}
				}
*/			}
		}

		scroll_x = (start_x < SBX_Left  + MIDPOSX ? SBX_Left :
					start_x > SBX_Right + MIDPOSX ? SBX_Right :
					start_x - MIDPOSX);

		scroll_y = (start_y < SBY_Upper + MIDPOSY ? SBY_Upper :
					start_y > SBY_Lower + MIDPOSY ? SBY_Lower :
					start_y - MIDPOSY);
	}
	else
	{
		scroll_x = (local_player->jx < SBX_Left  + MIDPOSX ? SBX_Left :
					local_player->jx > SBX_Right + MIDPOSX ? SBX_Right :
					local_player->jx - MIDPOSX);

		scroll_y = (local_player->jy < SBY_Upper + MIDPOSY ? SBY_Upper :
					local_player->jy > SBY_Lower + MIDPOSY ? SBY_Lower :
					local_player->jy - MIDPOSY);
	}

	CloseDoor(DOOR_CLOSE_1);

	DrawLevel();
	DrawAllPlayers();

	// after drawing the level, correct some elements
	if (game.timegate_time_left == 0)
		CloseAllOpenTimegates();
/*
  if (setup.soft_scrolling)
    BlitBitmap(fieldbuffer, backbuffer, FX, FY, SXSIZE, SYSIZE, SX, SY);
*/
	redraw_mask |= REDRAW_FROM_BACKBUFFER;
	FadeToFront();
/*
  // copy default game door content to main double buffer 
  BlitBitmap(graphic_info[IMG_GLOBAL_DOOR].bitmap, drawto,
	     DOOR_GFX_PAGEX5, DOOR_GFX_PAGEY1, DXSIZE, DYSIZE, DX, DY);

*/
	EndShowingLevelName();
	DrawGameDoorValues();
/*
	UnmapGameButtons();
	UnmapTapeButtons();
	MapGameButtons();
	MapTapeButtons();

	// copy actual game door content to door double buffer for OpenDoor()
	BlitBitmap(drawto, bitmap_db_door,
				DX, DY, DXSIZE, DYSIZE, DOOR_GFX_PAGEX1, DOOR_GFX_PAGEY1);
*/
	OpenDoor(DOOR_OPEN_ALL);

//	KeyboardAutoRepeatOffUnlessAutoplay();
	ResetTicks();//PALM!!!
}

void InitMovDir(int x, int y)
{
	short int xy=A2TO1(x,y);
	int i, element = Feld[xy];
	const int xy_x[4]={0,1,0,-1};
	const int xy_y[4]={1,0,-1,0}; //for palm only const , was static
/*	const int direction[3][4] = 
	{
		{ MV_RIGHT, MV_UP,   MV_LEFT,  MV_DOWN },
		{ MV_LEFT,  MV_DOWN, MV_RIGHT, MV_UP },
		{ MV_LEFT,  MV_RIGHT, MV_UP, MV_DOWN }
	};*/
	const int dir_RULD[4]={ MV_RIGHT, MV_UP,   MV_LEFT,  MV_DOWN };
	const int dir_LDRU[4]={ MV_LEFT,  MV_DOWN, MV_RIGHT, MV_UP };
	const int dir_LRUD[4]={ MV_LEFT,  MV_RIGHT, MV_UP, MV_DOWN };

	switch(element)
	{
		case EL_BUG_RIGHT:
		case EL_BUG_UP:
		case EL_BUG_LEFT:
		case EL_BUG_DOWN:
			Feld[xy] = EL_BUG;
			MovDir[xy] = dir_RULD[element - EL_BUG_RIGHT];
			break;

		case EL_SPACESHIP_RIGHT:
		case EL_SPACESHIP_UP:
		case EL_SPACESHIP_LEFT:
		case EL_SPACESHIP_DOWN:
			Feld[xy] = EL_SPACESHIP;
			MovDir[xy] = dir_RULD[element - EL_SPACESHIP_RIGHT];
			break;

		case EL_BD_BUTTERFLY_RIGHT:
		case EL_BD_BUTTERFLY_UP:
		case EL_BD_BUTTERFLY_LEFT:
		case EL_BD_BUTTERFLY_DOWN:
			Feld[xy] = EL_BD_BUTTERFLY;
			MovDir[xy] = dir_RULD[element - EL_BD_BUTTERFLY_RIGHT];
			break;

		case EL_BD_FIREFLY_RIGHT:
		case EL_BD_FIREFLY_UP:
		case EL_BD_FIREFLY_LEFT:
		case EL_BD_FIREFLY_DOWN:
			Feld[xy] = EL_BD_FIREFLY;
			MovDir[xy] = dir_RULD[element - EL_BD_FIREFLY_RIGHT];
			break;

		case EL_PACMAN_RIGHT:
		case EL_PACMAN_UP:
		case EL_PACMAN_LEFT:
		case EL_PACMAN_DOWN:
			Feld[xy] = EL_PACMAN;
			MovDir[xy] = dir_RULD[element - EL_PACMAN_RIGHT];
			break;

		case EL_SP_SNIKSNAK:
			MovDir[xy] = MV_UP;
			break;

		case EL_SP_ELECTRON:
			MovDir[xy] = MV_LEFT;
			break;

		case EL_MOLE_LEFT:
		case EL_MOLE_RIGHT:
		case EL_MOLE_UP:
		case EL_MOLE_DOWN:
			Feld[xy] = EL_MOLE;
			MovDir[xy] = dir_LRUD[element - EL_MOLE_LEFT];
			break;

		default:
			if (IS_CUSTOM_ELEMENT(element))
			{
/*	struct ElementInfo *ei = &element_info[element];
	int move_direction_initial = ei->move_direction_initial;
	int move_pattern = ei->move_pattern;

	if (move_direction_initial == MV_START_PREVIOUS)
	{
	  if (MovDir[x][y] != MV_NO_MOVING)
	    return;

	  move_direction_initial = MV_START_AUTOMATIC;
	}

	if (move_direction_initial == MV_START_RANDOM)
	  MovDir[x][y] = 1 << RND(4);
	else if (move_direction_initial & MV_ANY_DIRECTION)
	  MovDir[x][y] = move_direction_initial;
	else if (move_pattern == MV_ALL_DIRECTIONS ||
		 move_pattern == MV_TURNING_LEFT ||
		 move_pattern == MV_TURNING_RIGHT ||
		 move_pattern == MV_TURNING_LEFT_RIGHT ||
		 move_pattern == MV_TURNING_RIGHT_LEFT ||
		 move_pattern == MV_TURNING_RANDOM)
	  MovDir[x][y] = 1 << RND(4);
	else if (move_pattern == MV_HORIZONTAL)
	  MovDir[x][y] = (RND(2) ? MV_LEFT : MV_RIGHT);
	else if (move_pattern == MV_VERTICAL)
	  MovDir[x][y] = (RND(2) ? MV_UP : MV_DOWN);
	else if (move_pattern & MV_ANY_DIRECTION)
	  MovDir[x][y] = element_info[element].move_pattern;
	else if (move_pattern == MV_ALONG_LEFT_SIDE ||
		 move_pattern == MV_ALONG_RIGHT_SIDE)
	{
	  // use random direction as default start direction
	  if (game.engine_version >= VERSION_IDENT(3,1,0,0))
	    MovDir[x][y] = 1 << RND(4);

	  for (i = 0; i < NUM_DIRECTIONS; i++)
	  {
	    int x1 = x + xy[i][0];
	    int y1 = y + xy[i][1];

	    if (!IN_LEV_FIELD(x1, y1) || !IS_FREE(x1, y1))
	    {
	      if (move_pattern == MV_ALONG_RIGHT_SIDE)
		MovDir[x][y] = direction[0][i];
	      else
		MovDir[x][y] = direction[1][i];

	      break;
	    }
	  }
	}		 
*/			}
			else
			{
				MovDir[xy] = 1 << RND(4);

				if (element != EL_BUG &&
								element != EL_SPACESHIP &&
								element != EL_BD_BUTTERFLY &&
								element != EL_BD_FIREFLY)
					break;

				for (i = 0; i < NUM_DIRECTIONS; i++)
				{
					int x1 = x + xy_x[i];
					int y1 = y + xy_y[i];

					if (!IN_LEV_FIELD(x1, y1) || !IS_FREE(A2TO1(x1,y1)))
					{
						if (element == EL_BUG || element == EL_BD_BUTTERFLY)
						{
							MovDir[xy] = dir_RULD[i];
							break;
						}
						else if (element == EL_SPACESHIP || element == EL_BD_FIREFLY ||
								element == EL_SP_SNIKSNAK || element == EL_SP_ELECTRON)
						{
							MovDir[xy] = dir_LDRU[i];
							break;
						}
					}
				}
			}
			break;
	}

//	GfxDir[x][y] = MovDir[x][y];
}

void InitAmoebaNr(int x, int y)
{
	int i;
	int group_nr = AmoebeNachbarNr(x, y);

	if (group_nr == 0)
	{
		for (i = 1; i < MAX_NUM_AMOEBA; i++)
		{
			if (AmoebaCnt[i] == 0)
			{
				group_nr = i;
				break;
			}
		}
	}

	AmoebaNr[A2TO1(x,y)] = group_nr;
	AmoebaCnt[group_nr]++;
	AmoebaCnt2[group_nr]++;
}

void GameWon()
{
/*	int hi_pos;
	boolean raise_level = FALSE;
*/
	short int ExitXY;
	
	if (local_player->MovPos)
		return;

	local_player->LevelSolved = FALSE;

	if (TimeLeft)
	{
		while (TimeLeft > 0)
		{
			if (TimeLeft > 0 && !(TimeLeft % 10))
				RaiseScore(level.score[SC_TIME_BONUS]);
			if (TimeLeft > 100 && !(TimeLeft % 10))
				TimeLeft -= 10;
			else
				TimeLeft--;
			
			DrawGameValue_Time(TimeLeft);
			
			BackToFront();

			if (!tape.playing)
				DelayX10(1);
		}
	}
	else if (level.time == 0)		// level without time limit 
	{
		while (TimePlayed < 999)
		{
			if (TimePlayed < 999 && !(TimePlayed % 10))
				RaiseScore(level.score[SC_TIME_BONUS]);
			if (TimePlayed < 900 && !(TimePlayed % 10))
				TimePlayed += 10;
			else
				TimePlayed++;

			DrawGameValue_Time(TimePlayed);

			BackToFront();

			if (!tape.playing)
				DelayX10(1);
		}
	}

	ExitXY=A2TO1(ExitX, ExitY);
	
	// close exit door after last player
	if ((Feld[ExitXY] == EL_EXIT_OPEN ||
			Feld[ExitXY] == EL_SP_EXIT_OPEN) && AllPlayersGone)
	{
		int element = Feld[ExitXY];
		if(element == EL_EXIT_OPEN)
		{
			InitGfxValueAndFeld(ExitXY,EL_EXIT_CLOSING);
//			AddDelayChanging(ExitXY,29,EL_EXIT_CLOSED);//PALM!!!
		}
		else
		{
			InitGfxValueAndFeld(ExitXY,EL_SP_EXIT_CLOSING);
//			AddDelayChanging(ExitXY,29,EL_SP_EXIT_CLOSED);//PALM!!!
		}
	}

	// Hero disappears
	DrawLevelField(ExitX, ExitY);
	BackToFront();

	if (tape.playing)
		return;

	CloseDoor(DOOR_CLOSE_1);

	if (tape.recording)
	{
		TapeStop();
//		SaveTape(tape.level_nr);		// Ask to save tape
	}

/*  if (level_nr == leveldir_current->handicap_level)
  {
    leveldir_current->handicap_level++;
    SaveLevelSetup_SeriesInfo();
  }

  if (level_editor_test_game)
    local_player->score = -1;	// no highscore when playing from editor
  else if (level_nr < leveldir_current->last_level)
    raise_level = TRUE;		// advance to next level

  if ((hi_pos = NewHiScore()) >= 0) 
  {
    game_status = GAME_MODE_SCORES;
    DrawHallOfFame(hi_pos);
    if (raise_level)
    {
      level_nr++;
      TapeErase();
    }
  }
  else
  {
    game_status = GAME_MODE_MAIN;
    if (raise_level)
    {
      level_nr++;
      TapeErase();
    }
    DrawMainMenu();
  }
*/
	BackToFront();
	//PALM!!!
	TellGameWon();
}

void InitPlayerGfxAnimation(struct PlayerInfo *player, int action, int dir)
{
	if (player->GfxAction != action || player->GfxDir != dir)
	{
		player->GfxAction = action;
		player->GfxDir = dir;
		player->Frame = 0;
		player->StepFrame = 0;
	}
};

void InitMovingField(short int xy, char direction)
{
	int dx = (direction == MV_LEFT ? -1 : direction == MV_RIGHT ? +1 : 0);
	int dy = (direction == MV_UP   ? -1 : direction == MV_DOWN  ? +1 : 0);

	short int newxy=xy+A2TO1(dx,dy);
	
	if (!IsWasJustMoving(xy) || direction != MovDir[xy])
		ResetGfxAnimation(xy);

	MovDir[newxy] = MovDir[xy] = direction;
//	GfxDir[x][y] = direction;

	if (Feld[newxy] == EL_EMPTY)
		Feld[newxy] = EL_BLOCKED;

	if (direction == MV_DOWN && CAN_FALL(Feld[xy]))
		GfxAction[xy] = ACTION_FALLING;
	else
		GfxAction[xy] = ACTION_MOVING;

	GfxSFrame[newxy] = GfxSFrame[xy];//GfxFrame[newx][newy] = GfxFrame[x][y];
	GfxRandom[newxy] = GfxRandom[xy];
	GfxAction[newxy] = GfxAction[xy];
//	GfxDir[newx][newy] = GfxDir[x][y];
}

void Moving2Blocked(short int xy, short int *goes_to_xy)
{
	char direction = MovDir[xy];

	*goes_to_xy = xy+A2TO1(
				(direction == MV_LEFT ? -1 : direction == MV_RIGHT ? +1 : 0),
				(direction == MV_UP   ? -1 : direction == MV_DOWN  ? +1 : 0));
} 

void Blocked2Moving(short int xy, short int *comes_from_xy)
{
	int oldxy = xy;
	int direction = MovDir[xy];

	if (direction == MV_LEFT)
		oldxy+=A2TO1(1,0);
	else if (direction == MV_RIGHT)
		oldxy-=A2TO1(1,0);
	else if (direction == MV_UP)
		oldxy+=A2TO1(0,1);
	else if (direction == MV_DOWN)
		oldxy-=A2TO1(0,1);

	*comes_from_xy = oldxy;
}

int MovingOrBlocked2Element(int x, int y)
{
	short int xy=A2TO1(x,y);
	int element = Feld[xy];

	if (element == EL_BLOCKED)
	{
		short int oldxy;

		Blocked2Moving(xy, &oldxy);
		return Feld[oldxy];
	}
	else
		return element;
}

static int MovingOrBlocked2ElementIfNotLeaving(short int xy)
{
	// like MovingOrBlocked2Element(), but if element is moving
	//   and (x,y) is the field the moving element is just leaving,
	//   return EL_BLOCKED instead of the element value
	int element = Feld[xy];

	if (IS_MOVING(xy))
	{
		if (element == EL_BLOCKED)
		{
			short int oldxy;

			Blocked2Moving(xy, &oldxy);
			return Feld[oldxy];
		}
		else
			return EL_BLOCKED;
	}
	else
		return element;
};

LOGIC_SECTION static void RemoveField(short int xy)
{
	Feld[xy] = EL_EMPTY;

	MovPos[xy] = 0;
	MovDir[xy] = 0;
	MovDelay[xy] = 0;

	AmoebaNr[xy] = 0;
/*	ChangeDelay[x][y] = 0;
	ChangePage[x][y] = -1;*/
	setPushedFALSE(xy);//Pushed[x][y] = FALSE;

	GfxElement[xy] = EL_UNDEFINED;
	GfxAction[xy] = ACTION_DEFAULT;
//	Gfx[xy] = IMG_EMPTY;
//	GfxDir[x][y] = MV_NO_MOVING;
}

void RemoveMovingField(int x, int y)
{
	short int xy=A2TO1(x,y);
	short int newxy=xy;
	short int oldxy=xy;
	int element = Feld[xy];
	int next_element = EL_UNDEFINED;

	if (element != EL_BLOCKED && !IS_MOVING(xy))
		return;

	if (IS_MOVING(xy))
	{
		Moving2Blocked(xy, &newxy);
		if (Feld[newxy] != EL_BLOCKED)
		{
			//element is moving, but target field is not free (blocked), but
			//already occupied by something different (example: acid pool);
			//in this case, only remove the moving field, but not the target

			RemoveField(oldxy);
			
			Store[oldxy] = Store2[oldxy] = 0;

			DrawLevelField(A1TOX(oldxy), A1TOY(oldxy));

			return;
		}
	}
	else if (element == EL_BLOCKED)
	{
		Blocked2Moving(xy, &oldxy);
		if (!IS_MOVING(oldxy))
			return;
	}

	if (element == EL_BLOCKED)
	{
		short int el=Feld[oldxy];
		if(el== EL_QUICKSAND_EMPTYING)
			next_element=EL_QUICKSAND_EMPTY;
		else if(el==EL_MAGIC_WALL_EMPTYING)
			next_element=EL_MAGIC_WALL_ACTIVE;
		else if(el==EL_BD_MAGIC_WALL_EMPTYING)
			next_element=EL_BD_MAGIC_WALL_ACTIVE;
		else if(el==EL_AMOEBA_DROPPING)
			next_element = EL_AMOEBA_WET;
	}

	RemoveField(oldxy);
	RemoveField(newxy);
	Store[oldxy]=0;
	Store2[oldxy] = 0;

	if (next_element != EL_UNDEFINED)
		Feld[oldxy]=next_element;//InitGfxValueAndFeld(oldxy,next_element);

	DrawLevelField(A1TOX(oldxy), A1TOY(oldxy));
	DrawLevelField(A1TOX(newxy), A1TOY(newxy));
};

void DrawDynamite(int x, int y)
{
	int sx = SCREENX(x), sy = SCREENY(y);
	short int xy=A2TO1(x,y);
	int graphic = el2img(Feld[xy]);
	int frame;

	if (!IN_SCR_FIELD(sx, sy) || IS_PLAYER(xy))
		return;

	if (IS_WALKABLE_INSIDE(Back[xy]))
		return;

	if (Back[xy])
		DrawGraphic(sx, sy, el2img(Back[xy]), 0);
	else if (Store[xy])
		DrawGraphic(sx, sy, el2img(Store[xy]), 0);

	frame = getGraphicAnimationFrame(graphic, FrameCounter - GfxSFrame[xy]);
														//was GfxFrame[x][y]

	if (Back[xy] || Store[xy])
		DrawGraphicThruMask(sx, sy, graphic, frame);
	else
		DrawGraphic(sx, sy, graphic, frame);
};

LOGIC_SECTION void CheckDynamite(int x, int y)
{
	short int xy=A2TO1(x,y);
	if (MovDelay[xy] != 0)	// dynamite is still waiting to explode
	{
		MovDelay[xy]--;

		if (MovDelay[xy] != 0)
		{
			DrawDynamite(x, y);

			return;
		}
	}

	Bang(x, y);
};

void RelocatePlayer(int x, int y, int element)
{
	struct PlayerInfo *player = &stored_player_;//was [element - EL_PLAYER_1]

	RemoveField(A2TO1(x, y));		// temporarily remove newly placed player
	DrawLevelField(x, y);

	if (player->present)
	{
		while (player->MovPos)
		{
			ScrollPlayer(player, SCROLL_GO_ON);
			ScrollScreen(NULL, SCROLL_GO_ON);
			FrameCounter++;

			DrawPlayer(player);

			BackToFront();
			DelayX10(GAME_FRAME_DELAY/10);
		}

		DrawPlayer(player);		// needed here only to cleanup last field
		DrawLevelField(player->jx, player->jy);	// remove player graphic

		player->is_moving = FALSE;
	}

	InitGfxValueAndFeld(A2TO1(x,y),element);
	InitPlayerField(x, y, element, TRUE);

	if (player == local_player)
	{
		int scroll_xx = -999, scroll_yy = -999;

		while (scroll_xx != scroll_x || scroll_yy != scroll_y)
		{
			int dx = 0, dy = 0;
			int fx = FX, fy = FY;

			scroll_xx = (local_player->jx < SBX_Left  + MIDPOSX ? SBX_Left :
						local_player->jx > SBX_Right + MIDPOSX ? SBX_Right :
						local_player->jx - MIDPOSX);

			scroll_yy = (local_player->jy < SBY_Upper + MIDPOSY ? SBY_Upper :
						local_player->jy > SBY_Lower + MIDPOSY ? SBY_Lower :
						local_player->jy - MIDPOSY);

			dx = (scroll_xx < scroll_x ? +1 : scroll_xx > scroll_x ? -1 : 0);
			dy = (scroll_yy < scroll_y ? +1 : scroll_yy > scroll_y ? -1 : 0);

			scroll_x -= dx;
			scroll_y -= dy;

			fx += dx * TILEX / 2;
			fy += dy * TILEY / 2;

			ScrollLevel(dx, dy);
			DrawAllPlayers();

			// scroll in two steps of half tile size to make things smoother
//			BlitBitmap(drawto_field, window, fx, fy, SXSIZE, SYSIZE, SX, SY);
			FlushDisplay();
			DelayX10(GAME_FRAME_DELAY/10);

			// scroll second step to align at full tile size
			BackToFront();
			DelayX10(GAME_FRAME_DELAY/10);
		}
	}
}

void Explode(int ex, int ey, int phase, int mode)
{
	short int exy=A2TO1(ex,ey);
	short int xy;
	int x, y;
	int num_phase = 9;
	int delay = (game.emulation == EMU_SUPAPLEX ? 3 : 2);
	int last_phase = num_phase * delay;
	int half_phase = (num_phase / 2) * delay;
	int first_phase_after_start = EX_PHASE_START + 1;

	if (game.explosions_delayed)
	{
		setExplodeField(exy,ey,mode);	//	ExplodeField[ex][ey] = mode;
		return;
	}

	if (phase == EX_PHASE_START)		// initialize 'Store[][]' field
	{
		int center_element = Feld[exy];

		// remove things displayed in background while burning dynamite
		if (Back[exy] != EL_EMPTY && !IS_INDESTRUCTIBLE(Back[exy]))
			Back[exy] = 0;

		if (IS_MOVING(exy) || IS_BLOCKED(exy))
		{
			// put moving element to center field (and let it explode there)
			center_element = MovingOrBlocked2Element(ex, ey);
			RemoveMovingField(ex, ey);
			InitGfxValueAndFeld(exy,center_element);
		}

		for (y = ey - 1; y <= ey + 1; y++) for (x = ex - 1; x <= ex + 1; x++)
		{
			int xx = x - ex + 1;
			int yy = y - ey + 1;
			int element;
			short int xy=A2TO1(x,y);

			if (!IN_LEV_FIELD(x, y) ||
					((mode != EX_NORMAL || center_element == EL_AMOEBA_TO_DIAMOND) &&
																(x != ex || y != ey)))
				continue;

			element = Feld[xy];

			if (IS_MOVING(xy) || IS_BLOCKED(xy))
			{
				element = MovingOrBlocked2Element(x, y);

				if (!IS_EXPLOSION_PROOF(element))
					RemoveMovingField(x, y);
			}

			// indestructible elements can only explode in center (but not flames)
			if ((IS_EXPLOSION_PROOF(element) && (x != ex || y != ey)) ||
													element == EL_FLAMES)
				continue;

			if (IS_PLAYER(xy) && SHIELD_ON(PLAYERINFO(x, y)))
			{
				if (IS_ACTIVE_BOMB(element))
				{
					// re-activate things under the bomb like gate or penguin
					InitGfxValueAndFeld(xy,(Back[xy] ? Back[xy] : EL_EMPTY));
					Back[xy]=0;
				}

				continue;
			}

			// save walkable background elements while explosion on same tile
			if (IS_WALKABLE(element) && IS_INDESTRUCTIBLE(element))
				Back[xy] = element;

			// ignite explodable elements reached by other explosion 
			if (element == EL_EXPLOSION)
				element = Store2[xy];
			else
				FeldYExplosion[y]++;

			if (AmoebaNr[xy] &&
								 (element == EL_AMOEBA_FULL ||
								    element == EL_BD_AMOEBA ||
								element == EL_AMOEBA_GROWING))
			{
				AmoebaCnt[AmoebaNr[xy]]--;
				AmoebaCnt2[AmoebaNr[xy]]--;
			}

			RemoveField(xy);

			if (IS_PLAYER(A2TO1(ex,ey)) && !PLAYER_EXPLOSION_PROTECTED(ex, ey, A2TO1(ex,ey)))
			{
/*				switch(StorePlayer[ex][ey])
				{
					case EL_PLAYER_2:
						Store[x][y] = EL_PLAYER_IS_EXPLODING_2;
						break;
					case EL_PLAYER_3:
						Store[x][y] = EL_PLAYER_IS_EXPLODING_3;
						break;
					case EL_PLAYER_4:
						Store[x][y] = EL_PLAYER_IS_EXPLODING_4;
						break;
					case EL_PLAYER_1:
					default:*/
						Store[xy] = EL_PLAYER_IS_EXPLODING_1;
/*						break;
				}
*/
				if (PLAYERINFO(ex, ey)->use_murphy_graphic)
					Store[xy] = EL_EMPTY;
			}
			else if (center_element == EL_MOLE)
				Store[xy] = EL_EMERALD_RED;
			else if (center_element == EL_PENGUIN)
				Store[xy] = EL_EMERALD_PURPLE;
			else if (center_element == EL_BUG)
				Store[xy] = ((x == ex && y == ey) ? EL_DIAMOND : EL_EMERALD);
			else if (center_element == EL_BD_BUTTERFLY)
				Store[xy] = EL_BD_DIAMOND;
			else if (center_element == EL_SP_ELECTRON)
				Store[xy] = EL_SP_INFOTRON;
			else if (center_element == EL_AMOEBA_TO_DIAMOND)
				Store[xy] = level.amoeba_content;
			else if (center_element == EL_YAMYAM)
				Store[xy] = level.yamyam_content[game.yamyam_content_nr][xx][yy];
/*			else if (IS_CUSTOM_ELEMENT(center_element) &&
							element_info[center_element].content[xx][yy] != EL_EMPTY)
				Store[x][y] = element_info[center_element].content[xx][yy];
*/			else if (element == EL_WALL_EMERALD)
				Store[xy] = EL_EMERALD;
			else if (element == EL_WALL_DIAMOND)
				Store[xy] = EL_DIAMOND;
			else if (element == EL_WALL_BD_DIAMOND)
				Store[xy] = EL_BD_DIAMOND;
			else if (element == EL_WALL_EMERALD_YELLOW)
				Store[xy] = EL_EMERALD_YELLOW;
			else if (element == EL_WALL_EMERALD_RED)
				Store[xy] = EL_EMERALD_RED;
			else if (element == EL_WALL_EMERALD_PURPLE)
				Store[xy] = EL_EMERALD_PURPLE;
			else if (element == EL_WALL_PEARL)
				Store[xy] = EL_PEARL;
			else if (element == EL_WALL_CRYSTAL)
				Store[xy] = EL_CRYSTAL;
/*			else if (IS_CUSTOM_ELEMENT(element) && !CAN_EXPLODE(element))
				Store[x][y] = element_info[element].content[1][1];
*/			else
				Store[xy] = EL_EMPTY;

			if (x != ex || y != ey ||
						center_element == EL_AMOEBA_TO_DIAMOND || mode == EX_BORDER)
				Store2[xy] = element;

			Feld[xy] = EL_EXPLOSION;
			
			GfxElement[xy] = center_element;
			setExplodePhase(xy,1);	//ExplodePhase[x][y] = 1;

			setStopTRUE(x,y);	//Stop[x][y]=TRUE;
		}

		if (center_element == EL_YAMYAM)
			game.yamyam_content_nr =
							(game.yamyam_content_nr + 1) % level.num_yamyam_contents;

		return;
	}

	if (getStop(exy))//Stop[ex][ey]
		return;

	x = ex;
	y = ey;
	xy=A2TO1(x,y);

	{
		unsigned char df = (phase < last_phase ? phase + 1 : 0);
		setExplodePhase(xy,df);		//ExplodePhase[x][y] = df;
	}
/*
  if (GfxElement[x][y] == EL_UNDEFINED)
  {
    printf("\n\n");
    printf("Explode(): x = %d, y = %d: GfxElement == EL_UNDEFINED\n", x, y);
    printf("Explode(): This should never happen!\n");
    printf("\n\n");

    GfxElement[x][y] = EL_EMPTY;
  }
*/
	if (phase == first_phase_after_start)
	{
		int element = Store2[xy];

		if (element == EL_BLACK_ORB)
		{
			InitGfxValueAndFeld(xy,Store2[xy]);
			Store2[xy] = 0;
			Bang(x, y);
		}
	}
	else if (phase == half_phase)
	{
		int element = Store2[xy];

		if (IS_PLAYER(xy))
			{KillHeroUnlessExplosionProtected(x, y, xy);goto buggy_else1;}
		if (CAN_EXPLODE_BY_FIRE(element))
		{
			InitGfxValueAndFeld(xy,Store2[xy]);
			Store2[xy] = 0;
			Bang(x, y);
		}
		else if (element == EL_AMOEBA_TO_DIAMOND)
			AmoebeUmwandeln(x, y);
buggy_else1:
	}

	if (phase == last_phase)
	{
		int element;

		element = Feld[xy] = Store[xy];
		Store[xy] = Store2[xy] = 0;
		GfxElement[xy] = EL_UNDEFINED;

		// player can escape from explosions and might therefore be still alive
		if (element == EL_PLAYER_IS_EXPLODING_1)
			Feld[xy] = (stored_player_.active ?
					EL_EMPTY :
/*					element == EL_PLAYER_IS_EXPLODING_1 ? */EL_EMERALD_YELLOW /*:
		    element == EL_PLAYER_IS_EXPLODING_2 ? EL_EMERALD_RED :
		    element == EL_PLAYER_IS_EXPLODING_3 ? EL_EMERALD :
		    EL_EMERALD_PURPLE*/);

		// restore probably existing indestructible background element
		if (Back[xy] && IS_INDESTRUCTIBLE(Back[xy]))
			element = Feld[xy] = Back[xy];
		Back[xy] = 0;

		if(element!=EL_EXPLOSION)
		{
			if(FeldYExplosion[y]>0)
				FeldYExplosion[y]--;
			else
				bugd("Unbeleivable! FeldYExplosion=0",0);
		}
		
		MovDir[xy] = MovPos[xy] = MovDelay[xy] = 0;
/*		GfxDir[x][y] = MV_NO_MOVING;
		ChangeDelay[x][y] = 0;
		ChangePage[x][y] = -1;
*/
		InitField(x, y, FALSE);
		if (CAN_MOVE(element))
			InitMovDir(x, y);

		InitGfxValue(xy);

		DrawLevelField(x, y);

//		TestIfElementTouchesCustomElement(x, y);

		if (GFX_CRUMBLED(element))
			DrawLevelFieldCrumbledSandNeighbours(x, y);

		if (IS_PLAYER(xy) && !PLAYERINFO(x,y)->present)
			StorePlayerXY=-1;//[x][y]=0;

		if (ELEM_IS_PLAYER(element))
			RelocatePlayer(x, y, element);
	}
	else if (phase >= delay && IN_SCR_FIELD(SCREENX(x), SCREENY(y)))
	{
		int graphic = GetExplodingGfx(GfxElement[xy]);
		//was el_act2img( GfxElement[x][y], ACTION_EXPLODING);
		int frame = getGraphicAnimationFrame(graphic, phase - delay);

		if (phase == delay)
			DrawLevelFieldCrumbledSand(x, y);

		if (IS_WALKABLE_OVER(Back[xy]) && Back[xy] != EL_EMPTY)
		{
			DrawLevelElement(x, y, Back[xy]);
			DrawGraphicThruMask(SCREENX(x), SCREENY(y), graphic, frame);
		}
		else if (IS_WALKABLE_UNDER(Back[xy]))
		{
			DrawGraphic(SCREENX(x), SCREENY(y), graphic, frame);
			DrawLevelElementThruMask(x, y, Back[xy]);
		}
		else if (!IS_WALKABLE_INSIDE(Back[xy]))
			DrawGraphic(SCREENX(x), SCREENY(y), graphic, frame);
	}
}

void DynaExplode(int ex, int ey)
{
	int i, j;
	int dynabomb_element = Feld[A2TO1(ex,ey)];
	int dynabomb_size = 1;
	boolean dynabomb_xl = FALSE;
	struct PlayerInfo *player;
	const int xy_x[4]={0,-1,+1,0}; //for palm only, was static
	const int xy_y[4]={-1,0,0,+1}; //for palm only, was static

	if (IS_ACTIVE_BOMB(dynabomb_element))
	{
		player = &stored_player_;//was [dynabomb_element - EL_DYNABOMB_PLAYER_1_ACTIVE]
		dynabomb_size = player->dynabomb_size;
		dynabomb_xl = player->dynabomb_xl;
		player->dynabombs_left++;
	}

	Explode(ex, ey, EX_PHASE_START, EX_CENTER);
	//new Explode(ex, ey, EX_PHASE_START, EX_TYPE_CENTER);

	for (i = 0; i < NUM_DIRECTIONS; i++)
	{
		for (j = 1; j <= dynabomb_size; j++)
		{
			int x = ex + j * xy_x[i];
			int y = ey + j * xy_y[i];
			int element;
			short int xy=A2TO1(x,y);

			if (!IN_LEV_FIELD(x, y) || IS_INDESTRUCTIBLE(Feld[xy]))
				break;

			element = Feld[xy];

			// do not restart explosions of fields with active bombs 
			if (element == EL_EXPLOSION && IS_ACTIVE_BOMB(Store2[xy]))
				continue;

			Explode(x, y, EX_PHASE_START, EX_BORDER);
			//new Explode(x, y, EX_PHASE_START, EX_TYPE_BORDER);

			if (element != EL_EMPTY && element != EL_EXPLOSION &&
							!IS_DIGGABLE(element) && !dynabomb_xl)
				break;
		}
	}
}

void Bang(int x, int y)
{
	int element = MovingOrBlocked2Element(x, y);

	if (IS_PLAYER(A2TO1(x,y)) && !PLAYER_EXPLOSION_PROTECTED(x, y, A2TO1(x,y)))
	{
		struct PlayerInfo *player = PLAYERINFO(x, y);

		element = Feld[A2TO1(x,y)] = (player->use_murphy_graphic ? EL_SP_MURPHY :
														player->element_nr);
	}

	switch(element)
	{
		case EL_BUG:
		case EL_SPACESHIP:
		case EL_BD_BUTTERFLY:
		case EL_BD_FIREFLY:
		case EL_YAMYAM:
		case EL_DARK_YAMYAM:
		case EL_ROBOT:
		case EL_PACMAN:
		case EL_MOLE:
			RaiseScoreElement(element);
			Explode(x, y, EX_PHASE_START, EX_NORMAL);
			//new Explode(x, y, EX_PHASE_START, EX_TYPE_NORMAL);
			break;
		case EL_DYNABOMB_PLAYER_1_ACTIVE:
		case EL_DYNABOMB_PLAYER_2_ACTIVE:
		case EL_DYNABOMB_PLAYER_3_ACTIVE:
		case EL_DYNABOMB_PLAYER_4_ACTIVE:
		case EL_DYNABOMB_INCREASE_NUMBER:
		case EL_DYNABOMB_INCREASE_SIZE:
		case EL_DYNABOMB_INCREASE_POWER:
			DynaExplode(x, y);
			break;
		case EL_PENGUIN:
		case EL_LAMP:
		case EL_LAMP_ACTIVE:
		case EL_AMOEBA_TO_DIAMOND:
			if (IS_PLAYER(A2TO1(x,y)))
				Explode(x, y, EX_PHASE_START, EX_NORMAL);
				//new Explode(x, y, EX_PHASE_START, EX_TYPE_NORMAL);
			else
				Explode(x, y, EX_PHASE_START, EX_CENTER);
				//new Explode(x, y, EX_PHASE_START, EX_TYPE_CENTER);
			break;
		default:
/*new      if (element_info[element].explosion_type == EXPLODES_CROSS)
	Explode(x, y, EX_PHASE_START, EX_TYPE_CROSS);
      else if (element_info[element].explosion_type == EXPLODES_1X1)
	Explode(x, y, EX_PHASE_START, EX_TYPE_CENTER);
      else
	Explode(x, y, EX_PHASE_START, EX_TYPE_NORMAL);*/
			Explode(x, y, EX_PHASE_START, EX_NORMAL);
			break;
	}

//	CheckTriggeredElementChange(x, y, element, CE_OTHER_IS_EXPLODING);
}

void SplashAcid(int x, int y)
{
	short int xy=A2TO1(x,y);

	if (IN_LEV_FIELD(x - 1, y - 1) && IS_FREE(xy+A2TO1(-1,-1)) &&
				(!IN_LEV_FIELD(x - 1, y - 2) ||
				!CAN_FALL(MovingOrBlocked2Element(x - 1, y - 2))))
	{
		InitGfxValueAndFeld(xy+A2TO1(-1,-1),EL_ACID_SPLASH_LEFT);
//		AddDelayChanging(xy+A2TO1(-1,-1),8,EL_EMPTY);//PALM!!!
//		GfxSFrame[xy+A2TO1(-1,-1)]=FrameCounter;//absolutely HACKING!
	}

	if (IN_LEV_FIELD(x + 1, y - 1) && IS_FREE(xy+A2TO1(1,-1)) &&
				(!IN_LEV_FIELD(x + 1, y - 2) ||
				!CAN_FALL(MovingOrBlocked2Element(x + 1, y - 2))))
	{
		InitGfxValueAndFeld(xy+A2TO1(1,-1),EL_ACID_SPLASH_RIGHT);
//		AddDelayChanging(xy+A2TO1(1,-1),8,EL_EMPTY);//PALM!!!
//		GfxSFrame[xy+A2TO1(1,-1)]=FrameCounter;//absolutely HACKING!
	}
}

INIT_SECTION static void InitBeltMovement()
{
	const int belt_base_element[4] = //for palm only, was static
	{
		EL_CONVEYOR_BELT_1_LEFT,
		EL_CONVEYOR_BELT_2_LEFT,
		EL_CONVEYOR_BELT_3_LEFT,
		EL_CONVEYOR_BELT_4_LEFT
	};
	const int belt_base_active_element[4] = //for palm only, was static
	{
		EL_CONVEYOR_BELT_1_LEFT_ACTIVE,
		EL_CONVEYOR_BELT_2_LEFT_ACTIVE,
		EL_CONVEYOR_BELT_3_LEFT_ACTIVE,
		EL_CONVEYOR_BELT_4_LEFT_ACTIVE
	};

	int x, y, i, j;

	// set frame order for belt animation graphic according to belt direction
	for (i = 0; i < NUM_BELTS; i++)
	{
		int belt_nr = i;

		for (j = 0; j < NUM_BELT_PARTS; j++)
		{
			int element = belt_base_active_element[belt_nr] + j;
			byte anim = ANIM_INDEX(el2img(element));
			if(anim!=1)
			{
				if (game.belt_dir[i] == MV_LEFT)
					anim_infoF->anim_mode[anim] &= ~ANIM_REVERSE;
				else
					anim_infoF->anim_mode[anim] |=  ANIM_REVERSE;
			}
		}
	}

	for (y = 0; y < lev_fieldy; y++)
	{
		register short int xy=A2TO1(0,y);
		for (x = 0; x < lev_fieldx; x++,xy+=A2TO1(1,0))
		{
			int element = Feld[xy];

			for (i = 0; i < NUM_BELTS; i++)
			{
				if (IS_BELT(element) && game.belt_dir[i] != MV_NO_MOVING)
				{
					int e_belt_nr = getBeltNrFromBeltElement(element);
					int belt_nr = i;

					if (e_belt_nr == belt_nr)
					{
						int belt_part = Feld[xy] - belt_base_element[belt_nr];

						InitGfxValueAndFeld(xy,
									belt_base_active_element[belt_nr] + belt_part);
					}
				}
			}
		}
	}
};

LOGIC_SECTION static void ToggleBeltSwitch(int x, int y)
{
	const int belt_base_element[4] = //for palm only, was static
	{
		EL_CONVEYOR_BELT_1_LEFT,
		EL_CONVEYOR_BELT_2_LEFT,
		EL_CONVEYOR_BELT_3_LEFT,
		EL_CONVEYOR_BELT_4_LEFT
	};
	const int belt_base_active_element[4] = //for palm only, was static
	{
		EL_CONVEYOR_BELT_1_LEFT_ACTIVE,
		EL_CONVEYOR_BELT_2_LEFT_ACTIVE,
		EL_CONVEYOR_BELT_3_LEFT_ACTIVE,
		EL_CONVEYOR_BELT_4_LEFT_ACTIVE
	};
	const int belt_base_switch_element[4] = //for palm only, was static
	{
		EL_CONVEYOR_BELT_1_SWITCH_LEFT,
		EL_CONVEYOR_BELT_2_SWITCH_LEFT,
		EL_CONVEYOR_BELT_3_SWITCH_LEFT,
		EL_CONVEYOR_BELT_4_SWITCH_LEFT
	};
	const int belt_move_dir[4] = //for palm only, was static
	{
		MV_LEFT,
		MV_NO_MOVING,
		MV_RIGHT,
		MV_NO_MOVING,
	};

	int element = Feld[A2TO1(x,y)];
	int belt_nr = getBeltNrFromBeltSwitchElement(element);
	int belt_dir_nr = (game.belt_dir_nr[belt_nr] + 1) % 4;
	int belt_dir = belt_move_dir[belt_dir_nr];
	int xx, yy, i;

	if (!IS_BELT_SWITCH(element))
		return;

	game.belt_dir_nr[belt_nr] = belt_dir_nr;
	game.belt_dir[belt_nr] = belt_dir;

	if (belt_dir_nr == 3)
		belt_dir_nr = 1;

	// set frame order for belt animation graphic according to belt direction 
	for (i = 0; i < NUM_BELT_PARTS; i++)
	{
		int element = belt_base_active_element[belt_nr] + i;
		byte anim = ANIM_INDEX(el2img(element));

		if(anim!=1)
		{
			if (belt_dir == MV_LEFT)
				anim_infoF->anim_mode[anim] &= ~ANIM_REVERSE;
			else
				anim_infoF->anim_mode[anim] |=  ANIM_REVERSE;
		}
	}

	for (yy = 0; yy < lev_fieldy; yy++)
	{
		register short int xxyy=A2TO1(0,yy);
		for (xx = 0; xx < lev_fieldx; xx++,xxyy+=A2TO1(1,0))
		{
			int element = Feld[xxyy];

			if (IS_BELT_SWITCH(element))
			{
				int e_belt_nr = getBeltNrFromBeltSwitchElement(element);

				if (e_belt_nr == belt_nr)
				{
					InitGfxValueAndFeld(xxyy,
								belt_base_switch_element[belt_nr] + belt_dir_nr);
					DrawLevelField(xx, yy);
				}
			}
			else if (IS_BELT(element) && belt_dir != MV_NO_MOVING)
			{
				int e_belt_nr = getBeltNrFromBeltElement(element);

				if (e_belt_nr == belt_nr)
				{
					int belt_part = Feld[xxyy] - belt_base_element[belt_nr];

					InitGfxValueAndFeld(xxyy,
								belt_base_active_element[belt_nr] + belt_part);
					DrawLevelField(xx, yy);
				}
			}
			else if (IS_BELT_ACTIVE(element) && belt_dir == MV_NO_MOVING)
			{
				int e_belt_nr = getBeltNrFromBeltActiveElement(element);

				if (e_belt_nr == belt_nr)
				{
					int belt_part = Feld[xxyy] - belt_base_active_element[belt_nr];

					InitGfxValueAndFeld(xxyy, belt_base_element[belt_nr] + belt_part);
					DrawLevelField(xx, yy);
				}
			}
		}
	}
};

LOGIC_SECTION static void ToggleSwitchgateSwitch(int x, int y)
{
	int xx, yy;

	game.switchgate_pos = !game.switchgate_pos;

	for (yy = 0; yy < lev_fieldy; yy++)
	{
		register short int xxyy=A2TO1(0,yy);
		for (xx = 0; xx < lev_fieldx; xx++,xxyy+=A2TO1(1,0))
		{
			int element = Feld[xxyy];

			if (element == EL_SWITCHGATE_SWITCH_UP ||
							element == EL_SWITCHGATE_SWITCH_DOWN)
			{
				InitGfxValueAndFeld(xxyy,
							EL_SWITCHGATE_SWITCH_UP + game.switchgate_pos);
				DrawLevelField(xx, yy);
			}
			else if (element == EL_SWITCHGATE_OPEN ||
							element == EL_SWITCHGATE_OPENING)
			{
				InitGfxValueAndFeld(xxyy,EL_SWITCHGATE_CLOSING);
//				AddDelayChanging(xxyy,29,EL_SWITCHGATE_CLOSED);//PALM!!!
			}
			else if (element == EL_SWITCHGATE_CLOSED ||
							element == EL_SWITCHGATE_CLOSING)
			{
				InitGfxValueAndFeld(xxyy,EL_SWITCHGATE_OPENING);
//				AddDelayChanging(xxyy,29,EL_SWITCHGATE_OPEN);//PALM!!!
			}
		}
	}
};

LOGIC_SECTION static int getInvisibleActiveFromInvisibleElement(int element)
{
	return (element == EL_INVISIBLE_STEELWALL ? EL_INVISIBLE_STEELWALL_ACTIVE :
			element == EL_INVISIBLE_WALL      ? EL_INVISIBLE_WALL_ACTIVE :
			element == EL_INVISIBLE_SAND      ? EL_INVISIBLE_SAND_ACTIVE :
			element);
};

LOGIC_SECTION static int getInvisibleFromInvisibleActiveElement(int element)
{
	return (element == EL_INVISIBLE_STEELWALL_ACTIVE ? EL_INVISIBLE_STEELWALL :
			element == EL_INVISIBLE_WALL_ACTIVE      ? EL_INVISIBLE_WALL :
			element == EL_INVISIBLE_SAND_ACTIVE      ? EL_INVISIBLE_SAND :
			element);
};

LOGIC_SECTION static void RedrawAllLightSwitchesAndInvisibleElements()
{
	int x, y;

	for (y = 0; y < lev_fieldy; y++)
	{
		register short int xy=A2TO1(0,y);
		for (x = 0; x < lev_fieldx; x++, xy+=A2TO1(1,0))
		{
			int element = Feld[xy];

			if (element == EL_LIGHT_SWITCH &&
						game.light_time_left > 0)
			{
				InitGfxValueAndFeld(xy,EL_LIGHT_SWITCH_ACTIVE);
				DrawLevelField(x, y);
			}
			else if (element == EL_LIGHT_SWITCH_ACTIVE &&
						game.light_time_left == 0)
			{
				InitGfxValueAndFeld(xy,EL_LIGHT_SWITCH);
				DrawLevelField(x, y);
			}
			else if (element == EL_INVISIBLE_STEELWALL ||
							element == EL_INVISIBLE_WALL ||
							element == EL_INVISIBLE_SAND)
			{
				if (game.light_time_left > 0)
					InitGfxValueAndFeld(xy,
							getInvisibleActiveFromInvisibleElement(element));

				DrawLevelField(x, y);
			}
			else if (element == EL_INVISIBLE_STEELWALL_ACTIVE ||
							element == EL_INVISIBLE_WALL_ACTIVE ||
							element == EL_INVISIBLE_SAND_ACTIVE)
			{
				if (game.light_time_left == 0)
					InitGfxValueAndFeld(xy,
								getInvisibleFromInvisibleActiveElement(element));

				DrawLevelField(x, y);
			}
		}
	}
};

LOGIC_SECTION static void ToggleLightSwitch(int x, int y)
{
	int element = Feld[A2TO1(x,y)];

	game.light_time_left =
			(element == EL_LIGHT_SWITCH ?
			level.time_light * FRAMES_PER_SECOND : 0);

	RedrawAllLightSwitchesAndInvisibleElements();
};

LOGIC_SECTION static void ActivateTimegateSwitch(int x, int y)
{
	int xx, yy;

	game.timegate_time_left = level.time_timegate * FRAMES_PER_SECOND;

	for (yy = 0; yy < lev_fieldy; yy++)
	{
		register short int xxyy=A2TO1(0,yy);
		for (xx = 0; xx < lev_fieldx; xx++, xxyy+=A2TO1(1,0))
		{
			int element = Feld[xxyy];

			if (element == EL_TIMEGATE_CLOSED ||
						element == EL_TIMEGATE_CLOSING)
			{
				InitGfxValueAndFeld(xxyy,EL_TIMEGATE_OPENING);
//				AddDelayChanging(xxyy,29,EL_TIMEGATE_OPEN);//PALM!!!
			}

		}
	}

	InitGfxValueAndFeld(A2TO1(x,y),EL_TIMEGATE_SWITCH_ACTIVE);
//	AddDelayChanging(A2TO1(x,y), level.time_timegate * FRAMES_PER_SECOND,
//					EL_TIMEGATE_SWITCH);//PALM!!!
};

//PALM only!!!!
LOGIC_SECTION static byte FindElementMoveStepsize(short int element)
{
	if( element==EL_AMOEBA_DROP || element==EL_AMOEBA_DROPPING ||
			element==EL_MAGIC_WALL_FILLING || element==EL_BD_MAGIC_WALL_FILLING ||
			element==EL_MAGIC_WALL_EMPTYING || element==EL_BD_MAGIC_WALL_EMPTYING)
		return 1;//was 2 -> so 50% smaller
	if( element==EL_QUICKSAND_FILLING || element==EL_QUICKSAND_EMPTYING )
		return (FrameCounter&1);//was 1 -> so 50% smaller
	return MOVE_STEPSIZE_NORMAL;
}

LOGIC_SECTION static int getElementMoveStepsize(short int xy)
{
	int element = Feld[xy];
	int direction = MovDir[xy];
	int dx = (direction == MV_LEFT ? -1 : direction == MV_RIGHT ? +1 : 0);
	int dy = (direction == MV_UP   ? -1 : direction == MV_DOWN  ? +1 : 0);
	int horiz_move = (dx != 0);
	int sign = (horiz_move ? dx : dy);
	int step = sign * FindElementMoveStepsize(element);
//PALM !!!
	if (step==0 && !IS_MOVING(xy)) step = sign;//always start moving for 0.5 step

	// special values for move stepsize for spring and things on conveyor belt
	if (horiz_move)
	{
		if (CAN_FALL(element) && A1TOY(xy) < lev_fieldy - 1 && 
											IS_BELT_ACTIVE(Feld[xy+A2TO1(0,1)]))
			step = sign * MOVE_STEPSIZE_NORMAL / 2;
		else if (element == EL_SPRING)
			step = sign * MOVE_STEPSIZE_NORMAL * 2;
	}

	return step;
}

void Impact(int x, int y)
{
	short int xy=A2TO1(x,y);
	boolean lastline = (y == lev_fieldy-1);
	boolean object_hit = FALSE;
	boolean impact = (lastline || object_hit);
	int element = Feld[xy];
	int smashed = EL_STEELWALL;

	if (!lastline)	// check if element below was hit
	{
		if (Feld[xy+A2TO1(0,1)] == EL_PLAYER_IS_LEAVING)
			return;

		object_hit = (!IS_FREE(xy+A2TO1(0,1)) && (!IS_MOVING(xy + A2TO1(0,1)) ||
										MovDir[xy + A2TO1(0,1)] != MV_DOWN ||
										MovPos[xy + A2TO1(0,1)] <= TILEY / 2));

		// do not smash moving elements that left the smashed field in time 
		if (game.engine_version >= VERSION_IDENT(2,2,0,7) && IS_MOVING(xy + A2TO1(0,1)) &&
					ABS(MovPos[xy + A2TO1(0,1)]
							 + getElementMoveStepsize(xy + A2TO1(0,1))) >= TILEX)
			object_hit = FALSE;

		if (object_hit)
			smashed = MovingOrBlocked2Element(x, y + 1);

		impact = (lastline || object_hit);
	}

	if (!lastline && smashed == EL_ACID)	// element falls into acid 
	{
		SplashAcid(x, y + 1);
		return;
	}

	//!!! not sufficient for all cases -- see EL_PEARL below !!!
	// only reset graphic animation if graphic really changes after impact
	if (impact &&
			el_act_dir2img(element, GfxAction[xy], MV_DOWN) != el2img(element))
	{
		ResetGfxAnimation(xy);
		InitGfxValue(xy);
		DrawLevelField(x, y);
	}

	if (impact && CAN_EXPLODE_IMPACT(element))
	{
		Bang(x, y);
		return;
	}
	else if (impact && element == EL_PEARL)
	{
		ResetGfxAnimation(xy);

		InitGfxValueAndFeld(xy,EL_PEARL_BREAKING);
//		AddDelayChanging(xy,8,EL_EMPTY);//PALM!!!
		return;
	}
/*	else if (impact && CheckElementChange(x, y, element, smashed, CE_IMPACT))
	{
		return;
	}
*/
	if (impact && element == EL_AMOEBA_DROP)
	{
		if (object_hit && IS_PLAYER(xy+A2TO1(0,1)))
			{KillHeroUnlessEnemyProtected(x, y + 1, xy+A2TO1(0,1));return;}
		if (object_hit && smashed == EL_PENGUIN)
			Bang(x, y + 1);
		else
		{
			InitGfxValueAndFeld(xy,EL_AMOEBA_GROWING);
			Store[xy] = EL_AMOEBA_WET;

			ResetRandomAnimationValue(xy);
		}
		return;
	}

	if (object_hit)		// check which object was hit
	{
		if (CAN_PASS_MAGIC_WALL(element) && 
						(smashed == EL_MAGIC_WALL ||
						 smashed == EL_BD_MAGIC_WALL))
		{
			int xx, yy;
			int activated_magic_wall =
							(smashed == EL_MAGIC_WALL ? EL_MAGIC_WALL_ACTIVE :
										EL_BD_MAGIC_WALL_ACTIVE);

			// activate magic wall / mill
			for (yy = 0; yy < lev_fieldy; yy++)
			{
				register short int xxyy=A2TO1(0,yy);
				for (xx = 0; xx < lev_fieldx; xx++, xxyy+=A2TO1(1,0))
					if (Feld[xxyy] == smashed)
						InitGfxValueAndFeld(xxyy,activated_magic_wall);
			}

			game.magic_wall_time_left = level.time_magic_wall * FRAMES_PER_SECOND;
			game.magic_wall_active = TRUE;

		}

		if (IS_PLAYER(xy+A2TO1(0,1)))
		{
			if (CAN_SMASH_PLAYER(element))
			{
				KillHeroUnlessEnemyProtected(x, y + 1, xy+A2TO1(0,1));
				return;
			}
		}
		else if (smashed == EL_PENGUIN)
		{
			if (CAN_SMASH_PLAYER(element))
			{
				Bang(x, y + 1);
				return;
			}
		}
		else if (element == EL_BD_DIAMOND)
		{
			if (IS_CLASSIC_ENEMY(smashed) && IS_BD_ELEMENT(smashed))
			{
				Bang(x, y + 1);
				return;
			}
		}
		else if (((element == EL_SP_INFOTRON ||
					element == EL_SP_ZONK) &&
					(smashed == EL_SP_SNIKSNAK ||
					 smashed == EL_SP_ELECTRON ||
					 smashed == EL_SP_DISK_ORANGE)) ||
					(element == EL_SP_INFOTRON &&
					 smashed == EL_SP_DISK_YELLOW))
		{
			Bang(x, y + 1);
			return;
		}
		else if (CAN_SMASH_EVERYTHING(element))
		{
			if (IS_CLASSIC_ENEMY(smashed) || CAN_EXPLODE_SMASHED(smashed))
			{
				Bang(x, y + 1);
				return;
			}
			else if (!IS_MOVING(xy + A2TO1(0,1)) && !IS_BLOCKED(xy + A2TO1(0,1)))
			{
				if (smashed == EL_LAMP || smashed == EL_LAMP_ACTIVE)
				{
					Bang(x, y + 1);
					return;
				}
				else if (smashed == EL_NUT)
				{
					InitGfxValueAndFeld(xy+A2TO1(0,1),EL_NUT_BREAKING);
//					AddDelayChanging(xy+A2TO1(0,1),6,EL_EMERALD);//PALM!!!
					RaiseScoreElement(EL_NUT);
					return;
				}
				else if (smashed == EL_PEARL)
				{
					ResetGfxAnimation(xy);

					InitGfxValueAndFeld(xy+A2TO1(0,1),EL_PEARL_BREAKING);
//					AddDelayChanging(xy+A2TO1(0,1),8,EL_EMPTY);//PALM!!!
					return;
				}
				else if (smashed == EL_DIAMOND)
				{
					InitGfxValueAndFeld(xy+A2TO1(0,1),EL_DIAMOND_BREAKING);
					return;
				}
				else if (IS_BELT_SWITCH(smashed))
				{
					ToggleBeltSwitch(x, y + 1);
				}
				else if (smashed == EL_SWITCHGATE_SWITCH_UP ||
							smashed == EL_SWITCHGATE_SWITCH_DOWN)
				{
					ToggleSwitchgateSwitch(x, y + 1);
				}
				else if (smashed == EL_LIGHT_SWITCH ||
							smashed == EL_LIGHT_SWITCH_ACTIVE)
				{
					ToggleLightSwitch(x, y + 1);
				}
				else
				{
/*					CheckElementChange(x, y + 1, smashed, element, CE_SMASHED);

					// !!! TEST ONLY !!!
					CheckElementChangeBySide(x, y + 1, smashed, element,
												CE_SWITCHED, CH_SIDE_TOP);
					CheckTriggeredElementChangeBySide(x, y + 1, smashed,
										CE_OTHER_IS_SWITCHING,CH_SIDE_TOP);
*/				}
			}
			else
			{
//				CheckElementChange(x, y + 1, smashed, element, CE_SMASHED);
			}
		}
	}
};
/*
//PALM only!!!!
LOGIC_SECTION static byte FindElementMovePattern(short int element)
{
	if( element==EL_BUG || element==EL_BD_BUTTERFLY)
		return MV_ALONG_RIGHT_SIDE;
	if( element==EL_SPACESHIP || element==EL_BD_FIREFLY ||
				element==EL_SP_SNIKSNAK || element==EL_SP_ELECTRON)
		return MV_ALONG_LEFT_SIDE;
	return MV_ALL_DIRECTIONS;
}
*/
LOGIC_SECTION inline static void TurnRoundExt(int x, int y)
{
	short int xy=A2TO1(x,y); 
	const struct
	{
		int x, y;
	} move_xy1[] =
	{
		{  0,  0 },
		{ -1,  0 },
		{ +1,  0 },
		{  0,  0 },
		{  0, -1 },
		{  0,  0 }, { 0, 0 }, { 0, 0 },
		{  0, +1 }
	};
	const struct
	{
		int left, right, back;
	} turn[] =
	{
		{ 0,	0,		0	 },
		{ MV_DOWN,	MV_UP,		MV_RIGHT },
		{ MV_UP,	MV_DOWN,	MV_LEFT	 },
		{ 0,	0,		0	 },
		{ MV_LEFT,	MV_RIGHT,	MV_DOWN	 },
		{ 0,	0,		0	 },
		{ 0,	0,		0	 },
		{ 0,	0,		0	 },
		{ MV_RIGHT,	MV_LEFT,	MV_UP	 }
	};

	int element = Feld[xy];
//	int move_pattern = element_info[element].move_pattern;

	int old_move_dir = MovDir[xy];
	int left_dir  = turn[old_move_dir].left;
	int right_dir = turn[old_move_dir].right;
	int back_dir  = turn[old_move_dir].back;

	int left_dx  = move_xy1[left_dir].x,     left_dy  = move_xy1[left_dir].y;
	int right_dx = move_xy1[right_dir].x,    right_dy = move_xy1[right_dir].y;
	int move_dx  = move_xy1[old_move_dir].x, move_dy  = move_xy1[old_move_dir].y;
	int back_dx  = move_xy1[back_dir].x,     back_dy  = move_xy1[back_dir].y;

	int left_x  = x + left_dx,  left_y  = y + left_dy;
	int right_x = x + right_dx, right_y = y + right_dy;
	int move_x  = x + move_dx,  move_y  = y + move_dy;

	int xx, yy;
	
	short int left_xy=A2TO1(left_x,left_y);
	short int right_xy=A2TO1(right_x,right_y);
	short int move_xy=A2TO1(move_x,move_y);

	if (element == EL_BUG || element == EL_BD_BUTTERFLY)
	{
		TestIfBadThingTouchesOtherBadThing(x, y);

		if (ENEMY_CAN_ENTER_FIELD(element, right_x, right_y, right_xy))
			MovDir[xy] = right_dir;
		else if (!ENEMY_CAN_ENTER_FIELD(element, move_x, move_y, move_xy))
			MovDir[xy] = left_dir;

		if (element == EL_BUG && MovDir[xy] != old_move_dir)
			MovDelay[xy] = 9;
		else if (element == EL_BD_BUTTERFLY)     // && MovDir[x][y] == left_dir)
			MovDelay[xy] = 1;
	}
	else if (element == EL_SPACESHIP || element == EL_BD_FIREFLY)
	{
		TestIfBadThingTouchesOtherBadThing(x, y);

		if (ENEMY_CAN_ENTER_FIELD(element, left_x, left_y, left_xy))
			MovDir[xy] = left_dir;
		else if (!ENEMY_CAN_ENTER_FIELD(element, move_x, move_y, move_xy))
			MovDir[xy] = right_dir;

		if (element == EL_SPACESHIP	&& MovDir[xy] != old_move_dir)
			MovDelay[xy] = 9;
		else if (element == EL_BD_FIREFLY)	    // && MovDir[x][y] == right_dir)
			MovDelay[xy] = 1;
	}
	else if (element == EL_SP_SNIKSNAK || element == EL_SP_ELECTRON)
	{
		TestIfBadThingTouchesOtherBadThing(x, y);

		if (ELEMENT_CAN_ENTER_FIELD_BASE_4(element, left_x, left_y, 0, left_xy))
			MovDir[xy] = left_dir;
		else if (!ELEMENT_CAN_ENTER_FIELD_BASE_4(element, move_x, move_y, 0, move_xy))
			MovDir[xy] = right_dir;

		if (MovDir[xy] != old_move_dir)
			MovDelay[xy] = 9;
	}
	else if (element == EL_YAMYAM)
	{
		boolean can_turn_left  = YAMYAM_CAN_ENTER_FIELD(element,left_x, left_y, left_xy);
		boolean can_turn_right = YAMYAM_CAN_ENTER_FIELD(element,right_x, right_y, right_xy);

		if (can_turn_left && can_turn_right)
			MovDir[xy] = (RND(3) ? (RND(2) ? left_dir : right_dir) : back_dir);
		else if (can_turn_left)
			MovDir[xy] = (RND(2) ? left_dir : back_dir);
		else if (can_turn_right)
			MovDir[xy] = (RND(2) ? right_dir : back_dir);
		else
			MovDir[xy] = back_dir;

		MovDelay[xy] = 16 + 16 * RND(3);
	}
	else if (element == EL_DARK_YAMYAM)
	{
		boolean can_turn_left  = DARK_YAMYAM_CAN_ENTER_FIELD(element,left_x, left_y, left_xy);
		boolean can_turn_right = DARK_YAMYAM_CAN_ENTER_FIELD(element,right_x, right_y, right_xy);

		if (can_turn_left && can_turn_right)
			MovDir[xy] = (RND(3) ? (RND(2) ? left_dir : right_dir) : back_dir);
		else if (can_turn_left)
			MovDir[xy] = (RND(2) ? left_dir : back_dir);
		else if (can_turn_right)
			MovDir[xy] = (RND(2) ? right_dir : back_dir);
		else
			MovDir[xy] = back_dir;

		MovDelay[xy] = 16 + 16 * RND(3);
	}
	else if (element == EL_PACMAN)
	{
		boolean can_turn_left  = PACMAN_CAN_ENTER_FIELD(element, left_x, left_y, left_xy);
		boolean can_turn_right = PACMAN_CAN_ENTER_FIELD(element, right_x, right_y, right_xy);

		if (can_turn_left && can_turn_right)
			MovDir[xy] = (RND(3) ? (RND(2) ? left_dir : right_dir) : back_dir);
		else if (can_turn_left)
			MovDir[xy] = (RND(2) ? left_dir : back_dir);
		else if (can_turn_right)
			MovDir[xy] = (RND(2) ? right_dir : back_dir);
		else
			MovDir[xy] = back_dir;

		MovDelay[xy] = 6 + RND(40);
	}
	else if (element == EL_PIG)
	{
		boolean can_turn_left  = PIG_CAN_ENTER_FIELD(element,left_x, left_y, left_xy);
		boolean can_turn_right = PIG_CAN_ENTER_FIELD(element,right_x, right_y, right_xy);
		boolean can_move_on    = PIG_CAN_ENTER_FIELD(element,move_x, move_y, move_xy);
		boolean should_turn_left, should_turn_right, should_move_on;
		int rnd_value = 24;
		int rnd = RND(rnd_value);

		should_turn_left = (can_turn_left &&
							(!can_move_on ||
							IN_LEV_FIELD_AND_NOT_FREE(x + back_dx + left_dx,
													y + back_dy + left_dy)));
		should_turn_right = (can_turn_right &&
							(!can_move_on ||
							IN_LEV_FIELD_AND_NOT_FREE(x + back_dx + right_dx,
													y + back_dy + right_dy)));
		should_move_on = (can_move_on &&
							(!can_turn_left ||
							!can_turn_right ||
							IN_LEV_FIELD_AND_NOT_FREE(x + move_dx + left_dx,
													y + move_dy + left_dy) ||
							IN_LEV_FIELD_AND_NOT_FREE(x + move_dx + right_dx,
													y + move_dy + right_dy)));

		if (should_turn_left || should_turn_right || should_move_on)
		{
			if (should_turn_left && should_turn_right && should_move_on)
				MovDir[xy] = (rnd < rnd_value / 3     ? left_dir :
								rnd < 2 * rnd_value / 3 ? right_dir :
								old_move_dir);
			else if (should_turn_left && should_turn_right)
				MovDir[xy] = (rnd < rnd_value / 2 ? left_dir : right_dir);
			else if (should_turn_left && should_move_on)
				MovDir[xy] = (rnd < rnd_value / 2 ? left_dir : old_move_dir);
			else if (should_turn_right && should_move_on)
				MovDir[xy] = (rnd < rnd_value / 2 ? right_dir : old_move_dir);
			else if (should_turn_left)
				MovDir[xy] = left_dir;
			else if (should_turn_right)
				MovDir[xy] = right_dir;
			else if (should_move_on)
				MovDir[xy] = old_move_dir;
		}
		else if (can_move_on && rnd > rnd_value / 8)
			MovDir[xy] = old_move_dir;
		else if (can_turn_left && can_turn_right)
			MovDir[xy] = (rnd < rnd_value / 2 ? left_dir : right_dir);
		else if (can_turn_left && rnd > rnd_value / 8)
			MovDir[xy] = left_dir;
		else if (can_turn_right && rnd > rnd_value / 8)
			MovDir[xy] = right_dir;
		else
			MovDir[xy] = back_dir;

		xx = x + move_xy1[MovDir[xy]].x;
		yy = y + move_xy1[MovDir[xy]].y;

		if (!IS_FREE(A2TO1(xx,yy)) && !IS_FOOD_PIG(Feld[A2TO1(xx,yy)]))
			MovDir[xy] = old_move_dir;

		MovDelay[xy] = 0;
	}
	else if (element == EL_DRAGON)
	{
		boolean can_turn_left  = DRAGON_CAN_ENTER_FIELD(element,left_x,left_y,left_xy);
		boolean can_turn_right = DRAGON_CAN_ENTER_FIELD(element,right_x,right_y,right_xy);
		boolean can_move_on    = DRAGON_CAN_ENTER_FIELD(element,move_x,move_y,move_xy);
		int rnd_value = 24;
		int rnd = RND(rnd_value);

		if (can_move_on && rnd > rnd_value / 8)
			MovDir[xy] = old_move_dir;
		else if (can_turn_left && can_turn_right)
			MovDir[xy] = (rnd < rnd_value / 2 ? left_dir : right_dir);
		else if (can_turn_left && rnd > rnd_value / 8)
			MovDir[xy] = left_dir;
		else if (can_turn_right && rnd > rnd_value / 8)
			MovDir[xy] = right_dir;
		else
			MovDir[xy] = back_dir;

		xx = x + move_xy1[MovDir[xy]].x;
		yy = y + move_xy1[MovDir[xy]].y;

		if (!IN_LEV_FIELD_AND_IS_FREE(xx, yy))
			MovDir[xy] = old_move_dir;

		MovDelay[xy] = 0;
	}
	else if (element == EL_MOLE)
	{
		boolean can_move_on =
				(MOLE_CAN_ENTER_FIELD(element,move_x, move_y,
					IS_AMOEBOID(Feld[move_xy]) || Feld[move_xy] == EL_AMOEBA_SHRINKING,
					move_xy));
		if (!can_move_on)
		{
			boolean can_turn_left =
							(MOLE_CAN_ENTER_FIELD(element,left_x, left_y,
											IS_AMOEBOID(Feld[left_xy]), left_xy));

			boolean can_turn_right =
							(MOLE_CAN_ENTER_FIELD(element,right_x, right_y,
											IS_AMOEBOID(Feld[right_xy]), right_xy));

			if (can_turn_left && can_turn_right)
				MovDir[xy] = (RND(2) ? left_dir : right_dir);
			else if (can_turn_left)
				MovDir[xy] = left_dir;
			else
				MovDir[xy] = right_dir;
		}

		if (MovDir[xy] != old_move_dir)
			MovDelay[xy] = 9;
	}
	else if (element == EL_BALLOON)
	{
		MovDir[xy] = game.balloon_dir;
		MovDelay[xy] = 0;
	}
	else if (element == EL_SPRING)
	{
		if (MovDir[xy] & MV_HORIZONTAL &&
					(!SPRING_CAN_ENTER_FIELD(element, move_x, move_y, move_xy) ||
					 SPRING_CAN_ENTER_FIELD(element, x, y + 1, xy+A2TO1(0,1))))
			MovDir[xy] = MV_NO_MOVING;

		MovDelay[xy] = 0;
	}
	else if (element == EL_ROBOT ||
				element == EL_SATELLITE ||
				element == EL_PENGUIN)
	{
		int attr_x = -1, attr_y = -1;

		if (AllPlayersGone)
		{
			attr_x = ExitX;
			attr_y = ExitY;
		}
		else
		{

			struct PlayerInfo *player = &stored_player_;
			int jx = player->jx, jy = player->jy;

			if (player->active)
				if (attr_x == -1 ||
						ABS(jx - x) + ABS(jy - y) < ABS(attr_x - x) + ABS(attr_y - y))
				{
					attr_x = jx;
					attr_y = jy;
				}
		}

		if (element == EL_ROBOT && ZX >= 0 && ZY >= 0 &&
					(Feld[A2TO1(ZX,ZY)] == EL_ROBOT_WHEEL_ACTIVE ||
					 game.engine_version < VERSION_IDENT(3,1,0,0)))
		{
			attr_x = ZX;
			attr_y = ZY;
		}

		if (element == EL_PENGUIN)
		{
			int i;
			const int xy_x[4]={0,-1,+1,0};
			const int xy_y[4]={-1,0,0,+1};

			for (i = 0; i < NUM_DIRECTIONS; i++)
			{
				int ex = x + xy_x[i];
				int ey = y + xy_y[i];

				if (IN_LEV_FIELD(ex, ey) && Feld[A2TO1(ex,ey)] == EL_EXIT_OPEN)
				{
					attr_x = ex;
					attr_y = ey;
					break;
				}
			}
		}

		MovDir[xy] = MV_NO_MOVING;
		if (attr_x < x)
			MovDir[xy] |= (AllPlayersGone ? MV_RIGHT : MV_LEFT);
		else if (attr_x > x)
			MovDir[xy] |= (AllPlayersGone ? MV_LEFT : MV_RIGHT);
		if (attr_y < y)
			MovDir[xy] |= (AllPlayersGone ? MV_DOWN : MV_UP);
		else if (attr_y > y)
			MovDir[xy] |= (AllPlayersGone ? MV_UP : MV_DOWN);

		if (element == EL_ROBOT)
		{
			short int newxy;
			int newx, newy;

			if (MovDir[xy] & MV_HORIZONTAL && MovDir[xy] & MV_VERTICAL)
				MovDir[xy] &= (RND(2) ? MV_HORIZONTAL : MV_VERTICAL);
			Moving2Blocked(xy, &newxy);
			newx=A1TOX(newxy);
			newy=A1TOY(newxy);

			if (IN_LEV_FIELD(newx, newy) && IS_FREE_OR_PLAYER(newxy))
				MovDelay[xy] = 8 + 8 * !RND(3);
			else
				MovDelay[xy] = 16;
		}
		else if (element == EL_PENGUIN)
		{
			MovDelay[xy] = 1;

			if (MovDir[xy] & MV_HORIZONTAL && MovDir[xy] & MV_VERTICAL)
			{
				boolean first_horiz = RND(2);
				int new_move_dir = MovDir[xy];
				short int newxy;
				int newx, newy;

				MovDir[xy] =
						new_move_dir & (first_horiz ? MV_HORIZONTAL : MV_VERTICAL);
				Moving2Blocked(xy, &newxy);
				newx=A1TOX(newxy);
				newy=A1TOY(newxy);

				if (PENGUIN_CAN_ENTER_FIELD(EL_PENGUIN, newx, newy, newxy))
					return;

				MovDir[xy] =
						new_move_dir & (!first_horiz ? MV_HORIZONTAL : MV_VERTICAL);
				Moving2Blocked(xy, &newxy);
				newx=A1TOX(newxy);
				newy=A1TOY(newxy);

				if (PENGUIN_CAN_ENTER_FIELD(EL_PENGUIN, newx, newy, newxy))
					return;

				MovDir[xy] = old_move_dir;
				return;
			}
		}
		else	// (element == EL_SATELLITE)
		{
			MovDelay[xy] = 1;

			if (MovDir[xy] & MV_HORIZONTAL && MovDir[xy] & MV_VERTICAL)
			{
				boolean first_horiz = RND(2);
				int new_move_dir = MovDir[xy];
				int newx, newy;
				short int newxy;

				MovDir[xy] =
						new_move_dir & (first_horiz ? MV_HORIZONTAL : MV_VERTICAL);
				Moving2Blocked(xy, &newxy);
				newx=A1TOX(newxy);
				newy=A1TOY(newxy);

				if (SATELLITE_CAN_ENTER_FIELD(newx, newy, newxy))
					return;

				MovDir[xy] =
						new_move_dir & (!first_horiz ? MV_HORIZONTAL : MV_VERTICAL);
				Moving2Blocked(xy, &newxy);
				newx=A1TOX(newxy);
				newy=A1TOY(newxy);

				if (SATELLITE_CAN_ENTER_FIELD(newx, newy, newxy))
					return;

				MovDir[xy] = old_move_dir;
				return;
			}
		}
	}
/*  else if (move_pattern == MV_TURNING_LEFT ||
	   move_pattern == MV_TURNING_RIGHT ||
	   move_pattern == MV_TURNING_LEFT_RIGHT ||
	   move_pattern == MV_TURNING_RIGHT_LEFT ||
	   move_pattern == MV_TURNING_RANDOM ||
	   move_pattern == MV_ALL_DIRECTIONS)
  {
    boolean can_turn_left =
      CUSTOM_ELEMENT_CAN_ENTER_FIELD(element, left_x, left_y);
    boolean can_turn_right =
      CUSTOM_ELEMENT_CAN_ENTER_FIELD(element, right_x,right_y);

    if (move_pattern == MV_TURNING_LEFT)
      MovDir[x][y] = left_dir;
    else if (move_pattern == MV_TURNING_RIGHT)
      MovDir[x][y] = right_dir;
    else if (move_pattern == MV_TURNING_LEFT_RIGHT)
      MovDir[x][y] = (can_turn_left || !can_turn_right ? left_dir : right_dir);
    else if (move_pattern == MV_TURNING_RIGHT_LEFT)
      MovDir[x][y] = (can_turn_right || !can_turn_left ? right_dir : left_dir);
    else if (move_pattern == MV_TURNING_RANDOM)
      MovDir[x][y] = (can_turn_left && !can_turn_right ? left_dir :
		      can_turn_right && !can_turn_left ? right_dir :
		      RND(2) ? left_dir : right_dir);
    else if (can_turn_left && can_turn_right)
      MovDir[x][y] = (RND(3) ? (RND(2) ? left_dir : right_dir) : back_dir);
    else if (can_turn_left)
      MovDir[x][y] = (RND(2) ? left_dir : back_dir);
    else if (can_turn_right)
      MovDir[x][y] = (RND(2) ? right_dir : back_dir);
    else
      MovDir[x][y] = back_dir;

    MovDelay[x][y] = GET_NEW_MOVE_DELAY(element);
  }
  else if (move_pattern == MV_HORIZONTAL ||
	   move_pattern == MV_VERTICAL)
  {
    if (move_pattern & old_move_dir)
      MovDir[x][y] = back_dir;
    else if (move_pattern == MV_HORIZONTAL)
      MovDir[x][y] = (RND(2) ? MV_LEFT : MV_RIGHT);
    else if (move_pattern == MV_VERTICAL)
      MovDir[x][y] = (RND(2) ? MV_UP : MV_DOWN);

    MovDelay[x][y] = GET_NEW_MOVE_DELAY(element);
  }
  else if (move_pattern & MV_ANY_DIRECTION)
  {
    MovDir[x][y] = move_pattern;
    MovDelay[x][y] = GET_NEW_MOVE_DELAY(element);
  }
  else if (move_pattern == MV_ALONG_LEFT_SIDE)
  {
    if (CUSTOM_ELEMENT_CAN_ENTER_FIELD(element, left_x, left_y))
      MovDir[x][y] = left_dir;
    else if (!CUSTOM_ELEMENT_CAN_ENTER_FIELD(element, move_x, move_y))
      MovDir[x][y] = right_dir;

    if (MovDir[x][y] != old_move_dir)
      MovDelay[x][y] = GET_NEW_MOVE_DELAY(element);
  }
  else if (move_pattern == MV_ALONG_RIGHT_SIDE)
  {
    if (CUSTOM_ELEMENT_CAN_ENTER_FIELD(element, right_x, right_y))
      MovDir[x][y] = right_dir;
    else if (!CUSTOM_ELEMENT_CAN_ENTER_FIELD(element, move_x, move_y))
      MovDir[x][y] = left_dir;

    if (MovDir[x][y] != old_move_dir)
      MovDelay[x][y] = GET_NEW_MOVE_DELAY(element);
  }
  else if (move_pattern == MV_TOWARDS_PLAYER ||
	   move_pattern == MV_AWAY_FROM_PLAYER)
  {
    int attr_x = -1, attr_y = -1;
    int newx, newy;
    boolean move_away = (move_pattern == MV_AWAY_FROM_PLAYER);

    if (AllPlayersGone)
    {
      attr_x = ExitX;
      attr_y = ExitY;
    }
    else
    {
      int i;

      for (i = 0; i < MAX_PLAYERS; i++)
      {
	struct PlayerInfo *player = &stored_player[i];
	int jx = player->jx, jy = player->jy;

	if (!player->active)
	  continue;

	if (attr_x == -1 ||
	    ABS(jx - x) + ABS(jy - y) < ABS(attr_x - x) + ABS(attr_y - y))
	{
	  attr_x = jx;
	  attr_y = jy;
	}
      }
    }

    MovDir[x][y] = MV_NO_MOVING;
    if (attr_x < x)
      MovDir[x][y] |= (move_away ? MV_RIGHT : MV_LEFT);
    else if (attr_x > x)
      MovDir[x][y] |= (move_away ? MV_LEFT : MV_RIGHT);
    if (attr_y < y)
      MovDir[x][y] |= (move_away ? MV_DOWN : MV_UP);
    else if (attr_y > y)
      MovDir[x][y] |= (move_away ? MV_UP : MV_DOWN);

    MovDelay[x][y] = GET_NEW_MOVE_DELAY(element);

    if (MovDir[x][y] & MV_HORIZONTAL && MovDir[x][y] & MV_VERTICAL)
    {
      boolean first_horiz = RND(2);
      int new_move_dir = MovDir[x][y];

      MovDir[x][y] =
	new_move_dir & (first_horiz ? MV_HORIZONTAL : MV_VERTICAL);
      Moving2Blocked(x, y, &newx, &newy);

      if (CUSTOM_ELEMENT_CAN_ENTER_FIELD(element, newx, newy))
	return;

      MovDir[x][y] =
	new_move_dir & (!first_horiz ? MV_HORIZONTAL : MV_VERTICAL);
      Moving2Blocked(x, y, &newx, &newy);

      if (CUSTOM_ELEMENT_CAN_ENTER_FIELD(element, newx, newy))
	return;

      MovDir[x][y] = old_move_dir;
    }
  }
  else if (move_pattern == MV_WHEN_PUSHED ||
	   move_pattern == MV_WHEN_DROPPED)
  {
    if (!CUSTOM_ELEMENT_CAN_ENTER_FIELD(element, move_x, move_y))
      MovDir[x][y] = MV_NO_MOVING;

    MovDelay[x][y] = 0;
  }
  else if (move_pattern & MV_MAZE_RUNNER_STYLE)
  {
    static int test_xy[7][2] =
    {
      { 0, -1 },
      { -1, 0 },
      { +1, 0 },
      { 0, +1 },
      { 0, -1 },
      { -1, 0 },
      { +1, 0 },
    };
    static int test_dir[7] =
    {
      MV_UP,
      MV_LEFT,
      MV_RIGHT,
      MV_DOWN,
      MV_UP,
      MV_LEFT,
      MV_RIGHT,
    };
    boolean hunter_mode = (move_pattern == MV_MAZE_HUNTER);
    int move_preference = -1000000;	// start with very low preference
    int new_move_dir = MV_NO_MOVING;
    int start_test = RND(4);
    int i;

    for (i = 0; i < NUM_DIRECTIONS; i++)
    {
      int move_dir = test_dir[start_test + i];
      int move_dir_preference;

      xx = x + test_xy[start_test + i][0];
      yy = y + test_xy[start_test + i][1];

      if (hunter_mode && IN_LEV_FIELD(xx, yy) &&
	  (IS_PLAYER(xx, yy) || Feld[xx][yy] == EL_PLAYER_IS_LEAVING))
      {
	new_move_dir = move_dir;

	break;
      }

      if (!CUSTOM_ELEMENT_CAN_ENTER_FIELD(element, xx, yy))
	continue;

      move_dir_preference = -1 * RunnerVisit[xx][yy];
      if (hunter_mode && PlayerVisit[xx][yy] > 0)
	move_dir_preference = PlayerVisit[xx][yy];

      if (move_dir_preference > move_preference)
      {
	// prefer field that has not been visited for the longest time
	move_preference = move_dir_preference;
	new_move_dir = move_dir;
      }
      else if (move_dir_preference == move_preference &&
	       move_dir == old_move_dir)
      {
	// prefer last direction when all directions are preferred equally
	move_preference = move_dir_preference;
	new_move_dir = move_dir;
      }
    }

    MovDir[x][y] = new_move_dir;
    if (old_move_dir != new_move_dir)
      MovDelay[x][y] = 9;
  }*/
};

LOGIC_SECTION static void TurnRound(int x, int y)
{
	short int xy=A2TO1(x,y);
	int direction = MovDir[xy];

	TurnRoundExt(x, y);

//	GfxDir[x][y] = MovDir[x][y];

	if (direction != MovDir[xy])
		GfxSFrame[xy]=FrameCounter;//GfxFrame[x][y] = 0;

	if (MovDelay[xy])
		GfxAction[xy] = ACTION_TURNING_FROM_LEFT + MV_DIR_BIT(direction);
		
	InitGfxValue(xy);
};

LOGIC_SECTION static boolean JustBeingPushed(int x, int y)
{
	struct PlayerInfo *player = &stored_player_;

	if (player->active && player->is_pushing && player->MovPos)
	{
		int next_jx = player->jx + (player->jx - player->last_jx);
		int next_jy = player->jy + (player->jy - player->last_jy);

		if (x == next_jx && y == next_jy)
			return TRUE;
	}

	return FALSE;
}

void StartMoving(int x, int y)
{
	boolean started_moving = FALSE;	// some elements can fall _and_ move
	short int xy=A2TO1(x,y);
	int element = Feld[xy];

countBigStartMoving++;

	if (getStop(xy))//Stop[x][y]
		return;

	if (MovDelay[xy] == 0)
		InitGfxValueAndGfxAction(xy,ACTION_DEFAULT);

	if (CAN_FALL(element) && y < lev_fieldy - 1)
	{
		if ((x > 0 && IS_PLAYER(xy-A2TO1(1,0))) ||
						(x < lev_fieldx - 1 && IS_PLAYER(xy+A2TO1(1,0))))
			if (JustBeingPushed(x, y))
				return;
//bugd("StartMoving fall",x*256+y);
		if (element == EL_QUICKSAND_FULL)
		{
			if (IS_FREE(xy+A2TO1(0,1)))
			{
				InitMovingField(xy, MV_DOWN);
				started_moving = TRUE;
				
				InitGfxValueAndFeld(xy,EL_QUICKSAND_EMPTYING);
				Store[xy] = EL_ROCK;

			}
			else if (Feld[xy+A2TO1(0,1)] == EL_QUICKSAND_EMPTY)
			{
				if (!MovDelay[xy])
					MovDelay[xy] = TILEY + 1;

				if (MovDelay[xy])
				{
					MovDelay[xy]--;
					if (MovDelay[xy])
						return;
				}

				InitGfxValueAndFeld(xy,EL_QUICKSAND_EMPTY);
				InitGfxValueAndFeld(xy+A2TO1(0,1),EL_QUICKSAND_FULL);

				Store[xy+A2TO1(0,1)] = Store[xy];
				Store[xy] = 0;
			}
		}
		else if ((element == EL_ROCK || element == EL_BD_ROCK) &&
								Feld[xy+A2TO1(0,1)] == EL_QUICKSAND_EMPTY)
		{

			InitMovingField(xy, MV_DOWN);
			started_moving = TRUE;

			InitGfxValueAndFeld(xy,EL_QUICKSAND_FILLING);
			Store[xy] = element;
		}
		else if (element == EL_MAGIC_WALL_FULL)
		{

			if (IS_FREE(xy+A2TO1(0,1)))
			{
				InitMovingField(xy, MV_DOWN);
				started_moving = TRUE;

				InitGfxValueAndFeld(xy,EL_MAGIC_WALL_EMPTYING);
				Store[xy] = EL_CHANGED(Store[xy]);
			}
			else if (Feld[xy+A2TO1(0,1)] == EL_MAGIC_WALL_ACTIVE)
			{
				if (!MovDelay[xy])
					MovDelay[xy] = TILEY/4 + 1;

				if (MovDelay[xy])
				{
					MovDelay[xy]--;
					if (MovDelay[xy])
						return;
				}

				InitGfxValueAndFeld(xy,EL_MAGIC_WALL_ACTIVE);
				InitGfxValueAndFeld(xy+A2TO1(0,1),EL_MAGIC_WALL_FULL);
				Store[xy+A2TO1(0,1)] = EL_CHANGED(Store[xy]);
				Store[xy] = 0;
			}
		}
		else if (element == EL_BD_MAGIC_WALL_FULL)
		{

			if (IS_FREE(xy+A2TO1(0,1)))
			{
				InitMovingField(xy, MV_DOWN);
				started_moving = TRUE;

				InitGfxValueAndFeld(xy,EL_BD_MAGIC_WALL_EMPTYING);
				Store[xy] = EL_CHANGED2(Store[xy]);
			}
			else if (Feld[xy+A2TO1(0,1)] == EL_BD_MAGIC_WALL_ACTIVE)
			{
				if (!MovDelay[xy])
					MovDelay[xy] = TILEY/4 + 1;

				if (MovDelay[xy])
				{
					MovDelay[xy]--;
					if (MovDelay[xy])
						return;
				}

				InitGfxValueAndFeld(xy,EL_BD_MAGIC_WALL_ACTIVE);
				InitGfxValueAndFeld(xy+A2TO1(0,1),EL_BD_MAGIC_WALL_FULL);
				Store[xy+A2TO1(0,1)] = EL_CHANGED2(Store[xy]);
				Store[xy] = 0;
			}
		}
		else if (CAN_PASS_MAGIC_WALL(element) &&
						(Feld[xy+A2TO1(0,1)] == EL_MAGIC_WALL_ACTIVE ||
						 Feld[xy+A2TO1(0,1)] == EL_BD_MAGIC_WALL_ACTIVE))
		{

			InitMovingField(xy, MV_DOWN);
			started_moving = TRUE;

			InitGfxValueAndFeld(xy,	(Feld[xy+A2TO1(0,1)] == EL_MAGIC_WALL_ACTIVE 
							? EL_MAGIC_WALL_FILLING : EL_BD_MAGIC_WALL_FILLING));
			Store[xy] = element;
		}
		else if (CAN_FALL(element) && Feld[xy+A2TO1(0,1)] == EL_ACID)
		{

			SplashAcid(x, y + 1);

			InitMovingField(xy, MV_DOWN);
			started_moving = TRUE;
			InitGfxValue(xy);
			Store[xy] = EL_ACID;
		}
		else if (/*(game.engine_version >= VERSION_IDENT(3,1,0,0) &&
					CheckCollision[x][y] && !IS_FREE(x, y + 1)) ||*/
					(game.engine_version >= VERSION_IDENT(3,0,7,0) &&
					CAN_SMASH(element) && IsWasJustFalling(xy) &&
					(Feld[xy+A2TO1(0,1)] == EL_BLOCKED || IS_PLAYER(xy+A2TO1(0,1)))) ||
					(game.engine_version < VERSION_IDENT(2,2,0,7) &&
					CAN_SMASH(element) && IsWasJustMoving(xy) && !getPushed(xy+A2TO1(0,1)) &&
					(Feld[xy + A2TO1(0,1)] == EL_BLOCKED)))

		{
			if (getStop(xy))//Stop[x][y]
				return;

			// this is needed for a special case not covered by calling "Impact()"
			// from "ContinueMoving()": if an element moves to a tile directly below
			// another element which was just falling on that tile (which was empty
			// in the previous frame), the falling element above would just stop
			// instead of smashing the element below (in previous version, the above
			// element was just checked for "moving" instead of "falling", resulting
			// in incorrect smashes caused by horizontal movement of the above
			// element; also, the case of the player being the element to smash was
			// simply not covered here... :-/ )

//			CheckCollision[x][y] = 0;

			Impact(x, y);
		}
		else if (element == EL_SPRING && IS_FREE(xy+A2TO1(0,1)) && 
															level.use_spring_bug)
		{

			if(MovDir[xy] == MV_NO_MOVING)
			{
				InitMovingField(xy, MV_DOWN);
				started_moving = TRUE;
				InitGfxValue(xy);
			}
		}
		else if (IS_FREE(xy+A2TO1(0,1)) || 
							Feld[xy+A2TO1(0,1)] == EL_DIAMOND_BREAKING)
		{

			if (IsWasJustFalling(xy))	// prevent animation from being restarted
				MovDir[xy] = MV_DOWN;

			InitMovingField(xy, MV_DOWN);
			InitGfxValue(xy);
			started_moving = TRUE;
		}
		else if (element == EL_AMOEBA_DROP)
		{

			InitGfxValueAndFeld(xy,EL_AMOEBA_GROWING);
			Store[xy] = EL_AMOEBA_WET;
		}
		// Store[x][y + 1] must be zero, because:
		//   (EL_QUICKSAND_FULL -> EL_ROCK): Store[x][y + 1] == EL_QUICKSAND_EMPTY
		else if (((IS_SLIPPERY(Feld[xy+A2TO1(0,1)]) && !IS_PLAYER(xy+A2TO1(0,1))) ||
						(IS_EM_SLIPPERY_WALL(Feld[xy+A2TO1(0,1)]) && IS_GEM(element))) &&
						 !IS_FALLING(xy+A2TO1(0,1)) && !IsWasJustMoving(xy+A2TO1(0,1)) &&
						 element != EL_DX_SUPABOMB && element != EL_SP_DISK_ORANGE)
		{
			boolean can_fall_left  = (x > 0 && IS_FREE(xy+A2TO1(-1,0)) &&
										(IS_FREE(xy+A2TO1(-1,1)) ||
										Feld[xy+A2TO1(-1,1)] == EL_ACID));
			boolean can_fall_right = (x < lev_fieldx - 1 && 
										IS_FREE(xy+A2TO1(1,0)) &&
										(IS_FREE(xy+A2TO1(1,1)) ||
										Feld[xy+A2TO1(1,1)] == EL_ACID));
			boolean can_fall_any  = (can_fall_left || can_fall_right);
			boolean can_fall_both = (can_fall_left && can_fall_right);
/*
			if (can_fall_any && IS_CUSTOM_ELEMENT(Feld[x][y + 1]))
			{
				int slippery_type = element_info[Feld[x][y + 1]].slippery_type;

				if (slippery_type == SLIPPERY_ONLY_LEFT)
					can_fall_right = FALSE;
				else if (slippery_type == SLIPPERY_ONLY_RIGHT)
					can_fall_left = FALSE;
				else if (slippery_type == SLIPPERY_ANY_LEFT_RIGHT && can_fall_both)
					can_fall_right = FALSE;
				else if (slippery_type == SLIPPERY_ANY_RIGHT_LEFT && can_fall_both)
					can_fall_left = FALSE;

				can_fall_any  = (can_fall_left || can_fall_right);
				can_fall_both = (can_fall_left && can_fall_right);
			}
*/
			if (can_fall_any)
			{
				if (can_fall_both &&
								(game.emulation != EMU_BOULDERDASH &&
								element != EL_BD_ROCK && element != EL_BD_DIAMOND))
					can_fall_left = !(can_fall_right = RND(2));

				InitMovingField(xy, can_fall_left ? MV_LEFT : MV_RIGHT);
				InitGfxValue(xy);
				started_moving = TRUE;
			}
		}
		else if (IS_BELT_ACTIVE(Feld[xy+A2TO1(0,1)]))
		{
			boolean left_is_free  = (x > 0 && IS_FREE(xy+A2TO1(-1,0)));
			boolean right_is_free = (x < lev_fieldx - 1 && IS_FREE(xy+A2TO1(1,0)));
			int belt_nr = getBeltNrFromBeltActiveElement(Feld[xy+A2TO1(0,1)]);
			int belt_dir = game.belt_dir[belt_nr];

			if ((belt_dir == MV_LEFT  && left_is_free) ||
							(belt_dir == MV_RIGHT && right_is_free))
			{
				int addnextx = (belt_dir == MV_LEFT ? - 1 : + 1);

				started_moving = TRUE;
				InitMovingField(xy, belt_dir);

				setPushedTRUE(xy);//Pushed[x][y] = TRUE;
				setPushedTRUE(xy+A2TO1(addnextx,0));//Pushed[nextx][y] = TRUE;

				InitGfxValueAndGfxAction(xy,ACTION_DEFAULT);
			}
			else
			{
				MovDir[xy] = 0;	// if element was moving, stop it
				InitGfxValue(xy);
			}
		}
	}

	// not "else if" because of elements that can fall and move (EL_SPRING)
	if (CAN_MOVE(element) && !started_moving)
	{
//		int move_pattern = FindElementMovePattern(element); 
		// palm, was element_info[element].move_pattern;
		int newx, newy;
		short int newxy;

		if (IS_PUSHABLE(element) && JustBeingPushed(x, y))
			return;
/*
		Moving2Blocked(xy, &newxy);
		newx=A1TOX(newxy);
		newy=A1TOY(newxy);

    if (game.engine_version >= VERSION_IDENT(3,1,0,0) &&
	CheckCollision[x][y] && !IN_LEV_FIELD_AND_IS_FREE(newx, newy))
    {

      WasJustMoving[x][y] = 0;

      CheckCollision[x][y] = 0;

      TestIfElementHitsCustomElement(x, y, MovDir[x][y]);

      if (Feld[x][y] != element)	// element has changed
	return;
    }
*/
		if (!MovDelay[xy])	// start new movement phase
		{
			// all objects that can change their move direction after each step
			// YAMYAM, DARK_YAMYAM and PACMAN go straight until they hit a wall

			if (element != EL_YAMYAM &&
								element != EL_DARK_YAMYAM &&
								element != EL_PACMAN /*&&
								!(move_pattern & MV_ANY_DIRECTION) &&
								move_pattern != MV_TURNING_LEFT &&
								move_pattern != MV_TURNING_RIGHT &&
								move_pattern != MV_TURNING_LEFT_RIGHT &&
								move_pattern != MV_TURNING_RIGHT_LEFT &&
								move_pattern != MV_TURNING_RANDOM*/)
			{
				TurnRound(x, y);

				if (MovDelay[xy] && (element == EL_BUG ||
										element == EL_SPACESHIP ||
										element == EL_SP_SNIKSNAK ||
										element == EL_SP_ELECTRON ||
										element == EL_MOLE))
					DrawLevelField(x, y);
			}
		}

		if (MovDelay[xy])		// wait some time before next movement
		{
			MovDelay[xy]--;

			if (element == EL_ROBOT ||
					element == EL_YAMYAM ||
					element == EL_DARK_YAMYAM)
			{
				DrawLevelElementAnimationIfNeeded(x, y, element);
			}
			else if (element == EL_SP_ELECTRON)
				DrawLevelElementAnimationIfNeeded(x, y, element);
			else if (element == EL_DRAGON)
			{
				int i;
				int dir = MovDir[xy];
				int dx = (dir == MV_LEFT ? -1 : dir == MV_RIGHT ? +1 : 0);
				int dy = (dir == MV_UP   ? -1 : dir == MV_DOWN  ? +1 : 0);
				int graphic = (dir == MV_LEFT	? IMG_FLAMES_1_LEFT :
								dir == MV_RIGHT	? IMG_FLAMES_1_RIGHT :
								dir == MV_UP	? IMG_FLAMES_1_UP :
								dir == MV_DOWN	? IMG_FLAMES_1_DOWN : IMG_EMPTY);
				int frame = getGraphicAnimationFrame(graphic, 
								FrameCounter-GfxSFrame[xy]);//was GfxFrame[x][y]

				InitGfxValueAndGfxAction(xy, ACTION_ATTACKING);

				if (IS_PLAYER(xy))
					DrawPlayerField(x, y);
				else
					DrawLevelField(x, y);

				for (i = 1; i <= 3; i++)
				{
					int xx = x + i * dx;
					int yy = y + i * dy;
					int sx = SCREENX(xx);
					int sy = SCREENY(yy);
					int flame_graphic = graphic + (i - 1);

					if (!IN_LEV_FIELD(xx, yy) || IS_DRAGONFIRE_PROOF(Feld[A2TO1(xx,yy)]))
						break;

					if (MovDelay[xy])
					{
						int flamed = MovingOrBlocked2Element(xx, yy);

						// !!!
						if (IS_CLASSIC_ENEMY(flamed) || CAN_EXPLODE_BY_FIRE(flamed))
							Bang(xx, yy);
						else
							RemoveMovingField(xx, yy);

						//ChangeDelay[xx][yy] = 0;
						InitGfxValueAndFeld(A2TO1(xx,yy),EL_FLAMES);
						if (IN_SCR_FIELD(sx, sy))
						{
							DrawLevelFieldCrumbledSand(xx, yy);
							DrawGraphic(sx, sy, flame_graphic, frame);
						}
					}
					else
					{
						if (Feld[A2TO1(xx,yy)] == EL_FLAMES)
							InitGfxValueAndFeld(A2TO1(xx,yy),EL_EMPTY);
						DrawLevelField(xx, yy);
					}
				}
			}

			if (MovDelay[xy])	// element still has to wait some time
			{
				return;
			}

		}

		// now make next step

		Moving2Blocked(xy, &newxy);	// get next screen position
		newx=A1TOX(newxy);
		newy=A1TOY(newxy);

		if (DONT_COLLIDE_WITH(element) &&
							IN_LEV_FIELD(newx, newy) && IS_PLAYER(newxy) &&
										!PLAYER_ENEMY_PROTECTED(newx, newy, newxy))
		{
			TestIfBadThingRunsIntoHero(x, y, MovDir[xy]);

			return;
		}
		else if (CAN_MOVE_INTO_ACID(element) &&
					IN_LEV_FIELD(newx, newy) && Feld[newxy] == EL_ACID &&
					(MovDir[xy] == MV_DOWN ||
					game.engine_version >= VERSION_IDENT(3,1,0,0)))
		{
			SplashAcid(newx, newy);
			Store[xy] = EL_ACID;
		}
		else if (element == EL_PENGUIN && IN_LEV_FIELD(newx, newy))
		{
			if (Feld[newxy] == EL_EXIT_OPEN)
			{
				RemoveField(xy);
				DrawLevelField(x, y);

				if (IN_SCR_FIELD(SCREENX(newx), SCREENY(newy)))
					DrawGraphicThruMask(SCREENX(newx),SCREENY(newy), el2img(element), 0);

				local_player->friends_still_needed--;
				if (!local_player->friends_still_needed &&
								!local_player->GameOver && AllPlayersGone)
					local_player->LevelSolved = local_player->GameOver = TRUE;

				return;
			}
			else if (IS_FOOD_PENGUIN(Feld[newxy]))
			{
				if (DigField(local_player, x, y, newx, newy, 0, 0, DF_DIG) == MF_MOVING)
					DrawLevelField(newx, newy);
				else
				/*	GfxDir[x][y] =*/ MovDir[xy] = MV_NO_MOVING;
				InitGfxValue(xy);
			}
			else if (!IS_FREE(newxy))
			{
				InitGfxValueAndGfxAction(xy,ACTION_WAITING);

				if (IS_PLAYER(xy))
					DrawPlayerField(x, y);
				else
					DrawLevelField(x, y);

				return;
			}
		}
		else if (element == EL_PIG && IN_LEV_FIELD(newx, newy))
		{
			if (IS_FOOD_PIG(Feld[newxy]))
			{
				if (IS_MOVING(newxy))
					RemoveMovingField(newx, newy);
				else
				{
					InitGfxValueAndFeld(newxy,EL_EMPTY);
					DrawLevelField(newx, newy);
				}

			}
			else if (!IS_FREE(newxy))
			{
				if (IS_PLAYER(xy))
					DrawPlayerField(x, y);
				else
					DrawLevelField(x, y);

				return;
			}
		}
/*
    else if (IS_CUSTOM_ELEMENT(element) &&
	     CUSTOM_ELEMENT_CAN_ENTER_FIELD(element, newx, newy)

)
    {
      int new_element = Feld[newx][newy];

      if (!IS_FREE(newx, newy))
      {
	int action = (IS_DIGGABLE(new_element) ? ACTION_DIGGING :
		      IS_COLLECTIBLE(new_element) ? ACTION_COLLECTING :
		      ACTION_BREAKING);

	// no element can dig solid indestructible elements
	if (IS_INDESTRUCTIBLE(new_element) &&
	    !IS_DIGGABLE(new_element) &&
	    !IS_COLLECTIBLE(new_element))
	  return;

	if (AmoebaNr[newx][newy] &&
	    (new_element == EL_AMOEBA_FULL ||
	     new_element == EL_BD_AMOEBA ||
	     new_element == EL_AMOEBA_GROWING))
	{
	  AmoebaCnt[AmoebaNr[newx][newy]]--;
	  AmoebaCnt2[AmoebaNr[newx][newy]]--;
	}

	if (IS_MOVING(newx, newy))
	  RemoveMovingField(newx, newy);
	else
	{
	  RemoveField(newx, newy);
	  DrawLevelField(newx, newy);
	}

	// if digged element was about to explode, prevent the explosion
	ExplodeField[newx][newy] = EX_TYPE_NONE;

      }

      Store[newx][newy] = EL_EMPTY;
      if (IS_EQUAL_OR_IN_GROUP(new_element, MOVE_ENTER_EL(element)))
	Store[newx][newy] = element_info[element].move_leave_element;

      if (move_pattern & MV_MAZE_RUNNER_STYLE)
      {
	RunnerVisit[x][y] = FrameCounter;
	PlayerVisit[x][y] /= 8;		// expire player visit path
      }
    }
*/		else if (element == EL_DRAGON && IN_LEV_FIELD(newx, newy))
		{
			if (!IS_FREE(newxy))
			{
				if (IS_PLAYER(xy))
					DrawPlayerField(x, y);
				else
					DrawLevelField(x, y);

				return;
			}
			else
			{
				boolean wanna_flame = !RND(10);
				int dx = newx - x, dy = newy - y;
				int newx1 = newx + 1 * dx, newy1 = newy + 1 * dy;
				int newx2 = newx + 2 * dx, newy2 = newy + 2 * dy;
				int element1 = (IN_LEV_FIELD(newx1, newy1) ?
							MovingOrBlocked2Element(newx1, newy1) : EL_STEELWALL);
				int element2 = (IN_LEV_FIELD(newx2, newy2) ?
							MovingOrBlocked2Element(newx2, newy2) : EL_STEELWALL);

				if ((wanna_flame ||
							IS_CLASSIC_ENEMY(element1) ||
							IS_CLASSIC_ENEMY(element2)) &&
							element1 != EL_DRAGON && element2 != EL_DRAGON &&
							element1 != EL_FLAMES && element2 != EL_FLAMES)
				{
					ResetGfxAnimation(xy);
					InitGfxValueAndGfxAction(xy,ACTION_ATTACKING);

					if (IS_PLAYER(xy))
						DrawPlayerField(x, y);
					else
						DrawLevelField(x, y);

					MovDelay[xy] = 50;

					// !!!
					InitGfxValueAndFeld(newxy,EL_FLAMES);
					if (IN_LEV_FIELD(newx1, newy1) && 
											Feld[A2TO1(newx1,newy1)] == EL_EMPTY)
						InitGfxValueAndFeld(A2TO1(newx1,newy1),EL_FLAMES);
					if (IN_LEV_FIELD(newx2, newy2) &&
											Feld[A2TO1(newx2,newy2)] == EL_EMPTY)
						InitGfxValueAndFeld(A2TO1(newx2,newy2),EL_FLAMES);

					return;
				}
			}
		}
		else if (element == EL_YAMYAM && IN_LEV_FIELD(newx, newy) &&
									Feld[newxy] == EL_DIAMOND)
		{
			if (IS_MOVING(newxy))
				RemoveMovingField(newx, newy);
			else
			{
				InitGfxValueAndFeld(newxy,EL_EMPTY);
				DrawLevelField(newx, newy);
			}

		}
		else if (element == EL_DARK_YAMYAM && IN_LEV_FIELD(newx, newy) &&
								IS_FOOD_DARK_YAMYAM(Feld[newxy]))
		{
			if (AmoebaNr[newxy])
			{
				AmoebaCnt2[AmoebaNr[newxy]]--;
				if (Feld[newxy] == EL_AMOEBA_FULL ||
						Feld[newxy] == EL_BD_AMOEBA)
					AmoebaCnt[AmoebaNr[newxy]]--;
			}

			if (IS_MOVING(newxy))
			{
				RemoveMovingField(newx, newy);
			}
			else
			{
				InitGfxValueAndFeld(newxy,EL_EMPTY);
				DrawLevelField(newx, newy);
			}

		}
		else if ((element == EL_PACMAN || element == EL_MOLE)
					&& IN_LEV_FIELD(newx, newy) && IS_AMOEBOID(Feld[newxy]))
		{
			if (AmoebaNr[newxy])
			{
				AmoebaCnt2[AmoebaNr[newxy]]--;
				if (Feld[newxy] == EL_AMOEBA_FULL ||
									Feld[newxy] == EL_BD_AMOEBA)
					AmoebaCnt[AmoebaNr[newxy]]--;
			}

			if (element == EL_MOLE)
			{
				Feld[newxy] = EL_AMOEBA_SHRINKING;

				ResetGfxAnimation(xy);
				InitGfxValueAndGfxAction(xy,ACTION_DIGGING);
				DrawLevelField(x, y);

				MovDelay[newxy] = 0;	// start amoeba shrinking delay

				return;				// wait for shrinking amoeba
			}
			else	// element == EL_PACMAN
			{
				InitGfxValueAndFeld(newxy,EL_EMPTY);
				DrawLevelField(newx, newy);
			}
		}
		else if (element == EL_MOLE && IN_LEV_FIELD(newx, newy) &&
							(Feld[newxy] == EL_AMOEBA_SHRINKING ||
							(Feld[newxy] == EL_EMPTY && 
							getStop(newxy))))
							//Stop[newx][newy]
		{
			// wait for shrinking amoeba to completely disappear
			return;
		}
		else if (!IN_LEV_FIELD(newx, newy) || !IS_FREE(newxy))
		{
			// object was running against a wall

			TurnRound(x, y);

			if (GFX_ELEMENT(element) != EL_SAND)     // !!! FIX THIS (crumble) !!!
				DrawLevelElementAnimation(x, y, element);

			if (DONT_TOUCH(element))
				TestIfBadThingTouchesHero(x, y);

			return;
		}

		InitMovingField(xy, MovDir[xy]);
		InitGfxValue(xy);
	}

	if (MovDir[xy])
	{
		ContinueMoving(x, y);
	}

}

void ContinueMoving(int x, int y)
{
	short int xy=A2TO1(x,y);
	short int newxy;
	int element = Feld[xy];
	int stored = Store[xy];
//	struct ElementInfo *ei = &element_info[element];
	int direction = MovDir[xy];
	int dx = (direction == MV_LEFT ? -1 : direction == MV_RIGHT ? +1 : 0);
	int dy = (direction == MV_UP   ? -1 : direction == MV_DOWN  ? +1 : 0);
	int newx = x + dx, newy = y + dy;
	boolean pushed_by_player   = (getPushed(xy) && IS_PLAYER(xy));//Pushed[x][y]
	boolean pushed_by_conveyor = (getPushed(xy) && !IS_PLAYER(xy));//Pushed[x][y]

	MovPos[xy] += getElementMoveStepsize(xy);

	if (pushed_by_player)		// special case: moving object pushed by player
		MovPos[xy] = SIGN(MovPos[xy]) * (TILEX - ABS(PLAYERINFO(x,y)->MovPos));

	if (ABS(MovPos[xy]) < TILEX)
	{
		DrawLevelField(x, y);
		return;	// element is still moving
	}

	// element reached destination field

	newxy=A2TO1(newx,newy);

	Feld[xy] = EL_EMPTY;
	Feld[newxy] = element;
	MovPos[xy] = 0;	// force "not moving" for "crumbled sand"

	if (stored == EL_ACID)	// element is moving into acid pool
	{//was Store[x][y]
		element = Feld[newxy] = EL_ACID;
	}
	if (element == EL_MOLE)
	{
		Feld[xy] = EL_SAND;

		DrawLevelFieldCrumbledSandNeighbours(x, y);
	}
	else if (element == EL_QUICKSAND_FILLING)
	{
		element = Feld[newxy] = EL_QUICKSAND_FULL;
		Store[newxy] = Store[xy];
	}
	else if (element == EL_QUICKSAND_EMPTYING)
	{
		Feld[xy] = EL_QUICKSAND_EMPTY;
		element = Feld[newxy] = Store[xy];
	}
	else if (element == EL_MAGIC_WALL_FILLING)
	{
		element = Feld[newxy] = EL_MAGIC_WALL_FULL;
		if (!game.magic_wall_active)
			element = Feld[newxy] = EL_MAGIC_WALL_DEAD;
		Store[newxy] = Store[xy];
	}
	else if (element == EL_MAGIC_WALL_EMPTYING)
	{
		Feld[xy] = EL_MAGIC_WALL_ACTIVE;
		if (!game.magic_wall_active)
			Feld[xy] = EL_MAGIC_WALL_DEAD;
		element = Feld[newxy] = Store[xy];
	}
	else if (element == EL_BD_MAGIC_WALL_FILLING)
	{
		element = Feld[newxy] = EL_BD_MAGIC_WALL_FULL;
		if (!game.magic_wall_active)
			element = Feld[newxy] = EL_BD_MAGIC_WALL_DEAD;
		Store[newxy] = Store[xy];
	}
	else if (element == EL_BD_MAGIC_WALL_EMPTYING)
	{
		Feld[xy] = EL_BD_MAGIC_WALL_ACTIVE;
		if (!game.magic_wall_active)
			Feld[xy] = EL_BD_MAGIC_WALL_DEAD;
		element = Feld[newxy] = Store[xy];
	}
	else if (element == EL_AMOEBA_DROPPING)
	{
		Feld[xy] = EL_AMOEBA_WET;
		element = Feld[newxy] = Store[xy];
	}
	else if (element == EL_SOKOBAN_OBJECT)
	{
		if (Back[xy])
			Feld[xy] = Back[xy];

		if (Back[newxy])
			Feld[newxy] = EL_SOKOBAN_FIELD_FULL;

		Back[xy] = Back[newxy] = 0;
	}

	Store[xy] = EL_EMPTY;
	MovPos[xy] = MovDir[xy] = MovDelay[xy] = 0;
	MovDelay[newxy] = 0;
/*
	if (CAN_CHANGE(element))
	{
		// copy element change control values to new field
		ChangeDelay[newx][newy] = ChangeDelay[x][y];
		ChangePage[newx][newy] = ChangePage[x][y];
		Changed[newx][newy] = Changed[x][y];
		ChangeEvent[newx][newy] = ChangeEvent[x][y];
	}
	
	ChangeDelay[x][y] = 0;
	ChangePage[x][y] = -1;
	Changed[x][y] = CE_BITMASK_DEFAULT;
	ChangeEvent[x][y] = CE_BITMASK_DEFAULT;
*/

	// copy animation control values to new field
	GfxSFrame[newxy] = GfxSFrame[xy];//GfxFrame[newx][newy]  = GfxFrame[x][y];
	GfxRandom[newxy] = GfxRandom[xy];	// keep same random value
	GfxAction[newxy] = GfxAction[xy];	// keep action one frame
//	GfxDir[newx][newy]    = GfxDir[x][y];		// keep element direction

	setPushedFALSE(xy);
	setPushedFALSE(newxy);//Pushed[x][y] = Pushed[newx][newy] = FALSE;

	ResetGfxAnimation(xy);	// reset animation values for old field
/*
  // some elements can leave other elements behind after moving
  if (IS_CUSTOM_ELEMENT(element) && ei->move_leave_element != EL_EMPTY &&
      (ei->move_leave_type == LEAVE_TYPE_UNLIMITED || stored != EL_EMPTY) &&
      (!IS_PLAYER(x, y) || IS_WALKABLE(ei->move_leave_element)))
  {
    int move_leave_element = ei->move_leave_element;

    Feld[x][y] = move_leave_element;
    InitField(x, y, FALSE);

    if (GFX_CRUMBLED(Feld[x][y]))
      DrawLevelFieldCrumbledSandNeighbours(x, y);

    if (ELEM_IS_PLAYER(move_leave_element))
      RelocatePlayer(x, y, move_leave_element);
  }
*/
	if (!CAN_MOVE(element) ||
				(CAN_FALL(element) && direction == MV_DOWN &&
				(element == EL_SPRING /*||
				element_info[element].move_pattern == MV_WHEN_PUSHED ||
				element_info[element].move_pattern == MV_WHEN_DROPPED*/)))
/*		GfxDir[x][y] =*/ MovDir[newxy] = 0;

	{//PALM new
		short int zz;
		//there is no elements with 0 push delay in standart elements
		if (GfxAction[newxy] == ACTION_PUSHING)
			ResetGfxAnimation(newxy);
		//init global_sync elements
		if((zz=ANIM_INDEX(FindGfxValue(Feld[xy],GfxAction[xy],MovDir[xy])))!=0)
			if(anim_infoF->anim_global_sync[zz])
				GfxSFrame[xy]=0;
		if((zz=ANIM_INDEX(FindGfxValue(Feld[newxy],GfxAction[newxy],MovDir[newxy])))!=0)
			if(anim_infoF->anim_global_sync[zz])
				GfxSFrame[newxy]=0;
	}

	InitGfxValue(xy); 
	InitGfxValue(newxy);

	DrawLevelField(x, y);
	DrawLevelField(newx, newy);

	setStopTRUE(newx,newy);	// ignore this element until the next frame
	//was Stop[newx][newy]=TRUE

	// prevent pushed element from moving on in pushed direction
	if (pushed_by_player && CAN_MOVE(element)/* &&
				element_info[element].move_pattern & MV_ANY_DIRECTION &&
				!(element_info[element].move_pattern & direction)*/)
		TurnRound(newx, newy);

	// prevent elements on conveyor belt from moving on in last direction
	if (pushed_by_conveyor && CAN_FALL(element) &&
				direction & MV_HORIZONTAL)
	{
		MovDir[newxy] = 0;
	}

	if (!pushed_by_player)	// special case: moving object pushed by player
	{
/*		int nextx = newx + dx, nexty = newy + dy;
		boolean check_collision_again = IN_LEV_FIELD_AND_IS_FREE(nextx, nexty);
*/
//bugd("b.Set3WasJust x,y",(long)FFlags[newxy]*256*256+newx*256+newy);
		FFlags[newxy]|= FF_WASJUST_MASK;
		FFlags[newxy]&= 0xDF;//~FF_WASJUST_FALL;
//bugd("Set3WasJust x,y",(long)FFlags[newxy]*256*256+newx*256+newy);
		if(WasJustYBegin[newy]>newx)
			WasJustYBegin[newy]=newx;
		if(WasJustYEnd[newy]<newx)
			WasJustYEnd[newy]=newx;
		//WasJustMoving[newx][newy] = 3;

		if (CAN_FALL(element) && direction == MV_DOWN)
			FFlags[newxy]|=FF_WASJUST_FALL;//WasJustFalling[newx][newy] = 3;
			
/*
		if ((!CAN_FALL(element) || direction == MV_DOWN) && check_collision_again)
			CheckCollision[newx][newy] = 2;
*/	}

	if (DONT_TOUCH(element))	// object may be nasty to player or others
	{
		TestIfBadThingTouchesHero(newx, newy);
		TestIfBadThingTouchesFriend(newx, newy);

		if (!IS_CUSTOM_ELEMENT(element))
			TestIfBadThingTouchesOtherBadThing(newx, newy);
	}
	else if (element == EL_PENGUIN)
		TestIfFriendTouchesBadThing(newx, newy);

	if (CAN_FALL(element) && direction == MV_DOWN &&
					(newy == lev_fieldy - 1 || !IS_FREE(A2TO1(x, newy+1))))
		Impact(x, newy);
/*
	if (pushed_by_player)
	{
		int dig_side = MV_DIR_OPPOSITE(direction);
		struct PlayerInfo *player = PLAYERINFO(x, y);

		CheckElementChangeByPlayer(newx, newy, element, CE_PUSHED_BY_PLAYER,
													player->index_bit, dig_side);
		CheckTriggeredElementChangeByPlayer(newx,newy,element,CE_OTHER_GETS_PUSHED,
													player->index_bit, dig_side);
	}

	TestIfElementTouchesCustomElement(x, y);	// empty or new element

	TestIfElementHitsCustomElement(newx, newy, direction);

	TestIfPlayerTouchesCustomElement(newx, newy);
	TestIfElementTouchesCustomElement(newx, newy);*/
}

int AmoebeNachbarNr(int ax, int ay)
{
	int i;
	int element = Feld[A2TO1(ax,ay)];
	int group_nr = 0;
	const int xy_x[4]={0,-1,+1,0};
	const int xy_y[4]={-1,0,0,+1};

	for (i = 0; i < NUM_DIRECTIONS; i++)
	{
		int x = ax + xy_x[i];
		int y = ay + xy_y[i];
		short int xy0;

		if (!IN_LEV_FIELD(x, y))
			continue;
		
		xy0=A2TO1(x,y);
		if (Feld[xy0] == element && AmoebaNr[xy0] > 0)
			group_nr = AmoebaNr[xy0];
	}

	return group_nr;
};

LOGIC_SECTION void AmoebenVereinigen(int ax, int ay)
{
	int i, x, y, xx, yy;
	int new_group_nr = AmoebaNr[A2TO1(ax,ay)];
	const int xy_x[4]={0,-1,+1,0};
	const int xy_y[4]={-1,0,0,+1};

	if (new_group_nr == 0)
		return;

	for (i = 0; i < NUM_DIRECTIONS; i++)
	{
		short int xy0;
		x = ax + xy_x[i];
		y = ay + xy_y[i];

		if (!IN_LEV_FIELD(x, y))
			continue;

		xy0=A2TO1(x,y);
		if ((Feld[xy0] == EL_AMOEBA_FULL ||
				Feld[xy0] == EL_BD_AMOEBA ||
				Feld[xy0] == EL_AMOEBA_DEAD) &&
				AmoebaNr[xy0] != new_group_nr)
		{
			int old_group_nr = AmoebaNr[xy0];

			if (old_group_nr == 0)
				return;

			AmoebaCnt[new_group_nr] += AmoebaCnt[old_group_nr];
			AmoebaCnt[old_group_nr] = 0;
			AmoebaCnt2[new_group_nr] += AmoebaCnt2[old_group_nr];
			AmoebaCnt2[old_group_nr] = 0;

			for (yy = 0; yy < lev_fieldy; yy++)
			{
				register short int xxyy=A2TO1(0,yy);
				for (xx = 0; xx < lev_fieldx; xx++,xxyy+=A2TO1(1,0))
				{
					if (AmoebaNr[xxyy] == old_group_nr)
						AmoebaNr[xxyy] = new_group_nr;
				}
			}
		}
	}
}

void AmoebeUmwandeln(int ax, int ay)
{
	int i, x, y;
	short int axy=A2TO1(ax,ay);

	if (Feld[axy] == EL_AMOEBA_DEAD)
	{
		int group_nr = AmoebaNr[axy];

		for (y = 0; y < lev_fieldy; y++)
		{
			register short int xy=A2TO1(0,y);
			for (x = 0; x < lev_fieldx; x++,xy+=A2TO1(1,0))
			{
				if (Feld[xy] == EL_AMOEBA_DEAD && AmoebaNr[xy] == group_nr)
				{
					AmoebaNr[xy] = 0;
					InitGfxValueAndFeld(xy,EL_AMOEBA_TO_DIAMOND);
				}
			}
		}
		Bang(ax, ay);
	}
	else
	{
		const int xy_x[4]={0,-1,+1,0};
		const int xy_y[4]={-1,0,0,+1};

		for (i = 0; i < NUM_DIRECTIONS; i++)
		{
			x = ax + xy_x[i];
			y = ay + xy_y[i];

			if (!IN_LEV_FIELD(x, y))
				continue;

			if (Feld[A2TO1(x,y)] == EL_AMOEBA_TO_DIAMOND)
			{
				Bang(x, y);
			}
		}
	}
}

void AmoebeUmwandelnBD(int ax, int ay, int new_element)
{
	int x, y;
	int group_nr = AmoebaNr[A2TO1(ax,ay)];
	boolean done = FALSE;

	for (y = 0; y < lev_fieldy; y++)
	{
		register short int xy=A2TO1(0,y);
		for (x = 0; x < lev_fieldx; x++,xy+=A2TO1(1,0))
		{
			if (AmoebaNr[xy] == group_nr &&
					(Feld[xy] == EL_AMOEBA_DEAD ||
					Feld[xy] == EL_BD_AMOEBA ||
					Feld[xy] == EL_AMOEBA_GROWING))
			{
				AmoebaNr[xy] = 0;
				Feld[xy] = new_element;
				InitField(x, y, FALSE);
				DrawLevelField(x, y);
				done = TRUE;
			}
		}
	}

}

void AmoebeWaechst(int x, int y)
{
	short int xy=A2TO1(x,y);
	if (!MovDelay[xy])		// start new growing cycle
	{
		MovDelay[xy] = 7;
	}

	if (MovDelay[xy])		// wait some time before growing bigger //
	{
		MovDelay[xy]--;
		if (!(MovDelay[xy]&1) && IN_SCR_FIELD(SCREENX(x), SCREENY(y)))//was MovDelay[xy]/2
		{
			int frame = getGraphicAnimationFrame(IMG_AMOEBA_GROWING,
												6 - MovDelay[xy]);

			DrawGraphic(SCREENX(x), SCREENY(y), IMG_AMOEBA_GROWING, frame);
		}

		if (!MovDelay[xy])
		{
			InitGfxValueAndFeld(xy,Store[xy]);
			Store[xy] = 0;
			DrawLevelField(x, y);
		}
	}
};

LOGIC_SECTION void AmoebaDisappearing(int x, int y)
{
	short int xy=A2TO1(x,y);
	if (!MovDelay[xy])		// start new shrinking cycle
	{
		MovDelay[xy] = 7;
	}

	if (MovDelay[xy])		// wait some time before shrinking
	{
		MovDelay[xy]--;
		if (!(MovDelay[xy]&1) && IN_SCR_FIELD(SCREENX(x), SCREENY(y)))//was MovDelay[xy]/2
		{
			int frame = getGraphicAnimationFrame(IMG_AMOEBA_SHRINKING,
												6 - MovDelay[xy]);

			DrawGraphic(SCREENX(x), SCREENY(y), IMG_AMOEBA_SHRINKING, frame);
		}

		if (!MovDelay[xy])
		{
			InitGfxValueAndFeld(xy,EL_EMPTY);
			DrawLevelField(x, y);

			// don't let mole enter this field in this cycle;
			// (give priority to objects falling to this field from above)
			setStopTRUE(x,y);//was Stop[x][y] = TRUE;
		}
	}
}

void AmoebeAbleger(int ax, int ay)
{
	short int axy=A2TO1(ax,ay);
	short int newaxy;
	int i;
	int element = Feld[axy];
	int graphic = el2img(element);
	int newax = ax, neway = ay;
	const int xy_x[4]={0,-1,+1,0};
	const int xy_y[4]={-1,0,0,+1};

	if (!level.amoeba_speed)
	{
		InitGfxValueAndFeld(axy,EL_AMOEBA_DEAD);
		DrawLevelField(ax, ay);
		return;
	}

	if (IS_ANIMATED(graphic))
		DrawLevelGraphicAnimationIfNeeded(ax, ay, graphic);

	if (!MovDelay[axy])	// start making new amoeba field
		MovDelay[axy] = RND(FRAMES_PER_SECOND * 25 / (1 + level.amoeba_speed));

	if (MovDelay[axy])		// wait some time before making new amoeba
	{
		MovDelay[axy]--;
		if (MovDelay[axy])
			return;
	}

	if (element == EL_AMOEBA_WET)	// object is an acid / amoeba drop 
	{
		int start = RND(4);
		int x = ax + xy_x[start];
		int y = ay + xy_y[start];
		short int xy1=A2TO1(x,y);

		if (!IN_LEV_FIELD(x, y))
			return;

		if (IS_FREE(xy1) ||
					CAN_GROW_INTO(Feld[xy1]) ||
					Feld[xy1] == EL_QUICKSAND_EMPTY)
		{
			newax = x;
			neway = y;
		}

		if (newax == ax && neway == ay)
			return;
	}
	else				// normal or "filled" (BD style) amoeba
	{
		int start = RND(4);
		boolean waiting_for_player = FALSE;

		for (i = 0; i < NUM_DIRECTIONS; i++)
		{
			int j = (start + i) % 4;
			int x = ax + xy_x[j];
			int y = ay + xy_y[j];
			short int xy1=A2TO1(x,y);

			if (!IN_LEV_FIELD(x, y))
				continue;

			if (IS_FREE(xy1) ||
							CAN_GROW_INTO(Feld[xy1]) ||
							Feld[xy1] == EL_QUICKSAND_EMPTY)
			{
				newax = x;
				neway = y;
				break;
			}
			else if (IS_PLAYER(xy1))
				waiting_for_player = TRUE;
		}

		if (newax == ax && neway == ay)		// amoeba cannot grow
		{
			if (i == 4 && (!waiting_for_player || element == EL_BD_AMOEBA))
			{
				InitGfxValueAndFeld(axy,EL_AMOEBA_DEAD);
				DrawLevelField(ax, ay);
				AmoebaCnt[AmoebaNr[axy]]--;

				if (AmoebaCnt[AmoebaNr[axy]] <= 0)	// amoeba is completely dead
				{
					if (element == EL_AMOEBA_FULL)
						AmoebeUmwandeln(ax, ay);
					else if (element == EL_BD_AMOEBA)
						AmoebeUmwandelnBD(ax, ay, level.amoeba_content);
				}
			}
			return;
		}
		else if (element == EL_AMOEBA_FULL || element == EL_BD_AMOEBA)
		{
			// amoeba gets larger by growing in some direction

			int new_group_nr = AmoebaNr[axy];

			AmoebaNr[A2TO1(newax,neway)] = new_group_nr;
			AmoebaCnt[new_group_nr]++;
			AmoebaCnt2[new_group_nr]++;

			// if amoeba touches other amoeba(s) after growing, unify them
			AmoebenVereinigen(newax, neway);

			if (element == EL_BD_AMOEBA && AmoebaCnt2[new_group_nr] >= 200)
			{
				AmoebeUmwandelnBD(newax, neway, EL_BD_ROCK);
				return;
			}
		}
	}
	
	newaxy=A2TO1(newax,neway);
	
	if (element != EL_AMOEBA_WET || neway < ay || !IS_FREE(newaxy) ||
								(neway == lev_fieldy - 1 && newax != ax))
	{
		InitGfxValueAndFeld(newaxy,EL_AMOEBA_GROWING);	// creation of new amoeba
		Store[newaxy] = element;
	}
	else if (neway == ay)
	{
		InitGfxValueAndFeld(newaxy,EL_AMOEBA_DROP);	// drop left/right of amoeba
	}
	else
	{
		InitMovingField(axy, MV_DOWN);		// drop dripping from amoeba
		Feld[axy]=EL_AMOEBA_DROPPING;//InitGfxValueAndFeld(axy,EL_AMOEBA_DROPPING);
		Store[axy] = EL_AMOEBA_DROP;
		ContinueMoving(ax, ay);
		return;
	}

	DrawLevelField(newax, neway);
}

void Life(int ax, int ay)
{
	short int axy=A2TO1(ax,ay);
	int x1, y1, x2, y2;
	const int life[4] = { 2, 3, 3, 3 };	// parameters for "game of life"
	int life_time = 40;
	int element = Feld[axy];
	int graphic = el2img(element);
	boolean changed = FALSE;

	if (IS_ANIMATED(graphic))
		DrawLevelGraphicAnimationIfNeeded(ax, ay, graphic);

	if (getStop(axy))//was Stop[ax][ay]
		return;

	if (!MovDelay[axy])	// start new "game of life" cycle
		MovDelay[axy] = life_time;

	if (MovDelay[axy])		// wait some time before next cycle
	{
		MovDelay[axy]--;
		if (MovDelay[axy])
			return;
	}

	for (y1 = -1; y1 < 2; y1++) for (x1 = -1; x1 < 2; x1++)
	{
		int xx = ax+x1, yy = ay+y1;
		short int xxyy=A2TO1(xx,yy);
		int nachbarn = 0;

		if (!IN_LEV_FIELD(xx, yy))
			continue;

		for (y2 = -1; y2 < 2; y2++) for (x2 = -1; x2 < 2; x2++)
		{
			int x = xx+x2, y = yy+y2;
			short int xy1=A2TO1(x,y);

			if (!IN_LEV_FIELD(x, y) || (x == xx && y == yy))
				continue;

			if (((Feld[xy1] == element ||
						(element == EL_GAME_OF_LIFE && IS_PLAYER(xy1))) &&
						!getStop(xy1)) ||
						(IS_FREE(xy1) && getStop(xy1)))//was Stop[x][y]
				nachbarn++;
		}

		if (xx == ax && yy == ay)		// field in the middle
		{
			if (nachbarn < life[0] || nachbarn > life[1])
			{
				InitGfxValueAndFeld(xxyy,EL_EMPTY);
				if (!getStop(xxyy))//was Stop[xx][yy]
					DrawLevelField(xx, yy);
				setStopTRUE(xx,yy);//was Stop[xx][yy] = TRUE;
				changed = TRUE;
			}
		}
		else if (IS_FREE(xxyy) || CAN_GROW_INTO(Feld[xxyy]))
		{					// free border field
			if (nachbarn >= life[2] && nachbarn <= life[3])
			{
				InitGfxValueAndFeld(xxyy,element);
				MovDelay[xxyy] = (element == EL_GAME_OF_LIFE ? 0 : life_time-1);
				if (!getStop(xxyy))//was Stop[xx][yy]
					DrawLevelField(xx, yy);
				setStopTRUE(xx,yy);//was Stop[xx][yy] = TRUE;
				changed = TRUE;
			}
		}
	}
};

LOGIC_SECTION void CheckExit(int x, int y)
{
	if (local_player->gems_still_needed > 0 ||
				local_player->sokobanfields_still_needed > 0 ||
				local_player->lights_still_needed > 0)
	{
		int element = Feld[A2TO1(x,y)];
		int graphic = el2img(element);

		if (IS_ANIMATED(graphic))
			DrawLevelGraphicAnimationIfNeeded(x, y, graphic);

		return;
	}

	if (AllPlayersGone)	// do not re-open exit door closed after last player
		return;

	InitGfxValueAndFeld(A2TO1(x,y),EL_EXIT_OPENING);
//	AddDelayChanging(A2TO1(x,y),29,EL_EXIT_OPEN);//PALM!!!
};

LOGIC_SECTION void CheckExitSP(int x, int y)
{
	if (local_player->gems_still_needed > 0)
	{
		int element = Feld[A2TO1(x,y)];
		int graphic = el2img(element);

		if (IS_ANIMATED(graphic))
			DrawLevelGraphicAnimationIfNeeded(x, y, graphic);

		return;
	}

	if (AllPlayersGone)	// do not re-open exit door closed after last player
		return;

	InitGfxValueAndFeld(A2TO1(x,y),EL_SP_EXIT_OPENING);
//	AddDelayChanging(A2TO1(x,y),29,EL_SP_EXIT_OPEN);//PALM!!!
}

static void CloseAllOpenTimegates()
{
	int x, y;

	for (y = 0; y < lev_fieldy; y++)
	{
		register short int xy=A2TO1(0,y);
		for (x = 0; x < lev_fieldx; x++,xy=A2TO1(1,0))
		{
			int element = Feld[xy];

			if (element == EL_TIMEGATE_OPEN || element == EL_TIMEGATE_OPENING)
			{
				InitGfxValueAndFeld(xy,EL_TIMEGATE_CLOSING);
//				AddDelayChanging(xy,29,EL_TIMEGATE_CLOSED);//PALM!!!
			}
		}
	}
};

LOGIC_SECTION inline static void EdelsteinFunkeln(int x, int y)
{
	short int xy=A2TO1(x,y);

	if (MovDelay[xy] == 0)	// next animation frame
		MovDelay[xy] = 11 * !SimpleRND(500);

	if (MovDelay[xy] != 0)	// wait some time before next frame 
	{
		MovDelay[xy]--;
/*
		if (setup.direct_draw && MovDelay[xy])
			SetDrawtoField(DRAW_BUFFERED);
*/
		DrawLevelElementAnimation(x, y, Feld[xy]);

		if (MovDelay[xy] != 0)
		{
			int frame = getGraphicAnimationFrame(IMG_TWINKLE_WHITE,
												10 - MovDelay[xy]);

			DrawGraphicThruMask(SCREENX(x), SCREENY(y), IMG_TWINKLE_WHITE, frame);
/*
			if (setup.direct_draw)
			{
				int dest_x, dest_y;

				dest_x = FX + SCREENX(x) * TILEX;
				dest_y = FY + SCREENY(y) * TILEY;

				BlitBitmap(drawto_field, window,
									dest_x, dest_y, TILEX, TILEY, dest_x, dest_y);
				SetDrawtoField(DRAW_DIRECT);
			}
*/		}
	}
}

void MauerWaechst(int x, int y)
{
	short int xy=A2TO1(x,y);
	int delay = 6;

	if (!MovDelay[xy])		// next animation frame
		MovDelay[xy] = 3 * delay;

	if (MovDelay[xy])		// wait some time before next frame 
	{
		MovDelay[xy]--;

		if (IN_SCR_FIELD(SCREENX(x), SCREENY(y)))
		{
			int graphic = FindGfxValue(Feld[xy], ACTION_DEFAULT, MovDir[xy]); 
			//Gfx[xy];//was el_dir2img(Feld[xy], GfxDir[x][y]);
			int frame = getGraphicAnimationFrame(graphic, 17 - MovDelay[xy]);

			DrawGraphic(SCREENX(x), SCREENY(y), graphic, frame);
		}

		if (!MovDelay[xy])
		{
			if (MovDir[xy] == MV_LEFT)
			{
				if (IN_LEV_FIELD(x - 1, y) && IS_WALL(Feld[xy-A2TO1(1,0)]))
					DrawLevelField(x - 1, y);
			}
			else if (MovDir[xy] == MV_RIGHT)
			{
				if (IN_LEV_FIELD(x + 1, y) && IS_WALL(Feld[xy+A2TO1(1,0)]))
					DrawLevelField(x + 1, y);
			}
			else if (MovDir[xy] == MV_UP)
			{
				if (IN_LEV_FIELD(x, y - 1) && IS_WALL(Feld[xy-A2TO1(0,1)]))
					DrawLevelField(x, y - 1);
			}
			else
			{
				if (IN_LEV_FIELD(x, y + 1) && IS_WALL(Feld[xy+A2TO1(0,1)]))
					DrawLevelField(x, y + 1);
			}

			Feld[xy] = Store[xy];
			Store[xy] = 0;
/*			GfxDir[x][y] =*/ MovDir[xy] = MV_NO_MOVING;
			InitGfxValue(xy);
			DrawLevelField(x, y);
		}
	}
}

void MauerAbleger(int ax, int ay)
{
	short int axy=A2TO1(ax,ay);
	int element = Feld[axy];
	int graphic = el2img(element);
	boolean oben_frei = FALSE, unten_frei = FALSE;
	boolean links_frei = FALSE, rechts_frei = FALSE;
	boolean oben_massiv = FALSE, unten_massiv = FALSE;
	boolean links_massiv = FALSE, rechts_massiv = FALSE;
	boolean new_wall = FALSE;

	if (IS_ANIMATED(graphic))
		DrawLevelGraphicAnimationIfNeeded(ax, ay, graphic);

	if (!MovDelay[axy])	// start building new wall
		MovDelay[axy] = 6;

	if (MovDelay[axy])		// wait some time before building new wall
	{
		MovDelay[axy]--;
		if (MovDelay[axy])
			return;
	}

	if (IN_LEV_FIELD(ax, ay-1) && IS_FREE(axy+A2TO1(0,-1)))
		oben_frei = TRUE;
	if (IN_LEV_FIELD(ax, ay+1) && IS_FREE(axy+A2TO1(0,1)))
		unten_frei = TRUE;
	if (IN_LEV_FIELD(ax-1, ay) && IS_FREE(axy+A2TO1(-1,0)))
		links_frei = TRUE;
	if (IN_LEV_FIELD(ax+1, ay) && IS_FREE(axy+A2TO1(1,0)))
		rechts_frei = TRUE;

	if (element == EL_EXPANDABLE_WALL_VERTICAL ||
					element == EL_EXPANDABLE_WALL_ANY)
	{
		if (oben_frei)
		{
			Feld[axy+A2TO1(0,-1)] = EL_EXPANDABLE_WALL_GROWING;
			Store[axy+A2TO1(0,-1)] = element;
/*			GfxDir[ax][ay-1] = */MovDir[axy+A2TO1(0,-1)] = MV_UP;
			InitGfxValue(axy+A2TO1(0,-1));
			if (IN_SCR_FIELD(SCREENX(ax), SCREENY(ay-1)))
				DrawGraphic(SCREENX(ax), SCREENY(ay - 1),
											IMG_EXPANDABLE_WALL_GROWING_UP, 0);
			new_wall = TRUE;
		}
		if (unten_frei)
		{
			Feld[axy+A2TO1(0,1)] = EL_EXPANDABLE_WALL_GROWING;
			Store[axy+A2TO1(0,1)] = element;
/*			GfxDir[ax][ay+1] =*/ MovDir[axy+A2TO1(0,1)] = MV_DOWN;
			InitGfxValue(axy+A2TO1(0,1));
			if (IN_SCR_FIELD(SCREENX(ax), SCREENY(ay+1)))
				DrawGraphic(SCREENX(ax), SCREENY(ay + 1),
											IMG_EXPANDABLE_WALL_GROWING_DOWN, 0);
			new_wall = TRUE;
		}
	}

	if (element == EL_EXPANDABLE_WALL_HORIZONTAL ||
					element == EL_EXPANDABLE_WALL_ANY ||
					element == EL_EXPANDABLE_WALL)
	{
		if (links_frei)
		{
			Feld[axy-A2TO1(1,0)] = EL_EXPANDABLE_WALL_GROWING;
			Store[axy-A2TO1(1,0)] = element;
/*			GfxDir[ax-1][ay] = */MovDir[axy-A2TO1(1,0)] = MV_LEFT;
			InitGfxValue(axy-A2TO1(1,0));
			if (IN_SCR_FIELD(SCREENX(ax-1), SCREENY(ay)))
				DrawGraphic(SCREENX(ax - 1), SCREENY(ay),
											IMG_EXPANDABLE_WALL_GROWING_LEFT, 0);
			new_wall = TRUE;
		}

		if (rechts_frei)
		{
			Feld[axy+A2TO1(1,0)] = EL_EXPANDABLE_WALL_GROWING;
			Store[axy+A2TO1(1,0)] = element;
/*			GfxDir[ax+1][ay] = */MovDir[axy+A2TO1(1,0)] = MV_RIGHT;
			InitGfxValue(axy+A2TO1(1,0));
			if (IN_SCR_FIELD(SCREENX(ax+1), SCREENY(ay)))
				DrawGraphic(SCREENX(ax + 1), SCREENY(ay),
											IMG_EXPANDABLE_WALL_GROWING_RIGHT, 0);
			new_wall = TRUE;
		}
	}

	if (element == EL_EXPANDABLE_WALL && (links_frei || rechts_frei))
		DrawLevelField(ax, ay);

	if (!IN_LEV_FIELD(ax, ay-1) || IS_WALL(Feld[axy-A2TO1(0,1)]))
		oben_massiv = TRUE;
	if (!IN_LEV_FIELD(ax, ay+1) || IS_WALL(Feld[axy+A2TO1(0,1)]))
		unten_massiv = TRUE;
	if (!IN_LEV_FIELD(ax-1, ay) || IS_WALL(Feld[axy-A2TO1(1,0)]))
		links_massiv = TRUE;
	if (!IN_LEV_FIELD(ax+1, ay) || IS_WALL(Feld[axy+A2TO1(1,0)]))
		rechts_massiv = TRUE;

	if (((oben_massiv && unten_massiv) ||
				element == EL_EXPANDABLE_WALL_HORIZONTAL ||
				element == EL_EXPANDABLE_WALL) &&
				((links_massiv && rechts_massiv) ||
				element == EL_EXPANDABLE_WALL_VERTICAL))
		InitGfxValueAndFeld(axy,EL_WALL);
};

LOGIC_SECTION void CheckForDragon(int x, int y)
{
	int i, j;
	boolean dragon_found = FALSE;
	const int xy_x[4]={0,-1,+1,0};
	const int xy_y[4]={-1,0,0,-+1};

	for (i = 0; i < NUM_DIRECTIONS; i++)
	{
		for (j = 0; j < 4; j++)
		{
			int xx = x + j * xy_x[i], yy = y + j * xy_y[i];
			register short int xxyy=A2TO1(xx,yy);

			if (IN_LEV_FIELD(xx, yy) &&
						(Feld[xxyy] == EL_FLAMES || Feld[xxyy] == EL_DRAGON))
			{
				if (Feld[xxyy] == EL_DRAGON)
					dragon_found = TRUE;
			}
			else
				break;
		}
	}

	if (!dragon_found)
	{
		for (i = 0; i < NUM_DIRECTIONS; i++)
		{
			for (j = 0; j < 3; j++)
			{
				int xx = x + j * xy_x[i], yy = y + j * xy_y[i];
				register short int xxyy=A2TO1(xx,yy);
  
				if (IN_LEV_FIELD(xx, yy) && Feld[xxyy] == EL_FLAMES)
				{
					InitGfxValueAndFeld(xxyy,EL_EMPTY);
					DrawLevelField(xx, yy);
				}
				else
					break;
			}
		}
	}
}

static void ChangeElementNowExt(int x, int y, int target_element)
{
	short int xy=A2TO1(x,y);
	// check if element under player changes from accessible to unaccessible
	//   (needed for special case of dropping element which then changes)
	if (IS_PLAYER(xy) && !PLAYER_EXPLOSION_PROTECTED(x, y, xy) &&
				IS_ACCESSIBLE(Feld[xy]) && !IS_ACCESSIBLE(target_element))
	{
		Bang(x, y);
		return;
	}

	RemoveField(xy);
	Feld[xy] = target_element;

//	Changed[x][y] |= ChangeEvent[x][y];	// ignore same changes in this frame

	ResetGfxAnimation(xy);
	ResetRandomAnimationValue(xy);

	InitField(x, y, FALSE);
	if (CAN_MOVE(Feld[xy]))
		InitMovDir(x, y);

	InitGfxValue(xy);
	DrawLevelField(x, y);

	if (GFX_CRUMBLED(Feld[xy]))
		DrawLevelFieldCrumbledSandNeighbours(x, y);

	TestIfBadThingTouchesHero(x, y);
//	TestIfPlayerTouchesCustomElement(x, y);
//	TestIfElementTouchesCustomElement(x, y);

	if (ELEM_IS_PLAYER(target_element))
		RelocatePlayer(x, y, target_element);
}

static void ChangeElement(int x, int y, int page)
{
/*	int element = MovingOrBlocked2Element(x, y);
	struct ElementInfo *ei = &element_info[element];
	struct ElementChangeInfo *change = &ei->change_page[page];

  if (ChangeDelay[x][y] == 0)		// initialize element change
  {
    ChangeDelay[x][y] = (    change->delay_fixed  * change->delay_frames +
			 RND(change->delay_random * change->delay_frames)) + 1;

    ResetGfxAnimation(x, y);
    ResetRandomAnimationValue(x, y);

    if (change->pre_change_function)
      change->pre_change_function(x, y);
  }

  ChangeDelay[x][y]--;

  if (ChangeDelay[x][y] != 0)		// continue element change 
  {
    int graphic = el_act_dir2img(element, GfxAction[x][y], GfxDir[x][y]);

    if (IS_ANIMATED(graphic))
      DrawLevelGraphicAnimationIfNeeded(x, y, graphic);

    if (change->change_function)
      change->change_function(x, y);
  }
  else					// finish element change
  {
    if (ChangePage[x][y] != -1)		// remember page from delayed change
    {
      page = ChangePage[x][y];
      ChangePage[x][y] = -1;
    }

    if (IS_MOVING(x, y))		// never change a running system ;-) 
    {
      ChangeDelay[x][y] = 1;		// try change after next move step
      ChangePage[x][y] = page;		// remember page to use for change

      return;
    }

    if (ChangeElementNow(x, y, element, page))
    {
      if (change->post_change_function)
	change->post_change_function(x, y);
    }
  }*/
}

static void SetPlayerWaiting(struct PlayerInfo *player, boolean is_waiting)
{
	boolean last_waiting = player->is_waiting;
	int move_dir = player->MovDir;

	player->last_action_waiting = player->action_waiting;

	if (is_waiting)
	{
		if (!last_waiting)		// not waiting -> waiting
		{
			player->is_waiting = TRUE;

			player->frame_counter_bored =
					FrameCounter +
					game.player_boring_delay_fixed +
					SimpleRND(game.player_boring_delay_random);
			player->frame_counter_sleeping =
					FrameCounter +
					game.player_sleeping_delay_fixed +
					SimpleRND(game.player_sleeping_delay_random);

			InitPlayerGfxAnimation(player, ACTION_WAITING, player->MovDir);
		}

		if (game.player_sleeping_delay_fixed +
							game.player_sleeping_delay_random > 0 &&
							player->anim_delay_counter == 0 &&
							player->post_delay_counter == 0 &&
							FrameCounter >= player->frame_counter_sleeping)
			player->is_sleeping = TRUE;
		else if (game.player_boring_delay_fixed +
							game.player_boring_delay_random > 0 &&
							FrameCounter >= player->frame_counter_bored)
			player->is_bored = TRUE;

		player->action_waiting = (player->is_sleeping ? ACTION_SLEEPING :
									   player->is_bored ? ACTION_BORING :
														 ACTION_WAITING);

		if (player->is_sleeping)
		{
			if (player->num_special_action_sleeping > 0)
			{
				if (player->anim_delay_counter == 0 && player->post_delay_counter == 0)
				{
					int last_special_action = player->special_action_sleeping;
					int num_special_action = player->num_special_action_sleeping;
					int special_action =
							(last_special_action == ACTION_DEFAULT ? ACTION_SLEEPING_1 :
							last_special_action == ACTION_SLEEPING ? ACTION_SLEEPING :
							last_special_action < ACTION_SLEEPING_1 + num_special_action - 1 ?
							last_special_action + 1 : ACTION_SLEEPING);
					int special_graphic =
							el_act_dir2img(player->element_nr, special_action, move_dir);

/*					player->anim_delay_counter =
							graphic_info[special_graphic].anim_delay_fixed +
							SimpleRND(graphic_info[special_graphic].anim_delay_random);
					player->post_delay_counter =
							graphic_info[special_graphic].post_delay_fixed +
							SimpleRND(graphic_info[special_graphic].post_delay_random);
*/
					player->special_action_sleeping = special_action;
				}

				if (player->anim_delay_counter > 0)
				{
					player->action_waiting = player->special_action_sleeping;
					player->anim_delay_counter--;
				}
				else if (player->post_delay_counter > 0)
				{
					player->post_delay_counter--;
				}
			}
		}
		else if (player->is_bored)
		{
			if (player->num_special_action_bored > 0)
			{
				if (player->anim_delay_counter == 0 && player->post_delay_counter == 0)
				{
					int special_action =
							ACTION_BORING_1 + SimpleRND(player->num_special_action_bored);
					int special_graphic =
							el_act_dir2img(player->element_nr, special_action, move_dir);

/*					player->anim_delay_counter =
							graphic_info[special_graphic].anim_delay_fixed +
							SimpleRND(graphic_info[special_graphic].anim_delay_random);
					player->post_delay_counter =
							graphic_info[special_graphic].post_delay_fixed +
							SimpleRND(graphic_info[special_graphic].post_delay_random);
*/
					player->special_action_bored = special_action;
				}

				if (player->anim_delay_counter > 0)
				{
					player->action_waiting = player->special_action_bored;
					player->anim_delay_counter--;
				}
				else if (player->post_delay_counter > 0)
				{
					player->post_delay_counter--;
				}
			}
		}
	}
	else if (last_waiting)	// waiting -> not waiting
	{
		player->is_waiting = FALSE;
		player->is_bored = FALSE;
		player->is_sleeping = FALSE;

		player->frame_counter_bored = -1;
		player->frame_counter_sleeping = -1;

		player->anim_delay_counter = 0;
		player->post_delay_counter = 0;

		player->action_waiting = ACTION_DEFAULT;

		player->special_action_bored = ACTION_DEFAULT;
		player->special_action_sleeping = ACTION_DEFAULT;
	}
}; 

LOGIC_SECTION static byte PlayerActions(struct PlayerInfo *player, byte player_action)
{
	boolean moved = FALSE, snapped = FALSE, dropped = FALSE;
	int left	= player_action & MV_LEFT;
	int right	= player_action & MV_RIGHT;
	int up	= player_action & MV_UP;
	int down	= player_action & MV_DOWN;
	int button1	= player_action & KEY_BUTTON_1;
	int button2	= player_action & KEY_BUTTON_2;
	int dx	= (left ? -1	: right ? 1	: 0);
	int dy	= (up   ? -1	: down  ? 1	: 0);

	if (!player->active || tape.pausing)
		return 0;

	if (player_action)
	{
		if (button1)
			snapped = SnapField(player, dx, dy);
		else
		{
			if (button2)
				dropped = DropElement(player);

			moved = MovePlayer(player, dx, dy);
		}

		if (tape.single_step && tape.recording && !tape.pausing)
		{
			if (button1 || (dropped && !moved))
			{
				TapeTogglePause(TAPE_TOGGLE_AUTOMATIC);
				SnapField(player, 0, 0);		// stop snapping
			}
		}

		SetPlayerWaiting(player, FALSE);

		return player_action;
	}
	else
	{
		// no actions for this player (no input at player's configured device)

		DigField(player, 0, 0, 0, 0, 0, 0, DF_NO_PUSH);
		SnapField(player, 0, 0);
		CheckGravityMovementWhenNotMoving(player);

		if (player->MovPos == 0)
			SetPlayerWaiting(player, TRUE);

		if (player->MovPos == 0)	// needed for tape.playing
			player->is_moving = FALSE;

		player->is_dropping = FALSE;

		return 0;
	}
}

//PALM!!!!
LOGIC_SECTION static short int FindChangeDelay(short int element)
{
	switch(element)
	{
		case EL_SP_BUGGY_BASE:
			return 2*FRAMES_PER_SECOND + RND(5 * FRAMES_PER_SECOND) -
						FRAMES_PER_SECOND/4;
		case EL_SP_BUGGY_BASE_ACTIVATING:
			return FRAMES_PER_SECOND/4;
		case EL_SP_BUGGY_BASE_ACTIVE:
			return FRAMES_PER_SECOND + RND(FRAMES_PER_SECOND);
		case EL_TRAP:
			return 2*FRAMES_PER_SECOND + RND(5*FRAMES_PER_SECOND);
		case EL_TRAP_ACTIVE:
			return 31;
		case EL_EXIT_OPENING:
		case EL_EXIT_CLOSING:
		case EL_SP_EXIT_OPENING:
		case EL_SP_EXIT_CLOSING:
		case EL_SWITCHGATE_OPENING:
		case EL_SWITCHGATE_CLOSING:
		case EL_TIMEGATE_OPENING:
		case EL_TIMEGATE_CLOSING:
			return 29;
		case EL_ACID_SPLASH_LEFT:
		case EL_ACID_SPLASH_RIGHT:
		case EL_PEARL_BREAKING:
			return 8;
		case EL_NUT_BREAKING:
			return 6;
		case EL_TIMEGATE_SWITCH_ACTIVE:
			return level.time_timegate * FRAMES_PER_SECOND;
		case EL_ROBOT_WHEEL_ACTIVE:
			return level.time_wheel * FRAMES_PER_SECOND;
		default:
			bugd("FindChangeDelay cannot find!",element);
			return 1;
	}
}

//PALM!!!!
LOGIC_SECTION static short int FindTargetChangeElement(short int element)
{
	switch(element)
	{
		case EL_SP_BUGGY_BASE:
			return EL_SP_BUGGY_BASE_ACTIVATING;
		case EL_SP_BUGGY_BASE_ACTIVATING:
			return EL_SP_BUGGY_BASE_ACTIVE;
		case EL_SP_BUGGY_BASE_ACTIVE:
			return EL_SP_BUGGY_BASE;
		case EL_TRAP:
			return EL_TRAP_ACTIVE;
		case EL_TRAP_ACTIVE:
			return EL_TRAP;
		case EL_EXIT_OPENING:
			return EL_EXIT_OPEN;
		case EL_EXIT_CLOSING:
			return EL_EXIT_CLOSED;
		case EL_SP_EXIT_OPENING:
			return EL_SP_EXIT_OPEN;
		case EL_SP_EXIT_CLOSING:
			return EL_SP_EXIT_CLOSED;
		case EL_ACID_SPLASH_LEFT:
		case EL_ACID_SPLASH_RIGHT:
		case EL_PEARL_BREAKING:
			return EL_EMPTY;
		case EL_SWITCHGATE_CLOSING:
			return EL_SWITCHGATE_CLOSED;
		case EL_SWITCHGATE_OPENING:
			return EL_SWITCHGATE_OPEN;
		case EL_TIMEGATE_OPENING:
			return EL_TIMEGATE_OPEN;
		case EL_TIMEGATE_CLOSING:
			return EL_TIMEGATE_CLOSED;
		case EL_NUT_BREAKING:
			return EL_EMERALD;
		case EL_TIMEGATE_SWITCH_ACTIVE:
			return EL_TIMEGATE_SWITCH;
		case EL_ROBOT_WHEEL_ACTIVE:
			return EL_ROBOT_WHEEL;
		default:
			bugd("FindTargetChangeElement cannot find!",element);
			return EL_EMPTY;
	}
}

void GameActions()
{
//	static unsigned long action_delay = 0;
	unsigned long action_delay_value;
	int y, element, graphic;
	register short int x;
	byte recorded_player_action_;
	byte summarized_player_action = 0;
	byte tape_action_;

	if (game_status != GAME_MODE_PLAYING)
		return;

	action_delay_value =
		(tape.playing && tape.fast_forward ? FfwdFrameDelay : GameFrameDelay);

	if (tape.playing && tape.warp_forward && !tape.pausing)
		action_delay_value = 0;

	// ---------- main game synchronization point ----------

//	WaitUntilDelayReached(&action_delay, action_delay_value);

	if (tape.pausing)
		return;

	recorded_player_action_ = (tape.playing ? TapePlayAction() : 0);

	if (recorded_player_action_ == 0 && tape.pausing)
		return;

	summarized_player_action |= stored_player_.action;
	stored_player_.effective_action = stored_player_.action;

	local_player->effective_action = summarized_player_action;
	
	if (recorded_player_action_ != 0)
		stored_player_.effective_action = recorded_player_action_;

	{
		tape_action_ = stored_player_.effective_action;

		if (tape.recording && tape_action_ && !tape.player_participates[0])
			tape.player_participates[0] = TRUE;    // player just appeared from CE 
	}

	// only save actions from input devices, but not programmed actions

	if (tape.recording)
		TapeRecordAction(tape_action_);

	{
		byte actual_player_action = stored_player_.effective_action;

		//!!! THIS BREAKS THE FOLLOWING TAPES: !!!
		// - rnd_equinox_tetrachloride 048
		// - rnd_equinox_tetrachloride_ii 096
		// - rnd_emanuel_schmieg 002
		// - doctor_sloan_ww 001, 020
		if (stored_player_.MovPos == 0)
			CheckGravityMovement(&stored_player_);

		// overwrite programmed action with tape action
		if (stored_player_.programmed_action)
			actual_player_action = stored_player_.programmed_action;

		if (recorded_player_action_)
		{
		}

		PlayerActions(&stored_player_, actual_player_action);

		ScrollPlayer(&stored_player_, SCROLL_GO_ON);
	}

	ScrollScreen(NULL, SCROLL_GO_ON);

	//for downwards compatibility, the following code emulates a fixed bug that
	//occured when pushing elements (causing elements that just made their last
	//pushing step to already (if possible) make their first falling step in the
	//same game frame, which is bad); this code is also needed to use the famous
	//"spring push bug" which is used in older levels and might be wanted to be
	//used also in newer levels, but in this case the buggy pushing code is only
	//affecting the "spring" element and no other elements

	if (game.engine_version < VERSION_IDENT(2,2,0,7) || level.use_spring_bug)
	{
		int x = stored_player_.jx;
		int y = stored_player_.jy;
		short int xy=A2TO1(x,y);

		if (stored_player_.active && stored_player_.is_pushing 
							&& stored_player_.is_moving && IS_MOVING(xy) &&
							(game.engine_version < VERSION_IDENT(2,2,0,7) ||
							Feld[xy] == EL_SPRING))
			{
				ContinueMoving(x, y);

				// continue moving after pushing (this is actually a bug) 
				if (!IS_MOVING(xy))
				{
					setStopFALSE(xy);//was Stop[x][y] = FALSE;
				}
			}
	}

	for (y = 0; y < lev_fieldy; y++)
	{
		if(StopYBegin[y]<=StopYEnd[y])
		{
			register short int bzi=A2TO1(StopYBegin[y],y); 
			for (x = StopYBegin[y]; x <=StopYEnd[y]; x++,bzi+=A2TO1(1,0))
			{
				FFlags[bzi]&=~FF_STOP;//was Stop[x][y] = FALSE;
			}
			StopYBegin[y] = MAX_LEV_FIELDX;
			StopYEnd[y] = 0; 
		}
	}

/*		Changed[x][y] = CE_BITMASK_DEFAULT;
		ChangeEvent[x][y] = CE_BITMASK_DEFAULT;
*/
	for (y = 0; y < lev_fieldy; y++)
	{
		if(WasJustYBegin[y]<=WasJustYEnd[y])
		{
			int f_value;
			register short int bzi=A2TO1(WasJustYBegin[y],y);
//bugd("WasJust",(long)y*256*256+WasJustYBegin[y]*256+WasJustYEnd[y]); 
			for (x = WasJustYBegin[y]; x <=WasJustYEnd[y]; x++,bzi+=A2TO1(1,0))
			{
/*		if (WasJustMoving[x][y] > 0)
			WasJustMoving[x][y]--;
		if (WasJustFalling[x][y] > 0)
			WasJustFalling[x][y]--;
		if (CheckCollision[x][y] > 0)
			CheckCollision[x][y]--;*/
				if(IsWasJustMoving(bzi))
					FFlags[bzi]-=FF_WASJUST_STEP;
//bugd("after Flags,x,y",(long)FFlags[bzi]*256*256+x*256+y);
			}
			f_value=MAX_LEV_FIELDX;
			bzi=A2TO1(WasJustYBegin[y],y);
			for(x=WasJustYBegin[y]; x<=WasJustYEnd[y]; x++,bzi+=A2TO1(1,0))
				if(IsWasJustMoving(bzi))
				{
					f_value=x;
					break;
				}
			WasJustYBegin[y]=f_value;

			f_value=0;
			bzi=A2TO1(WasJustYEnd[y],y);
			for(x=WasJustYEnd[y]; x>=WasJustYBegin[y]; x--,bzi-=A2TO1(1,0))
				if(IsWasJustMoving(bzi))
				{
					f_value=x;
					break;
				}
			WasJustYEnd[y]=f_value;
		}
	}
/*
		GfxFrame[x][y]++;

		// reset finished pushing action (not done in ContinueMoving() to allow
		// continous pushing animation for elements with zero push delay)
		if (GfxAction[x][y] == ACTION_PUSHING && !IS_MOVING(x, y))
		{
			ResetGfxAnimation(x, y);
			DrawLevelField(x, y);
		}
*/

//if(Feld[1134]==EL_SP_TERMINAL_ACTIVE)for_bugd++;
//	CheckDelayChanging();
do{DS_TestMainLoop--;
	for (y = 0; y < lev_fieldy; y++)
	{	
	unsigned char lastx=FeldYEnd[y];
	short int xy=A2TO1(FeldYStart[y],y);
	for (x = FeldYStart[y]; x <=lastx; x++,xy+=A2TO1(1,0))
	{
		register char MDL=MainDoLoop[Feld[xy]];

		if(MDL==MDL_INACTIVE_NOANIM)
			continue;
		
		if(MDL==MDL_CLASSIC_FALL)
		{
			if (!IS_MOVING(xy))
			{
				if((y + 1 < lev_fieldy && !CAN_HOLD_FALLING_DOWN(Feld[xy+A2TO1(0,1)])
											//this meant down=SAND,WALL, and so on
						&& (!CAN_HOLD_BLOCKED_FALLING(Feld[xy+A2TO1(0,1)])
						  ||(x!=0 && Feld[xy-A2TO1(1,0)]==EL_EMPTY 
						  			&& (Feld[xy+A2TO1(-1,1)]==EL_EMPTY
										|| Feld[xy+A2TO1(-1,1)]==EL_ACID))
						  ||(x+1<lev_fieldx && Feld[xy+A2TO1(1,0)]==EL_EMPTY 
						  			&& (Feld[xy+A2TO1(1,1)]==EL_EMPTY
										|| Feld[xy+A2TO1(1,1)]==EL_ACID)))))
						  			//this meant down,left,right=rocks,diamonds,sand...
					StartMoving(x, y);
				
				if(	!IN_SCR_FIELD(SCREENX(x), SCREENY(y)) || IS_MOVING(xy))
					continue;

				element = Feld[xy];

				graphic = FindGfxValue(element, GfxAction[xy], MovDir[xy]);//Gfx[xy];
//was			graphic = el_act_dir2img(element, GfxAction[xy], GfxDir[x][y]);

				if (IS_ANIMATED(graphic) &&
//already						!IS_MOVING(xy) &&
								!getStop(xy))//Stop[x][y]
					DrawLevelGraphicAnimationIfNeeded(x, y, graphic);

				if (IS_GEM(element) || element == EL_SP_INFOTRON)
//already			if(	IN_SCR_FIELD(SCREENX(x), SCREENY(y)) && !IS_MOVING(xy))
						if (element != EL_BD_DIAMOND)
							EdelsteinFunkeln(x, y);
				continue;
			}
			else//IS_MOVING
			{
				ContinueMoving(x, y);
				continue;
			}
		}

	switch(MDL)
	{
	case MDL_FALL_OR_MOVE:
		if (!IS_MOVING(xy))
		{
			StartMoving(x, y);

			if(	!IN_SCR_FIELD(SCREENX(x), SCREENY(y)) || IS_MOVING(xy))
				break;//was continue;

			element = Feld[xy];
			graphic = FindGfxValue(element, GfxAction[xy], MovDir[xy]);
//was		graphic = el_act_dir2img(element, GfxAction[x][y], GfxDir[x][y]);

			if (IS_ANIMATED(graphic) &&
//already					!IS_MOVING(xy) &&
							!getStop(xy))//Stop[x][y]
				DrawLevelGraphicAnimationIfNeeded(x, y, graphic);
		}
		else //if (IS_MOVING(xy))
			ContinueMoving(x, y);
		break;
	case MDL_ANIM_ONLY:
		if(	!IN_SCR_FIELD(SCREENX(x), SCREENY(y)))
			break;
		element = Feld[xy];
		graphic = FindGfxValue(element, GfxAction[xy], MovDir[xy]);
//was	graphic = el_act_dir2img(element, GfxAction[x][y], GfxDir[x][y]);
		if(IS_ANIMATED(graphic))
			DrawLevelGraphicAnimationIfNeeded(x, y, graphic);
		break;
	case MDL_CONT_MOVE_ONLY:
		if (IS_MOVING(xy))
			ContinueMoving(x, y);
		break;
	case MDL_GAME_OF_LIFE:
		Life(x, y);
		break;
	case MDL_EXP_WALL_GROW:
		MauerWaechst(x, y);
		break;
	case MDL_EXP_WALL:
		MauerAbleger(x, y);
		break;
	case MDL_EXIT_CLOSED:
		CheckExit(x, y);
		break;
	case MDL_SP_EXIT_CLOSED:
		CheckExitSP(x, y);
		break;
	case MDL_FLAMES:
		CheckForDragon(x, y);
		break;
	case MDL_EXPLOSION:
		// drawing of correct explosion animation is handled separately
		break;
	case MDL_AMOEBA_GROW:
		AmoebeWaechst(x, y);
		break;
	case MDL_AMOEBA_SHRINK:
		AmoebaDisappearing(x, y);
		break;
	case MDL_AMOEBALIVE:
	//level.amoeba_speed is used for MDL definition in InitGame
		//no animation - because 1000000 delay
		if (!MovDelay[xy])	// start making new amoeba field
			MovDelay[xy] = RND(FRAMES_PER_SECOND * 25 / (1 + level.amoeba_speed));

		if (MovDelay[xy])		// wait some time before making new amoeba
		{
			MovDelay[xy]--;
			if (MovDelay[xy])
				break;
		}
		MovDelay[xy]=1;
		AmoebeAbleger(x, y);
		break;
	case MDL_CHANGING:
		if(MovDelay[xy]<20000)// initialize element change
		{
			if(MovDelay[xy]!=0)
				bugd("Warning! MovDelay non-zero for changing",Feld[xy]);
			MovDelay[xy]=20000+FindChangeDelay(Feld[xy]);
			ResetGfxAnimation(xy);
			ResetRandomAnimationValue(xy); 
		}
		MovDelay[xy]--;
		if(MovDelay[xy]>20000)// continue element change
		{
			int graphic = FindGfxValue(Feld[xy], GfxAction[xy], MovDir[xy]);
			//was el_act_dir2img(element, GfxAction[x][y], GfxDir[x][y]);

			if (Feld[xy]==EL_SP_BUGGY_BASE_ACTIVE)
			//was (ANIM_MODE(graphic) == ANIM_RANDOM)
			{
//				if(IS_NEXT_FRAME(FrameCounter - GfxSFrame[xy], graphic))
				ResetRandomAnimationValue(xy);//delay=1 :)

				SetRandomAnimationValue(xy);
			}
			if (IS_ANIMATED(graphic))
				DrawLevelGraphicAnimationIfNeeded(x, y, graphic);
		}
		else// finish element change
		{
			short int target_element=FindTargetChangeElement(Feld[xy]);
			ChangeElementNowExt(x, y, target_element);

			if(target_element==EL_ROBOT_WHEEL)
			{
				if (ZX == x && ZY == y)
					ZX = ZY = -1;
			}
		}
		break;
	case MDL_OTHER:
	default:

		countStartMoving++;

		element = Feld[xy];
		if(DS_MDL_Other) bugd("MDL_OTHER at x,y",(long)element*256*256+x*256+y);
		graphic = FindGfxValue(element, GfxAction[xy], MovDir[xy]);
		if(graphic==-1)
			{bugd("MDL_OTHER gfx bad for element",element);graphic=IMG_UNKNOWN;}
//was	graphic = el_act_dir2img(element, GfxAction[xy], GfxDir[x][y]);
/*
		if (graphic_info[graphic].anim_global_sync)
			GfxFrame[x][y] = FrameCounter;

		if (ANIM_MODE(graphic) == ANIM_RANDOM &&
						IS_NEXT_FRAME(GfxFrame[x][y], graphic))
			ResetRandomAnimationValue(x, y);
*/
		SetRandomAnimationValue(xy);

		if (IS_INACTIVE(element))
		{
			if (IS_ANIMATED(graphic))
				DrawLevelGraphicAnimationIfNeeded(x, y, graphic);
			
			break;//was continue;
		}

		// this may take place after moving, so 'element' may have changed
/*		if (IS_CHANGING(x, y) &&
					(game.engine_version < VERSION_IDENT(3,0,7,1) || !Stop[x][y]))
		{
			ChangeElement(x, y, element_info[element].event_page_nr[CE_DELAY]);
			element = Feld[x][y];
			graphic = el_act_dir2img(element, GfxAction[x][y], GfxDir[x][y]);
		}
*/
/*		else*/ if (IS_MOVING(xy))
			ContinueMoving(x, y);
		else if (IS_ACTIVE_BOMB(element))
			CheckDynamite(x, y);

#if !USE_NEW_AMOEBA_CODE
		else if (IS_AMOEBALIVE(element))
			AmoebeAbleger(x, y);
#endif

		else if (IS_ANIMATED(graphic) /*&& !IS_CHANGING(x, y)*/)
			DrawLevelGraphicAnimationIfNeeded(x, y, graphic);

	}
	}
	}
}while(DS_TestMainLoop>0);DS_TestMainLoop++;
#if USE_NEW_AMOEBA_CODE
	// new experimental amoeba growth stuff 
	if (!(FrameCounter % 8))
	{
		static unsigned long random = 1684108901;

		for (i = 0; i < level.amoeba_speed * 28 / 8; i++)
		{
			x = RND(lev_fieldx);
			y = RND(lev_fieldy);
			element = Feld[x][y];

			if (!IS_PLAYER(x,y) &&
					(element == EL_EMPTY ||
					CAN_GROW_INTO(element) ||
					element == EL_QUICKSAND_EMPTY ||
					element == EL_ACID_SPLASH_LEFT ||
					element == EL_ACID_SPLASH_RIGHT))
			{
				if ((IN_LEV_FIELD(x, y-1) && Feld[x][y-1] == EL_AMOEBA_WET) ||
						(IN_LEV_FIELD(x-1, y) && Feld[x-1][y] == EL_AMOEBA_WET) ||
						(IN_LEV_FIELD(x+1, y) && Feld[x+1][y] == EL_AMOEBA_WET) ||
						(IN_LEV_FIELD(x, y+1) && Feld[x][y+1] == EL_AMOEBA_WET))
					Feld[x][y] = EL_AMOEBA_DROP;
			}
			// !!! extend EL_SAND to anything diggable (but maybe not SP_BASE) !!!
			if (!IS_PLAYER(x,y) &&
					(element == EL_EMPTY ||
					element == EL_SAND ||
					element == EL_QUICKSAND_EMPTY ||
					element == EL_ACID_SPLASH_LEFT ||
					element == EL_ACID_SPLASH_RIGHT))
			{
				if ((IN_LEV_FIELD(x, y-1) && Feld[x][y-1] == EL_AMOEBA_WET) ||
						(IN_LEV_FIELD(x-1, y) && Feld[x-1][y] == EL_AMOEBA_WET) ||
						(IN_LEV_FIELD(x+1, y) && Feld[x+1][y] == EL_AMOEBA_WET) ||
						(IN_LEV_FIELD(x, y+1) && Feld[x][y+1] == EL_AMOEBA_WET))
					Feld[x][y] = EL_AMOEBA_DROP;
			}

			random = random * 129 + 1;
		}
	}
#endif

	game.explosions_delayed = FALSE;

	for (y = 0; y < lev_fieldy; y++)
	{
		if(ExplodeFieldY[y]||FeldYExplosion[y])
		{
			short int xy=A2TO1(0,y);
			for (x = 0; x < lev_fieldx; x++,xy+=A2TO1(1,0))
			{
				if (ExplodeFlags[xy]&EF_EXP_FIELD)//ExplodeField[x][y]
				{
					Explode(x, y, EX_PHASE_START,getExplodeField(xy));
										// ExplodeField[x][y]
					ExplodeFlags[xy]&=EF_EXP_PHASE;//ExplodeField[x][y] = EX_NO_EXPLOSION;
				}
				else if (Feld[xy] == EL_EXPLOSION)
				{
					Explode(x, y, getExplodePhase(xy), EX_NORMAL);//ExplodePhase[x][y]
//new		Explode(x, y, ExplodePhase[x][y], EX_TYPE_NORMAL);
					ExplodeFlags[xy]&=EF_EXP_PHASE;//ExplodeField[x][y] = EX_NO_EXPLOSION;
				}
				//was ExplodeField[x][y] = EX_NO_EXPLOSION;
//new	    ExplodeField[x][y] = EX_TYPE_NONE;
				//but I replaced it in if-else
			}
			ExplodeFieldY[y]=0;
		}
	}

	game.explosions_delayed = TRUE;

	if (game.magic_wall_active)
	{
		if (game.magic_wall_time_left > 0)
		{
			game.magic_wall_time_left--;
			if (!game.magic_wall_time_left)
			{
				for (y = 0; y < lev_fieldy; y++) for (x = 0; x < lev_fieldx; x++)
				{
					element = Feld[A2TO1(x,y)];

					if (element == EL_MAGIC_WALL_ACTIVE ||
									element == EL_MAGIC_WALL_FULL)
					{
						InitGfxValueAndFeld(A2TO1(x,y),EL_MAGIC_WALL_DEAD);
						DrawLevelField(x, y);
					}
					else if (element == EL_BD_MAGIC_WALL_ACTIVE ||
									element == EL_BD_MAGIC_WALL_FULL)
					{
						InitGfxValueAndFeld(A2TO1(x,y),EL_BD_MAGIC_WALL_DEAD);
						DrawLevelField(x, y);
					}
				}

				game.magic_wall_active = FALSE;
			}
		}
	}

	if (game.light_time_left > 0)
	{
		game.light_time_left--;
		if (game.light_time_left == 0)
			RedrawAllLightSwitchesAndInvisibleElements();
	}

	if (game.timegate_time_left > 0)
	{
		game.timegate_time_left--;
		if (game.timegate_time_left == 0)
			CloseAllOpenTimegates();
	}

	if (TimeFrames >= FRAMES_PER_SECOND)
	{
		TimeFrames = 0;
		TapeTime++;

		if (SHIELD_ON(&stored_player_))
		{
			stored_player_.shield_normal_time_left--;

			if (stored_player_.shield_deadly_time_left > 0)
				stored_player_.shield_deadly_time_left--;
		}

		if (!level.use_step_counter)
		{
			TimePlayed++;

			if (TimeLeft > 0)
			{
				TimeLeft--;

				DrawGameValue_Time(TimeLeft);

				if (!TimeLeft && setup.time_limit)
					KillHero(&stored_player_);
			}
			else if (level.time == 0 && !AllPlayersGone) // level w/o time limit
				DrawGameValue_Time(TimePlayed);
    }

//		if (tape.recording || tape.playing)
//			DrawVideoDisplay(VIDEO_STATE_TIME_ON, TapeTime);
	}

	DrawAllPlayers();
/*
  if (options.debug)			// calculate frames per second 
  {
    static unsigned long fps_counter = 0;
    static int fps_frames = 0;
    unsigned long fps_delay_ms = Counter() - fps_counter;

    fps_frames++;

    if (fps_delay_ms >= 500)	// calculate fps every 0.5 seconds
    {
      global.frames_per_second = 1000 * (float)fps_frames / fps_delay_ms;

      fps_frames = 0;
      fps_counter = Counter();
    }

    redraw_mask |= REDRAW_FPS;
  }
*/
	FrameCounter++;
	TimeFrames++;

	{
		int move_frames =
			MOVE_DELAY_NORMAL_SPEED /  stored_player_.move_delay_value;

		stored_player_.Frame += move_frames;

		if (stored_player_.MovPos != 0)
			stored_player_.StepFrame += move_frames;

		if (stored_player_.drop_delay > 0)
			stored_player_.drop_delay--;
	}

	if (local_player->show_envelope != 0 && local_player->MovPos == 0)
	{
		ShowEnvelope(local_player->show_envelope - EL_ENVELOPE_1);

		local_player->show_envelope = 0;
	}

}

static boolean AllPlayersInSight(struct PlayerInfo *player, int x, int y)
{
	int min_x = x, min_y = y, max_x = x, max_y = y;

	int jx = stored_player_.jx, jy = stored_player_.jy;

	if (!stored_player_.active || &stored_player_ == player)
		;
	else
	{
		min_x = MIN(min_x, jx);
		min_y = MIN(min_y, jy);
		max_x = MAX(max_x, jx);
		max_y = MAX(max_y, jy);
	}

	return (max_x - min_x < SCR_FIELDX && max_y - min_y < SCR_FIELDY);
}

static boolean AllPlayersInVisibleScreen()
{
	int jx = stored_player_.jx, jy = stored_player_.jy;

	if (!stored_player_.active)
		return TRUE;

	if (!IN_VIS_FIELD(SCREENX(jx), SCREENY(jy)))
		return FALSE;

	return TRUE;
}

void ScrollLevel(int dx, int dy)
{
//	int softscroll_offset = (setup.soft_scrolling ? TILEX : 0);
	int x, y;
/*
	BlitBitmap(drawto_field, drawto_field,
				FX + TILEX * (dx == -1) - softscroll_offset,
				FY + TILEY * (dy == -1) - softscroll_offset,
				SXSIZE - TILEX * (dx!=0) + 2 * softscroll_offset,
				SYSIZE - TILEY * (dy!=0) + 2 * softscroll_offset,
				FX + TILEX * (dx == 1) - softscroll_offset,
				FY + TILEY * (dy == 1) - softscroll_offset);
*/
	if (dx)
	{
		x = (dx == 1 ? BX1 : BX2);
		for (y = BY1; y <= BY2; y++)
			DrawScreenField(x, y);
	}

	if (dy)
	{
		y = (dy == 1 ? BY1 : BY2);
		for (x = BX1; x <= BX2; x++)
			DrawScreenField(x, y);
	}

	redraw_mask |= REDRAW_FIELD;
}; 

//PALM
LOGIC_SECTION unsigned char PalmAccessDirection(short int element)
{
	switch(element)
	{
		case EL_TUBE_ANY:			return MV_LEFT | MV_RIGHT | MV_UP | MV_DOWN;
		case EL_TUBE_VERTICAL:		return                      MV_UP | MV_DOWN;
		case EL_TUBE_HORIZONTAL:	return MV_LEFT | MV_RIGHT                  ;
		case EL_TUBE_VERTICAL_LEFT:	return MV_LEFT |            MV_UP | MV_DOWN;
		case EL_TUBE_VERTICAL_RIGHT:return           MV_RIGHT | MV_UP | MV_DOWN;
		case EL_TUBE_HORIZONTAL_UP:	return MV_LEFT | MV_RIGHT | MV_UP          ;
		case EL_TUBE_HORIZONTAL_DOWN:return MV_LEFT| MV_RIGHT |         MV_DOWN;
		case EL_TUBE_LEFT_UP:		return MV_LEFT |            MV_UP          ;
		case EL_TUBE_LEFT_DOWN:		return MV_LEFT |                    MV_DOWN;
		case EL_TUBE_RIGHT_UP:		return           MV_RIGHT | MV_UP          ;
		case EL_TUBE_RIGHT_DOWN:	return           MV_RIGHT |         MV_DOWN;
		case EL_SP_PORT_LEFT:		return           MV_RIGHT                  ;
		case EL_SP_PORT_RIGHT:		return MV_LEFT                             ;
		case EL_SP_PORT_UP:			return                              MV_DOWN;
		case EL_SP_PORT_DOWN:		return                      MV_UP          ;
		case EL_SP_PORT_HORIZONTAL:	return MV_LEFT | MV_RIGHT                  ;
		case EL_SP_PORT_VERTICAL:	return                      MV_UP | MV_DOWN;
		case EL_SP_PORT_ANY:		return MV_LEFT | MV_RIGHT | MV_UP | MV_DOWN;
		case EL_SP_GRAVITY_PORT_LEFT:return          MV_RIGHT                  ;
		case EL_SP_GRAVITY_PORT_RIGHT:return MV_LEFT                           ;
		case EL_SP_GRAVITY_PORT_UP:	return                              MV_DOWN;
		case EL_SP_GRAVITY_PORT_DOWN:return                     MV_UP          ;
		case EL_SP_GRAVITY_ON_PORT_LEFT:return       MV_RIGHT                  ;
		case EL_SP_GRAVITY_ON_PORT_RIGHT:return MV_LEFT                        ;
		case EL_SP_GRAVITY_ON_PORT_UP:return                            MV_DOWN;
		case EL_SP_GRAVITY_ON_PORT_DOWN:return                  MV_UP          ;
		case EL_SP_GRAVITY_OFF_PORT_LEFT:return      MV_RIGHT                  ;
		case EL_SP_GRAVITY_OFF_PORT_RIGHT:return MV_LEFT                       ;
		case EL_SP_GRAVITY_OFF_PORT_UP:return                           MV_DOWN;
		case EL_SP_GRAVITY_OFF_PORT_DOWN:return                 MV_UP          ;
		default:					return MV_ALL_DIRECTIONS;
	}
};

#define ACCESS_FROM(e, d)			(PalmAccessDirection(e) &(d))
#define IS_WALKABLE_FROM(e, d)		(IS_WALKABLE(e)   && ACCESS_FROM(e, d))
#define IS_PASSABLE_FROM(e, d)		(IS_PASSABLE(e)   && ACCESS_FROM(e, d))
#define IS_ACCESSIBLE_FROM(e, d)	(IS_ACCESSIBLE(e) && ACCESS_FROM(e, d))

static boolean canFallDown(struct PlayerInfo *player)
{
	int jx = player->jx, jy = player->jy;
	short int jxy=A2TO1(jx,jy);

	return (IN_LEV_FIELD(jx, jy + 1) &&
				(IS_FREE(jxy + A2TO1(0,1)) ||
				(Feld[jxy+A2TO1(0,1)] == EL_ACID && player->can_fall_into_acid)) &&
				IS_WALKABLE_FROM(Feld[jxy], MV_DOWN) &&
				!IS_WALKABLE_INSIDE(Feld[jxy]));
}

static boolean canPassField(short int xy, int move_dir)
{
	int opposite_dir = MV_DIR_OPPOSITE(move_dir);
	int dx = (move_dir & MV_LEFT ? -1 : move_dir & MV_RIGHT ? +1 : 0);
	int dy = (move_dir & MV_UP   ? -1 : move_dir & MV_DOWN  ? +1 : 0);
	short int nextxy=xy+A2TO1(dx,dy);
	int element = Feld[xy];

  return (IS_PASSABLE_FROM(element, opposite_dir) &&
	  !CAN_MOVE(element) &&
	  IN_LEV_FIELD(A1TOX(nextxy), A1TOY(nextxy)) && !IS_PLAYER(nextxy) &&
	  IS_WALKABLE_FROM(Feld[nextxy], move_dir) &&
	  (level.can_pass_to_walkable || IS_FREE(nextxy)));
}

static boolean canMoveToValidFieldWithGravity(short int xy, int move_dir)
{
	int opposite_dir = MV_DIR_OPPOSITE(move_dir);
	int dx = (move_dir & MV_LEFT ? -1 : move_dir & MV_RIGHT ? +1 : 0);
	int dy = (move_dir & MV_UP   ? -1 : move_dir & MV_DOWN  ? +1 : 0);
	short int newxy=xy+A2TO1(dx,dy);

	return (IN_LEV_FIELD(A1TOX(newxy), A1TOY(newxy)) && !IS_FREE_OR_PLAYER(newxy) &&
				IS_GRAVITY_REACHABLE(Feld[newxy]) &&
				(IS_DIGGABLE(Feld[newxy]) ||
				IS_WALKABLE_FROM(Feld[newxy], opposite_dir) ||
				canPassField(newxy, move_dir)));
}

static void CheckGravityMovement(struct PlayerInfo *player)
{
	if (game.gravity && !player->programmed_action)
	{
		int move_dir_horizontal = player->effective_action & MV_HORIZONTAL;
		int move_dir_vertical   = player->effective_action & MV_VERTICAL;

		boolean player_is_snapping = player->effective_action & KEY_BUTTON_1;

		short int jxy = A2TO1(player->jx, player->jy);

		boolean player_is_moving_to_valid_field =
				(!player_is_snapping &&
				(canMoveToValidFieldWithGravity(jxy, move_dir_horizontal) ||
				canMoveToValidFieldWithGravity(jxy, move_dir_vertical)));

		boolean player_can_fall_down = canFallDown(player);

		if (player_can_fall_down &&
						!player_is_moving_to_valid_field)
		{

			player->programmed_action = MV_DOWN;
		}
	}
};

/*
  MovePlayerOneStep()
  -----------------------------------------------------------------------------
  dx, dy:		direction (non-diagonal) to try to move the player to
  real_dx, real_dy:	direction as read from input device (can be diagonal)
*/

boolean MovePlayerOneStep(struct PlayerInfo *player,
						  int dx, int dy, int real_dx, int real_dy)
{
	int jx = player->jx, jy = player->jy;
	short int jxy=A2TO1(jx,jy);
	int new_jx = jx + dx, new_jy = jy + dy;
	int element;
	int can_move;

	if (!player->active || (!dx && !dy))
		return MF_NO_ACTION;

	player->MovDir = (dx < 0 ? MV_LEFT :
					  dx > 0 ? MV_RIGHT :
					  dy < 0 ? MV_UP :
					  dy > 0 ? MV_DOWN :	MV_NO_MOVING);

	if (!IN_LEV_FIELD(new_jx, new_jy))
		return MF_NO_ACTION;

	if (!AllPlayersInSight(player, new_jx, new_jy))
		return MF_NO_ACTION;

	element = MovingOrBlocked2ElementIfNotLeaving(A2TO1(new_jx, new_jy));

	if (DONT_RUN_INTO(element))
	{
		if (element == EL_ACID && dx == 0 && dy == 1)
		{
			SplashAcid(new_jx, new_jy);
			Feld[jxy] = EL_PLAYER_1;
			InitMovingField(jxy, MV_DOWN);
			InitGfxValue(jxy);
			Store[jxy]=EL_ACID;
			ContinueMoving(jx, jy);
			BuryHero(player);
		}
		else
			TestIfHeroRunsIntoBadThing(jx, jy, player->MovDir);

		return MF_MOVING;
	}

	can_move = DigField(player, jx, jy, new_jx, new_jy, real_dx, real_dy, DF_DIG);
	if (can_move != MF_MOVING)
		return can_move;

	// check if DigField() has caused relocation of the player
	if (player->jx != jx || player->jy != jy)
		return MF_NO_ACTION;

//	SASetValue(&StorePlayer, jx, jy, 0);
	player->last_jx = jx;
	player->last_jy = jy;
	player->jx = new_jx;
	player->jy = new_jy;
//	SASetValue(&StorePlayer, new_jx, new_jy, player->element_nr);
	StorePlayerXY=A2TO1(new_jx,new_jy);

	player->MovPos =
			(dx > 0 || dy > 0 ? -1 : 1) * (TILEX - TILEX / player->move_delay_value);

	player->step_counter++;

	player->drop_delay = 0;

//	PlayerVisit[jx][jy] = FrameCounter;

	ScrollPlayer(player, SCROLL_INIT);

	return MF_MOVING;
};

boolean MovePlayer(struct PlayerInfo *player, int dx, int dy)
{
	int jx = player->jx, jy = player->jy;
	int old_jx = jx, old_jy = jy;
	int moved = MF_NO_ACTION;

	if (!player->active)
		return FALSE;

	if (!dx && !dy)
	{
		if (player->MovPos == 0)
		{
			player->is_moving = FALSE;
			player->is_digging = FALSE;
			player->is_collecting = FALSE;
			player->is_snapping = FALSE;
			player->is_pushing = FALSE;
		}

		return FALSE;
	}

	if (!FrameReached(&player->move_delay, player->move_delay_value))
		return FALSE;

	// store if player is automatically moved to next field
	player->is_auto_moving = (player->programmed_action != MV_NO_MOVING);

	// remove the last programmed player action
	player->programmed_action = 0;

	if (player->MovPos)
	{
		// should only happen if pre-1.2 tape recordings are played
		// this is only for backward compatibility

		int original_move_delay_value = player->move_delay_value;

		// scroll remaining steps with finest movement resolution
		player->move_delay_value = MOVE_DELAY_NORMAL_SPEED;

		while (player->MovPos)
		{
			ScrollPlayer(player, SCROLL_GO_ON);
			ScrollScreen(NULL, SCROLL_GO_ON);
			FrameCounter++;
			DrawAllPlayers();
			BackToFront();
		}

		player->move_delay_value = original_move_delay_value;
	}

	if (player->last_move_dir & MV_HORIZONTAL)
	{
		if (!(moved |= MovePlayerOneStep(player, 0, dy, dx, dy)))
			moved |= MovePlayerOneStep(player, dx, 0, dx, dy);
	}
	else
	{
		if (!(moved |= MovePlayerOneStep(player, dx, 0, dx, dy)))
			moved |= MovePlayerOneStep(player, 0, dy, dx, dy);
	}

	jx = player->jx;
	jy = player->jy;

	if (moved & MF_MOVING && !ScreenMovPos )
	{
		int old_scroll_x = scroll_x, old_scroll_y = scroll_y;
		int offset = (setup.scroll_delay ? 2 : 0);//was 3:0

		if (!IN_VIS_FIELD(SCREENX(jx), SCREENY(jy)))
		{
			// actual player has left the screen -- scroll in that direction
			if (jx != old_jx)		// player has moved horizontally
				scroll_x += (jx - old_jx);
			else			// player has moved vertically
				scroll_y += (jy - old_jy);
		}
		else
		{
			if (jx != old_jx)		// player has moved horizontally
			{
				if ((player->MovDir == MV_LEFT && scroll_x > jx - MIDPOSX + offset+1) ||
						(player->MovDir == MV_RIGHT && scroll_x < jx - MIDPOSX - offset))
					scroll_x = jx-MIDPOSX + (scroll_x < jx-MIDPOSX ? -offset : +offset);
//PALM!!!
				if(offset) 
					scroll_x = old_scroll_x + 
								SIGN(scroll_x - old_scroll_x)*SCROLL_SCREEN_X;

				// don't scroll over playfield boundaries
				if (scroll_x < SBX_Left || scroll_x > SBX_Right)
					scroll_x = (scroll_x < SBX_Left ? SBX_Left : SBX_Right);
//PALM!!!
				// don't scroll more than one field at a time
				if(!offset)
					scroll_x = old_scroll_x + SIGN(scroll_x - old_scroll_x);

				// don't scroll against the player's moving direction
				if ((player->MovDir == MV_LEFT && scroll_x > old_scroll_x) ||
						(player->MovDir == MV_RIGHT && scroll_x < old_scroll_x))
					scroll_x = old_scroll_x;
			}
			else			// player has moved vertically
			{
				if ((player->MovDir == MV_UP && scroll_y > jy - MIDPOSY + offset+1) ||
						(player->MovDir == MV_DOWN && scroll_y < jy - MIDPOSY - offset))
					scroll_y = jy-MIDPOSY + (scroll_y < jy-MIDPOSY ? -offset : +offset);
//PALM!!!
				if(offset) 
					scroll_y = old_scroll_y + 
								SIGN(scroll_y - old_scroll_y)*SCROLL_SCREEN_Y;

				// don't scroll over playfield boundaries
				if (scroll_y < SBY_Upper || scroll_y > SBY_Lower)
					scroll_y = (scroll_y < SBY_Upper ? SBY_Upper : SBY_Lower);
//PALM!!!
				// don't scroll more than one field at a time
				if(!offset)
					scroll_y = old_scroll_y + SIGN(scroll_y - old_scroll_y);

				// don't scroll against the player's moving direction
				if ((player->MovDir == MV_UP && scroll_y > old_scroll_y) ||
						(player->MovDir == MV_DOWN && scroll_y < old_scroll_y))
					scroll_y = old_scroll_y;
			}
		}

		if (scroll_x != old_scroll_x || scroll_y != old_scroll_y)
		{
			if (!AllPlayersInVisibleScreen())
			{
				scroll_x = old_scroll_x;
				scroll_y = old_scroll_y;
			}
			else
			{
/*PALM!!!		ScrollScreen(player, SCROLL_INIT);
				ScrollLevel(old_scroll_x - scroll_x, old_scroll_y - scroll_y);
*/				DrawLevel();
				DrawAllPlayers();
			}
		}
	}

	player->StepFrame = 0;

	if (moved & MF_MOVING)
	{
		if (old_jx != jx && old_jy == jy)
			player->MovDir = (old_jx < jx ? MV_RIGHT : MV_LEFT);
		else if (old_jx == jx && old_jy != jy)
			player->MovDir = (old_jy < jy ? MV_DOWN : MV_UP);

		DrawLevelField(jx, jy);	// for "crumbled sand"

		player->last_move_dir = player->MovDir;
		player->is_moving = TRUE;
		player->is_snapping = FALSE;

		player->is_switching = FALSE;

		player->is_dropping = FALSE;

	}
	else
	{
		CheckGravityMovementWhenNotMoving(player);
		player->is_moving = FALSE;
	}

	if (game.engine_version < VERSION_IDENT(3,0,7,0))
	{
		TestIfHeroTouchesBadThing(jx, jy);
//		TestIfPlayerTouchesCustomElement(jx, jy);
	}

	if (!player->active)
		RemoveHero(player);

	return moved;
}

void ScrollPlayer(struct PlayerInfo *player, int mode)
{
	int jx = player->jx, jy = player->jy;
	short int jxy=A2TO1(jx,jy);
	int last_jx = player->last_jx, last_jy = player->last_jy;
	short int last_jxy=A2TO1(last_jx,last_jy);
	int move_stepsize = TILEX / player->move_delay_value;

	if (!player->active || !player->MovPos)
		return;

	if (mode == SCROLL_INIT)
	{
		player->actual_frame_counter = FrameCounter;
//		player->GfxPos = move_stepsize * (player->MovPos / move_stepsize);

		if (Feld[last_jxy] == EL_EMPTY)
			InitGfxValueAndFeld(last_jxy,EL_PLAYER_IS_LEAVING);

		return;
	}
	else if (!FrameReached(&player->actual_frame_counter, 1))
		return;

	player->MovPos += (player->MovPos > 0 ? -1 : 1) * move_stepsize;
//	player->GfxPos = move_stepsize * (player->MovPos / move_stepsize);

	if (!player->block_last_field &&
					Feld[last_jxy] == EL_PLAYER_IS_LEAVING)
		InitGfxValueAndFeld(last_jxy,EL_EMPTY);

	// before DrawPlayer() to draw correct player graphic for this case
	if (player->MovPos == 0)
		CheckGravityMovement(player);

	if (player->MovPos == 0)	// player reached destination field
	{
		if (player->move_delay_reset_counter > 0)
		{
			player->move_delay_reset_counter--;

			if (player->move_delay_reset_counter == 0)
			{
				// continue with normal speed after quickly moving through gate
				HALVE_PLAYER_SPEED(player);

				// be able to make the next move without delay
				player->move_delay = 0;
			}
		}

		if (player->block_last_field &&
						Feld[last_jxy] == EL_PLAYER_IS_LEAVING)
			InitGfxValueAndFeld(last_jxy,EL_EMPTY);

		player->last_jx = jx;
		player->last_jy = jy;

		if (Feld[jxy] == EL_EXIT_OPEN ||
				Feld[jxy] == EL_SP_EXIT_OPEN ||
				Feld[jxy] == EL_SP_EXIT_OPENING)	// <-- special case
		{
			DrawPlayer(player);	// needed here only to cleanup last field
			RemoveHero(player);

			if (local_player->friends_still_needed == 0 ||
							  IS_SP_ELEMENT(Feld[jxy]))
				player->LevelSolved = player->GameOver = TRUE;
		}
/*
    // !!! ENABLE THIS FOR NEW VERSIONS !!!
    // this breaks one level: "machine", level 000
    {
      int move_direction = player->MovDir;
      int enter_side = MV_DIR_OPPOSITE(move_direction);
      int leave_side = move_direction;
      int old_jx = last_jx;
      int old_jy = last_jy;
      int old_element = Feld[old_jx][old_jy];
      int new_element = Feld[jx][jy];

      // !!! TEST ONLY !!!
      if (IS_CUSTOM_ELEMENT(old_element))
	CheckElementChangeByPlayer(old_jx, old_jy, old_element,
				   CE_LEFT_BY_PLAYER,
				   player->index_bit, leave_side);

      CheckTriggeredElementChangeByPlayer(old_jx, old_jy, old_element,
					  CE_OTHER_GETS_LEFT,
					  player->index_bit, leave_side);

      if (IS_CUSTOM_ELEMENT(new_element))
	CheckElementChangeByPlayer(jx, jy, new_element, CE_ENTERED_BY_PLAYER,
				   player->index_bit, enter_side);

      CheckTriggeredElementChangeByPlayer(jx, jy, new_element,
					  CE_OTHER_GETS_ENTERED,
					  player->index_bit, enter_side);

    }
*/
		if (game.engine_version >= VERSION_IDENT(3,0,7,0))
		{
			TestIfHeroTouchesBadThing(jx, jy);
/*			TestIfPlayerTouchesCustomElement(jx, jy);
			// needed because pushed element has not yet reached its destination,
			// so it would trigger a change event at its previous field location
			if (!player->is_pushing)
				TestIfElementTouchesCustomElement(jx, jy);	// for empty space
*/
			if (!player->active)
				RemoveHero(player);
		}

		if (level.use_step_counter)
		{
			int i;

			TimePlayed++;

			if (TimeLeft > 0)
			{
				TimeLeft--;

				DrawGameValue_Time(TimeLeft);

				if (!TimeLeft && setup.time_limit)
					KillHero(&stored_player_);
			}
			else if (level.time == 0 && !AllPlayersGone) // level w/o time limit
				DrawGameValue_Time(TimePlayed);
		}
	
		if (tape.single_step && tape.recording && !tape.pausing &&
				!player->programmed_action)
			TapeTogglePause(TAPE_TOGGLE_AUTOMATIC);
	}
}

void ScrollScreen(struct PlayerInfo *player, int mode)
{
	static unsigned long screen_frame_counter = 0;

	if (mode == SCROLL_INIT)
	{
		// set scrolling step size according to actual player's moving speed
		ScrollStepSize = TILEX / player->move_delay_value;

		screen_frame_counter = FrameCounter;
		ScreenMovDir = player->MovDir;
		ScreenMovPos = player->MovPos;
//		ScreenGfxPos = ScrollStepSize * (ScreenMovPos / ScrollStepSize);
		return;
	}
	else if (!FrameReached(&screen_frame_counter, 1))
		return;

	if (ScreenMovPos)
	{
		ScreenMovPos += (ScreenMovPos > 0 ? -1 : 1) * ScrollStepSize;
//		ScreenGfxPos = ScrollStepSize * (ScreenMovPos / ScrollStepSize);
		redraw_mask |= REDRAW_FIELD;
	}
	else
		ScreenMovDir = MV_NO_MOVING;
}

void TestIfGoodThingHitsBadThing(int good_x, int good_y, int good_move_dir)
{
	int i, kill_x = -1, kill_y = -1;
	int bad_element = -1;
	const int test_xy_x[4] = {0,-1,+1,0};// for palm only, was static
	const int test_xy_y[4] = {-1,0,0,+1};// for palm only, was static
	const int test_dir[4] = // for palm only, was static
	{
		MV_UP,
		MV_LEFT,
		MV_RIGHT,
		MV_DOWN
	};

	for (i = 0; i < NUM_DIRECTIONS; i++)
	{
		int test_x, test_y, test_move_dir, test_element;

		test_x = good_x + test_xy_x[i];
		test_y = good_y + test_xy_y[i];

		if (!IN_LEV_FIELD(test_x, test_y))
			continue;

		test_move_dir =
			(IS_MOVING(A2TO1(test_x, test_y)) ?
								MovDir[A2TO1(test_x, test_y)] : MV_NO_MOVING);

		test_element = MovingOrBlocked2ElementIfNotLeaving(A2TO1(test_x, test_y));

		// 1st case: good thing is moving towards DONT_RUN_INTO style bad thing;
		//   2nd case: DONT_TOUCH style bad thing does not move away from good thing
    
		if ((DONT_RUN_INTO(test_element) && good_move_dir == test_dir[i]) ||
				(DONT_TOUCH(test_element)    && test_move_dir != test_dir[i]))
		{
			kill_x = test_x;
			kill_y = test_y;
			bad_element = test_element;

			break;
		}
	}

	if (kill_x != -1 || kill_y != -1)
	{
		if (IS_PLAYER(A2TO1(good_x, good_y)))
		{
			struct PlayerInfo *player = PLAYERINFO(good_x, good_y);

			if (player->shield_deadly_time_left > 0 &&
								!IS_INDESTRUCTIBLE(bad_element))
				Bang(kill_x, kill_y);
			else if (!PLAYER_ENEMY_PROTECTED(good_x, good_y,A2TO1(good_x,good_y)))
				KillHero(player);
		}
		else
			Bang(good_x, good_y);
	}
}

void TestIfBadThingHitsGoodThing(int bad_x, int bad_y, int bad_move_dir)
{
	int i, kill_x = -1, kill_y = -1;
	int bad_element = Feld[A2TO1(bad_x,bad_y)];
	const int test_xy_x[4] = {0,-1,+1,0};// for palm only, was static
	const int test_xy_y[4] = {-1,0,0,+1};// for palm only, was static
	const int touch_dir[4] = //for palm only, was static
	{
		MV_LEFT | MV_RIGHT,
		MV_UP   | MV_DOWN,
		MV_UP   | MV_DOWN,
		MV_LEFT | MV_RIGHT
	};
	const int test_dir[4] = // for palm only, was static
	{
		MV_UP,
		MV_LEFT,
		MV_RIGHT,
		MV_DOWN
	};

	if (bad_element == EL_EXPLOSION)	// skip just exploding bad things
		return;

	for (i = 0; i < NUM_DIRECTIONS; i++)
	{
		int test_x, test_y, test_move_dir, test_element;
		short int test_xy;

		test_x = bad_x + test_xy_x[i];
		test_y = bad_y + test_xy_y[i];
		if (!IN_LEV_FIELD(test_x, test_y))
			continue;

		test_xy=A2TO1(test_x,test_y);
		
		test_move_dir =	(IS_MOVING(test_xy) ? MovDir[test_xy] : MV_NO_MOVING);

		test_element = Feld[test_xy];

		// 1st case: good thing is moving towards DONT_RUN_INTO style bad thing;
		// 2nd case: DONT_TOUCH style bad thing does not move away from good thing

		if ((DONT_RUN_INTO(bad_element) &&  bad_move_dir == test_dir[i]) ||
					(DONT_TOUCH(bad_element)    && test_move_dir != test_dir[i]))
		{
			// good thing is player or penguin that does not move away
			if (IS_PLAYER(A2TO1(test_x, test_y)))
			{
				struct PlayerInfo *player = PLAYERINFO(test_x, test_y);

				if (bad_element == EL_ROBOT && player->is_moving)
					continue;	// robot does not kill player if he is moving

				if (game.engine_version >= VERSION_IDENT(3,0,7,0))
				{
					if (player->MovPos != 0 && !(player->MovDir & touch_dir[i]))
						continue;		// center and border element do not touch
				}

				kill_x = test_x;
				kill_y = test_y;
				break;
			}
			else if (test_element == EL_PENGUIN)
			{
				kill_x = test_x;
				kill_y = test_y;
				break;
			}
		}
	}

	if (kill_x != -1 || kill_y != -1)
	{
		if (IS_PLAYER(A2TO1(kill_x, kill_y)))
		{
			struct PlayerInfo *player = PLAYERINFO(kill_x, kill_y);

			if (player->shield_deadly_time_left > 0 &&
								!IS_INDESTRUCTIBLE(bad_element))
				Bang(bad_x, bad_y); 
			else if (!PLAYER_ENEMY_PROTECTED(kill_x, kill_y,A2TO1(kill_x,kill_y)))
				KillHero(player);
		}
		else
			Bang(kill_x, kill_y);
	}
}

void TestIfBadThingTouchesOtherBadThing(int bad_x, int bad_y)
{
	int i, kill_x = bad_x, kill_y = bad_y;
	const int xy_x[4] = {0,-1,+1,0};//for palm only, was static
	const int xy_y[4] = {-1,0,0,+1};//for palm only, was static

	for (i = 0; i < NUM_DIRECTIONS; i++)
	{
		int x, y, element;

		x = bad_x + xy_x[i];
		y = bad_y + xy_y[i];
		if (!IN_LEV_FIELD(x, y))
			continue;

		element = Feld[A2TO1(x,y)];
		if (IS_AMOEBOID(element) || element == EL_GAME_OF_LIFE ||
					element == EL_AMOEBA_GROWING || element == EL_AMOEBA_DROP)
		{
			kill_x = x;
			kill_y = y;
			break;
		}
	}

	if (kill_x != bad_x || kill_y != bad_y)
		Bang(bad_x, bad_y);
}

void KillHero(struct PlayerInfo *player)
{
	int jx = player->jx, jy = player->jy;

	if (!player->active)
		return;

	// remove accessible field at the player's position
	InitGfxValueAndFeld(A2TO1(jx,jy),EL_EMPTY);

	// deactivate shield (else Bang()/Explode() would not work right)
	player->shield_normal_time_left = 0;
	player->shield_deadly_time_left = 0;

	Bang(jx, jy);
	BuryHero(player);
}

void BuryHero(struct PlayerInfo *player)
{
//	int jx = player->jx, jy = player->jy;

	if (!player->active)
		return;

	player->GameOver = TRUE;
	RemoveHero(player);
}

void RemoveHero(struct PlayerInfo *player)
{
	int jx = player->jx, jy = player->jy;
//	int found = FALSE;

	player->present = FALSE;
	player->active = FALSE;

	if (!getExplodeField(A2TO1(jx,jy)))
		StorePlayerXY=-1;//[jx][jy] = 0;

/*
	if (stored_player[i].active)
		found = TRUE;

	if (!found)
*/		AllPlayersGone = TRUE;

	ExitX = ZX = jx;
	ExitY = ZY = jy;
} 

/*
  =============================================================================
  checkDiagonalPushing()
  -----------------------------------------------------------------------------
  check if diagonal input device direction results in pushing of object
  (by checking if the alternative direction is walkable, diggable, ...)
  =============================================================================
*/

static boolean checkDiagonalPushing(struct PlayerInfo *player,
									int x, int y, int real_dx, int real_dy)
{
	int jx, jy, dx, dy, xx, yy;

	if (real_dx == 0 || real_dy == 0)	// no diagonal direction => push
		return TRUE;

	// diagonal direction: check alternative direction
	jx = player->jx;
	jy = player->jy;
	dx = x - jx;
	dy = y - jy;
	xx = jx + (dx == 0 ? real_dx : 0);
	yy = jy + (dy == 0 ? real_dy : 0);

	return (!IN_LEV_FIELD(xx, yy) || IS_SOLID_FOR_PUSHING(Feld[A2TO1(xx,yy)]));
};

//PALM only!!!!
LOGIC_SECTION static byte GetNewPushDelay(short int element)
{

	if( element==EL_SPRING || element==EL_BALLOON)
		return 0;
	if( element==EL_SOKOBAN_OBJECT || element==EL_SOKOBAN_FIELD_FULL ||
				element==EL_SATELLITE || element==EL_SP_DISK_YELLOW)
		return 2;
	return (game.default_push_delay_fixed + RND(game.default_push_delay_random)); 
};

/*
  =============================================================================
  DigField()
  -----------------------------------------------------------------------------
  x, y:			field next to player (non-diagonal) to try to dig to
  real_dx, real_dy:	direction as read from input device (can be diagonal)
  =============================================================================
*/

LOGIC_SECTION int DigField(struct PlayerInfo *player,
			int oldx, int oldy, int x, int y, int real_dx, int real_dy, int mode)
{
	short int oldxy=A2TO1(oldx,oldy);
	short int xy=A2TO1(x,y);

	boolean is_player = (IS_PLAYER(oldxy) || mode != DF_DIG);
	boolean player_was_pushing = player->is_pushing;
	int jx = oldx, jy = oldy;
	int dx = x - jx, dy = y - jy;
	int nextx = x + dx, nexty = y + dy;

	short int jxy=A2TO1(jx,jy);
	short int nextxy=A2TO1(nextx,nexty);

	int move_direction = (dx == -1 ? MV_LEFT :
						  dx == +1 ? MV_RIGHT :
						  dy == -1 ? MV_UP :
						  dy == +1 ? MV_DOWN : MV_NO_MOVING);
	int opposite_direction = MV_DIR_OPPOSITE(move_direction);
	int dig_side = MV_DIR_OPPOSITE(move_direction);
	int old_element = Feld[jxy];
	int element;

	if (is_player)		// function can also be called by EL_PENGUIN
	{
		if (player->MovPos == 0)
		{
			player->is_digging = FALSE;
			player->is_collecting = FALSE;
		}

		if (player->MovPos == 0)	// last pushing move finished
			player->is_pushing = FALSE;

		if (mode == DF_NO_PUSH)	// player just stopped pushing
		{
			player->is_switching = FALSE;
			player->push_delay = 0;

			return MF_NO_ACTION;
		}
	}

	if (IS_MOVING(xy) || IS_PLAYER(xy))
		return MF_NO_ACTION;

	if (IS_TUBE(Back[jxy]) && game.engine_version >= VERSION_IDENT(2,2,0,0))
		old_element = Back[jxy];

	if (IS_WALKABLE(old_element) && !ACCESS_FROM(old_element, move_direction))
		return MF_NO_ACTION;	// field has no opening in this direction

	if (IS_PASSABLE(old_element) && !ACCESS_FROM(old_element,opposite_direction))
		return MF_NO_ACTION;	// field has no opening in this direction

	element = Feld[xy];

	if (!is_player && !IS_COLLECTIBLE(element))	// penguin cannot collect it
		return MF_NO_ACTION;

	if (mode == DF_SNAP && !IS_SNAPPABLE(element) &&
			game.engine_version >= VERSION_IDENT(2,2,0,0))
		return MF_NO_ACTION;

	if (game.gravity && is_player && !player->is_auto_moving &&
			canFallDown(player) && move_direction != MV_DOWN &&
			!canMoveToValidFieldWithGravity(jxy, move_direction))
		return MF_NO_ACTION;	// player cannot walk here due to gravity

	switch (element)
	{

		default:

			if (IS_WALKABLE(element) && ACCESS_FROM(element, opposite_direction))
			{

				if (IS_GATE(element))
				{
					if (!player->key[element - EL_GATE_1])
						return MF_NO_ACTION;
				}
				else if (IS_GATE_GRAY(element))
				{
					if (!player->key[element - EL_GATE_1_GRAY])
						return MF_NO_ACTION;
				}

				break;
			}
			else if (IS_PASSABLE(element) && canPassField(xy, move_direction))
			{

				if (!ACCESS_FROM(element, opposite_direction))
					return MF_NO_ACTION;	// field not accessible from this direction

				if (IS_EM_GATE(element))
				{
					if (!player->key[element - EL_EM_GATE_1])
						return MF_NO_ACTION;
				}
				else if (IS_EM_GATE_GRAY(element))
				{
					if (!player->key[element - EL_EM_GATE_1_GRAY])
						return MF_NO_ACTION;
				}
				else if (IS_SP_PORT(element))
				{
					if (element == EL_SP_GRAVITY_PORT_LEFT ||
							element == EL_SP_GRAVITY_PORT_RIGHT ||
							element == EL_SP_GRAVITY_PORT_UP ||
							element == EL_SP_GRAVITY_PORT_DOWN)
						game.gravity = !game.gravity;
					else if (element == EL_SP_GRAVITY_ON_PORT_LEFT ||
							element == EL_SP_GRAVITY_ON_PORT_RIGHT ||
							element == EL_SP_GRAVITY_ON_PORT_UP ||
							element == EL_SP_GRAVITY_ON_PORT_DOWN)
						game.gravity = TRUE;
					else if (element == EL_SP_GRAVITY_OFF_PORT_LEFT ||
							element == EL_SP_GRAVITY_OFF_PORT_RIGHT ||
							element == EL_SP_GRAVITY_OFF_PORT_UP ||
							element == EL_SP_GRAVITY_OFF_PORT_DOWN)
						game.gravity = FALSE;
				}

				// automatically move to the next field with double speed
				player->programmed_action = move_direction;
				if (player->move_delay_reset_counter == 0)
				{
					player->move_delay_reset_counter = 2;	// two double speed steps

					DOUBLE_PLAYER_SPEED(player);
				}

				break;
			}
			else if (IS_DIGGABLE(element))
			{
				RemoveField(xy);

				if (mode != DF_SNAP)
				{
					if(ActionGraphicExist(element,ACTION_DIGGING))
						GfxElement[xy] = GFX_ELEMENT(element);
					player->is_digging = TRUE;
				}
/*
				CheckTriggeredElementChangeByPlayer(x, y, element,CE_OTHER_GETS_DIGGED,
														player->index_bit, dig_side);

				if (mode == DF_SNAP)
					TestIfElementTouchesCustomElement(x, y);	// for empty space
*/
				break;
			}
			else if (IS_COLLECTIBLE(element))
			{
				RemoveField(xy);

				if (is_player && mode != DF_SNAP)
				{
					if(ActionGraphicExist(element,ACTION_COLLECTING))
						GfxElement[xy] = element;

					player->is_collecting = TRUE;
				}

				if (element == EL_SPEED_PILL)
					player->move_delay_value = MOVE_DELAY_HIGH_SPEED;
				else if (element == EL_EXTRA_TIME && level.time > 0)
				{
					TimeLeft += 10;
					DrawGameValue_Time(TimeLeft);
				}
				else if (element == EL_SHIELD_NORMAL || element == EL_SHIELD_DEADLY)
				{
					player->shield_normal_time_left += 10;
					if (element == EL_SHIELD_DEADLY)
						player->shield_deadly_time_left += 10;
				}
				else if (element == EL_DYNAMITE || element == EL_SP_DISK_RED)
				{
					if (player->inventory_size < MAX_INVENTORY_SIZE)
						player->inventory_element[player->inventory_size++] = element;

					DrawGameValue_Dynamite(local_player->inventory_size);
				}
				else if (element == EL_DYNABOMB_INCREASE_NUMBER)
				{
					player->dynabomb_count++;
					player->dynabombs_left++;
				}
				else if (element == EL_DYNABOMB_INCREASE_SIZE)
				{
					player->dynabomb_size++;
				}
				else if (element == EL_DYNABOMB_INCREASE_POWER)
				{
					player->dynabomb_xl = TRUE;
				}
				else if ((element >= EL_KEY_1 && element <= EL_KEY_4) ||
						 (element >= EL_EM_KEY_1 && element <= EL_EM_KEY_4))
				{
					int key_nr = (element >= EL_KEY_1 && element <= EL_KEY_4 ?
								  element - EL_KEY_1 : element - EL_EM_KEY_1);

					player->key[key_nr] = TRUE;

					DrawGameValue_Keys(player);

					redraw_mask |= REDRAW_DOOR_1;
				}
				else if (IS_ENVELOPE(element))
				{
					player->show_envelope = element;
				}
				else if (IS_DROPPABLE(element) ||
						 IS_THROWABLE(element))	// can be collected and dropped
				{
					int i;

					int collect_count = FindElementCollectCount(element);
					
					if (collect_count == 0)
						player->inventory_infinite_element = element;
					else
						for (i = 0; i < collect_count; i++)
										//PALM, was element_info[element].collect_count
							if (player->inventory_size < MAX_INVENTORY_SIZE)
								player->inventory_element[player->inventory_size++] = element;

					DrawGameValue_Dynamite(local_player->inventory_size);
				}
				else if (FindElementCollectCount(element) > 0)
										//PALM, was element_info[element].collect_count
				{
					local_player->gems_still_needed -= FindElementCollectCount(element);
										//PALM, was element_info[element].collect_count;
					if (local_player->gems_still_needed < 0)
						local_player->gems_still_needed = 0;

					DrawGameValue_Emeralds(local_player->gems_still_needed);
				}

				RaiseScoreElement(element);
/*
				if (is_player)
					CheckTriggeredElementChangeByPlayer(x, y, element,
												CE_OTHER_GETS_COLLECTED,
											player->index_bit, dig_side);

				if (mode == DF_SNAP)
					TestIfElementTouchesCustomElement(x, y);	// for empty space
*/
				break;
			}
			else if (IS_PUSHABLE(element))
			{
				if (mode == DF_SNAP && element != EL_BD_ROCK)
					return MF_NO_ACTION;

				if (CAN_FALL(element) && dy)
					return MF_NO_ACTION;

				if (CAN_FALL(element) && IN_LEV_FIELD(x, y + 1) && 
													IS_FREE(xy+A2TO1(0,1)) &&
							 !(element == EL_SPRING && level.use_spring_bug))
					return MF_NO_ACTION;
/*
				if (CAN_MOVE(element) && GET_MAX_MOVE_DELAY(element) == 0 &&
											((move_direction & MV_VERTICAL &&
							((element_info[element].move_pattern & MV_LEFT &&
							  IN_LEV_FIELD(x - 1, y) && IS_FREE(x - 1, y)) ||
							(element_info[element].move_pattern & MV_RIGHT &&
							IN_LEV_FIELD(x + 1, y) && IS_FREE(x + 1, y)))) ||
										   (move_direction & MV_HORIZONTAL &&
							  ((element_info[element].move_pattern & MV_UP &&
							  IN_LEV_FIELD(x, y - 1) && IS_FREE(x, y - 1)) ||
							 (element_info[element].move_pattern & MV_DOWN &&
							 IN_LEV_FIELD(x, y + 1) && IS_FREE(x, y + 1))))))
					return MF_NO_ACTION;
*/				// do not push elements already moving away faster than player
				if (CAN_MOVE(element) && MovDir[xy] == move_direction &&
					   ABS(getElementMoveStepsize(xy)) > MOVE_STEPSIZE_NORMAL)
					return MF_NO_ACTION;

				if (game.engine_version >= VERSION_IDENT(3,1,0,0))
				{
					if (player->push_delay_value == -1 || !player_was_pushing)
						player->push_delay_value = GetNewPushDelay(element);
				}
				else if (game.engine_version >= VERSION_IDENT(3,0,7,1))
				{
					if (player->push_delay_value == -1)
						player->push_delay_value = GetNewPushDelay(element);
				}
				else if (game.engine_version >= VERSION_IDENT(2,2,0,7))
				{
					if (!player->is_pushing)
						player->push_delay_value = GetNewPushDelay(element);
				}

				player->is_pushing = TRUE;

				if (!(IN_LEV_FIELD(nextx, nexty) &&
						  (IS_FREE(nextxy) ||
						  (Feld[nextxy] == EL_SOKOBAN_FIELD_EMPTY &&
							IS_SB_ELEMENT(element)))))
					return MF_NO_ACTION;

				if (!checkDiagonalPushing(player, x, y, real_dx, real_dy))
					return MF_NO_ACTION;

				if (player->push_delay == 0)	// new pushing; restart delay
					player->push_delay = FrameCounter;

				if (!FrameReached(&player->push_delay, player->push_delay_value) &&
						 !(tape.playing && tape.file_version < FILE_VERSION_2_0) &&
									 element != EL_SPRING && element != EL_BALLOON)
				{
					// make sure that there is no move delay before next try to push
					if (game.engine_version >= VERSION_IDENT(3,0,7,1))
						player->move_delay = INITIAL_MOVE_DELAY_OFF;

					return MF_NO_ACTION;
				}

				if (IS_SB_ELEMENT(element))
				{
					if (element == EL_SOKOBAN_FIELD_FULL)
					{
						Back[xy] = EL_SOKOBAN_FIELD_EMPTY;
						local_player->sokobanfields_still_needed++;
					}

					if (Feld[nextxy] == EL_SOKOBAN_FIELD_EMPTY)
					{
						Back[nextxy] = EL_SOKOBAN_FIELD_EMPTY;
						local_player->sokobanfields_still_needed--;
					}

					InitGfxValueAndFeld(xy,EL_SOKOBAN_OBJECT);

					if (local_player->sokobanfields_still_needed == 0 &&
											game.emulation == EMU_SOKOBAN)
					{
						player->LevelSolved = player->GameOver = TRUE;
					}
				}

				InitMovingField(xy, move_direction);
				InitGfxValueAndGfxAction(xy, ACTION_PUSHING);

				if (mode == DF_SNAP)
					ContinueMoving(x, y);
				else
					MovPos[xy] = (dx != 0 ? dx : dy);

				setPushedTRUE(xy);//Pushed[x][y] = TRUE;
				setPushedTRUE(nextxy);//Pushed[nextx][nexty] = TRUE;

				if (game.engine_version < VERSION_IDENT(2,2,0,7))
					player->push_delay_value = GetNewPushDelay(element);
				else
					player->push_delay_value = -1;	// get new value later

				// check for element change _after_ element has been pushed!
				
				break;
			}
			else if (IS_SWITCHABLE(element))
			{
				if (PLAYER_SWITCHING(player, x, y))
				{
/*					CheckTriggeredElementChangeByPlayer(x,y, element,
												CE_OTHER_GETS_PRESSED,
										player->index_bit, dig_side);
*/
					return MF_ACTION;
				}

				player->is_switching = TRUE;
				player->switch_x = x;
				player->switch_y = y;

				if (element == EL_ROBOT_WHEEL)
				{
					InitGfxValueAndFeld(xy,EL_ROBOT_WHEEL_ACTIVE);
//					AddDelayChanging(xy, level.time_wheel * FRAMES_PER_SECOND,
//							EL_ROBOT_WHEEL);//PALM!!!
					ZX = x;
					ZY = y;

					DrawLevelField(x, y);
				}
				else if (element == EL_SP_TERMINAL)
				{
					int xx, yy;

					for (yy = 0; yy < lev_fieldy; yy++) for (xx=0; xx < lev_fieldx; xx++)
					{
						short int el=Feld[A2TO1(xx,yy)];
						if ( el == EL_SP_DISK_YELLOW)
							Bang(xx, yy);
						else if (el == EL_SP_TERMINAL)
							InitGfxValueAndFeld(A2TO1(xx,yy),EL_SP_TERMINAL_ACTIVE);
					}
				}
				else if (IS_BELT_SWITCH(element))
				{
					ToggleBeltSwitch(x, y);
				}
				else if (element == EL_SWITCHGATE_SWITCH_UP ||
							element == EL_SWITCHGATE_SWITCH_DOWN)
				{
					ToggleSwitchgateSwitch(x, y);
				}
				else if (element == EL_LIGHT_SWITCH ||
							element == EL_LIGHT_SWITCH_ACTIVE)
				{
					ToggleLightSwitch(x, y);
				}
				else if (element == EL_TIMEGATE_SWITCH)
				{
					ActivateTimegateSwitch(x, y);
				}
				else if (element == EL_BALLOON_SWITCH_LEFT ||
							element == EL_BALLOON_SWITCH_RIGHT ||
							element == EL_BALLOON_SWITCH_UP ||
							element == EL_BALLOON_SWITCH_DOWN ||
							element == EL_BALLOON_SWITCH_ANY)
				{
					if (element == EL_BALLOON_SWITCH_ANY)
						game.balloon_dir = move_direction;
					else
						game.balloon_dir = (element == EL_BALLOON_SWITCH_LEFT  ? MV_LEFT :
											element == EL_BALLOON_SWITCH_RIGHT ? MV_RIGHT :
											element == EL_BALLOON_SWITCH_UP    ? MV_UP :
											element == EL_BALLOON_SWITCH_DOWN  ? MV_DOWN :
											MV_NO_MOVING);
				}
				else if (element == EL_LAMP)
				{
					InitGfxValueAndFeld(xy,EL_LAMP_ACTIVE);
					local_player->lights_still_needed--;

					DrawLevelField(x, y);
				}
				else if (element == EL_TIME_ORB_FULL)
				{
					InitGfxValueAndFeld(xy,EL_TIME_ORB_EMPTY);
					TimeLeft += 10;
					DrawGameValue_Time(TimeLeft);

					DrawLevelField(x, y);

				}
/*
				CheckTriggeredElementChangeByPlayer(x, y, element,
											CE_OTHER_IS_SWITCHING,
									player->index_bit, dig_side);

				CheckTriggeredElementChangeByPlayer(x,y, element,CE_OTHER_GETS_PRESSED,
														player->index_bit, dig_side);
*/
				return MF_ACTION;
			}
			else
			{
				if (!PLAYER_SWITCHING(player, x, y))
				{
					player->is_switching = TRUE;
					player->switch_x = x;
					player->switch_y = y;
/*
					// !!! TEST ONLY !!!
					CheckElementChangeByPlayer(x, y, element, CE_SWITCHED,
											player->index_bit, dig_side);
					CheckTriggeredElementChangeByPlayer(x, y, element,
												CE_OTHER_IS_SWITCHING,
										player->index_bit, dig_side);
*/				}
/*
				// !!! TEST ONLY !!! (this breaks "machine", level 000)
				CheckElementChangeByPlayer(x, y, element, CE_PRESSED_BY_PLAYER,
												player->index_bit, dig_side);
				CheckTriggeredElementChangeByPlayer(x,y, element,CE_OTHER_GETS_PRESSED,
														player->index_bit, dig_side);
*/			}

			return MF_NO_ACTION;
	}

	player->push_delay = 0;

	if (Feld[xy] != element)		// really digged/collected something 
		player->is_collecting = !player->is_digging;

	return MF_MOVING;
}

boolean SnapField(struct PlayerInfo *player, int dx, int dy)
{
	int jx = player->jx, jy = player->jy;
	int x = jx + dx, y = jy + dy;
	int snap_direction = (dx == -1 ? MV_LEFT :
						dx == +1 ? MV_RIGHT :
						dy == -1 ? MV_UP :
						dy == +1 ? MV_DOWN : MV_NO_MOVING);

	if (player->MovPos != 0 && game.engine_version >= VERSION_IDENT(2,2,0,0))
		return FALSE;

	if (!player->active || !IN_LEV_FIELD(x, y))
		return FALSE;

	if (dx && dy)
		return FALSE;

	if (!dx && !dy)
	{
		if (player->MovPos == 0)
			player->is_pushing = FALSE;

		player->is_snapping = FALSE;

		if (player->MovPos == 0)
		{
			player->is_moving = FALSE;
			player->is_digging = FALSE;
			player->is_collecting = FALSE;
		}

		return FALSE;
	}

	if (player->is_snapping)
		return FALSE;

	player->MovDir = snap_direction;

	if (player->MovPos == 0)
	{
		player->is_moving = FALSE;
		player->is_digging = FALSE;
		player->is_collecting = FALSE;
	}

	player->is_dropping = FALSE;

	if (DigField(player, jx, jy, x, y, 0, 0, DF_SNAP) == MF_NO_ACTION)
		return FALSE;

	player->is_snapping = TRUE;

	if (player->MovPos == 0)
	{
		player->is_moving = FALSE;
		player->is_digging = FALSE;
		player->is_collecting = FALSE;
	}

	if (player->MovPos != 0)	// prevent graphic bugs in versions < 2.2.0
		DrawLevelField(player->last_jx, player->last_jy);

	DrawLevelField(x, y);

	return TRUE;
}

boolean DropElement(struct PlayerInfo *player)
{
	int old_element, new_element;
	int dropx = player->jx, dropy = player->jy;
	short int dropxy;
	int drop_direction = player->MovDir;
	int drop_side = drop_direction;
	int drop_element = (player->inventory_size > 0 ?
					player->inventory_element[player->inventory_size - 1] :
					player->inventory_infinite_element != EL_UNDEFINED ?
					player->inventory_infinite_element :
					player->dynabombs_left > 0 ?
					EL_DYNABOMB_PLAYER_1_ACTIVE + player->index_nr :
					EL_UNDEFINED);

	if (IS_THROWABLE(drop_element))
	{
		dropx += GET_DX_FROM_DIR(drop_direction);
		dropy += GET_DY_FROM_DIR(drop_direction);

		if (!IN_LEV_FIELD(dropx, dropy))
			return FALSE;
	}

	dropxy=A2TO1(dropx,dropy);
	old_element = Feld[dropxy];	// old element at dropping position
	new_element = drop_element;		// default: no change when dropping

	// check if player is active, not moving and ready to drop
	if (!player->active || player->MovPos || player->drop_delay > 0)
		return FALSE;

	// check if player has anything that can be dropped
	if (new_element == EL_UNDEFINED)
		return FALSE;

	// check if anything can be dropped at the current position
	if (IS_ACTIVE_BOMB(old_element) || old_element == EL_EXPLOSION)
		return FALSE;

	// collected custom elements can only be dropped on empty fields
	if (IS_CUSTOM_ELEMENT(new_element) && old_element != EL_EMPTY)
		return FALSE;

	if (old_element != EL_EMPTY)
		Back[dropxy] = old_element;	// store old element on this field

	ResetGfxAnimation(dropxy);
	ResetRandomAnimationValue(dropxy);

	if (player->inventory_size > 0 ||
				player->inventory_infinite_element != EL_UNDEFINED)
	{
		if (player->inventory_size > 0)
		{
			player->inventory_size--;

			DrawGameValue_Dynamite(local_player->inventory_size);

			if (new_element == EL_DYNAMITE)
				new_element = EL_DYNAMITE_ACTIVE;
			else if (new_element == EL_SP_DISK_RED)
				new_element = EL_SP_DISK_RED_ACTIVE;
		}

		Feld[dropxy] = new_element;

		if (IN_SCR_FIELD(SCREENX(dropx), SCREENY(dropy)))
			DrawGraphicThruMask(SCREENX(dropx), SCREENY(dropy),
								el2img(Feld[dropxy]), 0);
/*
		// needed if previous element just changed to "empty" in the last frame
		Changed[dropx][dropy] = 0;		// allow another change

		// !!! TEST ONLY !!!
		CheckElementChangeByPlayer(dropx, dropy, new_element, CE_DROPPED_BY_PLAYER,
													player->index_bit, drop_side);
		CheckTriggeredElementChangeByPlayer(dropx, dropy, new_element,
													CE_OTHER_GETS_DROPPED,
													player->index_bit, drop_side);

		TestIfElementTouchesCustomElement(dropx, dropy);
*/	}
	else		// player is dropping a dyna bomb
	{
		player->dynabombs_left--;

		Feld[dropxy] = new_element;

		if (IN_SCR_FIELD(SCREENX(dropx), SCREENY(dropy)))
			DrawGraphicThruMask(SCREENX(dropx), SCREENY(dropy),
												el2img(Feld[dropxy]), 0);

	}

	if (Feld[dropxy] == new_element) // uninitialized unless CE change
	{
		InitField_WithBug1(dropx, dropy, FALSE, dropxy);
	}

	new_element = Feld[dropxy];	// element might have changed
/*
	if (IS_CUSTOM_ELEMENT(new_element) && CAN_MOVE(new_element) &&
				element_info[new_element].move_pattern == MV_WHEN_DROPPED)
	{
		int move_direction, nextx, nexty;

		if (element_info[new_element].move_direction_initial == MV_START_AUTOMATIC)
			MovDir[dropx][dropy] = drop_direction;

		move_direction = MovDir[dropx][dropy];
		nextx = dropx + GET_DX_FROM_DIR(move_direction);
		nexty = dropy + GET_DY_FROM_DIR(move_direction);

		Changed[dropx][dropy] = 0;		// allow another change
		CheckCollision[dropx][dropy] = 2;
	}
*/
	player->drop_delay = 0;//GET_NEW_DROP_DELAY(drop_element);

	player->is_dropping = TRUE;

	return TRUE;
}

void RaiseScore(int value)
{
	local_player->score += value;

	DrawGameValue_Score(local_player->score);
}

void RaiseScoreElement(int element)
{
	switch(element)
	{
		case EL_EMERALD:
		case EL_BD_DIAMOND:
		case EL_EMERALD_YELLOW:
		case EL_EMERALD_RED:
		case EL_EMERALD_PURPLE:
		case EL_SP_INFOTRON:
			RaiseScore(level.score[SC_EMERALD]);
			break;
		case EL_DIAMOND:
			RaiseScore(level.score[SC_DIAMOND]);
			break;
		case EL_CRYSTAL:
			RaiseScore(level.score[SC_CRYSTAL]);
			break;
		case EL_PEARL:
			RaiseScore(level.score[SC_PEARL]);
			break;
		case EL_BUG:
		case EL_BD_BUTTERFLY:
		case EL_SP_ELECTRON:
			RaiseScore(level.score[SC_BUG]);
			break;
		case EL_SPACESHIP:
		case EL_BD_FIREFLY:
		case EL_SP_SNIKSNAK:
			RaiseScore(level.score[SC_SPACESHIP]);
			break;
		case EL_YAMYAM:
		case EL_DARK_YAMYAM:
			RaiseScore(level.score[SC_YAMYAM]);
			break;
		case EL_ROBOT:
			RaiseScore(level.score[SC_ROBOT]);
			break;
		case EL_PACMAN:
			RaiseScore(level.score[SC_PACMAN]);
			break;
		case EL_NUT:
			RaiseScore(level.score[SC_NUT]);
			break;
		case EL_DYNAMITE:
		case EL_SP_DISK_RED:
		case EL_DYNABOMB_INCREASE_NUMBER:
		case EL_DYNABOMB_INCREASE_SIZE:
		case EL_DYNABOMB_INCREASE_POWER:
			RaiseScore(level.score[SC_DYNAMITE]);
			break;
		case EL_SHIELD_NORMAL:
		case EL_SHIELD_DEADLY:
			RaiseScore(level.score[SC_SHIELD]);
			break;
		case EL_EXTRA_TIME:
			RaiseScore(level.score[SC_TIME_BONUS]);
			break;
		case EL_KEY_1:
		case EL_KEY_2:
		case EL_KEY_3:
		case EL_KEY_4:
			RaiseScore(level.score[SC_KEY]);
			break;
		default:
//			RaiseScore(element_info[element].collect_score);
			break;
	}
}